(*******************************************************)
(*                                                     *)
(*         Formalization of the Algebraic Laws         *)
(*            of Strong Late Bisimilarity.             *)
(*                                                     *)
(*******************************************************)
(*                                                     *)
(* File   : alglaws.v                                  *)
(* Author : Ivan Scagnetto (scagnett@dimi.uniud.it),   *)
(*          Dipartimento di Matematica e Informatica,  *)
(*          University of Udine                        *)
(* Date   : July 1998                                  *)
(*                                                     *)
(*******************************************************)
(*                                                     *)
(* This file is part of "Pi-calculus in (Co)Inductive  *)
(* Type Theory", a joint work with Furio Honsell and   *)
(* Marino Miculan (accepted for publication on TCS).   *)
(* If you are interested you can find more information *)
(* (and a gzipped PostScript version of the article)   *)
(* at the following URL:                               *)
(* http://www.dimi.uniud.it/~scagnett/pi-calculus.html *)
(*                                                     *)
(*******************************************************)

Require Export basiclemmata.

Section equivalence.

Lemma REF: (p:proc)(StBisim p p).

Proof. (* This lemma is proved from scratch: it is a coinductive proof. *)

Cofix; Intro; Apply sb; Do 3 Try (Split; Intros).
Exists p1; Split; Auto.
Exists q1; Split; Auto.
Exists p1; Split; [Assumption | Intro; Auto].
Exists q1; Split; [Assumption | Intro; Auto].
Exists p1; Split; [Assumption | Intros; Auto].
Exists q1; Split; [Assumption | Intros; Auto].

Qed.

Lemma SYM: (p,q:proc)(StBisim p q) -> (StBisim q p).

Proof. (* This lemma is proved from scratch: it is a coinductive proof. *)

Cofix; Intros; Apply sb; Do 3 Try (Split; Intros).

Inversion H.
Elim H1; Intros.
Elim (H4 a); Intros.
Elim (H7 p1 H0); Intros.
Elim H8; Intros; Exists x; Split; [Assumption | Apply SYM; Assumption].

Inversion H.
Elim H1; Intros.
Elim (H4 a); Intros.
Elim (H6 q1 H0); Intros.
Elim H8; Intros; Exists x; Split; [Assumption | Apply SYM; Assumption].

Inversion H.
Elim H1; Intros.
Elim H5; Intros.
Elim (H6 x); Intros.
Elim (H9 p1 H0); Intros.
Elim H10; Intros; Exists x0; Split; [Assumption | Intro; Apply SYM; Auto].

Inversion H.
Elim H1; Intros.
Elim H5; Intros.
Elim (H6 x); Intros.
Elim (H8 q1 H0); Intros.
Elim H10; Intros; Exists x0; Split; [Assumption | Intro; Apply SYM; Auto].

Inversion H.
Elim H1; Intros.
Elim H5; Intros.
Elim (H7 x); Intros.
Elim (H9 p1 H0); Intros.
Elim H10; Intros; Exists x0; Split; [Assumption | Intros; Apply SYM; Apply H12; Auto].

Inversion H.
Elim H1; Intros.
Elim H5; Intros.
Elim (H7 x); Intros.
Elim (H8 q1 H0); Intros.
Elim H10; Intros; Exists x0; Split; [Assumption | Intros; Apply SYM; Apply H12; Auto].

Qed.

Lemma TRANS: (p,q,r:proc)(StBisim p q) -> (StBisim q r) -> (StBisim p r).

Proof. (* This lemma depends on Lemma L6_Light: it is a coinductive proof. *)

Cofix.
Do 4 Intro.
Inversion_clear H.
Intro.
Inversion_clear H.
Apply sb.
Split.
Intro.
Cut ((p1:proc)
      (ftrans p a p1)->(Ex [q1:proc](ftrans q a q1)/\(StBisim p1 q1)))
    /\((q1:proc)
        (ftrans q a q1)->(Ex [p1:proc](ftrans p a p1)/\(StBisim p1 q1))).
Intro.
Cut ((p1:proc)
      (ftrans q a p1)->(Ex [q1:proc](ftrans r a q1)/\(StBisim p1 q1)))
    /\((q1:proc)
        (ftrans r a q1)->(Ex [p1:proc](ftrans q a p1)/\(StBisim p1 q1))).
Intro.
Inversion_clear H.
Inversion_clear H2.
Split.
Intros.
Cut (Ex [q1:proc](ftrans q a q1)/\(StBisim p1 q1)).
Intro.
Inversion_clear H6.
Cut (Ex [q1:proc](ftrans r a q1)/\(StBisim x q1)).
Intro.
Inversion_clear H6.
Exists  x0.
Split.
Tauto.

Apply TRANS with x.
Tauto.

Tauto.

Apply H.
Tauto.

Apply H3.
Assumption.

Intros.
Cut (Ex [p1:proc](ftrans q a p1)/\(StBisim p1 q1)).
Intro.
Inversion_clear H6.
Cut (Ex [p1:proc](ftrans p a p1)/\(StBisim p1 x)).
Intro.
Inversion_clear H6.
Exists  x0.
Split.
Tauto.

Apply TRANS with x.
Tauto.

Tauto.

Apply H4.
Tauto.

Apply H5.
Assumption.

Elim H1.
Intros.
Apply H2.

Elim H0.
Intros.
Apply H.

Split.
Intro.
Cut ((p1:name->proc)
      (btrans p (In x) p1)
       ->(Ex
           [q1:name->proc]
            (btrans q (In x) q1)
            /\((y:name)(StBisim (p1 y) (q1 y)))))
    /\((q1:name->proc)
        (btrans q (In x) q1)
         ->(Ex
             [p1:name->proc]
              (btrans p (In x) p1)
              /\((y:name)(StBisim (p1 y) (q1 y))))).
Intro.
Cut ((p1:name->proc)
      (btrans q (In x) p1)
       ->(Ex
           [q1:name->proc]
            (btrans r (In x) q1)
            /\((y:name)(StBisim (p1 y) (q1 y)))))
    /\((q1:name->proc)
        (btrans r (In x) q1)
         ->(Ex
             [p1:name->proc]
              (btrans q (In x) p1)
              /\((y:name)(StBisim (p1 y) (q1 y))))).
Intro.
Inversion_clear H.
Inversion_clear H2.
Split.
Intros.

Cut (Ex
      [q1:name->proc]
       (btrans q (In x) q1)
       /\((y:name)(StBisim (p1 y) (q1 y)))).
Intro.
Inversion_clear H6.
Cut (Ex
      [q1:name->proc]
       (btrans r (In x) q1)
       /\((y:name)(StBisim (x0 y) (q1 y)))).
Intro.
Inversion_clear H6.
Exists  x1.
Split.
Tauto.

Intro.
Apply TRANS with (x0 y).
Generalize y .
Tauto.

Generalize y .
Tauto.

Apply H.
Tauto.

Apply H3.
Assumption.

Intros.
Cut (Ex
      [p1:name->proc]
       (btrans q (In x) p1)
       /\((y:name)(StBisim (p1 y) (q1 y)))).
Intro.
Inversion_clear H6.
Cut (Ex
      [p1:name->proc]
       (btrans p (In x) p1)
       /\((y:name)(StBisim (p1 y) (x0 y)))).
Intro.
Inversion_clear H6.
Exists  x1.
Split.
Tauto.

Intro.
Apply TRANS with (x0 y).
Generalize y .
Tauto.

Generalize y .
Tauto.

Apply H4.
Tauto.

Apply H5.
Assumption.

Elim H1.
Intros.
Elim H3.
Intros.
Apply H4.

Elim H0.
Intros.
Elim H2.
Intros.
Apply H3.

Intro.
Cut ((p1:name->proc)
      (btrans p (bOut x) p1)
       ->(Ex
           [q1:name->proc]
            (btrans q (bOut x) q1)
            /\((y:name)
                 (notin y (nu p1))
                  ->(notin y (nu q1))->(StBisim (p1 y) (q1 y)))))
    /\((q1:name->proc)
        (btrans q (bOut x) q1)
         ->(Ex
             [p1:name->proc]
              (btrans p (bOut x) p1)
              /\((y:name)
                 (notin y (nu p1))
                  ->(notin y (nu q1))->(StBisim (p1 y) (q1 y))))).
Intro.
Cut ((p1:name->proc)
      (btrans q (bOut x) p1)
       ->(Ex
           [q1:name->proc]
            (btrans r (bOut x) q1)
            /\((y:name)
                 (notin y (nu p1))
                  ->(notin y (nu q1))->(StBisim (p1 y) (q1 y)))))
    /\((q1:name->proc)
        (btrans r (bOut x) q1)
         ->(Ex
             [p1:name->proc]
              (btrans q (bOut x) p1)
              /\((y:name)
                 (notin y (nu p1))
                  ->(notin y (nu q1))->(StBisim (p1 y) (q1 y))))).
Intro.
Inversion_clear H.
Inversion_clear H2.
Split.
Intros.
Cut (Ex
      [q1:name->proc]
       (btrans q (bOut x) q1)
       /\((y:name)
                 (notin y (nu p1))
                  ->(notin y (nu q1))->(StBisim (p1 y) (q1 y)))).
Intro.
Inversion_clear H6.
Cut (Ex
      [q1:name->proc]
       (btrans r (bOut x) q1)
       /\((y:name)
                 (notin y (nu x0))
                  ->(notin y (nu q1))->(StBisim (x0 y) (q1 y)))).
Intro.
Inversion_clear H6.
Exists x1.
Split.
Tauto.

Intros.
(Elim (ho_proc_exp x0 y); Intros).
Inversion_clear H10.
Elim (unsat (par (nu p1) (par (nu x0) (par (nu x1) (nu [w:name](nu (x2 w)))))) (cons y empty)); Intros.
Inversion_clear H10.
Inversion_clear H13.
Inversion_clear H15.
Inversion_clear H16.
Inversion_clear H14.
Clear H18.

Apply TRANS with (x2 x3 y).
Elim (unsat (par (nu p1) (par (nu (x2 x3)) (nu x0))) (cons x3 (cons y empty))); Intros.
Inversion_clear H14.
Inversion_clear H18.
Inversion_clear H20.
Inversion_clear H19.
Inversion_clear H22.
Clear H23.
Apply L6_Light with x4.

Assumption.Assumption.

Change (StBisim ([z:name](p1 x4) x3) ([z:name](x2 z x4) x3)).
Apply L6_Light with y.

Apply notin_nu.
Intros.
Inversion H6.
Apply H24.
Auto.

Apply notin_nu.
Intros.
Inversion H11.
(Cut (notin y (nu (x2 z))); [ Intro | Auto ]).
Inversion H25.
(Cut (notin y (x2 z x4)); [ Intro | Auto ]).
Assumption.

Inversion_clear H7.
(Rewrite <- H12; Apply H23).
Assumption.
Assumption.

Apply notin_nu.
Intros.
Inversion H10.
(Apply H24; Auto).

Apply notin_nu.
Intros.
Inversion H17.
(Cut (notin x3 (nu (x2 z))); [ Intro | Auto ]).
Inversion H25.
(Cut (notin x3 (x2 z x4)); [ Intro | Auto ]).
Assumption.

Assumption.

Apply notin_nu.Intros.
Inversion H11.
Cut (notin y (nu (x2 x3))); [Intro | Auto].
Inversion H25.
Cut (notin y ((x2 x3) z)); [Intro | Auto].
Assumption.

Elim (unsat (par (nu x1) (par (nu (x2 x3)) (nu x0))) (cons x3 (cons y empty))); Intros.
Inversion_clear H14.
Inversion_clear H18.
Inversion_clear H20.
Inversion_clear H19.
Inversion_clear H22.
Clear H23.
Apply L6_Light with x4.

Assumption.Assumption.

Change (StBisim ([z:name](x2 z x4) x3) ([z:name](x1 x4) x3)).
Apply L6_Light with y.

Apply notin_nu.
Intros.
Inversion H11.
(Cut (notin y (nu (x2 z))); [ Intro | Auto ]).
Inversion H25.
(Cut (notin y (x2 z x4)); [ Intro | Auto ]).
Assumption.

Apply notin_nu.
Intros.
Inversion H9.
Apply H24.
Auto.

Inversion_clear H8.
(Rewrite <- H12; Apply H23).
Assumption.
Assumption.

Apply notin_nu.
Intros.
Inversion H17.
(Cut (notin x3 (nu (x2 z))); [ Intro | Auto ]).
Inversion H25.
(Cut (notin x3 (x2 z x4)); [ Intro | Auto ]).
Assumption.

Apply notin_nu.
Intros.
Inversion H15.
(Apply H24; Auto).

Apply notin_nu.Intros.
Inversion H11.
Cut (notin y (nu (x2 x3))); [Intro | Auto].
Inversion H25.
Cut (notin y ((x2 x3) z)); [Intro | Auto].
Assumption.

Assumption.

Apply H.Tauto.
Apply H3.Tauto.

Intros.
Cut (Ex
      [p1:name->proc]
       (btrans q (bOut x) p1)
       /\((y:name)
                 (notin y (nu p1))
                  ->(notin y (nu q1))->(StBisim (p1 y) (q1 y)))).
Intro.
Inversion_clear H6.
Cut (Ex
      [p1:name->proc]
       (btrans p (bOut x) p1)
       /\((y:name)
                 (notin y (nu p1))
                  ->(notin y (nu x0))->(StBisim (p1 y) (x0 y)))).
Intro.
Inversion_clear H6.
Exists  x1.
Split.
Tauto.

Intros.
(Elim (ho_proc_exp x0 y); Intros).
Inversion_clear H10.
Elim (unsat (par (nu x1) (par (nu x0) (par (nu q1) (nu [w:name](nu (x2 w)))))) (cons y empty)); Intros.
Inversion_clear H10.
Inversion_clear H13.
Inversion_clear H15.
Inversion_clear H16.
Inversion_clear H14.
Clear H18.
Apply TRANS with (x2 x3 y).
Elim (unsat (par (nu x1) (par (nu (x2 x3)) (nu x0))) (cons x3 (cons y empty))); Intros.
Inversion_clear H14.
Inversion_clear H18.
Inversion_clear H20.
Inversion_clear H19.
Inversion_clear H22.
Clear H23.
Apply L6_Light with x4.

Assumption.Assumption.

Change (StBisim ([z:name](x1 x4) x3) ([z:name](x2 z x4) x3)).
Apply L6_Light with y.

Apply notin_nu.
Intros.
Inversion H6.
Apply H24.
Auto.

Apply notin_nu.
Intros.
Inversion H11.
(Cut (notin y (nu (x2 z))); [ Intro | Auto ]).
Inversion H25.
(Cut (notin y (x2 z x4)); [ Intro | Auto ]).
Assumption.

Inversion_clear H8.
(Rewrite <- H12; Apply H23).
Assumption.

Assumption.

Apply notin_nu.
Intros.
Inversion H10.
(Apply H24; Auto).

Apply notin_nu.
Intros.
Inversion H17.
(Cut (notin x3 (nu (x2 z))); [ Intro | Auto ]).
Inversion H25.
(Cut (notin x3 (x2 z x4)); [ Intro | Auto ]).
Assumption.

Assumption.

Apply notin_nu.Intros.
Inversion H11.
Cut (notin y (nu (x2 x3))); [Intro | Auto].
Inversion H25.
Cut (notin y ((x2 x3) z)); [Intro | Auto].
Assumption.

Elim (unsat (par (nu q1) (par (nu (x2 x3)) (nu x0))) (cons x3 (cons y empty))); Intros.
Inversion_clear H14.
Inversion_clear H18.
Inversion_clear H20.
Inversion_clear H19.
Inversion_clear H22.
Clear H23.
Apply L6_Light with x4.

Assumption.Assumption.

Change (StBisim ([z:name](x2 z x4) x3) ([z:name](q1 x4) x3)).
Apply L6_Light with y.

Apply notin_nu.
Intros.
Inversion H11.
(Cut (notin y (nu (x2 z))); [ Intro | Auto ]).
Inversion H25.
(Cut (notin y (x2 z x4)); [ Intro | Auto ]).
Assumption.

Apply notin_nu.
Intros.
Inversion H9.
Apply H24.
Auto.

Inversion_clear H7.
(Rewrite <- H12; Apply H23).
Assumption.

Assumption.

Apply notin_nu.
Intros.
Inversion H17.
(Cut (notin x3 (nu (x2 z))); [ Intro | Auto ]).
Inversion H25.
(Cut (notin x3 (x2 z x4)); [ Intro | Auto ]).
Assumption.

Apply notin_nu.
Intros.
Inversion H15.
(Apply H24; Auto).

Apply notin_nu.Intros.
Inversion H11.
Cut (notin y (nu (x2 x3))); [Intro | Auto].
Inversion H25.
Cut (notin y ((x2 x3) z)); [Intro | Auto].
Assumption.

Assumption.

Apply H4.Tauto.
Apply H5.Tauto.

Elim H1.
Intros.
Elim H3.
Intros.
Apply H5.

Elim H0.
Intros.
Elim H2.
Intros.
Apply H4.

Qed.

End equivalence.

Section structural_congruence1.

Lemma TAU_S : (p,q:proc)(StBisim p q)->(StBisim (tau_pref p) (tau_pref q)).

Proof. (* This lemma is proved from scratch. *)

Intros; Apply sb; Do 3 Try (Split; Intros).
Inversion H0; Exists  q; Split; 
   [ Apply TAU | Rewrite <- H3]; Assumption.
Inversion H0; Exists  p; Split;
   [ Apply TAU | Rewrite <- H3 ]; Assumption.
Inversion_clear H0.
Inversion_clear H0.
Inversion_clear H0.
Inversion_clear H0.

Qed.

Lemma SUM_S: (p,q,r:proc)(StBisim p q)->(StBisim (sum p r) (sum q r)).

Proof. (* This lemma depends on Lemma REF. *)

(Intros; Apply sb; Do 3 Try (Split; Intros)).

Inversion H0.
Inversion H.
Inversion_clear H6.
(Elim (H9 a); Intros).
(Elim (H6 p1 H5); Intros).
(Exists  x; Split; [ Apply fSUM1; Tauto | Tauto ]).
(Exists p1; Split; [Apply fSUM2; Tauto | Apply REF]).

Inversion H0.
Inversion H.
Inversion_clear H6.
(Elim (H9 a); Intros).
(Elim (H11 q1 H5); Intros).
(Exists  x; Split; [ Apply fSUM1; Tauto | Tauto ]).
(Exists q1; Split; [Apply fSUM2; Tauto | Apply REF]).

Inversion H0.
Inversion H.
Inversion_clear H6.
Inversion_clear H10.
(Elim (H6 x); Intros).
(Elim (H10 p1 H5); Intros).
(Exists x0; Split; [ Apply bSUM1; Tauto | Tauto ]).
(Exists p1; Split; [Apply bSUM2; Tauto | Intro; Apply REF]).

Inversion H0.
Inversion H.
Inversion_clear H6.
Inversion_clear H10.
(Elim (H6 x); Intros).
(Elim (H12 q1 H5); Intros).
(Exists x0; Split; [ Apply bSUM1; Tauto | Tauto ]).
(Exists q1; Split; [Apply bSUM2; Tauto | Intro; Apply REF]).

Inversion H0.
Inversion H.
Inversion_clear H6.
Inversion_clear H10.
(Elim (H11 x); Intros).
(Elim (H10 p1 H5); Intros).
(Exists x0; Split; [ Apply bSUM1; Tauto | Tauto ]).
(Exists p1; Split; [Apply bSUM2; Tauto | Intros; Apply REF]).

Inversion H0.
Inversion H.
Inversion_clear H6.
Inversion_clear H10.
(Elim (H11 x); Intros).
(Elim (H12 q1 H5); Intros).
(Exists x0; Split; [ Apply bSUM1; Tauto | Tauto ]).
(Exists q1; Split; [Apply bSUM2; Tauto | Intros; Apply REF]).

Qed.

Lemma NU_S: (p,q:name->proc)
                  ((x:name)(notin x (nu p))->(notin x (nu q))
                    ->(StBisim (p x) (q x)))
                      ->(StBisim (nu p) (nu q)).

Proof. (* This lemma depends on Lemmata FTR_L3, BTR_L3 and L6_Light:
        * it is a coinductive proof.
        *)

(Cofix; Intros; Apply sb; Do 3 Try (Split; Intros)).
Inversion H0.
Generalize H2.
Generalize H0.
Elim a.
Intros.
Elim (unsat (par (nu p) (par (nu q) (nu p2))) l); Intros.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H10.
Cut (StBisim (p x) (q x)); [Intro | Auto].
Inversion_clear H10.
Inversion_clear H12.
Elim (H10 tau); Intros.
Cut (ftrans (p x) tau (p2 x)); [Intro | Apply H6; Auto].
Elim (H12 (p2 x) H15); Intros.
Inversion_clear H16.
Elim (proc_exp x0 x); Intros.
Inversion_clear H16.
Rewrite H20 in H17.
Rewrite H20 in H18.
Exists (nu x1); Split.

Apply fRES with empty; Intros; Auto.
Clear H22 H23.
(Change (ftrans (q y) ([_:name]tau y) (x1 y)); Apply FTR_L3 with x; Auto).
(Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau).
(Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau).
(Apply NU_S; Intros).
(Apply L6_Light with x; Auto).
Apply f_act_notin_tau.

Intros.
Elim (unsat (par (nu p) (par (nu q) (nu p2))) (cons n (cons n0 l))); Intros.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H10.
Inversion_clear H9.
Inversion_clear H12.
Cut (StBisim (p x) (q x)); [Intro | Auto].
Inversion_clear H12.
Inversion_clear H14.
Elim (H12 (Out n n0)); Intros.
Cut (ftrans (p x) (Out n n0) (p2 x)); [Intro | Apply H6; Auto].
Elim (H14 (p2 x) H17); Intros.
Inversion_clear H18.
Elim (proc_exp x0 x); Intros.
Inversion_clear H18.
Rewrite H22 in H19.
Rewrite H22 in H20.
Exists (nu x1); Split.
Apply fRES with empty; Intros; Auto.
(Change (ftrans (q y) ([_:name](Out n n0) y) (x1 y)); Apply FTR_L3 with x; Auto).
Clear H24.
(Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto).
Inversion_clear H25.
(Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto).
(Apply NU_S; Intros).
(Apply L6_Light with x; Auto).

Apply f_act_notin_Out; Auto.

Inversion H0.
Generalize H2.
Generalize H0.
Elim a.
Intros.
Elim (unsat (par (nu p) (par (nu q) (nu p2))) l); Intros.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H10.
Cut (StBisim (p x) (q x)); [Intro | Auto].
Inversion_clear H10.
Inversion_clear H12.
Elim (H10 tau); Intros.
Cut (ftrans (q x) tau (p2 x)); [Intro | Apply H6; Auto].
Elim (H14 (p2 x) H15); Intros.
Inversion_clear H16.
Elim (proc_exp x0 x); Intros.
Inversion_clear H16.
Rewrite H20 in H17.
Rewrite H20 in H18.
Exists (nu x1); Split.
Apply fRES with empty; Intros; Auto.
(Change (ftrans (p y) ([_:name]tau y) (x1 y)); Apply FTR_L3 with x; Auto).
(Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau).
(Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau).
(Apply NU_S; Intros).
(Apply L6_Light with x; Auto).
Apply f_act_notin_tau.

Intros.
Elim (unsat (par (nu p) (par (nu q) (nu p2))) (cons n (cons n0 l))); Intros.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H10.
Inversion_clear H9.
Inversion_clear H12.
Cut (StBisim (p x) (q x)); [Intro | Auto].
Inversion_clear H12.
Inversion_clear H14.
Elim (H12 (Out n n0)); Intros.
Cut (ftrans (q x) (Out n n0) (p2 x)); [Intro | Apply H6; Auto].
Elim (H16 (p2 x) H17); Intros.
Inversion_clear H18.
Elim (proc_exp x0 x); Intros.
Inversion_clear H18.
Rewrite H22 in H19.
Rewrite H22 in H20.
Exists (nu x1); Split.
Apply fRES with empty; Intros; Auto.
Clear H24.
(Change (ftrans (p y) ([_:name](Out n n0) y) (x1 y)); Apply FTR_L3 with x; Auto).
(Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto).
Inversion_clear H25.
(Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto).
(Apply NU_S; Intros).
(Apply L6_Light with x; Auto).
Apply f_act_notin_Out; Auto.

Inversion H0.
Elim (unsat (par (nu p) (par (nu q) (nu [z:name](nu (p2 z))))) (cons x l)); Intros.
Inversion_clear H5.
Inversion_clear H6.
Inversion_clear H8.
Inversion_clear H7.
Cut (StBisim (p x0) (q x0)); [Intro | Auto].
Inversion_clear H7.
Inversion_clear H11.
Inversion_clear H12.
Elim (H11 x); Intros.
Cut (btrans (p x0) (In x) (p2 x0)); [Intro | Apply H2; Auto].
Elim (H12 (p2 x0) H15); Intros.
Inversion_clear H16.
Elim (ho_proc_exp x1 x0); Intros.
Inversion_clear H16.
Rewrite H20 in H17.
Rewrite H20 in H18.
Exists [_:name](nu [z:name](x2 z _)); Split.
Apply bRES with empty; Intros; Auto.
Clear H23.
Inversion_clear H22.
(Change (btrans (q y) ([_:name](In x) y) (x2 y)); Apply BTR_L3 with x0; Auto).
(Unfold b_act_notin_ho; Intros; Apply b_act_notin_In).
Assumption.
(Unfold b_act_notin_ho; Intros; Apply b_act_notin_In).
Assumption.
Intro.
(Apply NU_S; Intros).
Elim (LEM_name x0 x3); Intros.
Rewrite <- H22; Auto.
Elim (LEM_name x0 y); Intros.
Rewrite <- H23; Auto.
Rewrite <- H23 in H16.
Rewrite <- H23 in H21.
Elim (unsat (par (nu [z:name](nu (p2 z))) (nu [z:name](nu (x2 z)))) (cons x0 (cons x3 empty))); Intros.
Inversion_clear H24.
Inversion_clear H25.
Inversion_clear H26.
Inversion_clear H28.
Clear H29.
Change (StBisim ([_:name](p2 _ x0) x3) ([_:name](x2 _ x0) x3)).
(Apply L6_Light with x4; Auto).
Apply notin_nu; Intros.
Inversion_clear H24.
Cut (notin x4 (nu (p2 z))); [Intro | Auto].
Inversion_clear H24.
Auto.
Apply notin_nu; Intros.
Inversion_clear H27.
Cut (notin x4 (nu (x2 z))); [Intro | Auto].
Inversion_clear H27.
Auto.
Elim (unsat (par (nu [z:name](nu (p2 z))) (nu [z:name](nu (x2 z)))) (cons x0 (cons x4 empty))); Intros.
Inversion_clear H28.
Inversion_clear H29.
Inversion_clear H30.
Inversion_clear H32.
Clear H33.
(Apply L6_Light with x5; Auto).
Apply notin_nu; Intros.
Inversion_clear H28.
Cut (notin x5 (nu (p2 x4))); [Intro | Auto].
Inversion_clear H28.
Auto.
Apply notin_nu; Intros.
Inversion_clear H31.
Cut (notin x5 (nu (x2 x4))); [Intro | Auto].
Inversion_clear H31.
Auto.
Change (StBisim ([_:name](p2 _ x5) x4) ([_:name](x2 _ x5) x4)).
(Apply L6_Light with x0; Auto).
Apply notin_nu; Intros.
Inversion_clear H9.
Cut (notin x0 (nu (p2 z))); [Intro | Auto].
Inversion_clear H9.
Auto.
Apply notin_nu; Intros.
Inversion_clear H19.
Cut (notin x0 (nu (x2 z))); [Intro | Auto].
Inversion_clear H19.
Auto.
Apply notin_nu; Intros.
Inversion_clear H24.
Cut (notin x4 (nu (p2 z))); [Intro | Auto].
Inversion_clear H24.
Auto.
Apply notin_nu; Intros.
Inversion_clear H27.
Cut (notin x4 (nu (x2 z))); [Intro | Auto].
Inversion_clear H27.
Auto.
Apply notin_nu; Intros.
Inversion_clear H9.
Cut (notin x0 (nu (p2 x4))); [Intro | Auto].
Inversion_clear H9.
Auto.
Apply notin_nu; Intros.
Inversion_clear H19.
Cut (notin x0 (nu (x2 x4))); [Intro | Auto].
Inversion_clear H19.
Auto.
Change (StBisim ([_:name](p2 _ y) x3) ([_:name](x2 _ y) x3)).
(Apply L6_Light with x0; Auto).
Apply notin_nu; Intros.
Inversion_clear H9.
Cut (notin x0 (nu (p2 z))); [Intro | Auto].
Inversion_clear H9.
Auto.
Apply notin_nu; Intros.
Inversion_clear H19.
Cut (notin x0 (nu (x2 z))); [Intro | Auto].
Inversion_clear H19.
Auto.
Apply b_act_notin_In; Auto.

Inversion H0.
Elim (unsat (par (nu p) (par (nu q) (nu [z:name](nu (p2 z))))) (cons x l)); Intros.
Inversion_clear H5.
Inversion_clear H6.
Inversion_clear H8.
Inversion_clear H7.
Cut (StBisim (p x0) (q x0)); [Intro | Auto].
Inversion_clear H7.
Inversion_clear H11.
Inversion_clear H12.
Elim (H11 x); Intros.
Cut (btrans (q x0) (In x) (p2 x0)); [Intro | Apply H2; Auto].
Elim (H14 (p2 x0) H15); Intros.
Inversion_clear H16.
Elim (ho_proc_exp x1 x0); Intros.
Inversion_clear H16.
Rewrite H20 in H17.
Rewrite H20 in H18.
Exists [_:name](nu [z:name](x2 z _)); Split.
Apply bRES with empty; Intros; Auto.
Clear H23.
Inversion_clear H22.
(Change (btrans (p y) ([_:name](In x) y) (x2 y)); Apply BTR_L3 with x0; Auto).
(Unfold b_act_notin_ho; Intros; Apply b_act_notin_In).
Assumption.
(Unfold b_act_notin_ho; Intros; Apply b_act_notin_In).
Assumption.
Intro.
(Apply NU_S; Intros).
Elim (LEM_name x0 x3); Intros.
Rewrite <- H22; Auto.
Elim (LEM_name x0 y); Intros.
Rewrite <- H23; Auto.
Rewrite <- H23 in H16.
Rewrite <- H23 in H21.

Elim (unsat (par (nu [z:name](nu (x2 z))) (nu [z:name](nu (p2 z)))) (cons x0 (cons x3 l))); Intros.
Inversion_clear H24.
Inversion_clear H25.
Inversion_clear H26.
Inversion_clear H28.
Change (StBisim ([_:name](x2 _ x0) x3) ([_:name](p2 _ x0) x3)).
(Apply L6_Light with x4; Auto).
Apply notin_nu; Intros.
Inversion_clear H24.
Cut (notin x4 (nu (x2 z))); [Intro | Auto].
Inversion_clear H24.
Auto.
Apply notin_nu; Intros.
Inversion_clear H27.
Cut (notin x4 (nu (p2 z))); [Intro | Auto].
Inversion_clear H27.
Auto.
Elim (unsat (par (nu [z:name](nu (x2 z))) (nu [z:name](nu (p2 z)))) (cons x0 (cons x4 l))); Intros.
Inversion_clear H28.
Inversion_clear H30.
Inversion_clear H31.
Inversion_clear H33.
(Apply L6_Light with x5; Auto).
Apply notin_nu; Intros.
Inversion_clear H28.
Cut (notin x5 (nu (x2 x4))); [Intro | Auto].
Inversion_clear H28.
Auto.
Apply notin_nu; Intros.
Inversion_clear H32.
Cut (notin x5 (nu (p2 x4))); [Intro | Auto].
Inversion_clear H32.
Auto.
Change (StBisim ([_:name](x2 _ x5) x4) ([_:name](p2 _ x5) x4)).
(Apply L6_Light with x0; Auto).
Apply notin_nu; Intros.
Inversion_clear H19.
Cut (notin x0 (nu (x2 z))); [Intro | Auto].
Inversion_clear H19.
Auto.
Apply notin_nu; Intros.
Inversion_clear H9.
Cut (notin x0 (nu (p2 z))); [Intro | Auto].
Inversion_clear H9.
Auto.
Apply notin_nu; Intros.
Inversion_clear H24.
Cut (notin x4 (nu (x2 z))); [Intro | Auto].
Inversion_clear H24.
Auto.
Apply notin_nu; Intros.
Inversion_clear H27.
Cut (notin x4 (nu (p2 z))); [Intro | Auto].
Inversion_clear H27.
Auto.
Apply notin_nu; Intros.
Inversion_clear H19.
Cut (notin x0 (nu (x2 x4))); [Intro | Auto].
Inversion_clear H19.
Auto.
Apply notin_nu; Intros.
Inversion_clear H9.
Cut (notin x0 (nu (p2 x4))); [Intro | Auto].
Inversion_clear H9.
Auto.
Change (StBisim ([_:name](x2 _ y) x3) ([_:name](p2 _ y) x3)).
(Apply L6_Light with x0; Auto).
Apply notin_nu; Intros.
Inversion_clear H19.
Cut (notin x0 (nu (x2 z))); [Intro | Auto].
Inversion_clear H19.
Auto.
Apply notin_nu; Intros.
Inversion_clear H9.
Cut (notin x0 (nu (p2 z))); [Intro | Auto].
Inversion_clear H9.
Auto.
Apply b_act_notin_In; Auto.

Inversion H0.
Elim (unsat (par (nu p) (par (nu q) (nu [z:name](nu (p2 z))))) (cons x l)); Intros.
Inversion_clear H5.
Inversion_clear H6.
Inversion_clear H8.
Inversion_clear H7.
Cut (StBisim (p x0) (q x0)); [Intro | Auto].
Inversion_clear H7.
Inversion_clear H11.
Inversion_clear H12.
Elim (H13 x); Intros.
Cut (btrans (p x0) (bOut x) (p2 x0)); [Intro | Apply H2; Auto].
Elim (H12 (p2 x0) H15); Intros.
Inversion_clear H16.
Elim (ho_proc_exp x1 x0); Intros.
Inversion_clear H16.
Rewrite H20 in H17.
Rewrite H20 in H18.
Exists [_:name](nu [z:name](x2 z _)); Split.
Apply bRES with empty; Intros; Auto.
Clear H23.
Inversion_clear H22.
(Change (btrans (q y) ([_:name](bOut x) y) (x2 y)); Apply BTR_L3 with x0; Auto).
(Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut).
Assumption.
(Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut).
Assumption.
Intros.
(Apply NU_S; Intros).
Elim (unsat (par (nu [z:name](nu (p2 z))) (nu [z:name](nu (x2 z)))) (cons x0 (cons x3 (cons y l)))); Intros.
Inversion_clear H24.
Inversion_clear H25.
Inversion_clear H26.
Inversion_clear H28.
Inversion_clear H29.
Change (StBisim ([_:name](p2 _ y) x3) ([_:name](x2 _ y) x3)).
(Apply L6_Light with x4; Auto).
Apply notin_nu; Intros.
Inversion_clear H24.
Cut (notin x4 (nu (p2 z))); [Intro | Auto].
Inversion_clear H24.
Auto.
Apply notin_nu; Intros.
Inversion_clear H27.
Cut (notin x4 (nu (x2 z))); [Intro | Auto].
Inversion_clear H27.
Auto.
Elim (unsat (par (nu [z:name](nu (p2 z))) (nu [z:name](nu (x2 z)))) (cons x0 (cons x4 (cons y l)))); Intros.
Inversion_clear H29.
Inversion_clear H31.
Inversion_clear H32.
Inversion_clear H34.
Inversion_clear H35.
(Apply L6_Light with x5; Auto).
Apply notin_nu; Intros.
Inversion_clear H29.
Cut (notin x5 (nu (p2 x4))); [Intro | Auto].
Inversion_clear H29.
Auto.
Apply notin_nu; Intros.
Inversion_clear H33.
Cut (notin x5 (nu (x2 x4))); [Intro | Auto].
Inversion_clear H33.
Auto.
Change (StBisim ([_:name](p2 _ x5) x4) ([_:name](x2 _ x5) x4)).
(Apply L6_Light with x0; Auto).
Apply notin_nu; Intros.
Inversion_clear H9.
Cut (notin x0 (nu (p2 z))); [Intro | Auto].
Inversion_clear H9.
Auto.
Apply notin_nu; Intros.
Inversion_clear H19.
Cut (notin x0 (nu (x2 z))); [Intro | Auto].
Inversion_clear H19.
Auto.
Apply H18.
Apply notin_nu; Intros.
Inversion_clear H29.
Cut (notin x5 (nu (p2 x0))); [Intro | Auto].
Inversion_clear H29.
Auto.
Apply notin_nu; Intros.
Inversion_clear H33.
Cut (notin x5 (nu (x2 x0))); [Intro | Auto].
Inversion_clear H33.
Auto.
Apply notin_nu; Intros.
Inversion_clear H24.
Cut (notin x4 (nu (p2 z))); [Intro | Auto].
Inversion_clear H24.
Auto.
Apply notin_nu; Intros.
Inversion_clear H27.
Cut (notin x4 (nu (x2 z))); [Intro | Auto].
Inversion_clear H27.
Auto.
Apply notin_nu; Intros.
Inversion_clear H16.
Cut (notin y (nu [z0:name](p2 z0 z))); [Intro | Auto].
Inversion_clear H16.
Auto.
Apply notin_nu; Intros.
Inversion_clear H21.
Cut (notin y (nu [z0:name](x2 z0 z))); [Intro | Auto].
Inversion_clear H21.
Auto.
Apply b_act_notin_bOut; Assumption.

Elim (unsat (par (nu p) (par (nu q) (nu p1))) (cons x l)); Intros.
Inversion_clear H5.
Inversion_clear H6.
Inversion_clear H8.
Inversion_clear H7.
Cut (StBisim (p x1) (q x1)); [Intro | Auto].
Inversion_clear H7.
Inversion_clear H11.
Elim (H7 (Out x x1)); Intros.
Cut (ftrans (p x1) (Out x x1) (p1 x1)); [Intro | Apply H3; Auto].
Elim (H11 (p1 x1) H14); Intros.
Inversion_clear H15.
Elim (proc_exp x2 x1); Intros.
Inversion_clear H15.
Rewrite H19 in H16.
Rewrite H19 in H17.
Exists x3; Split.
Apply OPEN with empty; Intros; Auto.
Clear H22.
(Change (ftrans (q y) ([_:name](Out x _) y) (x3 y)); Apply FTR_L3 with x1; Auto).
(Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out).
Assumption.
Assumption.
(Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out).
Auto.
Assumption.
Intros.
(Apply L6_Light with x1; Auto).

Inversion H0.
Elim (unsat (par (nu p) (par (nu q) (nu [z:name](nu (p2 z))))) (cons x l)); Intros.
Inversion_clear H5.
Inversion_clear H6.
Inversion_clear H8.
Inversion_clear H7.
Cut (StBisim (p x0) (q x0)); [Intro | Auto].
Inversion_clear H7.
Inversion_clear H11.
Inversion_clear H12.
Elim (H13 x); Intros.
Cut (btrans (q x0) (bOut x) (p2 x0)); [Intro | Apply H2; Auto].
Elim (H14 (p2 x0) H15); Intros.
Inversion_clear H16.
Elim (ho_proc_exp x1 x0); Intros.
Inversion_clear H16.
Rewrite H20 in H17.
Rewrite H20 in H18.
Exists [_:name](nu [z:name](x2 z _)); Split.
Apply bRES with empty; Intros; Auto.
Clear H23.
Inversion_clear H22.
(Change (btrans (p y) ([_:name](bOut x) y) (x2 y)); Apply BTR_L3 with x0; Auto).
(Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut).
Assumption.
(Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut).
Assumption.
Intros.
(Apply NU_S; Intros).
Elim (unsat (par (nu [z:name](nu (p2 z))) (nu [z:name](nu (x2 z)))) (cons x0 (cons x3 (cons y l)))); Intros.
Inversion_clear H24.
Inversion_clear H25.
Inversion_clear H26.
Inversion_clear H28.
Inversion_clear H29.
Change (StBisim ([_:name](x2 _ y) x3) ([_:name](p2 _ y) x3)).
(Apply L6_Light with x4; Auto).
Apply notin_nu; Intros.
Inversion_clear H27.
Cut (notin x4 (nu (x2 z))); [Intro | Auto].
Inversion_clear H27.
Auto.
Apply notin_nu; Intros.
Inversion_clear H24.
Cut (notin x4 (nu (p2 z))); [Intro | Auto].
Inversion_clear H24.
Auto.
Elim (unsat (par (nu [z:name](nu (p2 z))) (nu [z:name](nu (x2 z)))) (cons x0 (cons x4 (cons y l)))); Intros.
Inversion_clear H29.
Inversion_clear H31.
Inversion_clear H32.
Inversion_clear H34.
Inversion_clear H35.
(Apply L6_Light with x5; Auto).
Apply notin_nu; Intros.
Inversion_clear H33.
Cut (notin x5 (nu (x2 x4))); [Intro | Auto].
Inversion_clear H33.
Auto.
Apply notin_nu; Intros.
Inversion_clear H29.
Cut (notin x5 (nu (p2 x4))); [Intro | Auto].
Inversion_clear H29.
Auto.
Change (StBisim ([_:name](x2 _ x5) x4) ([_:name](p2 _ x5) x4)).
(Apply L6_Light with x0; Auto).
Apply notin_nu; Intros.
Inversion_clear H19.
Cut (notin x0 (nu (x2 z))); [Intro | Auto].
Inversion_clear H19.
Auto.
Apply notin_nu; Intros.
Inversion_clear H9.
Cut (notin x0 (nu (p2 z))); [Intro | Auto].
Inversion_clear H9.
Auto.
Apply H18.
Apply notin_nu; Intros.
Inversion_clear H33.
Cut (notin x5 (nu (x2 x0))); [Intro | Auto].
Inversion_clear H33.
Auto.
Apply notin_nu; Intros.
Inversion_clear H29.
Cut (notin x5 (nu (p2 x0))); [Intro | Auto].
Inversion_clear H29.
Auto.
Apply notin_nu; Intros.
Inversion_clear H27.
Cut (notin x4 (nu (x2 z))); [Intro | Auto].
Inversion_clear H27.
Auto.
Apply notin_nu; Intros.
Inversion_clear H24.
Cut (notin x4 (nu (p2 z))); [Intro | Auto].
Inversion_clear H24.
Auto.
Apply notin_nu; Intros.
Inversion_clear H16.
Cut (notin y (nu [z0:name](x2 z0 z))); [Intro | Auto].
Inversion_clear H16.
Auto.
Apply notin_nu; Intros.
Inversion_clear H21.
Cut (notin y (nu [z0:name](p2 z0 z))); [Intro | Auto].
Inversion_clear H21.
Auto.
Apply b_act_notin_bOut; Assumption.

Elim (unsat (par (nu p) (par (nu q) (nu q1))) (cons x l)); Intros.
Inversion_clear H5.
Inversion_clear H6.
Inversion_clear H8.
Inversion_clear H7.
Cut (StBisim (p x1) (q x1)); [Intro | Auto].
Inversion_clear H7.
Inversion_clear H11.
Elim (H7 (Out x x1)); Intros.
Cut (ftrans (q x1) (Out x x1) (q1 x1)); [Intro | Apply H3; Auto].
Elim (H13 (q1 x1) H14); Intros.
Inversion_clear H15.
Elim (proc_exp x2 x1); Intros.
Inversion_clear H15.
Rewrite H19 in H16.
Rewrite H19 in H17.
Exists x3; Split.
Apply OPEN with empty; Intros; Auto.
Clear H22.
(Change (ftrans (p y) ([_:name](Out x _) y) (x3 y)); Apply FTR_L3 with x1; Auto).
(Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out).
Assumption.
Assumption.
(Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out).
Auto.
Assumption.
Intros.
(Apply L6_Light with x1; Auto).

Qed.

(*****************************************************)
(* Congruence of StBisim w.r.t. parallel composition *)
(* is proved by means of internal cross adequacy.    *)
(*****************************************************)

(* The next four definitions allow to build the
 * strong late bisimulation needed to prove the
 * congruence of StBisim w.r.t. parallel composition.
 *)

Inductive BisPAR_S: proc -> proc -> Prop :=
  bispar_s: (p,q:proc)(StBisim p q) -> (r:proc)(BisPAR_S (par p r) (par q r)).

Inductive Op_PAR_S [R: proc -> proc -> Prop] : proc -> proc -> Prop :=
  op_par_s : (p,q:name->proc)((z:name)
             (notin z (nu p)) -> (notin z (nu q)) -> (R (p z) (q z))) ->
             ((Op_PAR_S R) (nu p) (nu q)).

Fixpoint PAR_S_fun [n:nat]: (proc->proc->Prop) :=
  Case n of
    BisPAR_S
    [m:nat](Op_PAR_S (PAR_S_fun m))
  end.

Inductive SBPAR_S [p,q:proc] : Prop :=
   sbpar_s : (n:nat)((PAR_S_fun n) p q) -> (SBPAR_S p q).

(* Next we establish a property of PAR_S_fun
 * similar to that of L6_Light for StBisim.
 *)

Lemma PAR_S_RW: (n:nat)(p,q:name->proc)(x:name)
                (notin x (nu p)) -> (notin x (nu q)) -> (PAR_S_fun n (p x) (q x)) ->
                (y:name)(notin y (nu p)) -> (notin y (nu q)) -> (PAR_S_fun n (p y) (q y)).

Proof.

Intros; Elim (LEM_name x y); Intro.

Rewrite <-H4; Assumption.

Generalize p q x H H0 y H4 H2 H3 H1; Clear H H0 H2 H3 H4 y H1 x p q; Induction n; Intros.

(* Base Case *)

Simpl in H1; Inversion H1; Simpl.
Elim (proc_exp p0 x); Elim (proc_exp q0 x); Elim (proc_exp r x); Intros.
Inversion_clear H8; Inversion_clear H9; Inversion_clear H10.
Rewrite H14 in H5; Rewrite H13 in H6; Rewrite H12 in H5; Rewrite H12 in H6; Cut p=([_:name](par (x2 _) (x0 _))); [Intro | Apply proc_ext with x; Auto].
Cut q=([_:name](par (x1 _) (x0 _))); [Intro | Apply proc_ext with x; Auto].
Rewrite H10; Rewrite H15; Apply bispar_s.
Rewrite H10 in H1; Rewrite H15 in H1; Inversion_clear H1.
Apply L6 with x; Auto.
Rewrite H10 in H2; Inversion_clear H2; Apply notin_nu; Intros.
Cut (notin y (par (x2 z) (x0 z))); [Intro | Auto].
Inversion_clear H17; Assumption.
Rewrite H15 in H3; Inversion_clear H3; Apply notin_nu; Intros.
Cut (notin y (par (x1 z) (x0 z))); [Intro | Auto].
Inversion_clear H17; Assumption.
Inversion_clear H8; Inversion_clear H11; Apply notin_nu; Intros; Apply notin_par; Auto.
Inversion_clear H9; Inversion_clear H11; Apply notin_nu; Intros; Apply notin_par; Auto.

(* Inductive Step *)

Simpl in H1; Inversion H1; Simpl.
Elim (ho_proc_exp p0 x); Elim (ho_proc_exp q0 x); Intros.
Inversion_clear H8; Inversion_clear H9.
Rewrite H12 in H5; Rewrite H11 in H6.
Cut p=([_:name](nu (x1 _))); [Intro | Apply proc_ext with x; Auto].
Cut q=([_:name](nu (x0 _))); [Intro | Apply proc_ext with x; Auto].
Rewrite H9; Rewrite H13; Apply op_par_s; Intros.
Elim (unsat (par (nu p) (par (nu q) (par (nu [_:name](nu (x0 _))) (nu [_:name](nu (x1 _)))))) (cons x (cons y (cons z empty)))); Intros.
Inversion_clear H16.
Inversion_clear H17; Inversion_clear H18.
Inversion_clear H19; Inversion_clear H20.
Inversion_clear H21; Inversion_clear H22.
Clear H24.
Apply Hrecn with x2; Auto.
Inversion_clear H23; Auto.
Inversion_clear H20; Auto.
Change (PAR_S_fun n ([_:name](x1 _ x2) y) ([_:name](x0 _ x2) y)); Apply Hrecn with x; Auto.
Inversion_clear H8; Apply notin_nu; Intros.
Cut (notin x (nu (x1 z0))); [Intro | Auto].
Inversion_clear H24; Auto.
Inversion_clear H10; Apply notin_nu; Intros.
Cut (notin x (nu (x0 z0))); [Intro | Auto].
Inversion_clear H24; Auto.
Apply proc_mono with x; Inversion_clear H2.
Cut (notin y (p x)); [Rewrite <-H5; Intro | Auto].
Inversion_clear H2; Auto.
Apply proc_mono with x; Inversion_clear H3.
Cut (notin y (q x)); [Rewrite <-H6; Intro | Auto].
Inversion_clear H3; Auto.
Rewrite <-H11; Rewrite <-H12; Apply H7; Auto.
Inversion_clear H23.
Cut (notin x2 (nu (x1 x))); [Intro | Auto]; Rewrite H12; Assumption.
Inversion_clear H20.
Cut (notin x2 (nu (x0 x))); [Intro | Auto]; Rewrite H11; Assumption.

Qed.

(* Now we prove that SBPAR_S is indeed a strong late bisimulation. *)

Lemma SBPAR_S_TO_SB': (Inclus SBPAR_S (Op_StBisim SBPAR_S)).

Proof.

Unfold Inclus; Intros.
Inversion_clear H.
Cut (PAR_S_fun n p1 p2);
 [Generalize p1 p2; Generalize n ; Induction n0; Intros | Assumption].

(* Base case *)

Simpl in H; Inversion H.
Inversion_clear H1.
Inversion_clear H4.
Inversion_clear H5.
Apply op_sb; Split; Intros.

Elim a; Intros; Split; Intros.

Inversion_clear H5.

Elim (H1 tau); Intros.
Elim (H5 p7 H7); Intros.
Inversion_clear H9.
Exists (par x r); Split; [Apply fPAR1; Assumption | Idtac].
Apply sbpar_s with O; Simpl; Apply bispar_s; Assumption.
Exists (par q p7); Split; [Apply fPAR2; Assumption | Idtac].
Apply sbpar_s with O; Simpl; Apply bispar_s; Rewrite <- H2 in H; Rewrite <- H3 in H; Inversion_clear H; Assumption.

Elim (H4 x); Intros.
Elim (H5 q1 H7); Intros.
Inversion_clear H10.
Exists (par (x0 y) q2); Split; [Apply COM1 with x; Assumption | Idtac].
Apply sbpar_s with O; Simpl; Apply bispar_s; Auto.
Elim (H1 (Out x y)); Intros.
Elim (H5 q1 H7); Intros.
Inversion_clear H10.
Exists (par x0 (q2 y)); Split; [Apply COM2 with x; Assumption | Idtac].
Apply sbpar_s with O; Simpl; Apply bispar_s; Assumption.

Elim (H4 x); Intros.
Elim (H5 q1 H7); Intros.
Inversion_clear H10.
Exists (nu [_:name](par (x0 _) (q2 _))); Split; [Apply CLOSE1 with x; Assumption | Idtac].
Apply sbpar_s with (S O); Simpl; Apply op_par_s; Intros; Apply bispar_s; Auto.

Elim (H6 x); Intros.
Elim (H5 q1 H7); Intros.
Inversion_clear H10.
Exists (nu [_:name](par (x0 _) (q2 _))); Split; [Apply CLOSE2 with x; Assumption | Idtac].
Apply sbpar_s with (S O); Simpl; Apply op_par_s; Intros; Apply bispar_s; Apply H12; Auto.
Inversion_clear H10; Apply notin_nu; Intros.
Cut (notin z (par (q1 z0) (q2 z0))); [Intro | Auto].
Inversion_clear H15; Assumption.
Inversion_clear H13; Apply notin_nu; Intros.
Cut (notin z (par (x0 z0) (q2 z0))); [Intro | Auto].
Inversion_clear H15; Assumption.

Inversion_clear H5.

Elim (H1 tau); Intros.
Elim (H8 p6 H7); Intros.
Inversion_clear H9.
Exists (par x r); Split; [Apply fPAR1; Assumption | Idtac].
Apply sbpar_s with O; Simpl; Apply bispar_s; Assumption.
Exists (par p p6); Split; [Apply fPAR2; Assumption | Idtac].
Apply sbpar_s with O; Simpl; Apply bispar_s; Rewrite <- H2 in H; Rewrite <- H3 in H; Inversion_clear H; Assumption.

Elim (H4 x); Intros.
Elim (H9 q0 H7); Intros.
Inversion_clear H10.
Exists (par (x0 y) q2); Split; [Apply COM1 with x; Assumption | Idtac].
Apply sbpar_s with O; Simpl; Apply bispar_s; Auto.
Elim (H1 (Out x y)); Intros.
Elim (H9 q0 H7); Intros.
Inversion_clear H10.
Exists (par x0 (q2 y)); Split; [Apply COM2 with x; Assumption | Idtac].
Apply sbpar_s with O; Simpl; Apply bispar_s; Assumption.

Elim (H4 x); Intros.
Elim (H9 q0 H7); Intros.
Inversion_clear H10.
Exists (nu [_:name](par (x0 _) (q2 _))); Split; [Apply CLOSE1 with x; Assumption | Idtac].
Apply sbpar_s with (S O); Simpl; Apply op_par_s; Intros; Apply bispar_s; Auto.

Elim (H6 x); Intros.
Elim (H9 q0 H7); Intros.
Inversion_clear H10.
Exists (nu [_:name](par (x0 _) (q2 _))); Split; [Apply CLOSE2 with x; Assumption | Idtac].
Apply sbpar_s with (S O); Simpl; Apply op_par_s; Intros; Apply bispar_s; Apply H12; Auto.
Inversion_clear H10; Apply notin_nu; Intros.
Cut (notin z (par (x0 z0) (q2 z0))); [Intro | Auto].
Inversion_clear H15; Assumption.
Inversion_clear H13; Apply notin_nu; Intros.
Cut (notin z (par (q0 z0) (q2 z0))); [Intro | Auto].
Inversion_clear H15; Assumption.

Inversion_clear H5.
Elim (H1 (Out n1 n2)); Intros.
Elim (H5 p7 H7); Intros.
Inversion_clear H9.
Exists (par x r); Split; [Apply fPAR1; Assumption | Idtac].
Apply sbpar_s with O; Simpl; Apply bispar_s; Assumption.
Exists (par q p7); Split; [Apply fPAR2; Assumption | Idtac].
Apply sbpar_s with O; Simpl; Apply bispar_s; Rewrite <- H2 in H; Rewrite <- H3 in H; Inversion_clear H; Assumption.

Inversion_clear H5.
Elim (H1 (Out n1 n2)); Intros.
Elim (H8 p6 H7); Intros.
Inversion_clear H9.
Exists (par x r); Split; [Apply fPAR1; Assumption | Idtac].
Apply sbpar_s with O; Simpl; Apply bispar_s; Assumption.
Exists (par p p6); Split; [Apply fPAR2; Assumption | Idtac].
Apply sbpar_s with O; Simpl; Apply bispar_s; Rewrite <- H2 in H; Rewrite <- H3 in H; Inversion_clear H; Assumption.

(* Input actions case *)

Split; Intros.

Elim (H4 x); Intros; Split; Intros.

Inversion_clear H8.
Elim (H5 p7 H9); Intros.
Inversion_clear H8.
Exists [_:name](par (x0 _) r); Split; [Apply bPAR1; Assumption | Idtac].
Intro; Apply sbpar_s with O; Simpl; Apply bispar_s; Auto.
Exists [_:name](par q (p7 _)); Split; [Apply bPAR2; Assumption | Idtac].
Intro; Apply sbpar_s with O; Simpl; Apply bispar_s; Rewrite <- H2 in H; Rewrite <- H3 in H; Inversion_clear H; Assumption.

Inversion_clear H8.
Elim (H7 p6 H9); Intros.
Inversion_clear H8.
Exists [_:name](par (x0 _) r); Split; [Apply bPAR1; Assumption | Idtac].
Intro; Apply sbpar_s with O; Simpl; Apply bispar_s; Auto.
Exists [_:name](par p (p6 _)); Split; [Apply bPAR2; Assumption | Idtac].
Intro; Apply sbpar_s with O; Simpl; Apply bispar_s; Rewrite <- H2 in H; Rewrite <- H3 in H; Inversion_clear H; Assumption.

(* Bound Output actions case *)

Elim (H6 x); Intros; Split; Intros.

Inversion_clear H8.
Elim (H5 p7 H9); Intros.
Inversion_clear H8.
Exists [_:name](par (x0 _) r); Split; [Apply bPAR1; Assumption | Idtac].
Intros; Apply sbpar_s with O; Simpl; Apply bispar_s; Apply H11; Auto.
Inversion_clear H8; Apply notin_nu; Intros.
Cut (notin y (par (p7 z) r)); [Intro | Auto].
Inversion_clear H14; Assumption.
Inversion_clear H12; Apply notin_nu; Intros.
Cut (notin y (par (x0 z) r)); [Intro | Auto].
Inversion_clear H14; Assumption.
Exists [_:name](par q (p7 _)); Split; [Apply bPAR2; Assumption | Idtac].
Intros; Apply sbpar_s with O; Simpl; Apply bispar_s; Rewrite <- H2 in H; Rewrite <- H3 in H; Inversion_clear H; Assumption.

Inversion_clear H8.
Elim (H7 p6 H9); Intros.
Inversion_clear H8.
Exists [_:name](par (x0 _) r); Split; [Apply bPAR1; Assumption | Idtac].
Intros; Apply sbpar_s with O; Simpl; Apply bispar_s; Apply H11; Auto.
Inversion_clear H8; Apply notin_nu; Intros.
Cut (notin y (par (x0 z) r)); [Intro | Auto].
Inversion_clear H14; Assumption.
Inversion_clear H12; Apply notin_nu; Intros.
Cut (notin y (par (p6 z) r)); [Intro | Auto].
Inversion_clear H14; Assumption.
Exists [_:name](par p (p6 _)); Split; [Apply bPAR2; Assumption | Idtac].
Intros; Apply sbpar_s with O; Simpl; Apply bispar_s; Rewrite <- H2 in H; Rewrite <- H3 in H; Inversion_clear H; Assumption.

(* Inductive Step *)

Simpl in H1; Inversion H1.
Apply op_sb; Split; Intros.

Elim a; Intros; Split; Intros.

Inversion_clear H5.
Cut (x:name)(notin x (nu p)) -> (notin x (nu q)) -> (Op_StBisim SBPAR_S (p x) (q x)); [Intro | Auto].
Elim (unsat (par (nu p) (par (nu q) (nu p6))) l); Intros.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H10.
Cut (ftrans (p x) tau (p6 x)); [Intro | Apply H6; Auto].
Cut (Op_StBisim SBPAR_S (p x) (q x)); [Intro | Apply H5; Auto].
Inversion_clear H12.
Inversion_clear H13.
Elim (H12 tau); Intros.
Elim (H13 (p6 x) H10); Intros.
Inversion_clear H16.
Elim (proc_exp x0 x); Intros.
Inversion_clear H16.
Rewrite H20 in H17; Rewrite H20 in H18.
Exists (nu x1); Split; [Apply fRES with l; Intros | Idtac].
Change (ftrans (q y) ([_:name]tau y) (x1 y)); Apply FTR_L3 with x; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Inversion_clear H18.
Apply sbpar_s with (S n2); Simpl; Apply op_par_s; Intros.
Apply PAR_S_RW with x; Auto.
Apply f_act_notin_tau.

Inversion_clear H5.
Cut (x:name)(notin x (nu p)) -> (notin x (nu q)) -> (Op_StBisim SBPAR_S (p x) (q x)); [Intro | Auto].
Elim (unsat (par (nu p) (par (nu q) (nu p5))) l); Intros.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H10.
Cut (ftrans (q x) tau (p5 x)); [Intro | Apply H6; Auto].
Cut (Op_StBisim SBPAR_S (p x) (q x)); [Intro | Apply H5; Auto].
Inversion_clear H12.
Inversion_clear H13.
Elim (H12 tau); Intros.
Elim (H15 (p5 x) H10); Intros.
Inversion_clear H16.
Elim (proc_exp x0 x); Intros.
Inversion_clear H16.
Rewrite H20 in H17; Rewrite H20 in H18.
Exists (nu x1); Split; [Apply fRES with l; Intros | Idtac].
Change (ftrans (p y) ([_:name]tau y) (x1 y)); Apply FTR_L3 with x; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Inversion_clear H18.
Apply sbpar_s with (S n2); Simpl; Apply op_par_s; Intros.
Apply PAR_S_RW with x; Auto.
Apply f_act_notin_tau.

Inversion_clear H5.
Cut (x:name)(notin x (nu p)) -> (notin x (nu q)) -> (Op_StBisim SBPAR_S (p x) (q x)); [Intro | Auto].
Elim (unsat (par (nu p) (par (nu q) (nu p6))) (cons n2 (cons n3 l))); Intros.
Inversion_clear H7.
Inversion_clear H8; Inversion_clear H9.
Inversion_clear H10; Inversion_clear H11.
Cut (ftrans (p x) (Out n2 n3) (p6 x)); [Intro | Apply H6; Auto].
Cut (Op_StBisim SBPAR_S (p x) (q x)); [Intro | Apply H5; Auto].
Inversion_clear H14.
Inversion_clear H15.
Elim (H14 (Out n2 n3)); Intros.
Elim (H15 (p6 x) H11); Intros.
Inversion_clear H18.
Elim (proc_exp x0 x); Intros.
Inversion_clear H18.
Rewrite H22 in H19; Rewrite H22 in H20.
Exists (nu x1); Split; [Apply fRES with l; Intros | Idtac].
Inversion_clear H25.
Change (ftrans (q y) ([_:name](Out n2 n3) y) (x1 y)); Apply FTR_L3 with x; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Inversion_clear H20.
Apply sbpar_s with (S n4); Simpl; Apply op_par_s; Intros.
Apply PAR_S_RW with x; Auto.
Apply f_act_notin_Out; Auto.

Inversion_clear H5.
Cut (x:name)(notin x (nu p)) -> (notin x (nu q)) -> (Op_StBisim SBPAR_S (p x) (q x)); [Intro | Auto].
Elim (unsat (par (nu p) (par (nu q) (nu p5))) (cons n2 (cons n3 l))); Intros.
Inversion_clear H7.
Inversion_clear H8; Inversion_clear H9.
Inversion_clear H10; Inversion_clear H11.
Cut (ftrans (q x) (Out n2 n3) (p5 x)); [Intro | Apply H6; Auto].
Cut (Op_StBisim SBPAR_S (p x) (q x)); [Intro | Apply H5; Auto].
Inversion_clear H14.
Inversion_clear H15.
Elim (H14 (Out n2 n3)); Intros.
Elim (H17 (p5 x) H11); Intros.
Inversion_clear H18.
Elim (proc_exp x0 x); Intros.
Inversion_clear H18.
Rewrite H22 in H19; Rewrite H22 in H20.
Exists (nu x1); Split; [Apply fRES with l; Intros | Idtac].
Inversion_clear H25.
Change (ftrans (p y) ([_:name](Out n2 n3) y) (x1 y)); Apply FTR_L3 with x; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Inversion_clear H20.
Apply sbpar_s with (S n4); Simpl; Apply op_par_s; Intros.
Apply PAR_S_RW with x; Auto.
Apply f_act_notin_Out; Auto.

Split; Intros.

Split; Intros.

Inversion_clear H5.
Cut (x:name)(notin x (nu p)) -> (notin x (nu q)) -> (Op_StBisim SBPAR_S (p x) (q x)); [Intro | Auto].
Elim (unsat (par (nu p) (par (nu q) (nu [_:name](nu (p6 _))))) (cons x l)); Intros.
Inversion_clear H7.
Inversion_clear H8; Inversion_clear H9.
Inversion_clear H10.
Cut (btrans (p x0) (In x) (p6 x0)); [Intro | Apply H6; Auto].
Cut (Op_StBisim SBPAR_S (p x0) (q x0)); [Intro | Apply H5; Auto].
Inversion_clear H13.
Inversion_clear H14.
Inversion_clear H15.
Elim (H14 x); Intros.
Elim (H15 (p6 x0) H10); Intros.
Inversion_clear H18.
Elim (ho_proc_exp x1 x0); Intros.
Inversion_clear H18.
Rewrite H22 in H19; Rewrite H22 in H20.
Exists [_:name](nu [u:name](x2 u _)); Split; [Apply bRES with l; Intros | Idtac].
Inversion_clear H24; Change (btrans (q y) ([_:name](In x) y) (x2 y)); Apply BTR_L3 with x0; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Intro; Elim (LEM_name x0 y); Intro.

Rewrite <-H18; Elim (unsat (par (nu [_:name](nu (p6 _))) (nu [_:name](nu (x2 _)))) (cons x0 empty)); Intros.
Inversion_clear H23.
Inversion_clear H24; Inversion_clear H25.
Clear H27; Cut (SBPAR_S (p6 x0 x3) (x2 x0 x3)); [Intro | Auto].
Inversion_clear H25.
Apply sbpar_s with (S n2); Simpl; Apply op_par_s; Intros.
Elim (unsat (par (nu [_:name](nu (p6 _))) (nu [_:name](nu (x2 _)))) (cons x0 (cons x3 (cons z empty)))); Intros.
Inversion_clear H29.
Inversion_clear H30; Inversion_clear H31.
Inversion_clear H33.
Inversion_clear H34.
Clear H35.
Cut (PAR_S_fun n2 (p6 x4 x3) (x2 x4 x3)); [Intro | Change (PAR_S_fun n2 ([_:name](p6 _ x3) x4) ([_:name](x2 _ x3) x4)); Apply PAR_S_RW with x0; Auto].
Change (PAR_S_fun n2 ([_:name](p6 _ x0) z) ([_:name](x2 _ x0) z)); Apply PAR_S_RW with x4; Auto.
Inversion_clear H29; Apply notin_nu; Intros.
Cut (notin x4 (nu (p6 z0))); [Intro | Auto].
Inversion_clear H36; Auto.
Inversion_clear H32; Apply notin_nu; Intros.
Cut (notin x4 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H36; Auto.
Apply PAR_S_RW with x3; Auto.
Inversion_clear H23; Auto.
Inversion_clear H26; Auto.
Inversion_clear H12; Auto.
Inversion_clear H21; Auto.
Inversion_clear H12; Apply notin_nu; Intros.
Cut (notin x0 (nu (p6 z0))); [Intro | Auto].
Inversion_clear H35; Auto.
Inversion_clear H21; Apply notin_nu; Intros.
Cut (notin x0 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H35; Auto.
Inversion_clear H29; Apply notin_nu; Intros.
Cut (notin x4 (nu (p6 z0))); [Intro | Auto].
Inversion_clear H35; Auto.
Inversion_clear H32; Apply notin_nu; Intros.
Cut (notin x4 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H35; Auto.

Cut (SBPAR_S (p6 x0 y) (x2 x0 y)); [Intro | Auto].
Inversion_clear H23.
Apply sbpar_s with (S n2); Simpl; Apply op_par_s; Intros.
Change (PAR_S_fun n2 ([_:name](p6 _ y) z) ([_:name](x2 _ y) z)); Apply PAR_S_RW with x0; Auto.
Inversion_clear H12; Apply notin_nu; Intros.
Cut (notin x0 (nu (p6 z0))); [Intro | Auto].
Inversion_clear H27; Auto.
Inversion_clear H21; Apply notin_nu; Intros.
Cut (notin x0 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H27; Auto.
Apply b_act_notin_In; Auto.

Inversion_clear H5.
Cut (x:name)(notin x (nu p)) -> (notin x (nu q)) -> (Op_StBisim SBPAR_S (p x) (q x)); [Intro | Auto].
Elim (unsat (par (nu p) (par (nu q) (nu [_:name](nu (p5 _))))) (cons x l)); Intros.
Inversion_clear H7.
Inversion_clear H8; Inversion_clear H9.
Inversion_clear H10.
Cut (btrans (q x0) (In x) (p5 x0)); [Intro | Apply H6; Auto].
Cut (Op_StBisim SBPAR_S (p x0) (q x0)); [Intro | Apply H5; Auto].
Inversion_clear H13.
Inversion_clear H14.
Inversion_clear H15.
Elim (H14 x); Intros.
Elim (H17 (p5 x0) H10); Intros.
Inversion_clear H18.
Elim (ho_proc_exp x1 x0); Intros.
Inversion_clear H18.
Rewrite H22 in H19; Rewrite H22 in H20.
Exists [_:name](nu [u:name](x2 u _)); Split; [Apply bRES with l; Intros | Idtac].
Inversion_clear H24; Change (btrans (p y) ([_:name](In x) y) (x2 y)); Apply BTR_L3 with x0; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Intro; Elim (LEM_name x0 y); Intro.

Rewrite <-H18; Elim (unsat (par (nu [_:name](nu (p5 _))) (nu [_:name](nu (x2 _)))) (cons x0 empty)); Intros.
Inversion_clear H23.
Inversion_clear H24; Inversion_clear H25.
Clear H27; Cut (SBPAR_S (x2 x0 x3) (p5 x0 x3)); [Intro | Auto].
Inversion_clear H25.
Apply sbpar_s with (S n2); Simpl; Apply op_par_s; Intros.
Elim (unsat (par (nu [_:name](nu (p5 _))) (nu [_:name](nu (x2 _)))) (cons x0 (cons x3 (cons z empty)))); Intros.
Inversion_clear H29.
Inversion_clear H30; Inversion_clear H31.
Inversion_clear H33.
Inversion_clear H34.
Clear H35.
Cut (PAR_S_fun n2 (x2 x4 x3) (p5 x4 x3)); [Intro | Change (PAR_S_fun n2 ([_:name](x2 _ x3) x4) ([_:name](p5 _ x3) x4)); Apply PAR_S_RW with x0; Auto].
Change (PAR_S_fun n2 ([_:name](x2 _ x0) z) ([_:name](p5 _ x0) z)); Apply PAR_S_RW with x4; Auto.
Inversion_clear H32; Apply notin_nu; Intros.
Cut (notin x4 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H36; Auto.
Inversion_clear H29; Apply notin_nu; Intros.
Cut (notin x4 (nu (p5 z0))); [Intro | Auto].
Inversion_clear H36; Auto.
Apply PAR_S_RW with x3; Auto.
Inversion_clear H26; Auto.
Inversion_clear H23; Auto.
Inversion_clear H21; Auto.
Inversion_clear H12; Auto.
Inversion_clear H21; Apply notin_nu; Intros.
Cut (notin x0 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H35; Auto.
Inversion_clear H12; Apply notin_nu; Intros.
Cut (notin x0 (nu (p5 z0))); [Intro | Auto].
Inversion_clear H35; Auto.
Inversion_clear H32; Apply notin_nu; Intros.
Cut (notin x4 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H35; Auto.
Inversion_clear H29; Apply notin_nu; Intros.
Cut (notin x4 (nu (p5 z0))); [Intro | Auto].
Inversion_clear H35; Auto.

Cut (SBPAR_S (x2 x0 y) (p5 x0 y)); [Intro | Auto].
Inversion_clear H23.
Apply sbpar_s with (S n2); Simpl; Apply op_par_s; Intros.
Change (PAR_S_fun n2 ([_:name](x2 _ y) z) ([_:name](p5 _ y) z)); Apply PAR_S_RW with x0; Auto.
Inversion_clear H21; Apply notin_nu; Intros.
Cut (notin x0 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H27; Auto.
Inversion_clear H12; Apply notin_nu; Intros.
Cut (notin x0 (nu (p5 z0))); [Intro | Auto].
Inversion_clear H27; Auto.
Apply b_act_notin_In; Auto.

Split; Intros.

Inversion_clear H5.

Cut (x:name)(notin x (nu p)) -> (notin x (nu q)) -> (Op_StBisim SBPAR_S (p x) (q x)); [Intro | Auto].
Elim (unsat (par (nu p) (par (nu q) (nu [_:name](nu (p6 _))))) (cons x l)); Intros.
Inversion_clear H7.
Inversion_clear H8; Inversion_clear H9.
Inversion_clear H10.
Cut (btrans (p x0) (bOut x) (p6 x0)); [Intro | Apply H6; Auto].
Cut (Op_StBisim SBPAR_S (p x0) (q x0)); [Intro | Apply H5; Auto].
Inversion_clear H13.
Inversion_clear H14.
Inversion_clear H15.
Elim (H16 x); Intros.
Elim (H15 (p6 x0) H10); Intros.
Inversion_clear H18.
Elim (ho_proc_exp x1 x0); Intros.
Inversion_clear H18.
Rewrite H22 in H19; Rewrite H22 in H20.
Exists [_:name](nu [u:name](x2 u _)); Split; [Apply bRES with l; Intros | Idtac].
Inversion_clear H24; Change (btrans (q y) ([_:name](bOut x) y) (x2 y)); Apply BTR_L3 with x0; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Intros; Elim (LEM_name x0 y); Intro.

Rewrite <-H24; Elim (unsat (par (nu [_:name](nu (p6 _))) (nu [_:name](nu (x2 _)))) (cons x0 empty)); Intros.
Inversion_clear H25.
Inversion_clear H26; Inversion_clear H27.
Clear H29; Cut (SBPAR_S (p6 x0 x3) (x2 x0 x3)); [Intro | Apply H20; Auto].
Inversion_clear H27.
Apply sbpar_s with (S n2); Simpl; Apply op_par_s; Intros.
Elim (unsat (par (nu [_:name](nu (p6 _))) (nu [_:name](nu (x2 _)))) (cons x0 (cons x3 (cons z empty)))); Intros.
Inversion_clear H31.
Inversion_clear H32; Inversion_clear H33.
Inversion_clear H35.
Inversion_clear H36.
Clear H37.
Cut (PAR_S_fun n2 (p6 x4 x3) (x2 x4 x3)); [Intro | Change (PAR_S_fun n2 ([_:name](p6 _ x3) x4) ([_:name](x2 _ x3) x4)); Apply PAR_S_RW with x0; Auto].
Change (PAR_S_fun n2 ([_:name](p6 _ x0) z) ([_:name](x2 _ x0) z)); Apply PAR_S_RW with x4; Auto.
Inversion_clear H31; Apply notin_nu; Intros.
Cut (notin x4 (nu (p6 z0))); [Intro | Auto].
Inversion_clear H38; Auto.
Inversion_clear H34; Apply notin_nu; Intros.
Cut (notin x4 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H38; Auto.
Apply PAR_S_RW with x3; Auto.
Inversion_clear H25; Auto.
Inversion_clear H28; Auto.
Inversion_clear H12; Auto.
Inversion_clear H21; Auto.
Inversion_clear H12; Apply notin_nu; Intros.
Cut (notin x0 (nu (p6 z0))); [Intro | Auto].
Inversion_clear H37; Auto.
Inversion_clear H21; Apply notin_nu; Intros.
Cut (notin x0 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H37; Auto.
Inversion_clear H31; Apply notin_nu; Intros.
Cut (notin x4 (nu (p6 z0))); [Intro | Auto].
Inversion_clear H37; Auto.
Inversion_clear H34; Apply notin_nu; Intros.
Cut (notin x4 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H37; Auto.
Inversion_clear H25; Auto.
Inversion_clear H28; Auto.

Cut (SBPAR_S (p6 x0 y) (x2 x0 y)); [Intro | Apply H20; Auto].
Inversion_clear H25.
Apply sbpar_s with (S n2); Simpl; Apply op_par_s; Intros.
Change (PAR_S_fun n2 ([_:name](p6 _ y) z) ([_:name](x2 _ y) z)); Apply PAR_S_RW with x0; Auto.
Inversion_clear H12; Apply notin_nu; Intros.
Cut (notin x0 (nu (p6 z0))); [Intro | Auto].
Inversion_clear H29; Auto.
Inversion_clear H21; Apply notin_nu; Intros.
Cut (notin x0 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H29; Auto.
Inversion_clear H18; Apply notin_nu; Intros.
Cut (notin y (nu [z0:name](p6 z0 z))); [Intro | Auto].
Inversion_clear H26; Auto.
Inversion_clear H23; Apply notin_nu; Intros.
Cut (notin y (nu [z0:name](x2 z0 z))); [Intro | Auto].
Inversion_clear H26; Auto.
Apply b_act_notin_bOut; Auto.

Cut (x:name)(notin x (nu p)) -> (notin x (nu q)) -> (Op_StBisim SBPAR_S (p x) (q x)); [Intro | Auto].
Elim (unsat (par (nu p) (par (nu q) (nu p4))) (cons x l)); Intros.
Inversion_clear H7.
Inversion_clear H8; Inversion_clear H9.
Inversion_clear H10.
Cut (ftrans (p x0) (Out x x0) (p4 x0)); [Intro | Apply H6; Auto].
Cut (Op_StBisim SBPAR_S (p x0) (q x0)); [Intro | Apply H5; Auto].
Inversion_clear H13.
Inversion_clear H14.
Elim (H13 (Out x x0)); Intros.
Elim (H14 (p4 x0) H10); Intros.
Inversion_clear H17.
Elim (proc_exp x1 x0); Intros.
Inversion_clear H17.
Rewrite H21 in H18; Rewrite H21 in H19.
Exists x2; Split; [Apply OPEN with l; Intros | Idtac].
Change (ftrans (q y) (Out x y) (x2 y)); Apply FTR_L3 with x0; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Intros; Inversion_clear H19.
Apply sbpar_s with n2; Apply PAR_S_RW with x0; Auto.

Inversion_clear H5.

Cut (x:name)(notin x (nu p)) -> (notin x (nu q)) -> (Op_StBisim SBPAR_S (p x) (q x)); [Intro | Auto].
Elim (unsat (par (nu p) (par (nu q) (nu [_:name](nu (p5 _))))) (cons x l)); Intros.
Inversion_clear H7.
Inversion_clear H8; Inversion_clear H9.
Inversion_clear H10.
Cut (btrans (q x0) (bOut x) (p5 x0)); [Intro | Apply H6; Auto].
Cut (Op_StBisim SBPAR_S (p x0) (q x0)); [Intro | Apply H5; Auto].
Inversion_clear H13.
Inversion_clear H14.
Inversion_clear H15.
Elim (H16 x); Intros.
Elim (H17 (p5 x0) H10); Intros.
Inversion_clear H18.
Elim (ho_proc_exp x1 x0); Intros.
Inversion_clear H18.
Rewrite H22 in H19; Rewrite H22 in H20.
Exists [_:name](nu [u:name](x2 u _)); Split; [Apply bRES with l; Intros | Idtac].
Inversion_clear H24; Change (btrans (p y) ([_:name](bOut x) y) (x2 y)); Apply BTR_L3 with x0; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Intros; Elim (LEM_name x0 y); Intro.

Rewrite <-H24; Elim (unsat (par (nu [_:name](nu (p5 _))) (nu [_:name](nu (x2 _)))) (cons x0 empty)); Intros.
Inversion_clear H25.
Inversion_clear H26; Inversion_clear H27.
Clear H29; Cut (SBPAR_S (x2 x0 x3) (p5 x0 x3)); [Intro | Apply H20; Auto].
Inversion_clear H27.
Apply sbpar_s with (S n2); Simpl; Apply op_par_s; Intros.
Elim (unsat (par (nu [_:name](nu (p5 _))) (nu [_:name](nu (x2 _)))) (cons x0 (cons x3 (cons z empty)))); Intros.
Inversion_clear H31.
Inversion_clear H32; Inversion_clear H33.
Inversion_clear H35.
Inversion_clear H36.
Clear H37.
Cut (PAR_S_fun n2 (x2 x4 x3) (p5 x4 x3)); [Intro | Change (PAR_S_fun n2 ([_:name](x2 _ x3) x4) ([_:name](p5 _ x3) x4)); Apply PAR_S_RW with x0; Auto].
Change (PAR_S_fun n2 ([_:name](x2 _ x0) z) ([_:name](p5 _ x0) z)); Apply PAR_S_RW with x4; Auto.
Inversion_clear H34; Apply notin_nu; Intros.
Cut (notin x4 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H38; Auto.
Inversion_clear H31; Apply notin_nu; Intros.
Cut (notin x4 (nu (p5 z0))); [Intro | Auto].
Inversion_clear H38; Auto.
Apply PAR_S_RW with x3; Auto.
Inversion_clear H28; Auto.
Inversion_clear H25; Auto.
Inversion_clear H21; Auto.
Inversion_clear H12; Auto.
Inversion_clear H21; Apply notin_nu; Intros.
Cut (notin x0 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H37; Auto.
Inversion_clear H12; Apply notin_nu; Intros.
Cut (notin x0 (nu (p5 z0))); [Intro | Auto].
Inversion_clear H37; Auto.
Inversion_clear H34; Apply notin_nu; Intros.
Cut (notin x4 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H37; Auto.
Inversion_clear H31; Apply notin_nu; Intros.
Cut (notin x4 (nu (p5 z0))); [Intro | Auto].
Inversion_clear H37; Auto.
Inversion_clear H28; Auto.
Inversion_clear H25; Auto.

Cut (SBPAR_S (x2 x0 y) (p5 x0 y)); [Intro | Apply H20; Auto].
Inversion_clear H25.
Apply sbpar_s with (S n2); Simpl; Apply op_par_s; Intros.
Change (PAR_S_fun n2 ([_:name](x2 _ y) z) ([_:name](p5 _ y) z)); Apply PAR_S_RW with x0; Auto.
Inversion_clear H21; Apply notin_nu; Intros.
Cut (notin x0 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H29; Auto.
Inversion_clear H12; Apply notin_nu; Intros.
Cut (notin x0 (nu (p5 z0))); [Intro | Auto].
Inversion_clear H29; Auto.
Inversion_clear H18; Apply notin_nu; Intros.
Cut (notin y (nu [z0:name](x2 z0 z))); [Intro | Auto].
Inversion_clear H26; Auto.
Inversion_clear H23; Apply notin_nu; Intros.
Cut (notin y (nu [z0:name](p5 z0 z))); [Intro | Auto].
Inversion_clear H26; Auto.
Apply b_act_notin_bOut; Auto.

Cut (x:name)(notin x (nu p)) -> (notin x (nu q)) -> (Op_StBisim SBPAR_S (p x) (q x)); [Intro | Auto].
Elim (unsat (par (nu p) (par (nu q) (nu q1))) (cons x l)); Intros.
Inversion_clear H7.
Inversion_clear H8; Inversion_clear H9.
Inversion_clear H10.
Cut (ftrans (q x0) (Out x x0) (q1 x0)); [Intro | Apply H6; Auto].
Cut (Op_StBisim SBPAR_S (p x0) (q x0)); [Intro | Apply H5; Auto].
Inversion_clear H13.
Inversion_clear H14.
Elim (H13 (Out x x0)); Intros.
Elim (H16 (q1 x0) H10); Intros.
Inversion_clear H17.
Elim (proc_exp x1 x0); Intros.
Inversion_clear H17.
Rewrite H21 in H18; Rewrite H21 in H19.
Exists x2; Split; [Apply OPEN with l; Intros | Idtac].
Change (ftrans (p y) (Out x y) (x2 y)); Apply FTR_L3 with x0; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Intros; Inversion_clear H19.
Apply sbpar_s with n2; Apply PAR_S_RW with x0; Auto.

Qed.

(* Finally, the main goal, i.e., the congruence of StBisim
 * w.r.t. parallel composition.
 *)

Lemma PAR_S:  (p,q,r:proc)(StBisim p q)->(StBisim (par p r) (par q r)).

Proof.

Intros; Apply Completeness; Apply Co_Ind with SBPAR_S; [Exact SBPAR_S_TO_SB' | Apply sbpar_s with O; Simpl; Apply bispar_s; Assumption].

Qed.

Lemma MATCH_S: (p,q:proc)(StBisim p q)->(x,y:name)(StBisim (match x y p) (match x y q)).

Proof. (* This lemma is proved from scratch. *)

(Intros; Apply sb; Do 3 Try (Split; Intros)).
Inversion H0.
Inversion H.
Inversion_clear H7.
(Elim (H10 a); Intros).
(Elim (H7 p1 H6); Intros).
(Exists  x1; Split; [ Apply fMATCH; Tauto | Tauto ]).

Inversion H0.
Inversion H.
Inversion_clear H7.
(Elim (H10 a); Intros).
(Elim (H12 q1 H6); Intros).
(Exists  x1; Split; [ Apply fMATCH; Tauto | Tauto ]).

Inversion H0.
Inversion H.
Inversion_clear H7.
Inversion_clear H11.
(Elim (H7 x0); Intros).
(Elim (H11 p1 H6); Intros).
Inversion_clear H14.
(Exists  x2; Split; [ Apply bMATCH; Assumption | Intro; Auto ]).

Inversion H0.
Inversion H.
Inversion_clear H7.
Inversion_clear H11.
(Elim (H7 x0); Intros).
(Elim (H13 q1 H6); Intros).
Inversion_clear H14.
(Exists  x2; Split; [ Apply bMATCH; Assumption | Intro; Auto ]).

Inversion H0.
Inversion H.
Inversion_clear H7.
Inversion_clear H11.
(Elim (H12 x0); Intros).
(Elim (H11 p1 H6); Intros).
Inversion_clear H14.
(Exists  x2; Split; [ Apply bMATCH; Assumption | Intros; Auto ]).

Inversion H0.
Inversion H.
Inversion_clear H7.
Inversion_clear H11.
(Elim (H12 x0); Intros).
(Elim (H13 q1 H6); Intros).
Inversion_clear H14.
(Exists  x2; Split; [ Apply bMATCH; Assumption | Intros; Auto ]).

Qed.

Lemma MISMATCH_S: (p,q:proc)(StBisim p q)->(x,y:name)(StBisim (mismatch x y p) (mismatch x y q)).

Proof. (* This lemma is proved from scratch. *)

(Intros; Apply sb; Do 3 Try (Split; Intros)).
Inversion H0.
Inversion H.
Inversion_clear H8.
(Elim (H11 a); Intros).
(Elim (H8 p1 H7); Intros).
(Exists  x1; Split; [ Apply fMISMATCH; Tauto | Tauto ]).

Inversion H0.
Inversion H.
Inversion_clear H8.
(Elim (H11 a); Intros).
(Elim (H13 q1 H7); Intros).
(Exists  x1; Split; [ Apply fMISMATCH; Tauto | Tauto ]).

Inversion H0.
Inversion H.
Inversion_clear H8.
Inversion_clear H12.
(Elim (H8 x0); Intros).
(Elim (H12 p1 H7); Intros).
Inversion_clear H15.
(Exists  x2; Split; [ Apply bMISMATCH; Assumption | Intro; Auto ]).

Inversion H0.
Inversion H.
Inversion_clear H8.
Inversion_clear H12.
(Elim (H8 x0); Intros).
(Elim (H14 q1 H7); Intros).
Inversion_clear H15.
(Exists  x2; Split; [ Apply bMISMATCH; Assumption | Intro; Auto ]).

Inversion H0.
Inversion H.
Inversion_clear H8.
Inversion_clear H12.
(Elim (H13 x0); Intros).
(Elim (H12 p1 H7); Intros).
Inversion_clear H15.
(Exists  x2; Split; [ Apply bMISMATCH; Assumption | Intros; Auto ]).

Inversion H0.
Inversion H.
Inversion_clear H8.
Inversion_clear H12.
(Elim (H13 x0); Intros).
(Elim (H14 q1 H7); Intros).
Inversion_clear H15.
(Exists  x2; Split; [ Apply bMISMATCH; Assumption | Intros; Auto ]).

Qed.

Lemma IN_S: (p',q':name->proc)(x,y:name)
            (notin y (nu p'))->(notin y (nu q'))->
            ((z:name)
             ((isin z (nu p'))\/(isin z (nu q'))\/z=y)->
             (StBisim (p' z) (q' z)))->
            (StBisim (in_pref x p') (in_pref x q')).

Proof. (* This lemma depends on Lemma L6_Light. *)

Intros; Apply sb; Do 3 Try (Split; Intros).

Inversion_clear H2.

Inversion_clear H2.

Inversion H2.
Rewrite <- H3 in H2; Rewrite <- H5 in H2; Rewrite <- H6 in H2.
Exists q'; Split.
Apply IN.
Rewrite <- H3; Intro.
Elim (LEM_OC (par (nu p') (par (nu q') (out_pref y y nil))) y0); Intro.
Inversion_clear H7; [Apply H1; Left; Assumption | Inversion_clear H8; [Apply H1; Right; Left; Assumption | Inversion H7; [Inversion_clear H9 | Apply H1; Right; Right; Trivial | Apply H1; Right; Right; Trivial]]].
Inversion_clear H7.
Inversion_clear H9.
Inversion_clear H10.
Clear H11 H12.
Apply L6_Light with y; Auto.

Inversion H2.
Rewrite <- H3 in H2; Rewrite <- H5 in H2; Rewrite <- H6 in H2.
Exists p'; Split.
Apply IN.
Rewrite <- H3; Intro.
Elim (LEM_OC (par (nu p') (par (nu q') (out_pref y y nil))) y0); Intro.
Inversion_clear H7; [Apply H1; Left; Assumption | Inversion_clear H8; [Apply H1; Right; Left; Assumption | Inversion H7; [Inversion_clear H9 | Apply H1; Right; Right; Trivial | Apply H1; Right; Right; Trivial]]].
Inversion_clear H7.
Inversion_clear H9.
Inversion_clear H10.
Clear H11 H12.
Apply L6_Light with y; Auto.

Inversion_clear H2.

Inversion_clear H2.

Qed.

Lemma OUT_S: (p,q:proc)(StBisim p q)->(x,y:name)(StBisim (out_pref x y p) (out_pref x y q)).

Proof. (* This lemma is proved from scratch. *)

(Intros; Apply sb; Do 3 Try (Split; Intros)).
Inversion H0.
(Exists  q; Split; [ Apply OUT | Rewrite <- H5; Assumption ]).

Inversion H0.
(Exists  p; Split; [ Apply OUT | Rewrite <- H5; Assumption ]).

Inversion_clear H0.

Inversion_clear H0.

Inversion_clear H0.

Inversion_clear H0.

Qed.

End structural_congruence1.

Section monoidal_sum.

Lemma ID_SUM: (p:proc)(StBisim (sum p nil) p).

Proof. (* This lemma depends on Lemma REF. *)

Intro; Apply sb; Do 3 Try (Split; Intros).

Inversion_clear H.
Exists  p1; Split; [Assumption | Apply REF].
Inversion_clear H0.
Exists q1; Split; [Apply fSUM1; Assumption | Apply REF].

Inversion_clear H.
Exists  p1; Split; [Assumption | Intro; Apply REF].
Inversion_clear H0.
Exists q1; Split; [Apply bSUM1; Assumption | Intro; Apply REF].

Inversion_clear H.
Exists  p1; Split; [Assumption | Intros; Apply REF].
Inversion_clear H0.
Exists q1; Split; [Apply bSUM1; Assumption | Intros; Apply REF].

Qed.

Lemma IDEM_SUM: (p:proc)(StBisim (sum p p) p).

Proof. (* This lemma depends on Lemma REF. *)

Intro; Apply sb; Do 3 Try (Split; Intros).

Inversion_clear H.
Exists  p1; Split; [Assumption | Apply REF].
Exists  p1; Split; [Assumption | Apply REF].
Exists  q1; Split; [Apply fSUM1; Assumption | Apply REF].

Inversion_clear H.
Exists  p1; Split; [Assumption | Intro; Apply REF].
Exists  p1; Split; [Assumption | Intro; Apply REF].
Exists  q1; Split; [Apply bSUM1; Assumption | Intro; Apply REF].

Inversion_clear H.
Exists  p1; Split; [Assumption | Intros; Apply REF].
Exists  p1; Split; [Assumption | Intros; Apply REF].
Exists  q1; Split; [Apply bSUM1; Assumption | Intros; Apply REF].

Qed.

Lemma COMM_SUM: (p,q:proc)(StBisim (sum p q) (sum q p)).

Proof. (* This lemma depends on Lemma REF. *)

Intros; Apply sb; Do 3 Try (Split; Intros).

Inversion_clear H.
Exists  p1; Split; [Apply fSUM2; Assumption | Apply REF].
Exists  p1; Split; [Apply fSUM1; Assumption | Apply REF].

Inversion_clear H.
Exists  q1; Split; [Apply fSUM2; Assumption | Apply REF].
Exists  q1; Split; [Apply fSUM1; Assumption | Apply REF].

Inversion_clear H.
Exists  p1; Split; [Apply bSUM2; Assumption | Intro; Apply REF].
Exists  p1; Split; [Apply bSUM1; Assumption | Intro; Apply REF].

Inversion_clear H.
Exists  q1; Split; [Apply bSUM2; Assumption | Intro; Apply REF].
Exists  q1; Split; [Apply bSUM1; Assumption | Intro; Apply REF].

Inversion_clear H.
Exists  p1; Split; [Apply bSUM2; Assumption | Intros; Apply REF].
Exists  p1; Split; [Apply bSUM1; Assumption | Intros; Apply REF].

Inversion_clear H.
Exists  q1; Split; [Apply bSUM2; Assumption | Intros; Apply REF].
Exists  q1; Split; [Apply bSUM1; Assumption | Intros; Apply REF].

Qed.

Lemma ASS_SUM: (p,q,r:proc)(StBisim (sum p (sum q r)) (sum (sum p q) r)).

Proof. (* This lemma depends on Lemma REF. *)

Intros; Apply sb; Do 3 Try (Split; Intros).

Inversion_clear H.
Exists  p1; Split; [Apply fSUM1; Apply fSUM1; Assumption | Apply REF].
Inversion_clear H0.
Exists  p1; Split; [Apply fSUM1; Apply fSUM2; Assumption | Apply REF].
Exists  p1; Split; [Apply fSUM2; Assumption | Apply REF].

Inversion_clear H.
Inversion_clear H0.
Exists  q1; Split; [Apply fSUM1; Assumption | Apply REF].
Exists  q1; Split; [Apply fSUM2; Apply fSUM1; Assumption | Apply REF].
Exists  q1; Split; [Apply fSUM2; Apply fSUM2; Assumption | Apply REF].

Inversion_clear H.
Exists  p1; Split; [Apply bSUM1; Apply bSUM1; Assumption | Intro; Apply REF].
Inversion_clear H0.
Exists  p1; Split; [Apply bSUM1; Apply bSUM2; Assumption | Intro; Apply REF].
Exists  p1; Split; [Apply bSUM2; Assumption | Intro; Apply REF].

Inversion_clear H.
Inversion_clear H0.
Exists  q1; Split; [Apply bSUM1; Assumption | Intro; Apply REF].
Exists  q1; Split; [Apply bSUM2; Apply bSUM1; Assumption | Intro; Apply REF].
Exists  q1; Split; [Apply bSUM2; Apply bSUM2; Assumption | Intro; Apply REF].

Inversion_clear H.
Exists  p1; Split; [Apply bSUM1; Apply bSUM1; Assumption | Intros; Apply REF].
Inversion_clear H0.
Exists  p1; Split; [Apply bSUM1; Apply bSUM2; Assumption | Intros; Apply REF].
Exists  p1; Split; [Apply bSUM2; Assumption | Intros; Apply REF].

Inversion_clear H.
Inversion_clear H0.
Exists  q1; Split; [Apply bSUM1; Assumption | Intros; Apply REF].
Exists  q1; Split; [Apply bSUM2; Apply bSUM1; Assumption | Intros; Apply REF].
Exists  q1; Split; [Apply bSUM2; Apply bSUM2; Assumption | Intros; Apply REF].

Qed.

End monoidal_sum.

Section bang_unfolding.

Lemma BANG_UNF : (p:proc)(StBisim (bang p) (par p (bang p))).

Proof. (* This lemma depends on Lemma REF. *)

Intro; Apply sb; Do 3 Try (Split; Intros).
Exists p1; Split; [Inversion_clear H; Assumption | Apply REF].
Exists q1; Split; [Apply fBANG; Assumption | Apply REF].
Exists p1; Split; [Inversion_clear H; Assumption | Intro; Apply REF].
Exists q1; Split; [Apply bBANG; Assumption | Intro; Apply REF].
Exists p1; Split; [Inversion_clear H; Assumption | Intros; Apply REF].
Exists q1; Split; [Apply bBANG; Assumption | Intros; Apply REF].

Qed.

End bang_unfolding.

Section restriction_laws1.

Lemma NU_NIL: (StBisim (nu [x:name]nil) nil).

Proof. (* This lemma is proved from scratch. *)

Apply sb; Do 3 Try (Split; Intros).

Generalize H; Clear H; Elim a; Intros.

Inversion_clear H.
Elim (unsat (par (nu [_:name]nil) (nu p2)) l); Intros.
Inversion_clear H.
Inversion_clear H1.
Cut (ftrans nil tau (p2 x)); [Intro | Apply H0; Auto].
Inversion_clear H1.
Apply f_act_notin_tau.

Inversion_clear H.
Elim (unsat (par (nu [_:name]nil) (nu p2)) (cons n (cons n0 l))); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H4.
Cut (ftrans nil (Out n n0) (p2 x)); [Intro | Apply H0; Auto].
Inversion_clear H4.
Apply f_act_notin_Out; Auto.

Inversion_clear H.

Inversion_clear H.
Elim (unsat (par (nu [_:name]nil) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans nil (In x) (p2 x0)); [Intro | Apply H0; Auto].
Inversion_clear H2.
Apply b_act_notin_In; Auto.

Inversion_clear H.

Inversion_clear H.
Elim (unsat (par (nu [_:name]nil) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans nil (bOut x) (p2 x0)); [Intro | Apply H0; Auto].
Inversion_clear H2.
Apply b_act_notin_bOut; Auto.

Elim (unsat (par (nu [_:name]nil) (nu p1)) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (ftrans nil (Out x x0) (p1 x0)); [Intro | Apply H0; Auto].
Inversion_clear H2.

Inversion_clear H.

Qed.

Lemma NU_NIL1: (p:proc)(x:name)(StBisim (nu [y:name](out_pref y x p)) nil).

Proof. (* This lemma is proved from scratch. *)

Intros; Apply sb; Do 3 Try (Split; Intros).
Generalize H; Clear H; Elim a; Intros.
Inversion_clear H.
Elim (unsat (par (nu [y0:name](out_pref y0 x p)) (nu p2)) l); Intros.
Inversion_clear H.
Inversion_clear H1.
Cut (ftrans (out_pref x0 x p) tau (p2 x0)); [Intro | Apply H0; Auto].
Inversion_clear H1.
Apply f_act_notin_tau.
Inversion_clear H.
Elim (unsat (par (nu [y0:name](out_pref y0 x p)) (nu p2)) (cons n (cons n0 l))); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H4.
Cut (ftrans (out_pref x0 x p) (Out n n0) (p2 x0)); [Intro | Apply H0; Auto].
Inversion H4.
Absurd x0=n; Auto.
Apply f_act_notin_Out; Auto.
Inversion_clear H.

Inversion_clear H.
Elim (unsat (par (nu [y0:name](out_pref y0 x p)) (nu [z:name](nu (p2 z)))) (cons x0 l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (out_pref x1 x p) (In x0) (p2 x1)); [Intro | Apply H0; Auto].
Inversion_clear H2.
Apply b_act_notin_In; Auto.
Inversion_clear H.

Inversion_clear H.
Elim (unsat (par (nu [y0:name](out_pref y0 x p)) (nu [z:name](nu (p2 z)))) (cons x0 l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (out_pref x1 x p) (bOut x0) (p2 x1)); [Intro | Apply H0; Auto].
Inversion_clear H2.
Apply b_act_notin_bOut; Auto.
Elim (unsat (par (nu [y0:name](out_pref y0 x p)) (nu p1)) (cons x0 l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (ftrans (out_pref x1 x p) (Out x0 x1) (p1 x1)); [Intro | Apply H0; Auto].
Inversion H2.
Absurd x1=x0; Auto.
Inversion_clear H.

Qed.

Lemma NU_NIL2: (p:name->proc)(StBisim (nu [y:name](in_pref y p)) nil).

Proof. (* This lemma is proved from scratch. *)

Intros; Apply sb; Do 3 Try (Split; Intros).
Generalize H; Clear H; Elim a; Intros.
Inversion_clear H.
Elim (unsat (par (nu [y0:name](in_pref y0 p)) (nu p2)) l); Intros.
Inversion_clear H.
Inversion_clear H1.
Cut (ftrans (in_pref x p) tau (p2 x)); [Intro | Apply H0; Auto].
Inversion_clear H1.
Apply f_act_notin_tau.
Inversion_clear H.
Elim (unsat (par (nu [y0:name](in_pref y0 p)) (nu p2)) (cons n (cons n0 l))); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H4.
Cut (ftrans (in_pref x p) (Out n n0) (p2 x)); [Intro | Apply H0; Auto].
Inversion H4.
Apply f_act_notin_Out; Auto.
Inversion_clear H.

Inversion_clear H.
Elim (unsat (par (nu [y0:name](in_pref y0 p)) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (in_pref x0 p) (In x) (p2 x0)); [Intro | Apply H0; Auto].
Inversion H2.
Absurd x0=x; Auto.
Apply b_act_notin_In; Auto.
Inversion_clear H.

Inversion_clear H.
Elim (unsat (par (nu [y0:name](in_pref y0 p)) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (in_pref x0 p) (bOut x) (p2 x0)); [Intro | Apply H0; Auto].
Inversion_clear H2.
Apply b_act_notin_bOut; Auto.
Elim (unsat (par (nu [y0:name](in_pref y0 p)) (nu p1)) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (ftrans (in_pref x0 p) (Out x x0) (p1 x0)); [Intro | Apply H0; Auto].
Inversion_clear H2.
Inversion_clear H.

Qed.

Lemma NU_COMM: (p:name->name->proc)(StBisim (nu [y:name](nu [z:name](p y z))) (nu [z:name](nu [y:name](p y z)))).

Proof. (* This lemma depends on Lemmata FTR_L3, BTR_L3:
        * it is a coinductive proof.
        *)

Cofix; Intro; Apply sb; Do 3 Try (Split; Intros).
Generalize H; Clear H; Elim a; Intros.

(* TAU action case *)

Inversion_clear H.
Elim (unsat (par (nu [y0:name](nu [z:name](p y0 z))) (nu p2)) l); Intros.
Inversion_clear H.
Inversion_clear H1.
Cut (ftrans (nu [z:name](p x z)) tau (p2 x)); [Intro | Apply H0; Auto].
Inversion H1.
Elim (unsat (par (nu [z:name](p x z)) (nu p3)) l0); Intros.
Inversion_clear H8.
Inversion_clear H9.
Cut (ftrans (p x x0) tau (p3 x0)); [Intro | Apply H7; Auto].
Elim (ho_proc_exp p3 x); Intros.
Inversion_clear H12.
Exists (nu [z:name](nu [y:name](x1 y z))); Split.
Apply fRES with (cons x (cons x0 empty)); Intros.
Inversion_clear H16.
Inversion_clear H19.
Clear H17 H20.
Apply fRES with (cons x (cons x0 empty)); Intros.
Inversion_clear H20.
Inversion_clear H23.
Clear H21 H24.
Change (ftrans ([_:name](p _ y) y0) ([_:name]tau y0) ([_:name](x1 _ y) y0)); Apply FTR_L3 with x; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x (nu [z0:name](p z z0))); [Intro | Auto].
Inversion_clear H23; Auto.
Inversion_clear H13; Apply notin_nu; Intros.
Cut (notin x (nu (x1 z))); [Intro | Auto].
Inversion_clear H23; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Change (ftrans (p x y) ([_:name]tau y) (x1 x y)); Apply FTR_L3 with x0; Auto.
Inversion_clear H8; Apply notin_nu; Intros; Auto.
Rewrite <- H14; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Rewrite <- H14; Assumption.
Inversion_clear H12; Apply notin_nu; Intros.
Cut (notin y (nu [y0:name](p y0 z))); [Intro | Auto].
Inversion_clear H23; Auto.
Inversion_clear H15; Apply notin_nu; Intros.
Cut (notin y (nu [y0:name](x1 y0 z))); [Intro | Auto].
Inversion_clear H23; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Rewrite H14 in H5; Cut p2=([_:name](nu (x1 _))); [Intro | Apply proc_ext with x; Auto].
Rewrite H12; Cut (nu [_:name](nu (x1 _)))=(nu [_:name](nu [u:name](x1 _ u))); [Intro | Apply ETA_EQ2].
Rewrite H15; Apply NU_COMM.
Apply f_act_notin_tau.
Apply f_act_notin_tau.

(* OUT action case *)

Inversion_clear H.
Elim (unsat (par (nu [y0:name](nu [z:name](p y0 z))) (nu p2)) (cons n (cons n0 l))); Intros.
Inversion_clear H.
Inversion_clear H1; Inversion_clear H2.
Inversion_clear H4.
Cut (ftrans (nu [z:name](p x z)) (Out n n0) (p2 x)); [Intro | Apply H0; Auto].
Inversion H4.
Elim (unsat (par (nu [z:name](p x z)) (nu p3)) (cons n (cons n0 l0))); Intros.
Inversion_clear H10.
Inversion_clear H11; Inversion_clear H12.
Inversion_clear H14.
Cut (ftrans (p x x0) (Out n n0) (p3 x0)); [Intro | Apply H9; Auto].
Elim (ho_proc_exp p3 x); Intros.
Inversion_clear H16.
Exists (nu [z:name](nu [y:name](x1 y z))); Split.
Apply fRES with (cons x (cons x0 empty)); Intros.
Inversion_clear H20; Inversion_clear H21.
Inversion_clear H23.
Clear H25.
Apply fRES with (cons x (cons x0 empty)); Intros.
Inversion_clear H26; Inversion_clear H27.
Inversion_clear H29.
Clear H31.
Change (ftrans ([_:name](p _ y) y0) ([_:name](Out n n0) y0) ([_:name](x1 _ y) y0)); Apply FTR_L3 with x; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x (nu [z0:name](p z z0))); [Intro | Auto].
Inversion_clear H31; Auto.
Inversion_clear H17; Apply notin_nu; Intros.
Cut (notin x (nu (x1 z))); [Intro | Auto].
Inversion_clear H31; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Change (ftrans (p x y) ([_:name](Out n n0) y) (x1 x y)); Apply FTR_L3 with x0; Auto.
Inversion_clear H10; Apply notin_nu; Intros; Auto.
Rewrite <- H18; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite <- H18; Assumption.
Inversion_clear H16; Apply notin_nu; Intros.
Cut (notin y (nu [y0:name](p y0 z))); [Intro | Auto].
Inversion_clear H31; Auto.
Inversion_clear H19; Apply notin_nu; Intros.
Cut (notin y (nu [y0:name](x1 y0 z))); [Intro | Auto].
Inversion_clear H31; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite H18 in H7; Cut p2=([_:name](nu (x1 _))); [Intro | Apply proc_ext with x; Auto].
Rewrite H16;  Cut (nu [_:name](nu (x1 _)))=(nu [_:name](nu [u:name](x1 _ u))); [Intro | Apply ETA_EQ2].
Rewrite H19; Apply NU_COMM.
Apply f_act_notin_Out; Auto.
Apply f_act_notin_Out; Auto.

(* The way back: symmetric case *)

Generalize H; Clear H; Elim a; Intros.

(* TAU action case *)

Inversion_clear H.
Elim (unsat (par (nu [z:name](nu [y0:name](p y0 z))) (nu p2)) l); Intros.
Inversion_clear H.
Inversion_clear H1.
Cut (ftrans (nu [y0:name](p y0 x)) tau (p2 x)); [Intro | Apply H0; Auto].
Inversion H1.
Elim (unsat (par (nu [y0:name](p y0 x)) (nu p0)) l0); Intros.
Inversion_clear H8.
Inversion_clear H9.
Cut (ftrans (p x0 x) tau (p0 x0)); [Intro | Apply H7; Auto].
Elim (ho_proc_exp p0 x); Intros.
Inversion_clear H12.
Exists (nu [y:name](nu [z:name](x1 z y))); Split.
Apply fRES with (cons x (cons x0 empty)); Intros.
Inversion_clear H16.
Inversion_clear H19.
Clear H17 H20.
Apply fRES with (cons x (cons x0 empty)); Intros.
Inversion_clear H20.
Inversion_clear H23.
Clear H21 H24.
Change (ftrans (p y y0) ([_:name]tau y0) ([_:name](x1 _ y) y0)); Apply FTR_L3 with x; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x (nu [y0:name](p y0 z))); [Intro | Auto].
Inversion_clear H23; Auto.
Inversion_clear H13; Apply notin_nu; Intros.
Cut (notin x (nu (x1 z))); [Intro | Auto].
Inversion_clear H23; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Change (ftrans ([_:name](p _ x) y) ([_:name]tau y) (x1 x y)); Apply FTR_L3 with x0; Auto.
Rewrite <- H14; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Rewrite <- H14; Assumption.
Inversion_clear H12; Apply notin_nu; Intros.
Cut (notin y (nu [z0:name](p z z0))); [Intro | Auto].
Inversion_clear H23; Auto.
Inversion_clear H15; Apply notin_nu; Intros.
Cut (notin y (nu [z0:name](x1 z0 z))); [Intro | Auto].
Inversion_clear H23; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Inversion_clear H17; Apply notin_nu; Intros ; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Rewrite H14 in H5; Cut p2=([_:name](nu (x1 _))); [Intro | Apply proc_ext with x; Auto].
Rewrite H12; Cut (nu [_:name](nu (x1 _)))=(nu [_:name](nu [u:name](x1 _ u))); [Intro | Apply ETA_EQ2].
Rewrite H15; Change (StBisim (nu [y:name](nu [z:name]([u:name][v:name](x1 v u) y z))) (nu [_:name](nu [u:name]([u':name][v:name](x1 v u') u _)))); Apply NU_COMM.
Apply f_act_notin_tau.
Apply f_act_notin_tau.

(* OUT action case *)

Inversion_clear H.
Elim (unsat (par (nu [z:name](nu [y0:name](p y0 z))) (nu p2)) (cons n (cons n0 l))); Intros.
Inversion_clear H.
Inversion_clear H1; Inversion_clear H2.
Inversion_clear H4.
Cut (ftrans (nu [y0:name](p y0 x)) (Out n n0) (p2 x)); [Intro | Apply H0; Auto].
Inversion H4.
Elim (unsat (par (nu [y0:name](p y0 x)) (nu p0)) (cons n (cons n0 l0))); Intros.
Inversion_clear H10.
Inversion_clear H11; Inversion_clear H12.
Inversion_clear H14.
Cut (ftrans (p x0 x) (Out n n0) (p0 x0)); [Intro | Apply H9; Auto].
Elim (ho_proc_exp p0 x); Intros.
Inversion_clear H16.
Exists (nu [y:name](nu [z:name](x1 z y))); Split.
Apply fRES with (cons x (cons x0 empty)); Intros.
Inversion_clear H20; Inversion_clear H21.
Inversion_clear H23.
Clear H25; Apply fRES with (cons x (cons x0 empty)); Intros.
Inversion_clear H26; Inversion_clear H27.
Inversion_clear H29.
Clear H31; Change (ftrans (p y y0) ([_:name](Out n n0) y0) ([_:name](x1 _ y) y0)); Apply FTR_L3 with x; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x (nu [z0:name](p z0 z))); [Intro | Auto].
Inversion_clear H31; Auto.
Inversion_clear H17; Apply notin_nu; Intros.
Cut (notin x (nu (x1 z))); [Intro | Auto].
Inversion_clear H31; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Change (ftrans ([_:name](p _ x) y) ([_:name](Out n n0) y) (x1 x y)); Apply FTR_L3 with x0; Auto.
Rewrite <- H18; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite <- H18; Assumption.
Inversion_clear H16; Apply notin_nu; Intros.
Cut (notin y (nu [z0:name](p z z0))); [Intro | Auto].
Inversion_clear H31; Auto.
Inversion_clear H19; Apply notin_nu; Intros.
Cut (notin y (nu [z0:name](x1 z0 z))); [Intro | Auto].
Inversion_clear H31; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Inversion_clear H23; Apply notin_nu; Intros; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite H18 in H7; Cut p2=([_:name](nu (x1 _))); [Intro | Apply proc_ext with x; Auto].
Rewrite H16; Cut (nu [_:name](nu (x1 _)))=(nu [_:name](nu [u:name](x1 _ u))); [Intro | Apply ETA_EQ2].
Rewrite H19; Change (StBisim (nu [y:name](nu [z:name]([u:name][v:name](x1 v u) y z))) (nu [_:name](nu [u:name]([u':name][v:name](x1 v u') u _)))); Apply NU_COMM.
Apply f_act_notin_Out; Auto.
Apply f_act_notin_Out; Auto.

(* INPUT action case *)

Inversion_clear H.
Elim (unsat (par (nu [y0:name](nu [z:name](p y0 z))) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1; Inversion_clear H2.
Cut (btrans (nu [z:name](p x0 z)) (In x) (p2 x0)); [Intro | Apply H0; Auto].
Inversion H2.
Elim (unsat (par (nu [z:name](p x0 z)) (nu [z:name](nu (p3 z)))) (cons x l0)); Intros.
Inversion_clear H9.
Inversion_clear H10; Inversion_clear H11.
Cut (btrans (p x0 x1) (In x) (p3 x1)); [Intro | Apply H8; Auto].
Elim (ho2_proc_exp p3 x0); Intros.
Inversion_clear H14.
Exists [_:name](nu [z:name](nu [y:name](x2 y z _))); Split.
Change (btrans (nu [z:name](nu [y:name](p y z))) (In x) [_:name](nu [z:name]([u:name][v:name](nu [y:name](x2 y u v)) z _))); Apply bRES with (cons x0 (cons x1 empty)); Intros.
Inversion_clear H18; Inversion_clear H19.
Inversion_clear H21.
Clear H22; Change  (btrans (nu [y0:name](p y0 y)) (In x) [v:name](nu [y0:name](([_:name](x2 _ y) y0) v))); Apply bRES with (cons x (cons x0 empty)); Intros.
Inversion_clear H23; Inversion_clear H24.
Inversion_clear H26.
Clear H27; Change (btrans ([_:name](p _ y) y0) ([_:name](In x) y0) ([_:name](x2 _ y) y0)); Apply BTR_L3 with x0; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x0 (nu [z0:name](p z z0))); [Intro | Auto].
Inversion_clear H27; Auto.
Inversion_clear H15; Apply notin_nu; Intros.
Cut (notin x0 (nu [z0:name](nu (x2 z z0)))); [Intro | Auto].
Inversion_clear H27; Apply notin_nu; Intros.
Cut (notin x0 (nu (x2 z y))); [Intro | Auto].
Inversion_clear H29; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Change (btrans (p x0 y) ([_:name](In x) y) (x2 x0 y)); Apply BTR_L3 with x1; Auto.
Inversion_clear H9; Apply notin_nu; Intros; Auto.
Rewrite <- H16; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Rewrite <- H16; Assumption.
Inversion_clear H14; Apply notin_nu; Intros; Cut (notin y (nu [y0:name](p y0 z))); [Intro | Auto].
Inversion_clear H27; Auto.
Inversion_clear H17; Apply notin_nu; Intros.
Cut (notin y (nu [v:name](nu [y0:name](x2 y0 z v)))); [Intro | Auto].
Inversion_clear H27; Apply notin_nu; Intros.
Cut (notin y (nu [y0:name](x2 y0 z z0))); [Intro | Auto].
Inversion_clear H29; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Rewrite H16 in H6; Cut ([_:name][y:name](nu [z:name](x2 _ z y)))=p2; [Intro | Apply ho_proc_ext with x0; Auto].
Rewrite <- H14; Intro; Change (StBisim (nu [z:name](nu [z0:name]([u:name][v:name](x2 u v y) z z0))) (nu [z:name](nu [y0:name]([u:name][v:name](x2 u v y) y0 z)))); Apply NU_COMM.
Inversion_clear H15; Do 3 (Apply notin_nu; Intros).
Cut (notin x0 (nu [z0:name](nu (x2 z z0)))); [Intro | Auto].
Inversion_clear H19.
Cut (notin x0 (nu (x2 z z1))); [Intro | Auto].
Inversion_clear H19; Auto.
Apply b_act_notin_In; Auto.
Apply b_act_notin_In; Auto.

(* The way back: symmetric case *)

Inversion_clear H.
Elim (unsat (par (nu [z:name](nu [y0:name](p y0 z))) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1; Inversion_clear H2.
Cut (btrans (nu [y0:name](p y0 x0)) (In x) (p2 x0)); [Intro | Apply H0; Auto].
Inversion H2.
Elim (unsat (par (nu [y0:name](p y0 x0)) (nu [z:name](nu (p0 z)))) (cons x l0)); Intros.
Inversion_clear H9.
Inversion_clear H10; Inversion_clear H11.
Cut (btrans (p x1 x0) (In x) (p0 x1)); [Intro | Apply H8; Auto].
Elim (ho2_proc_exp p0 x0); Intros.
Inversion_clear H14; Exists [_:name](nu [z:name](nu [y:name](x2 y z _))); Split.
Change (btrans (nu [z:name](nu [y:name](p z y))) (In x) [_:name](nu [z:name]([u:name][v:name](nu [y:name](x2 y u v)) z _))); Apply bRES with (cons x0 (cons x1 empty)); Intros.
Inversion_clear H18; Inversion_clear H19.
Inversion_clear H21.
Clear H22; Change  (btrans (nu [y0:name](p y y0)) (In x) [v:name](nu [y0:name](([_:name](x2 _ y) y0) v))); Apply bRES with (cons x (cons x0 empty)); Intros.
Inversion_clear H23; Inversion_clear H24.
Inversion_clear H26.
Clear H27; Change (btrans (p y y0) ([_:name](In x) y0) ([_:name](x2 _ y) y0)); Apply BTR_L3 with x0; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x0 (nu [y0:name](p y0 z))); [Intro | Auto].
Inversion_clear H27; Auto.
Inversion_clear H15; Apply notin_nu; Intros.
Cut (notin x0 (nu [z0:name](nu (x2 z z0)))); [Intro | Auto].
Inversion_clear H27; Apply notin_nu; Intros.
Cut (notin x0 (nu (x2 z y))); [Intro | Auto].
Inversion_clear H29; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Change (btrans ([_:name](p _ x0) y) ([_:name](In x) y) (x2 x0 y)); Apply BTR_L3 with x1; Auto.
Rewrite <- H16; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Rewrite <- H16; Assumption.
Inversion_clear H14; Apply notin_nu; Intros; Cut (notin y (nu [y0:name](p z y0))); [Intro | Auto].
Inversion_clear H27; Auto.
Inversion_clear H17; Apply notin_nu; Intros.
Cut (notin y (nu [v:name](nu [y0:name](x2 y0 z v)))); [Intro | Auto].
Inversion_clear H27; Apply notin_nu; Intros.
Cut (notin y (nu [y0:name](x2 y0 z z0))); [Intro | Auto].
Inversion_clear H29; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Inversion_clear H21; Apply notin_nu; Intros; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Rewrite H16 in H6; Cut p2=([_:name][y:name](nu [z:name](x2 _ z y))); [Intro | Apply ho_proc_ext with x0; Auto].
Rewrite H14; Intro; Change (StBisim (nu [z:name](nu [y0:name]([u:name][v:name](x2 v u y) z y0))) (nu [z:name](nu [z0:name]([u:name][v:name](x2 v u y) z0 z)))); Apply NU_COMM.
Inversion_clear H15; Do 3 (Apply notin_nu; Intros).
Cut (notin x0 (nu [z0:name](nu (x2 z z0)))); [Intro | Auto].
Inversion_clear H19.
Cut (notin x0 (nu (x2 z z1))); [Intro | Auto].
Inversion_clear H19; Auto.
Apply b_act_notin_In; Auto.
Apply b_act_notin_In; Auto.

(* BOUND OUTPUT action case *)

Inversion_clear H.
Elim (unsat (par (nu [y0:name](nu [z:name](p y0 z))) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1; Inversion_clear H2.
Cut (btrans (nu [z:name](p x0 z)) (bOut x) (p2 x0)); [Intro | Apply H0; Auto].
Inversion H2.
Elim (unsat (par (nu [z:name](p x0 z)) (nu [z:name](nu (p3 z)))) (cons x l0)); Intros.
Inversion_clear H9.
Inversion_clear H10; Inversion_clear H11.
Cut (btrans (p x0 x1) (bOut x) (p3 x1)); [Intro | Apply H8; Auto].
Elim (ho2_proc_exp p3 x0); Intros.
Inversion_clear H14; Exists [_:name](nu [z:name](nu [y:name](x2 y z _))); Split.
Change (btrans (nu [z:name](nu [y:name](p y z))) (bOut x) [_:name](nu [z:name]([u:name][v:name](nu [y:name](x2 y u v)) z _))); Apply bRES with (cons x0 (cons x1 empty)); Intros.
Inversion_clear H18; Inversion_clear H19.
Inversion_clear H21.
Clear H22; Change (btrans (nu [y0:name](p y0 y)) (bOut x) [v:name](nu [y0:name](([_:name](x2 _ y) y0) v))); Apply bRES with (cons x (cons x0 empty)); Intros.
Inversion_clear H23.
Inversion_clear H24.
Inversion_clear H26.
Clear H27; Change (btrans ([_:name](p _ y) y0) ([_:name](bOut x) y0) ([_:name](x2 _ y) y0)); Apply BTR_L3 with x0; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x0 (nu [z0:name](p z z0))); [Intro | Auto].
Inversion_clear H27; Auto.
Inversion_clear H15; Apply notin_nu; Intros.
Cut (notin x0 (nu [z0:name](nu (x2 z z0)))); [Intro | Auto].
Inversion_clear H27; Apply notin_nu; Intros.
Cut (notin x0 (nu (x2 z y))); [Intro | Auto].
Inversion_clear H29; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Change (btrans (p x0 y) ([_:name](bOut x) y) (x2 x0 y)); Apply BTR_L3 with x1; Auto.
Inversion_clear H9; Apply notin_nu; Intros; Auto.
Rewrite <- H16; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Rewrite <- H16; Assumption.
Inversion_clear H14; Apply notin_nu; Intros; Cut (notin y (nu [y0:name](p y0 z))); [Intro | Auto].
Inversion_clear H27; Auto.
Inversion_clear H17; Apply notin_nu; Intros.
Cut (notin y (nu [v:name](nu [y0:name](x2 y0 z v)))); [Intro | Auto].
Inversion_clear H27; Apply notin_nu; Intros.
Cut (notin y (nu [y0:name](x2 y0 z z0))); [Intro | Auto].
Inversion_clear H29; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Rewrite H16 in H6; Cut p2=([_:name][y:name](nu [z:name](x2 _ z y))); [Intro | Apply ho_proc_ext with x0; Auto].
Rewrite H14; Intros; Change (StBisim (nu [z:name](nu [z0:name]([u:name][v:name](x2 u v y) z z0))) (nu [z:name](nu [y0:name]([u:name][v:name](x2 u v y) y0 z)))); Apply NU_COMM.
Inversion_clear H15; Do 3 (Apply notin_nu; Intros).
Cut (notin x0 (nu [z0:name](nu (x2 z z0)))); [Intro | Auto].
Inversion_clear H19.
Cut (notin x0 (nu (x2 z z1))); [Intro | Auto].
Inversion_clear H19; Auto.
Apply b_act_notin_bOut; Auto.

Elim (unsat (par (nu [z:name](p x0 z)) (nu (p2 x0))) (cons x l0)); Intros.
Inversion_clear H9.
Inversion_clear H10; Inversion_clear H11.
Cut (ftrans (p x0 x2) (Out x x2) (p2 x0 x2)); [Intro | Apply H7; Auto].
Exists [_:name](nu [z:name](p2 z _)); Split.
Apply OPEN with (cons x (cons x0 empty)); Intros.
Inversion_clear H17.
Inversion_clear H19.
Clear H20; Apply fRES with (cons x0 (cons x1 empty)); Intros.
Inversion_clear H21; Inversion_clear H22.
Inversion_clear H24.
Clear H26; Change (ftrans ([_:name](p _ y) y0) ([_:name](Out x y) y0) ([_:name](p2 _ y) y0)); Apply FTR_L3 with x0; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x0 (nu [z0:name](p z z0))); [Intro | Auto].
Inversion_clear H26; Auto.
Inversion_clear H3; Apply notin_nu; Intros.
Cut (notin x0 (nu (p2 z))); [Intro | Auto].
Inversion_clear H26; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Change (ftrans (p x0 y) (Out x y) (p2 x0 y)); Apply FTR_L3 with x2; Auto.
Inversion_clear H9; Apply notin_nu; Intros; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Inversion_clear H14; Apply notin_nu; Intros; Cut (notin y (nu [y0:name](p y0 z))); [Intro | Auto].
Inversion_clear H26; Auto.
Inversion_clear H15; Apply notin_nu; Intros.
Cut (notin y (nu [z0:name](p2 z0 z))); [Intro | Auto].
Inversion_clear H26; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Intros; Apply REF.
Apply b_act_notin_bOut; Auto.

Elim (unsat (par (nu [y0:name](nu [z:name](p y0 z))) (nu p1)) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1; Inversion_clear H2.
Cut (ftrans (nu [z:name](p x0 z)) (Out x x0) (p1 x0)); [Intro | Apply H0; Auto].
Inversion H2.
Elim (unsat (par (nu [z:name](p x0 z)) (nu p2)) (cons x (cons x0 l0))); Intros.
Inversion_clear H9.
Inversion_clear H10; Inversion_clear H11.
Inversion_clear H13.
Cut (ftrans (p x0 x1) (Out x x0) (p2 x1)); [Intro | Apply H8; Auto].
Elim (ho_proc_exp p2 x0); Intros.
Inversion_clear H15; Exists [y:name](nu [z:name](x2 y z)); Split.
Change (btrans (nu [z:name](nu [y:name](p y z))) (bOut x) [y:name](nu [z:name]([u:name][_:name](x2 _ u) z y))); Apply bRES with (cons x0 (cons x1 empty)); Intros.
Inversion_clear H19; Inversion_clear H20.
Inversion_clear H22.
Clear H23; Apply OPEN with (cons x0 empty); Intros.
Inversion_clear H25.
Clear H27; Change (ftrans ([_:name](p _ y) y0) (Out x y0) ([_:name](x2 _ y) y0)); Apply FTR_L3 with x0; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x0 (nu [z0:name](p z z0))); [Intro | Auto].
Inversion_clear H27; Auto.
Inversion_clear H16; Apply notin_nu; Intros.
Cut (notin x0 (nu (x2 z))); [Intro | Auto].
Inversion_clear H27; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Change (ftrans (p x0 y) ([_:name](Out x x0) y) (x2 x0 y)); Apply FTR_L3 with x1; Auto.
Inversion_clear H9; Apply notin_nu; Intros; Auto.
Rewrite <- H17; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite <- H17; Assumption.
Inversion_clear H15; Apply notin_nu; Intros; Cut (notin y (nu [y0:name](p y0 z))); [Intro | Auto].
Inversion_clear H27; Auto.
Inversion_clear H18; Apply notin_nu; Intros.
Cut (notin y (nu [_:name](x2 _ z))); [Intro | Auto].
Inversion_clear H27; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite H17 in H6; Cut p1=([_:name](nu (x2 _))); [Intro | Apply proc_ext with x0; Auto].
Rewrite H15; Intros; Cut (nu (x2 y))=(nu [_:name](x2 y _)); [Intro | Apply ETA_EQ]; Rewrite H20; Apply REF.
Apply f_act_notin_Out; Auto.

(* The way back: symmetric case *)

Inversion_clear H.
Elim (unsat (par (nu [z:name](nu [y0:name](p y0 z))) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1; Inversion_clear H2.
Cut (btrans (nu [y0:name](p y0 x0)) (bOut x) (p2 x0)); [Intro | Apply H0; Auto].
Inversion H2.
Elim (unsat (par (nu [y0:name](p y0 x0)) (nu [z:name](nu (p0 z)))) (cons x l0)); Intros.
Inversion_clear H9.
Inversion_clear H10; Inversion_clear H11.
Cut (btrans (p x1 x0) (bOut x) (p0 x1)); [Intro | Apply H8; Auto].
Elim (ho2_proc_exp p0 x0); Intros.
Inversion_clear H14; Exists [_:name](nu [z:name](nu [y:name](x2 y z _))); Split.
Change (btrans (nu [y:name](nu [z:name](p y z))) (bOut x) [_:name](nu [z:name]([u:name][v:name](nu [y:name](x2 y u v)) z _))); Apply bRES with (cons x0 (cons x1 empty)); Intros.
Inversion_clear H18; Inversion_clear H19.
Inversion_clear H21.
Clear H22; Change (btrans (nu [z:name](p y z)) (bOut x) [v:name](nu [y0:name](([_:name](x2 _ y) y0) v))); Apply bRES with (cons x (cons x0 empty)); Intros.
Inversion_clear H23; Inversion_clear H24.
Inversion_clear H26.
Clear H27; Change (btrans (p y y0) ([_:name](bOut x) y0) ([_:name](x2 _ y) y0)); Apply BTR_L3 with x0; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x0 (nu [y0:name](p y0 z))); [Intro | Auto].
Inversion_clear H27; Auto.
Inversion_clear H15; Apply notin_nu; Intros.
Cut (notin x0 (nu [z0:name](nu (x2 z z0)))); [Intro | Auto].
Inversion_clear H27; Apply notin_nu; Intros.
Cut (notin x0 (nu (x2 z y))); [Intro | Auto].
Inversion_clear H29; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Change (btrans ([_:name](p _ x0) y) ([_:name](bOut x) y) (x2 x0 y)); Apply BTR_L3 with x1; Auto.
Rewrite <- H16; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Rewrite <- H16; Assumption.
Inversion_clear H14; Apply notin_nu; Intros; Cut (notin y (nu [z0:name](p z z0))); [Intro | Auto].
Inversion_clear H27; Auto.
Inversion_clear H17; Apply notin_nu; Intros.
Cut (notin y (nu [v:name](nu [y0:name](x2 y0 z v)))); [Intro | Auto].
Inversion_clear H27; Apply notin_nu; Intros.
Cut (notin y (nu [y0:name](x2 y0 z z0))); [Intro | Auto].
Inversion_clear H29; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Inversion_clear H21; Apply notin_nu; Intros; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Rewrite H16 in H6; Cut p2=([_:name][y:name](nu [z:name](x2 _ z y))); [Intro | Apply ho_proc_ext with x0; Auto].
Rewrite H14; Intros; Change (StBisim (nu [z:name](nu [y0:name]([u:name][v:name](x2 v u y) z y0))) (nu [y0:name](nu [z:name]([u:name][v:name](x2 v u y) z y0)))); Apply NU_COMM.
Inversion_clear H15; Do 3 (Apply notin_nu; Intros).
Cut (notin x0 (nu [z0:name](nu (x2 z z0)))); [Intro | Auto].
Inversion_clear H19; Cut (notin x0 (nu (x2 z z1))); [Intro | Auto].
Inversion_clear H19; Auto.
Apply b_act_notin_bOut; Auto.

Elim (unsat (par (nu [y0:name](p y0 x0)) (nu (p2 x0))) (cons x l0)); Intros.
Inversion_clear H9.
Inversion_clear H10; Inversion_clear H11.
Cut (ftrans (p x2 x0) (Out x x2) (p2 x0 x2)); [Intro | Apply H7; Auto].
Exists [_:name](nu [z:name](p2 z _)); Split.
Apply OPEN with (cons x (cons x0 empty)); Intros.
Inversion_clear H17.
Inversion_clear H19.
Clear H20; Apply fRES with (cons x0 (cons x1 empty)); Intros.
Inversion_clear H21; Inversion_clear H22.
Inversion_clear H24.
Clear H26; Change (ftrans (p y y0) ([_:name](Out x y) y0) ([_:name](p2 _ y) y0)); Apply FTR_L3 with x0; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x0 (nu [y0:name](p y0 z))); [Intro | Auto].
Inversion_clear H26; Auto.
Inversion_clear H3; Apply notin_nu; Intros.
Cut (notin x0 (nu (p2 z))); [Intro | Auto].
Inversion_clear H26; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Change (ftrans ([_:name](p _ x0) y) (Out x y) (p2 x0 y)); Apply FTR_L3 with x2; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Inversion_clear H14; Apply notin_nu; Intros; Cut (notin y (nu [z0:name](p z z0))); [Intro | Auto].
Inversion_clear H26; Auto.
Inversion_clear H15; Apply notin_nu; Intros.
Cut (notin y (nu [z0:name](p2 z0 z))); [Intro | Auto].
Inversion_clear H26; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Inversion_clear H19; Apply notin_nu; Intros; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Intros; Apply REF.
Apply b_act_notin_bOut; Auto.

Elim (unsat (par (nu [z:name](nu [y0:name](p y0 z))) (nu q1)) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1; Inversion_clear H2.
Cut (ftrans (nu [y0:name](p y0 x0)) (Out x x0) (q1 x0)); [Intro | Apply H0; Auto].
Inversion H2.
Elim (unsat (par (nu [y0:name](p y0 x0)) (nu p2)) (cons x (cons x0 l0))); Intros.
Inversion_clear H9.
Inversion_clear H10; Inversion_clear H11.
Inversion_clear H13.
Cut (ftrans (p x1 x0) (Out x x0) (p2 x1)); [Intro | Apply H8; Auto].
Elim (ho_proc_exp p2 x0); Intros.
Inversion_clear H15; Exists [y:name](nu [z:name](x2 y z)); Split.
Change (btrans (nu [y:name](nu [z:name](p y z))) (bOut x) [y:name](nu [z:name]([u:name][_:name](x2 _ u) z y))); Apply bRES with (cons x0 (cons x1 empty)); Intros.
Inversion_clear H19; Inversion_clear H20.
Inversion_clear H22.
Clear H23; Apply OPEN with (cons x0 empty); Intros.
Inversion_clear H25.
Clear H27; Change (ftrans (p y y0) (Out x y0) ([_:name](x2 _ y) y0)); Apply FTR_L3 with x0; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x0 (nu [y0:name](p y0 z))); [Intro | Auto].
Inversion_clear H27; Auto.
Inversion_clear H16; Apply notin_nu; Intros.
Cut (notin x0 (nu (x2 z))); [Intro | Auto].
Inversion_clear H27; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Change (ftrans ([_:name](p _ x0) y) ([_:name](Out x x0) y) (x2 x0 y)); Apply FTR_L3 with x1; Auto.
Rewrite <- H17; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite <- H17; Assumption.
Inversion_clear H15; Apply notin_nu; Intros; Cut (notin y (nu [z0:name](p z z0))); [Intro | Auto].
Inversion_clear H27; Auto.
Inversion_clear H18; Apply notin_nu; Intros.
Cut (notin y (nu [_:name](x2 _ z))); [Intro | Auto].
Inversion_clear H27; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Inversion_clear H22; Apply notin_nu; Intros; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite H17 in H6; Cut q1=([_:name](nu (x2 _))); [Intro | Apply proc_ext with x0; Auto].
Rewrite H15; Intros; Cut (nu (x2 y))=(nu [_:name](x2 y _)); [Intro | Apply ETA_EQ]; Rewrite H20; Apply REF.
Apply f_act_notin_Out; Auto.

Qed.

Lemma NU_SUM: (p,q:name->proc)(StBisim (nu [y:name](sum (p y) (q y))) (sum (nu p) (nu q))).

Proof. (* This lemma depends on Lemmata FTR_L3, BTR_L3, REF. *)

Intros; Apply sb; Do 3 Try (Split; Intros).

(* Free actions case *)

Generalize H; Clear H; Elim a; Intros.

Inversion_clear H.
Elim (unsat (par (nu [y0:name](sum (p y0) (q y0))) (nu p2)) l); Intros.
Inversion_clear H.
Inversion_clear H1.
Cut (ftrans (sum (p x) (q x)) tau (p2 x)); [Intro | Apply H0; Auto].
Inversion_clear H1.
Exists (nu p2); Split.
Apply fSUM1; Apply fRES with l; Intros.
Change (ftrans (p y) ([_:name]tau y) (p2 y)); Apply FTR_L3 with x; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H9; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply REF.
Exists (nu p2); Split.
Apply fSUM2; Apply fRES with l; Intros.
Change (ftrans (q y) ([_:name]tau y) (p2 y)); Apply FTR_L3 with x; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H9; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply REF.
Apply f_act_notin_tau.

Inversion_clear H.
Elim (unsat (par (nu [y0:name](sum (p y0) (q y0))) (nu p2)) (cons n (cons n0 l))); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H4.
Cut (ftrans (sum (p x) (q x)) (Out n n0) (p2 x)); [Intro | Apply H0; Auto].
Inversion_clear H4.
Exists (nu p2); Split.
Apply fSUM1; Apply fRES with l; Intros.
Inversion_clear H9.
Change (ftrans (p y) ([_:name](Out n n0) y) (p2 y)); Apply FTR_L3 with x; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H12; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Apply REF.
Exists (nu p2); Split.
Apply fSUM2; Apply fRES with l; Intros.
Inversion_clear H9.
Change (ftrans (q y) ([_:name](Out n n0) y) (p2 y)); Apply FTR_L3 with x; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H12; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Apply REF.
Apply f_act_notin_Out; Auto.

(* Symmetric case *)

Generalize H; Clear H; Elim a; Intros.

Inversion_clear H.
Exists q1; Split.
Inversion_clear H0.
Apply fRES with l; Intros.
Clear H3; Apply fSUM1; Apply H; Auto.
Inversion_clear H0; Apply notin_nu; Intros.
Cut (notin y (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H4; Assumption.
Apply f_act_notin_tau.
Apply REF.
Exists q1; Split.
Inversion_clear H0.
Apply fRES with l; Intros.
Clear H3; Apply fSUM2; Apply H; Auto.
Inversion_clear H0; Apply notin_nu; Intros.
Cut (notin y (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H4; Assumption.
Apply f_act_notin_tau.
Apply REF.

Inversion_clear H.
Exists q1; Split.
Inversion_clear H0.
Apply fRES with l; Intros.
Inversion_clear H3; Apply fSUM1; Apply H; Auto.
Inversion_clear H0; Apply notin_nu; Intros.
Cut (notin y (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H6; Assumption.
Apply f_act_notin_Out; Auto.
Apply REF.
Exists q1; Split.
Inversion_clear H0.
Apply fRES with l; Intros.
Inversion_clear H3; Apply fSUM2; Apply H; Auto.
Inversion_clear H0; Apply notin_nu; Intros.
Cut (notin y (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H6; Assumption.
Apply f_act_notin_Out; Auto.
Apply REF.

(* Input actions case *)

Inversion_clear H.
Elim (unsat (par (nu [y0:name](sum (p y0) (q y0))) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (sum (p x0) (q x0)) (In x) (p2 x0)); [Intro | Apply H0; Auto].
Inversion_clear H2.
Exists [u:name](nu [v:name](p2 v u)); Split.
Apply bSUM1; Apply bRES with l; Intros.
Inversion_clear H7.
Change (btrans (p y) ([_:name](In x) y) (p2 y)); Apply BTR_L3 with x0; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x0 (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H10; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Intro; Apply REF.

Exists [u:name](nu [v:name](p2 v u)); Split.
Apply bSUM2; Apply bRES with l; Intros.
Inversion_clear H7.
Change (btrans (q y) ([_:name](In x) y) (p2 y)); Apply BTR_L3 with x0; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x0 (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H10; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Intro; Apply REF.
Apply b_act_notin_In; Auto.

(* Symmetric case *)

Inversion_clear H.
Exists q1; Split.
Inversion_clear H0.
Apply bRES with l; Intros.
Apply bSUM1; Apply H; Auto.
Inversion_clear H0; Apply notin_nu; Intros.
Cut (notin y (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H5; Assumption.
Intro; Apply REF.
Exists q1; Split.
Inversion_clear H0.
Apply bRES with l; Intros.
Apply bSUM2; Apply H; Auto.
Inversion_clear H0; Apply notin_nu; Intros.
Cut (notin y (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H5; Assumption.
Intro; Apply REF.

(* Bound output actions case *)

Inversion_clear H.
Elim (unsat (par (nu [y0:name](sum (p y0) (q y0))) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (sum (p x0) (q x0)) (bOut x) (p2 x0)); [Intro | Apply H0; Auto].
Inversion_clear H2.
Exists [u:name](nu [v:name](p2 v u)); Split.
Apply bSUM1; Apply bRES with l; Intros.
Inversion_clear H7.
Change (btrans (p y) ([_:name](bOut x) y) (p2 y)); Apply BTR_L3 with x0; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x0 (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H10; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Intros; Apply REF.

Exists [u:name](nu [v:name](p2 v u)); Split.
Apply bSUM2; Apply bRES with l; Intros.
Inversion_clear H7.
Change (btrans (q y) ([_:name](bOut x) y) (p2 y)); Apply BTR_L3 with x0; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x0 (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H10; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Intros; Apply REF.
Apply b_act_notin_bOut; Auto.

Elim (unsat (par (nu [y0:name](sum (p y0) (q y0))) (nu p1)) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (ftrans (sum (p x0) (q x0)) (Out x x0) (p1 x0)); [Intro | Apply H0; Auto].
Inversion_clear H2.
Exists p1; Split.
Apply bSUM1; Apply OPEN with l; Intros.
Change (ftrans (p y) (Out x y) (p1 y)); Apply FTR_L3 with x0; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x0 (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H10; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Intros; Apply REF.

Exists p1; Split.
Apply bSUM2; Apply OPEN with l; Intros.
Change (ftrans (q y) (Out x y) (p1 y)); Apply FTR_L3 with x0; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x0 (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H10; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Intros; Apply REF.

(* Symmetric case *)

Inversion_clear H.
Exists q1; Split.
Inversion_clear H0.
Apply bRES with l; Intros.
Apply bSUM1; Apply H; Auto.
Inversion_clear H0; Apply notin_nu; Intros.
Cut (notin y (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H5; Assumption.
Apply OPEN with l; Intros.
Apply fSUM1; Apply H; Auto.
Inversion_clear H0; Apply notin_nu; Intros.
Cut (notin y (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H5; Assumption.
Intros; Apply REF.
Exists q1; Split.
Inversion_clear H0.
Apply bRES with l; Intros.
Apply bSUM2; Apply H; Auto.
Inversion_clear H0; Apply notin_nu; Intros.
Cut (notin y (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H5; Assumption.
Apply OPEN with l; Intros.
Apply fSUM2; Apply H; Auto.
Inversion_clear H0; Apply notin_nu; Intros.
Cut (notin y (sum (p z) (q z))); [Intro | Auto].
Inversion_clear H5; Assumption.
Intros; Apply REF.

Qed.

Lemma NU_TAU_PREF: (p:name->proc)(StBisim (nu [x:name](tau_pref (p x))) (tau_pref (nu p))).

Proof. (* This lemma depends on Lemmata REF, L6_Light, NU_S. *)

Intros; Apply sb; Do 3 Try (Split; Intros).

(* Free actions case *)

Generalize H; Clear H; Elim a; Intros.

Inversion_clear H.
Elim (unsat (par (nu [x:name](tau_pref (p x))) (nu p2)) l); Intros.
Inversion_clear H.
Inversion_clear H1.
Cut (ftrans (tau_pref (p x)) tau (p2 x)); [Intro | Apply H0; Auto].
Inversion H1.
Exists (nu p); Split.
Apply TAU.
Apply NU_S; Intros.
Apply L6_Light with x; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x (tau_pref (p z))); [Intro | Auto].
Inversion_clear H10; Assumption.
Rewrite H7; Apply REF.
Apply f_act_notin_tau.

Inversion_clear H.
Elim (unsat (par (nu [x:name](tau_pref (p x))) (nu p2)) (cons n (cons n0 l))); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H4.
Cut (ftrans (tau_pref (p x)) (Out n n0) (p2 x)); [Intro | Apply H0; Auto].
Inversion_clear H4.
Apply f_act_notin_Out; Auto.

(* Symmetric case *)

Inversion H.
Exists (nu p); Split.
Apply fRES with empty; Intros.
Clear H5 H6.
Apply TAU.
Apply REF.

(* Input actions case *)

Inversion_clear H.
Elim (unsat (par (nu [x0:name](tau_pref (p x0))) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (tau_pref (p x0)) (In x) (p2 x0)); [Intro | Apply H0; Auto].
Inversion_clear H2.
Apply b_act_notin_In; Auto.

Inversion_clear H.

(* Bound output actions case *)

Inversion_clear H.
Elim (unsat (par (nu [x0:name](tau_pref (p x0))) (nu [z:name](nu (p2 z)))) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (tau_pref (p x0)) (bOut x) (p2 x0)); [Intro | Apply H0; Auto].
Inversion_clear H2.
Apply b_act_notin_bOut; Auto.

Elim (unsat (par (nu [x0:name](tau_pref (p x0))) (nu p1)) (cons x l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (ftrans (tau_pref (p x0)) (Out x x0) (p1 x0)); [Intro | Apply H0; Auto].
Inversion_clear H2.

Inversion_clear H.

Qed.

Lemma NU_OUT_PREF: (p:name->proc)(x,y:name)
                   (StBisim (nu [z:name](out_pref x y (p z))) (out_pref x y (nu p))).

Proof. (* This lemma depends on Lemmata REF, L6_Light, NU_S. *)

Intros; Apply sb; Do 3 Try (Split; Intros).

(* Free actions case *)

Generalize H; Clear H; Elim a; Intros.

Inversion_clear H.
Elim (unsat (par (nu [z:name](out_pref x y (p z))) (nu p2)) l); Intros.
Inversion_clear H.
Inversion_clear H1.
Cut (ftrans (out_pref x y (p x0)) tau (p2 x0)); [Intro | Apply H0; Auto].
Inversion_clear H1.
Apply f_act_notin_tau.

Inversion_clear H.
Elim (unsat (par (nu [z:name](out_pref x y (p z))) (nu p2)) (cons n (cons n0 l))); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H4.
Cut (ftrans (out_pref x y (p x0)) (Out n n0) (p2 x0)); [Intro | Apply H0; Auto].
Inversion H4.
Exists (nu p); Split.
Apply OUT.
Apply NU_S; Intros.
Apply L6_Light with x0; Auto.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x0 (out_pref x y (p z))); [Intro | Auto].
Inversion_clear H15; Assumption.
Rewrite H12; Apply REF.
Apply f_act_notin_Out; Auto.

Inversion_clear H.
Exists (nu p); Split.
Apply fRES with empty; Intros; Apply OUT.
Apply REF.

(* Input actions case *)

Inversion_clear H.
Elim (unsat (par (nu [z:name](out_pref x y (p z))) (nu [z:name](nu (p2 z)))) (cons x0 l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (out_pref x y (p x1)) (In x0) (p2 x1)); [Intro | Apply H0; Auto].
Inversion_clear H2.
Apply b_act_notin_In; Auto.

Inversion_clear H.

(* Bound output actions case *)

Inversion_clear H.
Elim (unsat (par (nu [z:name](out_pref x y (p z))) (nu [z:name](nu (p2 z)))) (cons x0 l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (out_pref x y (p x1)) (bOut x0) (p2 x1)); [Intro | Apply H0; Auto].
Inversion_clear H2.
Apply b_act_notin_bOut; Auto.

Elim (unsat (par (nu [z:name](out_pref x y (p z))) (nu p1)) (cons x0 l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (ftrans (out_pref x y (p x1)) (Out x0 x1) (p1 x1)); [Intro | Apply H0; Auto].
Inversion H2.
Absurd y=x1; Auto.
Inversion_clear H.
Cut (notin x1 (out_pref x y (p x0))); [Intro | Auto].
Inversion_clear H; Auto.

Inversion_clear H.

Qed.

Lemma NU_IN_PREF: (p:name->name->proc)(x:name)
                  (StBisim (nu [y:name](in_pref x (p y))) (in_pref x [z:name](nu [y:name](p y z)))).

Proof. (* This lemma depends on Lemma REF. *)

Intros; Apply sb; Do 3 Try (Split; Intros).

(* Free actions case *)

Generalize H; Clear H; Elim a; Intros.

Inversion_clear H.
Elim (unsat (par (nu [y0:name](in_pref x (p y0))) (nu p2)) l); Intros.
Inversion_clear H.
Inversion_clear H1.
Cut (ftrans (in_pref x (p x0)) tau (p2 x0)); [Intro | Apply H0; Auto].
Inversion_clear H1.
Apply f_act_notin_tau.

Inversion_clear H.
Elim (unsat (par (nu [y0:name](in_pref x (p y0))) (nu p2)) (cons n (cons n0 l))); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Inversion_clear H4.
Cut (ftrans (in_pref x (p x0)) (Out n n0) (p2 x0)); [Intro | Apply H0; Auto].
Inversion_clear H4.
Apply f_act_notin_Out; Auto.

Inversion_clear H.

(* Input actions case *)

Inversion_clear H.
Elim (unsat (par (nu [y0:name](in_pref x (p y0))) (nu [z:name](nu (p2 z)))) (cons x0 l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (in_pref x (p x1)) (In x0) (p2 x1)); [Intro | Apply H0; Auto].
Inversion H2.
Exists [z:name](nu [y:name](p y z)); Split.
Apply IN.
Cut p=p2; [Intro | Apply ho_proc_ext with x1; Auto].
Intro; Rewrite <- H5; Apply REF.
Inversion_clear H; Apply notin_nu; Intros.
Cut (notin x1 (in_pref x (p z))); [Intro | Auto]; Apply notin_nu; Intros.
Inversion_clear H10; Auto.
Apply b_act_notin_In; Auto.

Inversion H.
Exists ([z:name](nu [y:name](p y z))); Split.
Apply bRES with empty; Intros.
Apply IN.
Intro; Apply REF.

(* Bound output actions case *)

Inversion_clear H.
Elim (unsat (par (nu [y0:name](in_pref x (p y0))) (nu [z:name](nu (p2 z)))) (cons x0 l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (btrans (in_pref x (p x1)) (bOut x0) (p2 x1)); [Intro | Apply H0; Auto].
Inversion_clear H2.
Apply b_act_notin_bOut; Auto.

Elim (unsat (par (nu [y0:name](in_pref x (p y0))) (nu p1)) (cons x0 l)); Intros.
Inversion_clear H.
Inversion_clear H1.
Inversion_clear H2.
Cut (ftrans (in_pref x (p x1)) (Out x0 x1) (p1 x1)); [Intro | Apply H0; Auto].
Inversion_clear H2.

Inversion_clear H.

Qed.

End restriction_laws1.

Section parallel.

Lemma ID_PAR: (p:proc)(StBisim (par p nil) p).

Proof. (* This lemma is proved from scratch: it is a coinductive proof. *)

Cofix; Intro; Apply sb; Do 3 Try (Split; Intros).

Inversion_clear H.
Exists p3; Split; [Assumption | Apply ID_PAR].
Inversion_clear H0.
Inversion_clear H1.
Inversion_clear H1.
Inversion_clear H1.
Inversion_clear H1.
Exists (par q1 nil); Split; [Apply fPAR1; Assumption | Apply ID_PAR].

Inversion_clear H.
Exists p3; Split; [Assumption | Intro; Apply ID_PAR].
Inversion_clear H0.
Exists [_:name](par (q1 _) nil); Split; [Apply bPAR1; Assumption | Intro; Apply ID_PAR].

Inversion_clear H.
Exists p3; Split; [Assumption | Intros; Apply ID_PAR].
Inversion_clear H0.
Exists [_:name](par (q1 _) nil); Split; [Apply bPAR1; Assumption | Intros; Apply ID_PAR].

Qed.

(*******************************************************************)
(* Commutativity of parallel composition is proved by means of     *)
(* internal cross adequacy, i.e., building a suitable bisimulation *)
(* and applying the Completeness lemma.                            *)
(*******************************************************************)

(* The next four definitions allow to build the
 * strong late bisimulation needed to prove the
 * commutativity of parallel composition.
 *)

Inductive BisPAR_C: proc -> proc -> Prop :=
  bispar_c: (p,q:proc)(BisPAR_C (par p q) (par q p)).

Inductive Op_PAR_C [R: proc -> proc -> Prop] : proc -> proc -> Prop :=
  op_par_c : (p,q:name->proc)((z:name)
             (notin z (nu p)) -> (notin z (nu q)) -> (R (p z) (q z))) ->
             ((Op_PAR_C R) (nu p) (nu q)).

Fixpoint PAR_C_fun [n:nat]: (proc->proc->Prop) :=
  Case n of
    BisPAR_C
    [m:nat](Op_PAR_C (PAR_C_fun m))
  end.

Inductive SBPAR_C [p,q:proc] : Prop :=
   sbpar_c : (n:nat)((PAR_C_fun n) p q) -> (SBPAR_C p q).

(* Next we establish a property of PAR_C_fun
 * similar to that of L6_Light for StBisim.
 *)

Lemma PAR_C_RW: (n:nat)(p,q:name->proc)(x:name)
                (notin x (nu p)) -> (notin x (nu q)) -> (PAR_C_fun n (p x) (q x)) ->
                (y:name)(notin y (nu p)) -> (notin y (nu q)) -> (PAR_C_fun n (p y) (q y)).

Proof.

Intros; Elim (LEM_name x y); Intro.

Rewrite <-H4; Assumption.

Generalize p q x H H0 y H4 H2 H3 H1; Clear H H0 H2 H3 H4 y H1 x p q; Induction n; Intros.

(* Base Case *)

Simpl in H1; Inversion H1; Simpl.
Elim (proc_exp p0 x); Elim (proc_exp q0 x); Intros.
Inversion_clear H5; Inversion_clear H8.
Rewrite H10 in H6; Rewrite H11 in H6; Rewrite H10 in H7; Rewrite H11 in H7; Cut p=([_:name](par (x1 _) (x0 _))); [Intro | Apply proc_ext with x; Auto].
Cut q=([_:name](par (x0 _) (x1 _))); [Intro | Apply proc_ext with x; Auto].
Rewrite H8; Rewrite H12; Apply bispar_c.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H9 | Inversion_clear H5]; Auto.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H5 | Inversion_clear H9]; Auto.

(* Inductive Step *)

Simpl in H1; Inversion H1; Simpl.
Elim (ho_proc_exp p0 x); Elim (ho_proc_exp q0 x); Intros.
Inversion_clear H8; Inversion_clear H9.
Rewrite H12 in H5; Rewrite H11 in H6.
Cut p=([_:name](nu (x1 _))); [Intro | Apply proc_ext with x; Auto].
Cut q=([_:name](nu (x0 _))); [Intro | Apply proc_ext with x; Auto].
Rewrite H9; Rewrite H13; Apply op_par_c; Intros.
Elim (unsat (par (nu p) (par (nu q) (par (nu [_:name](nu (x0 _))) (nu [_:name](nu (x1 _)))))) (cons x (cons y (cons z empty)))); Intros.
Inversion_clear H16.
Inversion_clear H17; Inversion_clear H18.
Inversion_clear H19; Inversion_clear H20.
Inversion_clear H21; Inversion_clear H22.
Clear H24.
Apply Hrecn with x2; Auto.
Inversion_clear H23; Auto.
Inversion_clear H20; Auto.
Change (PAR_C_fun n ([_:name](x1 _ x2) y) ([_:name](x0 _ x2) y)); Apply Hrecn with x; Auto.
Inversion_clear H8; Apply notin_nu; Intros.
Cut (notin x (nu (x1 z0))); [Intro | Auto].
Inversion_clear H24; Auto.
Inversion_clear H10; Apply notin_nu; Intros.
Cut (notin x (nu (x0 z0))); [Intro | Auto].
Inversion_clear H24; Auto.
Apply proc_mono with x; Inversion_clear H2.
Cut (notin y (p x)); [Rewrite <-H5; Intro | Auto].
Inversion_clear H2; Auto.
Apply proc_mono with x; Inversion_clear H3.
Cut (notin y (q x)); [Rewrite <-H6; Intro | Auto].
Inversion_clear H3; Auto.
Rewrite <-H11; Rewrite <-H12; Apply H7; Auto.
Inversion_clear H23.
Cut (notin x2 (nu (x1 x))); [Intro | Auto]; Rewrite H12; Assumption.
Inversion_clear H20.
Cut (notin x2 (nu (x0 x))); [Intro | Auto]; Rewrite H11; Assumption.

Qed.

(* SBPAR_C is a strong bisimulation *)

Lemma SBPAR_C_TO_SB': (Inclus SBPAR_C (Op_StBisim SBPAR_C)).

Proof.

Unfold Inclus; Intros.
Inversion_clear H.
Cut (PAR_C_fun n p1 p2);
 [Generalize p1 p2; Generalize n ; Induction n0; Intros | Assumption].

(* Base case *)

Simpl in H; Inversion H.
Apply op_sb; Split; Intros.

(* Free actions case *)

Split; Intros.

Inversion_clear H3.

Exists (par q p7); Split; [Apply fPAR2; Assumption | Apply sbpar_c with O; Apply bispar_c].

Exists (par p7 p); Split; [Apply fPAR1; Assumption | Apply sbpar_c with O; Apply bispar_c].

Exists (par q2 (q1 y)); Split; [Apply COM2 with x; Assumption | Apply sbpar_c with O; Apply bispar_c].

Exists (par (q2 y) q1); Split; [Apply COM1 with x; Assumption | Apply sbpar_c with O; Apply bispar_c].

Exists (nu [_:name](par (q2 _) (q1 _))); Split; [Apply CLOSE2 with x; Assumption | Apply sbpar_c with (S O); Simpl; Apply op_par_c; Intros; Apply bispar_c].

Exists (nu [_:name](par (q2 _) (q1 _))); Split; [Apply CLOSE1 with x; Assumption | Apply sbpar_c with (S O); Simpl; Apply op_par_c; Intros; Apply bispar_c].

Inversion_clear H3.

Exists (par p p6); Split; [Apply fPAR2; Assumption | Apply sbpar_c with O; Apply bispar_c].

Exists (par p6 q); Split; [Apply fPAR1; Assumption | Apply sbpar_c with O; Apply bispar_c].

Exists (par q2 (q0 y)); Split; [Apply COM2 with x; Assumption | Apply sbpar_c with O; Apply bispar_c].

Exists (par (q2 y) q0); Split; [Apply COM1 with x; Assumption | Apply sbpar_c with O; Apply bispar_c].

Exists (nu [_:name](par (q2 _) (q0 _))); Split; [Apply CLOSE2 with x; Assumption | Apply sbpar_c with (S O); Simpl; Apply op_par_c; Intros; Apply bispar_c].

Exists (nu [_:name](par (q2 _) (q0 _))); Split; [Apply CLOSE1 with x; Assumption | Apply sbpar_c with (S O); Simpl; Apply op_par_c; Intros; Apply bispar_c].

(* Input actions case *)

Split; Intros.

Split; Intros.

Inversion_clear H3.

Exists [_:name](par q (p7 _)); Split; [Apply bPAR2; Assumption | Intro; Apply sbpar_c with O; Simpl; Apply bispar_c].

Exists [_:name](par (p7 _) p); Split; [Apply bPAR1; Assumption | Intro; Apply sbpar_c with O; Simpl; Apply bispar_c].

Inversion_clear H3.

Exists [_:name](par p (p6 _)); Split; [Apply bPAR2; Assumption | Intro; Apply sbpar_c with O; Simpl; Apply bispar_c].

Exists [_:name](par (p6 _) q); Split; [Apply bPAR1; Assumption | Intro; Apply sbpar_c with O; Simpl; Apply bispar_c].

(* Bound Output actions case *)

Split; Intros.

Inversion_clear H3.

Exists [_:name](par q (p7 _)); Split; [Apply bPAR2; Assumption | Intros; Apply sbpar_c with O; Simpl; Apply bispar_c].

Exists [_:name](par (p7 _) p); Split; [Apply bPAR1; Assumption | Intros; Apply sbpar_c with O; Simpl; Apply bispar_c].

Inversion_clear H3.

Exists [_:name](par p (p6 _)); Split; [Apply bPAR2; Assumption | Intros; Apply sbpar_c with O; Simpl; Apply bispar_c].

Exists [_:name](par (p6 _) q); Split; [Apply bPAR1; Assumption | Intros; Apply sbpar_c with O; Simpl; Apply bispar_c].

(* Inductive Step *)

Simpl in H1; Inversion H1.
Apply op_sb; Split; Intros.

Elim a; Intros; Split; Intros.

Inversion_clear H5.
Cut (x:name)(notin x (nu p)) -> (notin x (nu q)) -> (Op_StBisim SBPAR_C (p x) (q x)); [Intro | Auto].
Elim (unsat (par (nu p) (par (nu q) (nu p6))) l); Intros.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H10.
Cut (ftrans (p x) tau (p6 x)); [Intro | Apply H6; Auto].
Cut (Op_StBisim SBPAR_C (p x) (q x)); [Intro | Apply H5; Auto].
Inversion_clear H12.
Inversion_clear H13.
Elim (H12 tau); Intros.
Elim (H13 (p6 x) H10); Intros.
Inversion_clear H16; Elim (proc_exp x0 x); Intros.
Inversion_clear H16.
Rewrite H20 in H17; Rewrite H20 in H18.
Exists (nu x1); Split; [Apply fRES with l; Intros | Idtac].
Change (ftrans (q y) ([_:name]tau y) (x1 y)); Apply FTR_L3 with x; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Inversion_clear H18; Apply sbpar_c with (S n2); Simpl; Apply op_par_c; Intros.
Apply PAR_C_RW with x; Auto.
Apply f_act_notin_tau.

Inversion_clear H5.
Cut (x:name)(notin x (nu p)) -> (notin x (nu q)) -> (Op_StBisim SBPAR_C (p x) (q x)); [Intro | Auto].
Elim (unsat (par (nu p) (par (nu q) (nu p5))) l); Intros.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H10; Cut (ftrans (q x) tau (p5 x)); [Intro | Apply H6; Auto].
Cut (Op_StBisim SBPAR_C (p x) (q x)); [Intro | Apply H5; Auto].
Inversion_clear H12.
Inversion_clear H13.
Elim (H12 tau); Intros.
Elim (H15 (p5 x) H10); Intros.
Inversion_clear H16; Elim (proc_exp x0 x); Intros.
Inversion_clear H16.
Rewrite H20 in H17; Rewrite H20 in H18.
Exists (nu x1); Split; [Apply fRES with l; Intros | Idtac].
Change (ftrans (p y) ([_:name]tau y) (x1 y)); Apply FTR_L3 with x; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Inversion_clear H18.
Apply sbpar_c with (S n2); Simpl; Apply op_par_c; Intros.
Apply PAR_C_RW with x; Auto.
Apply f_act_notin_tau.

Inversion_clear H5.
Cut (x:name)(notin x (nu p)) -> (notin x (nu q)) -> (Op_StBisim SBPAR_C (p x) (q x)); [Intro | Auto].
Elim (unsat (par (nu p) (par (nu q) (nu p6))) (cons n2 (cons n3 l))); Intros.
Inversion_clear H7.
Inversion_clear H8; Inversion_clear H9.
Inversion_clear H10; Inversion_clear H11.
Cut (ftrans (p x) (Out n2 n3) (p6 x)); [Intro | Apply H6; Auto].
Cut (Op_StBisim SBPAR_C (p x) (q x)); [Intro | Apply H5; Auto].
Inversion_clear H14.
Inversion_clear H15.
Elim (H14 (Out n2 n3)); Intros.
Elim (H15 (p6 x) H11); Intros.
Inversion_clear H18; Elim (proc_exp x0 x); Intros.
Inversion_clear H18.
Rewrite H22 in H19; Rewrite H22 in H20.
Exists (nu x1); Split; [Apply fRES with l; Intros | Idtac].
Inversion_clear H25; Change (ftrans (q y) ([_:name](Out n2 n3) y) (x1 y)); Apply FTR_L3 with x; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Inversion_clear H20.
Apply sbpar_c with (S n4); Simpl; Apply op_par_c; Intros.
Apply PAR_C_RW with x; Auto.
Apply f_act_notin_Out; Auto.

Inversion_clear H5.
Cut (x:name)(notin x (nu p)) -> (notin x (nu q)) -> (Op_StBisim SBPAR_C (p x) (q x)); [Intro | Auto].
Elim (unsat (par (nu p) (par (nu q) (nu p5))) (cons n2 (cons n3 l))); Intros.
Inversion_clear H7.
Inversion_clear H8; Inversion_clear H9.
Inversion_clear H10; Inversion_clear H11.
Cut (ftrans (q x) (Out n2 n3) (p5 x)); [Intro | Apply H6; Auto].
Cut (Op_StBisim SBPAR_C (p x) (q x)); [Intro | Apply H5; Auto].
Inversion_clear H14.
Inversion_clear H15.
Elim (H14 (Out n2 n3)); Intros.
Elim (H17 (p5 x) H11); Intros.
Inversion_clear H18; Elim (proc_exp x0 x); Intros.
Inversion_clear H18.
Rewrite H22 in H19; Rewrite H22 in H20.
Exists (nu x1); Split; [Apply fRES with l; Intros | Idtac].
Inversion_clear H25; Change (ftrans (p y) ([_:name](Out n2 n3) y) (x1 y)); Apply FTR_L3 with x; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Inversion_clear H20.
Apply sbpar_c with (S n4); Simpl; Apply op_par_c; Intros.
Apply PAR_C_RW with x; Auto.
Apply f_act_notin_Out; Auto.

Split; Intros.

Split; Intros.

Inversion_clear H5.
Cut (x:name)(notin x (nu p)) -> (notin x (nu q)) -> (Op_StBisim SBPAR_C (p x) (q x)); [Intro | Auto].
Elim (unsat (par (nu p) (par (nu q) (nu [_:name](nu (p6 _))))) (cons x l)); Intros.
Inversion_clear H7.
Inversion_clear H8; Inversion_clear H9.
Inversion_clear H10.
Cut (btrans (p x0) (In x) (p6 x0)); [Intro | Apply H6; Auto].
Cut (Op_StBisim SBPAR_C (p x0) (q x0)); [Intro | Apply H5; Auto].
Inversion_clear H13.
Inversion_clear H14.
Inversion_clear H15.
Elim (H14 x); Intros.
Elim (H15 (p6 x0) H10); Intros.
Inversion_clear H18; Elim (ho_proc_exp x1 x0); Intros.
Inversion_clear H18.
Rewrite H22 in H19; Rewrite H22 in H20.
Exists [_:name](nu [u:name](x2 u _)); Split; [Apply bRES with l; Intros | Idtac].
Inversion_clear H24; Change (btrans (q y) ([_:name](In x) y) (x2 y)); Apply BTR_L3 with x0; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Intro; Elim (LEM_name x0 y); Intro.

Rewrite <-H18; Elim (unsat (par (nu [_:name](nu (p6 _))) (nu [_:name](nu (x2 _)))) (cons x0 empty)); Intros.
Inversion_clear H23.
Inversion_clear H24; Inversion_clear H25.
Clear H27; Cut (SBPAR_C (p6 x0 x3) (x2 x0 x3)); [Intro | Auto].
Inversion_clear H25.
Apply sbpar_c with (S n2); Simpl; Apply op_par_c; Intros.
Elim (unsat (par (nu [_:name](nu (p6 _))) (nu [_:name](nu (x2 _)))) (cons x0 (cons x3 (cons z empty)))); Intros.
Inversion_clear H29.
Inversion_clear H30; Inversion_clear H31.
Inversion_clear H33.
Inversion_clear H34.
Clear H35.
Cut (PAR_C_fun n2 (p6 x4 x3) (x2 x4 x3)); [Intro | Change (PAR_C_fun n2 ([_:name](p6 _ x3) x4) ([_:name](x2 _ x3) x4)); Apply PAR_C_RW with x0; Auto].
Change (PAR_C_fun n2 ([_:name](p6 _ x0) z) ([_:name](x2 _ x0) z)); Apply PAR_C_RW with x4; Auto.
Inversion_clear H29; Apply notin_nu; Intros.
Cut (notin x4 (nu (p6 z0))); [Intro | Auto].
Inversion_clear H36; Auto.
Inversion_clear H32; Apply notin_nu; Intros.
Cut (notin x4 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H36; Auto.
Apply PAR_C_RW with x3; Auto.
Inversion_clear H23; Auto.
Inversion_clear H26; Auto.
Inversion_clear H12; Auto.
Inversion_clear H21; Auto.
Inversion_clear H12; Apply notin_nu; Intros.
Cut (notin x0 (nu (p6 z0))); [Intro | Auto].
Inversion_clear H35; Auto.
Inversion_clear H21; Apply notin_nu; Intros.
Cut (notin x0 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H35; Auto.
Inversion_clear H29; Apply notin_nu; Intros.
Cut (notin x4 (nu (p6 z0))); [Intro | Auto].
Inversion_clear H35; Auto.
Inversion_clear H32; Apply notin_nu; Intros.
Cut (notin x4 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H35; Auto.

Cut (SBPAR_C (p6 x0 y) (x2 x0 y)); [Intro | Auto].
Inversion_clear H23.
Apply sbpar_c with (S n2); Simpl; Apply op_par_c; Intros.
Change (PAR_C_fun n2 ([_:name](p6 _ y) z) ([_:name](x2 _ y) z)); Apply PAR_C_RW with x0; Auto.
Inversion_clear H12; Apply notin_nu; Intros.
Cut (notin x0 (nu (p6 z0))); [Intro | Auto].
Inversion_clear H27; Auto.
Inversion_clear H21; Apply notin_nu; Intros.
Cut (notin x0 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H27; Auto.
Apply b_act_notin_In; Auto.

Inversion_clear H5.
Cut (x:name)(notin x (nu p)) -> (notin x (nu q)) -> (Op_StBisim SBPAR_C (p x) (q x)); [Intro | Auto].
Elim (unsat (par (nu p) (par (nu q) (nu [_:name](nu (p5 _))))) (cons x l)); Intros.
Inversion_clear H7.
Inversion_clear H8; Inversion_clear H9.
Inversion_clear H10.
Cut (btrans (q x0) (In x) (p5 x0)); [Intro | Apply H6; Auto].
Cut (Op_StBisim SBPAR_C (p x0) (q x0)); [Intro | Apply H5; Auto].
Inversion_clear H13.
Inversion_clear H14.
Inversion_clear H15.
Elim (H14 x); Intros.
Elim (H17 (p5 x0) H10); Intros.
Inversion_clear H18; Elim (ho_proc_exp x1 x0); Intros.
Inversion_clear H18.
Rewrite H22 in H19; Rewrite H22 in H20.
Exists [_:name](nu [u:name](x2 u _)); Split; [Apply bRES with l; Intros | Idtac].
Inversion_clear H24; Change (btrans (p y) ([_:name](In x) y) (x2 y)); Apply BTR_L3 with x0; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Intro; Elim (LEM_name x0 y); Intro.

Rewrite <-H18; Elim (unsat (par (nu [_:name](nu (p5 _))) (nu [_:name](nu (x2 _)))) (cons x0 empty)); Intros.
Inversion_clear H23.
Inversion_clear H24; Inversion_clear H25.
Clear H27; Cut (SBPAR_C (x2 x0 x3) (p5 x0 x3)); [Intro | Auto].
Inversion_clear H25.
Apply sbpar_c with (S n2); Simpl; Apply op_par_c; Intros.
Elim (unsat (par (nu [_:name](nu (p5 _))) (nu [_:name](nu (x2 _)))) (cons x0 (cons x3 (cons z empty)))); Intros.
Inversion_clear H29.
Inversion_clear H30; Inversion_clear H31.
Inversion_clear H33.
Inversion_clear H34.
Clear H35; Cut (PAR_C_fun n2 (x2 x4 x3) (p5 x4 x3)); [Intro | Change (PAR_C_fun n2 ([_:name](x2 _ x3) x4) ([_:name](p5 _ x3) x4)); Apply PAR_C_RW with x0; Auto].
Change (PAR_C_fun n2 ([_:name](x2 _ x0) z) ([_:name](p5 _ x0) z)); Apply PAR_C_RW with x4; Auto.
Inversion_clear H32; Apply notin_nu; Intros.
Cut (notin x4 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H36; Auto.
Inversion_clear H29; Apply notin_nu; Intros.
Cut (notin x4 (nu (p5 z0))); [Intro | Auto].
Inversion_clear H36; Auto.
Apply PAR_C_RW with x3; Auto.
Inversion_clear H26; Auto.
Inversion_clear H23; Auto.
Inversion_clear H21; Auto.
Inversion_clear H12; Auto.
Inversion_clear H21; Apply notin_nu; Intros.
Cut (notin x0 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H35; Auto.
Inversion_clear H12; Apply notin_nu; Intros.
Cut (notin x0 (nu (p5 z0))); [Intro | Auto].
Inversion_clear H35; Auto.
Inversion_clear H32; Apply notin_nu; Intros.
Cut (notin x4 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H35; Auto.
Inversion_clear H29; Apply notin_nu; Intros.
Cut (notin x4 (nu (p5 z0))); [Intro | Auto].
Inversion_clear H35; Auto.

Cut (SBPAR_C (x2 x0 y) (p5 x0 y)); [Intro | Auto].
Inversion_clear H23.
Apply sbpar_c with (S n2); Simpl; Apply op_par_c; Intros.
Change (PAR_C_fun n2 ([_:name](x2 _ y) z) ([_:name](p5 _ y) z)); Apply PAR_C_RW with x0; Auto.
Inversion_clear H21; Apply notin_nu; Intros.
Cut (notin x0 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H27; Auto.
Inversion_clear H12; Apply notin_nu; Intros.
Cut (notin x0 (nu (p5 z0))); [Intro | Auto].
Inversion_clear H27; Auto.
Apply b_act_notin_In; Auto.

Split; Intros.

Inversion_clear H5.

Cut (x:name)(notin x (nu p)) -> (notin x (nu q)) -> (Op_StBisim SBPAR_C (p x) (q x)); [Intro | Auto].
Elim (unsat (par (nu p) (par (nu q) (nu [_:name](nu (p6 _))))) (cons x l)); Intros.
Inversion_clear H7.
Inversion_clear H8; Inversion_clear H9.
Inversion_clear H10.
Cut (btrans (p x0) (bOut x) (p6 x0)); [Intro | Apply H6; Auto].
Cut (Op_StBisim SBPAR_C (p x0) (q x0)); [Intro | Apply H5; Auto].
Inversion_clear H13.
Inversion_clear H14.
Inversion_clear H15.
Elim (H16 x); Intros.
Elim (H15 (p6 x0) H10); Intros.
Inversion_clear H18; Elim (ho_proc_exp x1 x0); Intros.
Inversion_clear H18.
Rewrite H22 in H19; Rewrite H22 in H20.
Exists [_:name](nu [u:name](x2 u _)); Split; [Apply bRES with l; Intros | Idtac].
Inversion_clear H24; Change (btrans (q y) ([_:name](bOut x) y) (x2 y)); Apply BTR_L3 with x0; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Intros; Elim (LEM_name x0 y); Intro.

Rewrite <-H24; Elim (unsat (par (nu [_:name](nu (p6 _))) (nu [_:name](nu (x2 _)))) (cons x0 empty)); Intros.
Inversion_clear H25.
Inversion_clear H26; Inversion_clear H27.
Clear H29; Cut (SBPAR_C (p6 x0 x3) (x2 x0 x3)); [Intro | Apply H20; Auto].
Inversion_clear H27.
Apply sbpar_c with (S n2); Simpl; Apply op_par_c; Intros.
Elim (unsat (par (nu [_:name](nu (p6 _))) (nu [_:name](nu (x2 _)))) (cons x0 (cons x3 (cons z empty)))); Intros.
Inversion_clear H31.
Inversion_clear H32; Inversion_clear H33.
Inversion_clear H35.
Inversion_clear H36.
Clear H37; Cut (PAR_C_fun n2 (p6 x4 x3) (x2 x4 x3)); [Intro | Change (PAR_C_fun n2 ([_:name](p6 _ x3) x4) ([_:name](x2 _ x3) x4)); Apply PAR_C_RW with x0; Auto].
Change (PAR_C_fun n2 ([_:name](p6 _ x0) z) ([_:name](x2 _ x0) z)); Apply PAR_C_RW with x4; Auto.
Inversion_clear H31; Apply notin_nu; Intros.
Cut (notin x4 (nu (p6 z0))); [Intro | Auto].
Inversion_clear H38; Auto.
Inversion_clear H34; Apply notin_nu; Intros.
Cut (notin x4 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H38; Auto.
Apply PAR_C_RW with x3; Auto.
Inversion_clear H25; Auto.
Inversion_clear H28; Auto.
Inversion_clear H12; Auto.
Inversion_clear H21; Auto.
Inversion_clear H12; Apply notin_nu; Intros.
Cut (notin x0 (nu (p6 z0))); [Intro | Auto].
Inversion_clear H37; Auto.
Inversion_clear H21; Apply notin_nu; Intros.
Cut (notin x0 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H37; Auto.
Inversion_clear H31; Apply notin_nu; Intros.
Cut (notin x4 (nu (p6 z0))); [Intro | Auto].
Inversion_clear H37; Auto.
Inversion_clear H34; Apply notin_nu; Intros.
Cut (notin x4 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H37; Auto.
Inversion_clear H25; Auto.
Inversion_clear H28; Auto.

Cut (SBPAR_C (p6 x0 y) (x2 x0 y)); [Intro | Apply H20; Auto].
Inversion_clear H25.
Apply sbpar_c with (S n2); Simpl; Apply op_par_c; Intros.
Change (PAR_C_fun n2 ([_:name](p6 _ y) z) ([_:name](x2 _ y) z)); Apply PAR_C_RW with x0; Auto.
Inversion_clear H12; Apply notin_nu; Intros.
Cut (notin x0 (nu (p6 z0))); [Intro | Auto].
Inversion_clear H29; Auto.
Inversion_clear H21; Apply notin_nu; Intros.
Cut (notin x0 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H29; Auto.
Inversion_clear H18; Apply notin_nu; Intros.
Cut (notin y (nu [z0:name](p6 z0 z))); [Intro | Auto].
Inversion_clear H26; Auto.
Inversion_clear H23; Apply notin_nu; Intros.
Cut (notin y (nu [z0:name](x2 z0 z))); [Intro | Auto].
Inversion_clear H26; Auto.
Apply b_act_notin_bOut; Auto.

Cut (x:name)(notin x (nu p)) -> (notin x (nu q)) -> (Op_StBisim SBPAR_C (p x) (q x)); [Intro | Auto].
Elim (unsat (par (nu p) (par (nu q) (nu p4))) (cons x l)); Intros.
Inversion_clear H7.
Inversion_clear H8; Inversion_clear H9.
Inversion_clear H10.
Cut (ftrans (p x0) (Out x x0) (p4 x0)); [Intro | Apply H6; Auto].
Cut (Op_StBisim SBPAR_C (p x0) (q x0)); [Intro | Apply H5; Auto].
Inversion_clear H13.
Inversion_clear H14.
Elim (H13 (Out x x0)); Intros.
Elim (H14 (p4 x0) H10); Intros.
Inversion_clear H17.
Elim (proc_exp x1 x0); Intros.
Inversion_clear H17.
Rewrite H21 in H18; Rewrite H21 in H19.
Exists x2; Split; [Apply OPEN with l; Intros | Idtac].
Change (ftrans (q y) (Out x y) (x2 y)); Apply FTR_L3 with x0; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Intros; Inversion_clear H19.
Apply sbpar_c with n2; Apply PAR_C_RW with x0; Auto.

Inversion_clear H5.

Cut (x:name)(notin x (nu p)) -> (notin x (nu q)) -> (Op_StBisim SBPAR_C (p x) (q x)); [Intro | Auto].
Elim (unsat (par (nu p) (par (nu q) (nu [_:name](nu (p5 _))))) (cons x l)); Intros.
Inversion_clear H7.
Inversion_clear H8; Inversion_clear H9.
Inversion_clear H10.
Cut (btrans (q x0) (bOut x) (p5 x0)); [Intro | Apply H6; Auto].
Cut (Op_StBisim SBPAR_C (p x0) (q x0)); [Intro | Apply H5; Auto].
Inversion_clear H13.
Inversion_clear H14.
Inversion_clear H15.
Elim (H16 x); Intros.
Elim (H17 (p5 x0) H10); Intros.
Inversion_clear H18.
Elim (ho_proc_exp x1 x0); Intros.
Inversion_clear H18.
Rewrite H22 in H19; Rewrite H22 in H20.
Exists [_:name](nu [u:name](x2 u _)); Split; [Apply bRES with l; Intros | Idtac].
Inversion_clear H24; Change (btrans (p y) ([_:name](bOut x) y) (x2 y)); Apply BTR_L3 with x0; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Intros; Elim (LEM_name x0 y); Intro.

Rewrite <-H24; Elim (unsat (par (nu [_:name](nu (p5 _))) (nu [_:name](nu (x2 _)))) (cons x0 empty)); Intros.
Inversion_clear H25.
Inversion_clear H26; Inversion_clear H27.
Clear H29; Cut (SBPAR_C (x2 x0 x3) (p5 x0 x3)); [Intro | Apply H20; Auto].
Inversion_clear H27.
Apply sbpar_c with (S n2); Simpl; Apply op_par_c; Intros.
Elim (unsat (par (nu [_:name](nu (p5 _))) (nu [_:name](nu (x2 _)))) (cons x0 (cons x3 (cons z empty)))); Intros.
Inversion_clear H31.
Inversion_clear H32; Inversion_clear H33.
Inversion_clear H35.
Inversion_clear H36.
Clear H37; Cut (PAR_C_fun n2 (x2 x4 x3) (p5 x4 x3)); [Intro | Change (PAR_C_fun n2 ([_:name](x2 _ x3) x4) ([_:name](p5 _ x3) x4)); Apply PAR_C_RW with x0; Auto].
Change (PAR_C_fun n2 ([_:name](x2 _ x0) z) ([_:name](p5 _ x0) z)); Apply PAR_C_RW with x4; Auto.
Inversion_clear H34; Apply notin_nu; Intros.
Cut (notin x4 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H38; Auto.
Inversion_clear H31; Apply notin_nu; Intros.
Cut (notin x4 (nu (p5 z0))); [Intro | Auto].
Inversion_clear H38; Auto.
Apply PAR_C_RW with x3; Auto.
Inversion_clear H28; Auto.
Inversion_clear H25; Auto.
Inversion_clear H21; Auto.
Inversion_clear H12; Auto.
Inversion_clear H21; Apply notin_nu; Intros.
Cut (notin x0 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H37; Auto.
Inversion_clear H12; Apply notin_nu; Intros.
Cut (notin x0 (nu (p5 z0))); [Intro | Auto].
Inversion_clear H37; Auto.
Inversion_clear H34; Apply notin_nu; Intros.
Cut (notin x4 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H37; Auto.
Inversion_clear H31; Apply notin_nu; Intros.
Cut (notin x4 (nu (p5 z0))); [Intro | Auto].
Inversion_clear H37; Auto.
Inversion_clear H28; Auto.
Inversion_clear H25; Auto.

Cut (SBPAR_C (x2 x0 y) (p5 x0 y)); [Intro | Apply H20; Auto].
Inversion_clear H25.
Apply sbpar_c with (S n2); Simpl; Apply op_par_c; Intros.
Change (PAR_C_fun n2 ([_:name](x2 _ y) z) ([_:name](p5 _ y) z)); Apply PAR_C_RW with x0; Auto.
Inversion_clear H21; Apply notin_nu; Intros.
Cut (notin x0 (nu (x2 z0))); [Intro | Auto].
Inversion_clear H29; Auto.
Inversion_clear H12; Apply notin_nu; Intros.
Cut (notin x0 (nu (p5 z0))); [Intro | Auto].
Inversion_clear H29; Auto.
Inversion_clear H18; Apply notin_nu; Intros.
Cut (notin y (nu [z0:name](x2 z0 z))); [Intro | Auto].
Inversion_clear H26; Auto.
Inversion_clear H23; Apply notin_nu; Intros.
Cut (notin y (nu [z0:name](p5 z0 z))); [Intro | Auto].
Inversion_clear H26; Auto.
Apply b_act_notin_bOut; Auto.

Cut (x:name)(notin x (nu p)) -> (notin x (nu q)) -> (Op_StBisim SBPAR_C (p x) (q x)); [Intro | Auto].
Elim (unsat (par (nu p) (par (nu q) (nu q1))) (cons x l)); Intros.
Inversion_clear H7.
Inversion_clear H8; Inversion_clear H9.
Inversion_clear H10.
Cut (ftrans (q x0) (Out x x0) (q1 x0)); [Intro | Apply H6; Auto].
Cut (Op_StBisim SBPAR_C (p x0) (q x0)); [Intro | Apply H5; Auto].
Inversion_clear H13.
Inversion_clear H14.
Elim (H13 (Out x x0)); Intros.
Elim (H16 (q1 x0) H10); Intros.
Inversion_clear H17.
Elim (proc_exp x1 x0); Intros.
Inversion_clear H17.
Rewrite H21 in H18; Rewrite H21 in H19.
Exists x2; Split; [Apply OPEN with l; Intros | Idtac].
Change (ftrans (p y) (Out x y) (x2 y)); Apply FTR_L3 with x0; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Intros; Inversion_clear H19.
Apply sbpar_c with n2; Apply PAR_C_RW with x0; Auto.

Qed.

(* Finally, the commutativity of parallel composition
 * is proved by means of internal cross adequacy.
 *)

Lemma COMM_PAR:  (p,q:proc)(StBisim (par p q) (par q p)).

Proof.

Intros; Apply Completeness; Apply Co_Ind with SBPAR_C; [Exact SBPAR_C_TO_SB' | Apply sbpar_c with O; Simpl; Apply bispar_c; Assumption].

Qed.

End parallel.

Section upto.

(***************************************************************************)
(* In this section the technique of strong bisimulation up to equivalence  *)
(* and restriction (UpTo for short) is exploited in order to prove the     *)
(* restriction operator extrusion law and the associativity of parallel    *)
(* composition.                                                            *)
(***************************************************************************)

Definition SB_R_SB:=[R:proc->proc->Prop][p,q:proc]
                    (Ex [p':proc](Ex [q':proc](StBisim p p')
                                           /\ (R p' q')
                                           /\ (StBisim q' q))).

(* Definition of being a strong bisimulation
 * up to equivalence and restriction.
 *)

Definition UpTo:= [R:proc->proc->Prop](p,q:proc)(R p q) ->
                  ((p',q':name->proc)(x:name)
                   (notin x (nu p')) -> (notin x (nu q')) ->
                   p=(p' x) -> q=(q' x) ->
                   (y:name)(notin y (nu p')) -> (notin y (nu q')) ->
                   (R (p' y) (q' y))
                  )
                  /\
                  ((x,y:name)
                   ((p1:proc)(ftrans p (Out x y) p1)->
                    (Ex [q1:proc]((ftrans q (Out x y) q1) /\ (SB_R_SB R p1 q1)))
                   )
                   /\
                   ((q1:proc)(ftrans q (Out x y) q1)->
                    (Ex [p1:proc]((ftrans p (Out x y) p1) /\ (SB_R_SB R p1 q1)))
                   )
                  )
                  /\
                  ((x:name)
                   ((p1:name->proc)(btrans p (In x) p1)->
                    (Ex [q1:name->proc]((btrans q (In x) q1) /\ (y:name)(SB_R_SB R (p1 y) (q1 y))))
                   )
                   /\
                   ((q1:name->proc)(btrans q (In x) q1)->
                    (Ex [p1:name->proc]((btrans p (In x) p1) /\ (y:name)(SB_R_SB R (p1 y) (q1 y))))
                   )
                  )
                  /\
                  ((x:name)
                   ((p1:name->proc)(btrans p (bOut x) p1)->
                    (Ex [q1:name->proc]((btrans q (bOut x) q1) /\ (y:name)(notin y (nu p1))->(notin y (nu q1))->(SB_R_SB R (p1 y) (q1 y))))
                   )
                   /\
                   ((q1:name->proc)(btrans q (bOut x) q1)->
                    (Ex [p1:name->proc]((btrans p (bOut x) p1) /\ (y:name)(notin y (nu p1))->(notin y (nu q1))->(SB_R_SB R (p1 y) (q1 y))))
                   )
                  )
                  /\
                  (((p1:proc)(ftrans p tau p1)->
                    (Ex [q1:proc](ftrans q tau q1) /\
                        ((SB_R_SB R p1 q1)
                         \/
                         (Ex [p1':name->proc](Ex [q1':name->proc](StBisim p1 (nu p1'))
                                                              /\ ((z:name)
                                                                  (notin z (nu p1'))->
                                                                  (notin z (nu q1'))->
                                                                  (R (p1' z) (q1' z))
                                                                 )
                                                              /\ (StBisim (nu q1') q1)))
                        )
                    )
                   )
                   /\
                   ((q1:proc)(ftrans q tau q1)->
                    (Ex [p1:proc](ftrans p tau p1) /\
                        ((SB_R_SB R p1 q1)
                         \/
                         (Ex [p1':name->proc](Ex [q1':name->proc](StBisim p1 (nu p1'))
                                                              /\ ((z:name)
                                                                  (notin z (nu p1'))->
                                                                  (notin z (nu q1'))->
                                                                  (R (p1' z) (q1' z))
                                                                 )
                                                              /\ (StBisim (nu q1') q1)))
                        )
                    )
                   )
                  ).

Inductive Op_NU [R: proc -> proc -> Prop] : proc -> proc -> Prop :=
  op_nu : (p,q:proc)(p',q':name->proc)
          (StBisim p (nu p')) -> (StBisim (nu q') q) ->
          ((z:name)
             (notin z (nu p')) -> (notin z (nu q')) -> (R (p' z) (q' z))
          ) ->
          ((Op_NU R) p q).

Fixpoint NU_fun [R:proc->proc->Prop; n:nat]: (proc->proc->Prop) :=
  Case n of
    (SB_R_SB R)
    [m:nat](Op_NU (NU_fun R m))
  end.

Inductive SBUPTO [R:proc->proc->Prop; p,q:proc] : Prop :=
   sbupto : (n:nat)((NU_fun R n) p q) -> (SBUPTO R p q).

(* Strong bisimulations UpTo enjoy a property similar to that established
 * by L6_Light for StBisim.
 *)

Lemma NU_FUN_RW: (R:proc->proc->Prop)(UpTo R) -> (n:nat)(p,q:name->proc)(x:name)
                 (notin x (nu p)) -> (notin x (nu q)) -> (NU_fun R n (p x) (q x)) ->
                 (y:name)(notin y (nu p)) -> (notin y (nu q)) -> (NU_fun R n (p y) (q y)).

Proof.

Intros; Elim (LEM_name x y); Intro.

Rewrite <-H5; Assumption.

Generalize p q x H0 H1 y H5 H3 H4 H2; Clear H0 H1 H3 H4 H5 y H2 x p q; Induction n; Intros.

(* Base Case *)

Simpl in H2; Inversion H2; Simpl; Unfold SB_R_SB.
Inversion_clear H6; Inversion_clear H7; Inversion_clear H8.
Elim (proc_exp x0 x); Elim (proc_exp x1 x); Intros.
Inversion_clear H8; Inversion_clear H10.
Rewrite H12 in H7; Rewrite H12 in H9; Rewrite H13 in H6; Rewrite H13 in H7.
Elim (ho_proc_exp x2 y); Elim (ho_proc_exp x3 y); Intros.
Inversion_clear H10; Inversion_clear H14.
Elim (unsat (par (nu p) (par (nu q) (par (nu [_:name](nu (x4 _))) (nu [_:name](nu (x5 _)))))) (cons x (cons y empty))); Intros.
Inversion_clear H14; Inversion_clear H18; Inversion_clear H19; Inversion_clear H20; Inversion_clear H21; Clear H23; Inversion_clear H22.
Exists (x4 x6 y); Exists (x5 x6 y); Split; [Idtac | Split].
Apply L6_Light with x; Auto.
Cut (notin x (nu [y:name](nu (x4 y)))); [Intro; Inversion_clear H22; Auto | Apply proc_mono with y; Rewrite <-H16; Assumption].
Change (StBisim ([_:name](p x) x6) ([_:name](x4 _ x) x6)); Apply L6_Light with y; Auto.
Inversion_clear H3; Apply notin_nu; Intros; Auto.
Inversion_clear H15; Apply notin_nu; Intros; Cut (notin y (nu (x4 z))); [Intro | Auto]; Inversion_clear H24; Auto.
Rewrite <-H16; Assumption.
Inversion_clear H14; Apply notin_nu; Intros; Auto.
Inversion_clear H21; Apply notin_nu; Intros; Cut (notin x6 (nu (x4 z))); [Intro | Auto]; Inversion_clear H24; Auto.
Inversion_clear H15; Auto.
Rewrite H16 in H7; Rewrite H17 in H7; Unfold UpTo in H; Elim (H (x4 x6 x) (x5 x6 x)); Intros; Try Clear H24.
Apply H22 with x; Auto.
Cut (notin x (nu [_:name](nu (x4 _)))); [Intro | Apply proc_mono with y; Rewrite <-H16; Assumption]; Inversion_clear H24; Auto.
Cut (notin x (nu [_:name](nu (x5 _)))); [Intro | Apply proc_mono with y; Rewrite <-H17; Assumption]; Inversion_clear H24; Auto.
Inversion_clear H15; Auto.
Inversion_clear H10; Auto.
Unfold UpTo in H; Elim (H (x4 y x) (x5 y x)); Intros; Auto; Clear H24.
Change (R ([_:name](x4 _ x) x6) ([_:name](x5 _ x) x6)); Apply H22 with y; Auto.
Inversion_clear H15; Apply notin_nu; Intros; Cut (notin y (nu (x4 z))); [Intro | Auto]; Inversion_clear H25; Auto.
Inversion_clear H10; Apply notin_nu; Intros; Cut (notin y (nu (x5 z))); [Intro | Auto]; Inversion_clear H25; Auto.
Inversion_clear H21; Apply notin_nu; Intros; Cut (notin x6 (nu (x4 z))); [Intro | Auto]; Inversion_clear H25; Auto.
Inversion_clear H23; Apply notin_nu; Intros; Cut (notin x6 (nu (x5 z))); [Intro | Auto]; Inversion_clear H25; Auto.
Apply L6_Light with x; Auto.
Cut (notin x (nu [y:name](nu (x5 y)))); [Intro; Inversion_clear H22; Auto | Apply proc_mono with y; Rewrite <-H17; Assumption].
Change (StBisim ([_:name](x5 _ x) x6) ([_:name](q x) x6)); Apply L6_Light with y; Auto.
Inversion_clear H10; Apply notin_nu; Intros; Cut (notin y (nu (x5 z))); [Intro | Auto]; Inversion_clear H24; Auto.
Inversion_clear H4; Apply notin_nu; Intros; Auto.
Rewrite <-H17; Assumption.
Inversion_clear H23; Apply notin_nu; Intros; Cut (notin x6 (nu (x5 z))); [Intro | Auto]; Inversion_clear H24; Auto.
Inversion_clear H19; Apply notin_nu; Intros; Auto.
Inversion_clear H10; Auto.

(* Inductive step *)

Simpl in H2; Inversion H2; Simpl.
Elim (ho_proc_exp p' x); Elim (ho_proc_exp q' x); Intros.
Inversion_clear H11; Inversion_clear H12.
Rewrite H14 in H7; Rewrite H15 in H6; Rewrite H14 in H8; Rewrite H15 in H8.
Elim (ho2_proc_exp x0 y); Elim (ho2_proc_exp x1 y); Intros.
Inversion_clear H12; Inversion_clear H16.
Elim (unsat (par (nu p) (par (nu q) (par (nu [u:name](nu [v:name](nu (x2 u v)))) (nu [u:name](nu [v:name](nu (x3 u v))))))) (cons x (cons y empty))); Intros.
Inversion_clear H16; Inversion_clear H20; Inversion_clear H21; Inversion_clear H22; Inversion_clear H23; Clear H25; Inversion_clear H24.
Apply op_nu with (x2 x4 y) (x3 x4 y); Intros.
Change (StBisim (p y) ([_:name](nu (x2 x4 _)) y)); Apply L6_Light with x; Auto.
Cut (notin x (nu [y:name](nu [z:name](nu (x2 y z))))); [Intro; Inversion_clear H24; Apply notin_nu; Intros; Cut (notin x (nu [z:name](nu (x2 x4 z)))); [Intro | Auto]; Inversion_clear H27; Auto | Apply proc_mono with y; Rewrite <-H18; Assumption].
Change (StBisim ([_:name](p x) x4) ([_:name](nu (x2 _ x)) x4)); Apply L6_Light with y; Auto.
Inversion_clear H3; Apply notin_nu; Intros; Auto.
Inversion_clear H17; Apply notin_nu; Intros; Cut (notin y (nu [_:name](nu (x2 z _)))); [Intro | Auto]; Inversion_clear H26; Auto.
Rewrite <-H18; Assumption.
Inversion_clear H16; Apply notin_nu; Intros; Auto.
Inversion_clear H23; Apply notin_nu; Intros; Cut (notin x4 (nu [_:name](nu (x2 z _)))); [Intro | Auto]; Inversion_clear H26; Auto.
Inversion_clear H17; Apply notin_nu; Intros; Cut (notin y (nu [_:name](nu (x2 x4 _)))); [Intro | Auto]; Inversion_clear H26; Auto.

Change (StBisim ([_:name](nu (x3 x4 _)) y) (q y)); Apply L6_Light with x; Auto.
Cut (notin x (nu [y:name](nu [z:name](nu (x3 y z))))); [Intro; Inversion_clear H24; Apply notin_nu; Intros; Cut (notin x (nu [z:name](nu (x3 x4 z)))); [Intro | Auto]; Inversion_clear H27; Auto | Apply proc_mono with y; Rewrite <-H19; Assumption].
Change (StBisim ([_:name](nu (x3 _ x)) x4) ([_:name](q x) x4)); Apply L6_Light with y; Auto.
Inversion_clear H12; Apply notin_nu; Intros; Cut (notin y (nu [_:name](nu (x3 z _)))); [Intro | Auto]; Inversion_clear H26; Auto.
Inversion_clear H4; Apply notin_nu; Intros; Auto.
Rewrite <-H19; Assumption.
Inversion_clear H25; Apply notin_nu; Intros; Cut (notin x4 (nu [_:name](nu (x3 z _)))); [Intro | Auto]; Inversion_clear H26; Auto.
Inversion_clear H21; Apply notin_nu; Intros; Auto.
Inversion_clear H12; Apply notin_nu; Intros; Cut (notin y (nu [_:name](nu (x3 x4 _)))); [Intro | Auto]; Inversion_clear H26; Auto.
Elim (unsat (par (nu [u:name](nu [v:name](nu (x2 u v)))) (nu [u:name](nu [v:name](nu (x3 u v))))) (cons x (cons y (cons z (cons x4 empty))))); Intros.
Inversion_clear H27; Inversion_clear H28; Inversion_clear H29; Inversion_clear H31; Inversion_clear H32; Inversion_clear H33; Clear H34.
Apply Hrecn with x5; Auto.
Inversion_clear H27; Apply notin_nu; Intros; Cut (notin x5 (nu [v:name](nu (x2 x4 v)))); [Intro | Auto]; Inversion_clear H34; Cut (notin x5 (nu (x2 x4 y))); [Intro | Auto]; Inversion_clear H34; Auto.
Inversion_clear H30; Apply notin_nu; Intros; Cut (notin x5 (nu [v:name](nu (x3 x4 v)))); [Intro | Auto]; Inversion_clear H34; Cut (notin x5 (nu (x3 x4 y))); [Intro | Auto]; Inversion_clear H34; Auto.
Change (NU_fun R n ([_:name](x2 x4 _ x5) y) ([_:name](x3 x4 _ x5) y)); Apply Hrecn with x; Auto.
Cut (notin x (nu [y:name](nu [z:name](nu (x2 y z))))); [Intro; Inversion_clear H33; Apply notin_nu; Intros; Cut (notin x (nu [z:name](nu (x2 x4 z)))); [Intro | Auto]; Inversion_clear H35; Cut (notin x (nu (x2 x4 z0))); [Intro | Auto]; Inversion_clear H35; Auto | Apply proc_mono with y; Rewrite <-H18; Assumption].
Cut (notin x (nu [y:name](nu [z:name](nu (x3 y z))))); [Intro; Inversion_clear H33; Apply notin_nu; Intros; Cut (notin x (nu [z:name](nu (x3 x4 z)))); [Intro | Auto]; Inversion_clear H35; Cut (notin x (nu (x3 x4 z0))); [Intro | Auto]; Inversion_clear H35; Auto | Apply proc_mono with y; Rewrite <-H19; Assumption].
Inversion_clear H17; Apply notin_nu; Intros; Cut (notin y (nu [z:name](nu (x2 x4 z)))); [Intro | Auto]; Inversion_clear H34; Cut (notin y (nu (x2 x4 z0))); [Intro | Auto]; Inversion_clear H34; Auto.
Inversion_clear H12; Apply notin_nu; Intros; Cut (notin y (nu [z:name](nu (x3 x4 z)))); [Intro | Auto]; Inversion_clear H34; Cut (notin y (nu (x3 x4 z0))); [Intro | Auto]; Inversion_clear H34; Auto.
Change (NU_fun R n ([_:name](x2 _ x x5) x4) ([_:name](x3 _ x x5) x4)); Apply Hrecn with y; Auto.
Inversion_clear H17; Apply notin_nu; Intros; Cut (notin y (nu [z:name](nu (x2 z0 z)))); [Intro | Auto]; Inversion_clear H34; Cut (notin y (nu (x2 z0 x))); [Intro | Auto]; Inversion_clear H34; Auto.
Inversion_clear H12; Apply notin_nu; Intros; Cut (notin y (nu [z:name](nu (x3 z0 z)))); [Intro | Auto]; Inversion_clear H34; Cut (notin y (nu (x3 z0 x))); [Intro | Auto]; Inversion_clear H34; Auto.
Inversion_clear H23; Apply notin_nu; Intros; Cut (notin x4 (nu [z:name](nu (x2 z0 z)))); [Intro | Auto]; Inversion_clear H34; Cut (notin x4 (nu (x2 z0 x))); [Intro | Auto]; Inversion_clear H34; Auto.
Inversion_clear H25; Apply notin_nu; Intros; Cut (notin x4 (nu [z:name](nu (x3 z0 z)))); [Intro | Auto]; Inversion_clear H34; Cut (notin x4 (nu (x3 z0 x))); [Intro | Auto]; Inversion_clear H34; Auto.
Rewrite <-H18; Rewrite <-H19; Apply H8; Auto.
Rewrite H18; Inversion_clear H27; Cut (notin x5 (nu [v:name](nu (x2 y v)))); [Intro | Auto]; Inversion_clear H27; Auto.
Rewrite H19; Inversion_clear H30; Cut (notin x5 (nu [v:name](nu (x3 y v)))); [Intro | Auto]; Inversion_clear H30; Auto.

Qed.

(* SBUPTO is a strong late bisimulation. *)

Lemma UPTO_SB': (R:proc->proc->Prop)(UpTo R) -> (Inclus (SBUPTO R) (Op_StBisim (SBUPTO R))).

Proof.

Unfold Inclus; Intros.
Inversion_clear H0.
Cut (NU_fun R n p1 p2);
 [Generalize p1 p2; Generalize n ; Induction n0; Intros | Assumption].

(* Base case *)

Simpl in H0; Inversion H0.
Apply op_sb; Split; Intros.

Inversion_clear H2; Elim a; Inversion_clear H3; Inversion_clear H4; Split; Intros.

Inversion_clear H2; Inversion_clear H6; Elim (H2 tau); Intros; Elim (H6 p4 H4); Intros; Inversion_clear H9.
Unfold UpTo in H; Elim (H x x0); Intros; Auto.
Inversion_clear H12; Inversion_clear H14; Inversion_clear H15; Inversion_clear H16.
Elim (H15 x1 H10); Intros.
Inversion_clear H16; Inversion_clear H5; Inversion_clear H16.
Elim (H5 tau); Intros.
Elim (H16 x2 H18); Intros.
Inversion_clear H22; Exists x3; Split; [Assumption | Inversion_clear H19].

Unfold SB_R_SB in H22; Inversion_clear H22; Inversion_clear H19; Apply sbupto with O; Simpl; Unfold SB_R_SB; Inversion_clear H22; Inversion_clear H25.
Exists x4; Exists x5; Split; [Apply TRANS with x1; Assumption | Split; [Assumption | Apply TRANS with x2; Assumption]].

Inversion_clear H22; Inversion_clear H19; Inversion_clear H22; Inversion_clear H25; Apply sbupto with (S O); Simpl.
Apply op_nu with x4 x5; [Apply TRANS with x1; Assumption | Apply TRANS with x2; Assumption | Intros; Unfold SB_R_SB; Exists (x4 z); Exists (x5 z); Split; [Apply REF | Split; [Auto | Apply REF]]].

Inversion_clear H5; Inversion_clear H6; Elim (H5 tau); Intros; Elim (H8 q1 H4); Intros; Inversion_clear H9.
Unfold UpTo in H; Elim (H x x0); Intros; Auto.
Inversion_clear H12; Inversion_clear H14; Inversion_clear H15; Inversion_clear H16.
Elim (H17 x1 H10); Intros.
Inversion_clear H16; Inversion_clear H2; Inversion_clear H16.
Elim (H2 tau); Intros.
Elim (H21 x2 H18); Intros.
Inversion_clear H22; Exists x3; Split; [Assumption | Inversion_clear H19].

Unfold SB_R_SB in H22; Inversion_clear H22; Inversion_clear H19; Apply sbupto with O; Simpl; Unfold SB_R_SB; Inversion_clear H22; Inversion_clear H25.
Exists x4; Exists x5; Split; [Apply TRANS with x2; Assumption | Split; [Assumption | Apply TRANS with x1; Assumption]].

Inversion_clear H22; Inversion_clear H19; Inversion_clear H22; Inversion_clear H25; Apply sbupto with (S O); Simpl.
Apply op_nu with x4 x5; [Apply TRANS with x2; Assumption | Apply TRANS with x1; Assumption | Intros; Unfold SB_R_SB; Exists (x4 z); Exists (x5 z); Split; [Apply REF | Split; [Auto | Apply REF]]].

Inversion_clear H2; Inversion_clear H6; Elim (H2 (Out n1 n2)); Intros; Elim (H6 p4 H4); Intros; Inversion_clear H9.
Unfold UpTo in H; Elim (H x x0); Intros; Auto.
Inversion_clear H12; Inversion_clear H14; Inversion_clear H15; Inversion_clear H16.
Elim (H13 n1 n2); Intros.
Elim (H16 x1 H10); Intros.
Inversion_clear H19; Inversion_clear H5; Inversion_clear H19.
Elim (H5 (Out n1 n2)); Intros.
Elim (H19 x2 H20); Intros.
Inversion_clear H24; Exists x3; Split; [Assumption | Inversion_clear H21; Inversion_clear H24; Inversion_clear H21; Inversion_clear H27].
Apply sbupto with O; Simpl; Unfold SB_R_SB; Exists x4; Exists x5; Split; [Apply TRANS with x1; Assumption | Split; [Assumption | Apply TRANS with x2; Assumption]].

Inversion_clear H5; Inversion_clear H6; Elim (H5 (Out n1 n2)); Intros; Elim (H8 q1 H4); Intros; Inversion_clear H9.
Unfold UpTo in H; Elim (H x x0); Intros; Auto.
Inversion_clear H12; Inversion_clear H14; Inversion_clear H15; Inversion_clear H16.
Elim (H13 n1 n2); Intros.
Elim (H18 x1 H10); Intros.
Inversion_clear H19; Inversion_clear H2; Inversion_clear H19.
Elim (H2 (Out n1 n2)); Intros.
Elim (H23 x2 H20); Intros.
Inversion_clear H24; Exists x3; Split; [Assumption | Inversion_clear H21; Inversion_clear H24; Inversion_clear H21; Inversion_clear H27].
Apply sbupto with O; Simpl; Unfold SB_R_SB; Exists x4; Exists x5; Split; [Apply TRANS with x2; Assumption | Split; [Assumption | Apply TRANS with x1; Assumption]].

Split; Intros.

Split; Intros.

Inversion_clear H2; Inversion_clear H4; Inversion_clear H5; Inversion_clear H2; Inversion_clear H5; Inversion_clear H7; Elim (H5 x0); Intros; Elim (H7 p4 H3); Intros; Inversion_clear H10.
Unfold UpTo in H; Elim (H x x1); Intros; Auto; Inversion_clear H13; Inversion_clear H15.
Elim (H13 x0); Intros.
Elim (H15 x2 H11); Intros.
Inversion_clear H18; Inversion_clear H6; Inversion_clear H18; Inversion_clear H21.
Elim (H18 x0); Intros.
Elim (H21 x3 H19); Intros.
Inversion_clear H24; Exists x4; Split; [Assumption | Intro; Cut (SB_R_SB R (x2 y) (x3 y)); [Intro | Auto]; Inversion_clear H24; Inversion_clear H27; Inversion_clear H24; Inversion_clear H28].
Apply sbupto with O; Simpl; Unfold SB_R_SB; Exists x5; Exists x6; Split; [Apply TRANS with (x2 y); Auto | Split; [Assumption | Apply TRANS with (x3 y); Auto]].

Inversion_clear H2; Inversion_clear H4; Inversion_clear H5; Inversion_clear H6; Inversion_clear H5; Inversion_clear H7; Elim (H5 x0); Intros; Elim (H9 q1 H3); Intros; Inversion_clear H10.
Unfold UpTo in H; Elim (H x x1); Intros; Auto; Inversion_clear H13; Inversion_clear H15.
Elim (H13 x0); Intros.
Elim (H17 x2 H11); Intros.
Inversion_clear H18; Inversion_clear H2; Inversion_clear H18; Inversion_clear H21.
Elim (H18 x0); Intros.
Elim (H23 x3 H19); Intros.
Inversion_clear H24; Exists x4; Split; [Assumption | Intro; Cut (SB_R_SB R (x3 y) (x2 y)); [Intro | Auto]; Inversion_clear H24; Inversion_clear H27; Inversion_clear H24; Inversion_clear H28].
Apply sbupto with O; Simpl; Unfold SB_R_SB; Exists x5; Exists x6; Split; [Apply TRANS with (x3 y); Auto | Split; [Assumption | Apply TRANS with (x2 y); Auto]].

Split; Intros.

Inversion_clear H2; Inversion_clear H4; Inversion_clear H5; Inversion_clear H2; Inversion_clear H5; Inversion_clear H7; Elim (H8 x0); Intros; Elim (H7 p4 H3); Intros; Inversion_clear H10.
Unfold UpTo in H; Elim (H x x1); Intros; Auto; Inversion_clear H13; Inversion_clear H15; Inversion_clear H16.
Elim (H15 x0); Intros.
Elim (H16 x2 H11); Intros.
Inversion_clear H19; Inversion_clear H6; Inversion_clear H19; Inversion_clear H22.
Elim (H23 x0); Intros.
Elim (H22 x3 H20); Intros.
Inversion_clear H25; Exists x4; Split; [Assumption | Intros; Elim (unsat (par (nu p4) (par (nu x2) (par (nu x3) (nu x4)))) (cons y empty)); Intros; Inversion_clear H29; Inversion_clear H30; Inversion_clear H31; Clear H33; Inversion_clear H32; Inversion_clear H33; Cut (SB_R_SB R (x2 x5) (x3 x5)); [Intro; Inversion_clear H33; Inversion_clear H35; Inversion_clear H33; Inversion_clear H36 | Auto]].
Elim (proc_exp x6 x5); Elim (proc_exp x7 x5); Intros; Inversion_clear H36; Inversion_clear H38; Rewrite H41 in H35; Rewrite H40 in H37.
Elim (ho_proc_exp x2 y); Elim (ho_proc_exp x3 y); Intros; Inversion_clear H38; Inversion_clear H42.
Elim (ho_proc_exp x8 y); Elim (ho_proc_exp x9 y); Intros; Inversion_clear H42; Inversion_clear H46.
Elim (unsat (par (nu p4) (par (nu x4) (par (nu [_:name](nu (x10 _))) (par (nu [_:name](nu (x11 _))) (par (nu [_:name](nu (x12 _))) (nu [_:name](nu (x13 _)))))))) (cons y (cons x5 empty))); Intros.
Inversion_clear H46; Inversion_clear H50; Inversion_clear H51; Inversion_clear H52; Inversion_clear H53; Inversion_clear H54; Clear H55; Inversion_clear H56; Inversion_clear H55.
Apply sbupto with O; Simpl; Unfold SB_R_SB; Exists (x12 x5 y); Exists (x13 x5 y); Split; [Apply TRANS with (x11 x5 y); Apply L6_Light with x14; Auto | Split].
Inversion_clear H54; Auto.
Change (StBisim ([_:name](p4 x14) x5) ([_:name](x11 _ x14) x5)); Apply L6_Light with y; Auto.
Inversion_clear H25; Apply notin_nu; Intros; Auto.
Inversion_clear H38; Apply notin_nu; Intros; Cut (notin y (nu (x11 z))); [Intro | Auto]; Inversion_clear H58; Auto.
Rewrite <-H45; Apply H12; Auto.
Rewrite H45; Inversion_clear H54; Auto.
Inversion_clear H29; Apply notin_nu; Intros; Auto.
Apply proc_mono with y; Rewrite <-H45; Inversion_clear H31; Auto.
Inversion_clear H38; Auto.
Inversion_clear H54; Auto.
Inversion_clear H56; Auto.
Change (StBisim ([_:name](x11 _ x14) x5) ([_:name](x12 _ x14) x5)); Apply L6_Light with y; Auto.
Inversion_clear H38; Apply notin_nu; Intros; Cut (notin y (nu (x11 z))); [Intro | Auto]; Inversion_clear H58; Auto.
Inversion_clear H47; Apply notin_nu; Intros; Cut (notin y (nu (x12 z))); [Intro | Auto]; Inversion_clear H58; Auto.
Rewrite <-H45; Rewrite <-H48; Apply L6_Light with x5; Auto.
Rewrite H45; Inversion_clear H54; Auto.
Rewrite H48; Inversion_clear H56; Auto.
Apply proc_mono with y; Rewrite <-H45; Inversion_clear H31; Auto.
Apply proc_mono with y; Rewrite <-H48; Inversion_clear H36; Auto.
Inversion_clear H38; Auto.
Inversion_clear H47; Auto.
Rewrite H40 in H33; Rewrite H41 in H33; Rewrite H48 in H33; Rewrite H49 in H33.
Elim (H (x12 y x5) (x13 y x5) H33); Intros; Clear H58.
Cut (R (x12 x14 x5) (x13 x14 x5)); [Intro | Change (R ([_:name](x12 _ x5) x14) ([_:name](x13 _ x5) x14)); Apply H55 with y; Auto].
Elim (H (x12 x14 x5) (x13 x14 x5) H58); Intros; Clear H60.
Cut (R (x12 x14 y) (x13 x14 y)); [Intro | Apply H59 with x5; Auto].
Elim (H (x12 x14 y) (x13 x14 y) H60); Intros; Clear H62.
Cut (R (x12 x5 y) (x13 x5 y)); [Intro; Assumption | Change (R ([_:name](x12 _ y) x5) ([_:name](x13 _ y) x5)); Apply H61 with x14; Auto].
Inversion_clear H56; Apply notin_nu; Intros; Cut (notin x14 (nu (x12 z))); [Intro | Auto]; Inversion_clear H63; Auto.
Inversion_clear H57; Apply notin_nu; Intros; Cut (notin x14 (nu (x13 z))); [Intro | Auto]; Inversion_clear H63; Auto.
Apply proc_mono with y; Rewrite <-H48; Inversion_clear H36; Auto.
Apply proc_mono with y; Rewrite <-H49; Inversion_clear H39; Auto.
Cut (notin x5 (nu [_:name](nu (x12 _)))); [Intro; Inversion_clear H60; Auto | Apply proc_mono with y; Rewrite <-H48; Assumption].
Cut (notin x5 (nu [_:name](nu (x13 _)))); [Intro; Inversion_clear H60; Auto | Apply proc_mono with y; Rewrite <-H49; Assumption].
Inversion_clear H47; Auto.
Inversion_clear H42; Auto.
Inversion_clear H47; Apply notin_nu; Intros; Cut (notin y (nu (x12 z))); [Intro | Auto]; Inversion_clear H59; Auto.
Inversion_clear H42; Apply notin_nu; Intros; Cut (notin y (nu (x13 z))); [Intro | Auto]; Inversion_clear H59; Auto.
Inversion_clear H56; Apply notin_nu; Intros; Cut (notin x14 (nu (x12 z))); [Intro | Auto]; Inversion_clear H59; Auto.
Inversion_clear H57; Apply notin_nu; Intros; Cut (notin x14 (nu (x13 z))); [Intro | Auto]; Inversion_clear H59; Auto.
Apply TRANS with (x10 x5 y); Apply L6_Light with x14; Auto.
Inversion_clear H57; Auto.
Inversion_clear H53; Auto.
Change (StBisim ([_:name](x13 _ x14) x5) ([_:name](x10 _ x14) x5)); Apply L6_Light with y; Auto.
Inversion_clear H42; Apply notin_nu; Intros; Cut (notin y (nu (x13 z))); [Intro | Auto]; Inversion_clear H58; Auto.
Inversion_clear H43; Apply notin_nu; Intros; Cut (notin y (nu (x10 z))); [Intro | Auto]; Inversion_clear H58; Auto.
Apply L6_Light with x5; Auto.
Rewrite <-H49; Assumption.
Rewrite <-H44; Assumption.
Rewrite <-H49; Rewrite <-H44; Assumption.
Inversion_clear H57; Auto.
Inversion_clear H53; Auto.
Apply proc_mono with y; Rewrite <-H49; Inversion_clear H39; Auto.
Apply proc_mono with y; Rewrite <-H44; Inversion_clear H32; Auto.
Inversion_clear H42; Auto.
Inversion_clear H43; Auto.
Inversion_clear H53; Auto.
Change (StBisim ([_:name](x10 _ x14) x5) ([_:name](x4 x14) x5)); Apply L6_Light with y; Auto.
Inversion_clear H43; Apply notin_nu; Intros; Cut (notin y (nu (x10 z))); [Intro | Auto]; Inversion_clear H58; Auto.
Inversion_clear H28; Apply notin_nu; Intros; Auto.
Rewrite <-H44; Apply H27; Auto.
Rewrite H44; Inversion_clear H53; Auto.
Apply proc_mono with y; Rewrite <-H44; Inversion_clear H32; Auto.
Inversion_clear H34; Apply notin_nu; Intros; Auto.
Inversion_clear H43; Auto.

Inversion_clear H2; Inversion_clear H4; Inversion_clear H5; Inversion_clear H6; Inversion_clear H5; Inversion_clear H7; Elim (H8 x0); Intros; Elim (H9 q1 H3); Intros; Inversion_clear H10.
Unfold UpTo in H; Elim (H x x1); Intros; Auto; Inversion_clear H13; Inversion_clear H15; Inversion_clear H16.
Elim (H15 x0); Intros.
Elim (H18 x2 H11); Intros.
Inversion_clear H19; Inversion_clear H2; Inversion_clear H19; Inversion_clear H22.
Elim (H23 x0); Intros.
Elim (H24 x3 H20); Intros.
Inversion_clear H25; Exists x4; Split; [Assumption | Intros; Elim (unsat (par (nu q1) (par (nu x2) (par (nu x3) (nu x4)))) (cons y empty)); Intros; Inversion_clear H29; Inversion_clear H30; Inversion_clear H31; Clear H33; Inversion_clear H32; Inversion_clear H33; Cut (SB_R_SB R (x3 x5) (x2 x5)); [Intro; Inversion_clear H33; Inversion_clear H35; Inversion_clear H33; Inversion_clear H36 | Auto]].
Elim (proc_exp x6 x5); Elim (proc_exp x7 x5); Intros; Inversion_clear H36; Inversion_clear H38; Rewrite H41 in H35; Rewrite H40 in H37.
Elim (ho_proc_exp x2 y); Elim (ho_proc_exp x3 y); Intros; Inversion_clear H38; Inversion_clear H42.
Elim (ho_proc_exp x8 y); Elim (ho_proc_exp x9 y); Intros; Inversion_clear H42; Inversion_clear H46.
Elim (unsat (par (nu q1) (par (nu x4) (par (nu [_:name](nu (x10 _))) (par (nu [_:name](nu (x11 _))) (par (nu [_:name](nu (x12 _))) (nu [_:name](nu (x13 _)))))))) (cons y (cons x5 empty))); Intros.
Inversion_clear H46; Inversion_clear H50; Inversion_clear H51; Inversion_clear H52; Inversion_clear H53; Inversion_clear H54; Clear H55; Inversion_clear H56; Inversion_clear H55.
Apply sbupto with O; Simpl; Unfold SB_R_SB; Exists (x12 x5 y); Exists (x13 x5 y); Split; [Apply TRANS with (x10 x5 y); Apply L6_Light with x14; Auto | Split].
Inversion_clear H53; Auto.
Change (StBisim ([_:name](x4 x14) x5) ([_:name](x10 _ x14) x5)); Apply L6_Light with y; Auto.
Inversion_clear H25; Apply notin_nu; Intros; Auto.
Inversion_clear H43; Apply notin_nu; Intros; Cut (notin y (nu (x10 z))); [Intro | Auto]; Inversion_clear H58; Auto.
Rewrite <-H44; Apply H27; Auto.
Rewrite H44; Inversion_clear H53; Auto.
Inversion_clear H34; Apply notin_nu; Intros; Auto.
Apply proc_mono with y; Rewrite <-H44; Inversion_clear H32; Auto.
Inversion_clear H43; Auto.
Inversion_clear H53; Auto.
Inversion_clear H56; Auto.
Change (StBisim ([_:name](x10 _ x14) x5) ([_:name](x12 _ x14) x5)); Apply L6_Light with y; Auto.
Inversion_clear H43; Apply notin_nu; Intros; Cut (notin y (nu (x10 z))); [Intro | Auto]; Inversion_clear H58; Auto.
Inversion_clear H47; Apply notin_nu; Intros; Cut (notin y (nu (x12 z))); [Intro | Auto]; Inversion_clear H58; Auto.
Rewrite <-H44; Rewrite <-H48; Apply L6_Light with x5; Auto.
Rewrite H44; Inversion_clear H53; Auto.
Rewrite H48; Inversion_clear H56; Auto.
Apply proc_mono with y; Rewrite <-H44; Inversion_clear H32; Auto.
Apply proc_mono with y; Rewrite <-H48; Inversion_clear H36; Auto.
Inversion_clear H43; Auto.
Inversion_clear H47; Auto.
Rewrite H40 in H33; Rewrite H41 in H33; Rewrite H48 in H33; Rewrite H49 in H33.
Elim (H (x12 y x5) (x13 y x5) H33); Intros; Clear H58.
Cut (R (x12 x14 x5) (x13 x14 x5)); [Intro | Change (R ([_:name](x12 _ x5) x14) ([_:name](x13 _ x5) x14)); Apply H55 with y; Auto].
Elim (H (x12 x14 x5) (x13 x14 x5) H58); Intros; Clear H60.
Cut (R (x12 x14 y) (x13 x14 y)); [Intro | Apply H59 with x5; Auto].
Elim (H (x12 x14 y) (x13 x14 y) H60); Intros; Clear H62.
Cut (R (x12 x5 y) (x13 x5 y)); [Intro; Assumption | Change (R ([_:name](x12 _ y) x5) ([_:name](x13 _ y) x5)); Apply H61 with x14; Auto].
Inversion_clear H56; Apply notin_nu; Intros; Cut (notin x14 (nu (x12 z))); [Intro | Auto]; Inversion_clear H63; Auto.
Inversion_clear H57; Apply notin_nu; Intros; Cut (notin x14 (nu (x13 z))); [Intro | Auto]; Inversion_clear H63; Auto.
Apply proc_mono with y; Rewrite <-H48; Inversion_clear H36; Auto.
Apply proc_mono with y; Rewrite <-H49; Inversion_clear H39; Auto.
Cut (notin x5 (nu [_:name](nu (x12 _)))); [Intro; Inversion_clear H60; Auto | Apply proc_mono with y; Rewrite <-H48; Assumption].
Cut (notin x5 (nu [_:name](nu (x13 _)))); [Intro; Inversion_clear H60; Auto | Apply proc_mono with y; Rewrite <-H49; Assumption].
Inversion_clear H47; Auto.
Inversion_clear H42; Auto.
Inversion_clear H47; Apply notin_nu; Intros; Cut (notin y (nu (x12 z))); [Intro | Auto]; Inversion_clear H59; Auto.
Inversion_clear H42; Apply notin_nu; Intros; Cut (notin y (nu (x13 z))); [Intro | Auto]; Inversion_clear H59; Auto.
Inversion_clear H56; Apply notin_nu; Intros; Cut (notin x14 (nu (x12 z))); [Intro | Auto]; Inversion_clear H59; Auto.
Inversion_clear H57; Apply notin_nu; Intros; Cut (notin x14 (nu (x13 z))); [Intro | Auto]; Inversion_clear H59; Auto.
Apply TRANS with (x11 x5 y); Apply L6_Light with x14; Auto.
Inversion_clear H57; Auto.
Inversion_clear H54; Auto.
Change (StBisim ([_:name](x13 _ x14) x5) ([_:name](x11 _ x14) x5)); Apply L6_Light with y; Auto.
Inversion_clear H42; Apply notin_nu; Intros; Cut (notin y (nu (x13 z))); [Intro | Auto]; Inversion_clear H58; Auto.
Inversion_clear H38; Apply notin_nu; Intros; Cut (notin y (nu (x11 z))); [Intro | Auto]; Inversion_clear H58; Auto.
Apply L6_Light with x5; Auto.
Rewrite <-H49; Assumption.
Rewrite <-H45; Assumption.
Rewrite <-H49; Rewrite <-H45; Assumption.
Inversion_clear H57; Auto.
Inversion_clear H54; Auto.
Apply proc_mono with y; Rewrite <-H49; Inversion_clear H39; Auto.
Apply proc_mono with y; Rewrite <-H45; Inversion_clear H31; Auto.
Inversion_clear H42; Auto.
Inversion_clear H38; Auto.
Inversion_clear H54; Auto.
Change (StBisim ([_:name](x11 _ x14) x5) ([_:name](q1 x14) x5)); Apply L6_Light with y; Auto.
Inversion_clear H38; Apply notin_nu; Intros; Cut (notin y (nu (x11 z))); [Intro | Auto]; Inversion_clear H58; Auto.
Inversion_clear H28; Apply notin_nu; Intros; Auto.
Rewrite <-H45; Apply H12; Auto.
Rewrite H45; Inversion_clear H54; Auto.
Apply proc_mono with y; Rewrite <-H45; Inversion_clear H31; Auto.
Inversion_clear H29; Apply notin_nu; Intros; Auto.
Inversion_clear H38; Auto.

(* Inductive step *)

Simpl in H2; Inversion H2.
Apply op_sb; Split; Intros.

Elim a; Intros; Split; Intros.

Inversion_clear H3; Inversion_clear H9.
Elim (H3 tau); Intros.
Elim (H9 p4 H8); Intros.
Inversion_clear H12; Inversion H13.
Elim (unsat (par (nu p') (par (nu q') (nu p6))) l); Intros.
Inversion_clear H18; Inversion_clear H19; Inversion_clear H21.
Cut (ftrans (p' x0) tau (p6 x0)); [Intro; Cut (NU_fun R n1 (p' x0) (q' x0)); [Intro | Auto] | Apply H15; Auto; Try (Apply f_act_notin_tau)].
Cut (Op_StBisim (SBUPTO R) (p' x0) (q' x0)); [Intro | Auto].
Inversion_clear H24; Inversion_clear H25.
Elim (H24 tau); Intros.
Elim (H25 (p6 x0) H21); Intros; Inversion_clear H28.
Elim (proc_exp x1 x0); Intros; Inversion_clear H28; Rewrite H32 in H29; Rewrite H32 in H30.
Inversion_clear H4; Inversion_clear H28.
Elim (H4 tau); Intros; Cut (ftrans (nu q') tau (nu x2)); [Intro | Apply fRES with empty; Intros; Change (ftrans (q' y) ([_:name]tau y) (x2 y)); Apply FTR_L3 with x0; Auto; Try (Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau)].
Elim (H28 (nu x2) H35); Intros; Inversion_clear H36.
Exists x3; Split; [Assumption | Inversion_clear H30].
Apply sbupto with (S n2); Simpl; Apply op_nu with p6 x2; [Rewrite H17; Assumption | Assumption | Intros; Apply NU_FUN_RW with x0; Auto].

Inversion_clear H4; Inversion_clear H9.
Elim (H4 tau); Intros.
Elim (H11 q1 H8); Intros.
Inversion_clear H12; Inversion H13.
Elim (unsat (par (nu p') (par (nu q') (nu p5))) l); Intros.
Inversion_clear H18; Inversion_clear H19; Inversion_clear H21.
Cut (ftrans (q' x0) tau (p5 x0)); [Intro; Cut (NU_fun R n1 (p' x0) (q' x0)); [Intro | Auto] | Apply H15; Auto; Try (Apply f_act_notin_tau)].
Cut (Op_StBisim (SBUPTO R) (p' x0) (q' x0)); [Intro | Auto].
Inversion_clear H24; Inversion_clear H25.
Elim (H24 tau); Intros.
Elim (H27 (p5 x0) H21); Intros; Inversion_clear H28.
Elim (proc_exp x1 x0); Intros; Inversion_clear H28; Rewrite H32 in H29; Rewrite H32 in H30.
Inversion_clear H3; Inversion_clear H28.
Elim (H3 tau); Intros; Cut (ftrans (nu p') tau (nu x2)); [Intro | Apply fRES with empty; Intros; Change (ftrans (p' y) ([_:name]tau y) (x2 y)); Apply FTR_L3 with x0; Auto; Try (Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau)].
Elim (H34 (nu x2) H35); Intros; Inversion_clear H36.
Exists x3; Split; [Assumption | Inversion_clear H30].
Apply sbupto with (S n2); Simpl; Apply op_nu with x2 p5; [Assumption | Rewrite H17; Assumption | Intros; Apply NU_FUN_RW with x0; Auto].

Inversion_clear H3; Inversion_clear H9.
Elim (H3 (Out n2 n3)); Intros.
Elim (H9 p4 H8); Intros.
Inversion_clear H12; Inversion H13.
Elim (unsat (par (nu p') (par (nu q') (nu p6))) (cons n2 (cons n3 l))); Intros.
Inversion_clear H18; Inversion_clear H19; Inversion_clear H20; Inversion_clear H21; Inversion_clear H22.
Cut (ftrans (p' x0) (Out n2 n3) (p6 x0)); [Intro; Cut (NU_fun R n1 (p' x0) (q' x0)); [Intro | Auto] | Apply H15; Auto; Try (Apply f_act_notin_Out; Auto)].
Cut (Op_StBisim (SBUPTO R) (p' x0) (q' x0)); [Intro | Auto].
Inversion_clear H26; Inversion_clear H27.
Elim (H26 (Out n2 n3)); Intros.
Elim (H27 (p6 x0) H22); Intros; Inversion_clear H30.
Elim (proc_exp x1 x0); Intros; Inversion_clear H30; Rewrite H34 in H31; Rewrite H34 in H32.
Inversion_clear H4; Inversion_clear H30.
Elim (H4 (Out n2 n3)); Intros; Cut (ftrans (nu q') (Out n2 n3) (nu x2)); [Intro | Apply fRES with empty; Intros; Change (ftrans (q' y) ([_:name](Out n2 n3) y) (x2 y)); Apply FTR_L3 with x0; Auto; Try (Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto)].
Elim (H30 (nu x2) H37); Intros; Inversion_clear H38.
Exists x3; Split; [Assumption | Inversion_clear H32].
Apply sbupto with (S n4); Simpl; Apply op_nu with p6 x2; [Rewrite H17; Assumption | Assumption | Intros; Apply NU_FUN_RW with x0; Auto].
Inversion_clear H40; Auto.
Inversion_clear H40; Auto.

Inversion_clear H4; Inversion_clear H9.
Elim (H4 (Out n2 n3)); Intros.
Elim (H11 q1 H8); Intros.
Inversion_clear H12; Inversion H13.
Elim (unsat (par (nu p') (par (nu q') (nu p5))) (cons n2 (cons n3 l))); Intros.
Inversion_clear H18; Inversion_clear H19; Inversion_clear H20; Inversion_clear H21; Inversion_clear H22.
Cut (ftrans (q' x0) (Out n2 n3) (p5 x0)); [Intro; Cut (NU_fun R n1 (p' x0) (q' x0)); [Intro | Auto] | Apply H15; Auto; Try (Apply f_act_notin_Out; Auto)].
Cut (Op_StBisim (SBUPTO R) (p' x0) (q' x0)); [Intro | Auto].
Inversion_clear H26; Inversion_clear H27.
Elim (H26 (Out n2 n3)); Intros.
Elim (H29 (p5 x0) H22); Intros; Inversion_clear H30.
Elim (proc_exp x1 x0); Intros; Inversion_clear H30; Rewrite H34 in H31; Rewrite H34 in H32.
Inversion_clear H3; Inversion_clear H30.
Elim (H3 (Out n2 n3)); Intros; Cut (ftrans (nu p') (Out n2 n3) (nu x2)); [Intro | Apply fRES with empty; Intros; Change (ftrans (p' y) ([_:name](Out n2 n3) y) (x2 y)); Apply FTR_L3 with x0; Auto; Try (Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto)].
Elim (H36 (nu x2) H37); Intros; Inversion_clear H38.
Exists x3; Split; [Assumption | Inversion_clear H32].
Apply sbupto with (S n4); Simpl; Apply op_nu with x2 p5; [Assumption | Rewrite H17; Assumption | Intros; Apply NU_FUN_RW with x0; Auto].
Inversion_clear H40; Auto.
Inversion_clear H40; Auto.

Split; Intros.

Split; Intros.

Inversion_clear H3; Inversion_clear H9; Inversion_clear H10.
Elim (H9 x); Intros.
Elim (H10 p4 H8); Intros.
Inversion_clear H13; Inversion H14.
Elim (unsat (par (nu p') (par (nu q') (nu [z:name](nu (p6 z))))) (cons x l)); Intros.
Inversion_clear H19; Inversion_clear H20; Inversion_clear H21; Inversion_clear H22.
Cut (btrans (p' x1) (In x) (p6 x1)); [Intro; Cut (NU_fun R n1 (p' x1) (q' x1)); [Intro | Auto] | Apply H16; Auto; Try (Apply b_act_notin_In; Auto)].
Cut (Op_StBisim (SBUPTO R) (p' x1) (q' x1)); [Intro | Auto].
Inversion_clear H26; Inversion_clear H27; Inversion_clear H28.
Elim (H27 x); Intros.
Elim (H28 (p6 x1) H22); Intros; Inversion_clear H31.
Elim (ho_proc_exp x2 x1); Intros; Inversion_clear H31; Rewrite H35 in H32; Rewrite H35 in H33.
Inversion_clear H4; Inversion_clear H31; Inversion_clear H36.
Elim (H31 x); Intros.
Cut (btrans (nu q') (In x) [u:name](nu [v:name](x3 v u))); [Intro | Apply bRES with (cons x empty); Intros; Inversion_clear H42; Change (btrans (q' y) ([_:name](In x) y) (x3 y)); Apply BTR_L3 with x1; Auto; Try (Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto)].
Elim (H36 [u:name](nu [v:name](x3 v u)) H39); Intros; Inversion_clear H40.
Exists x4; Split; [Assumption | Idtac].
Intro; Elim (LEM_name x1 y); Intros.

Rewrite <-H40; Elim (unsat (par (nu [u:name](nu [v:name](p6 v u))) (nu [u:name](nu [v:name](x3 v u)))) (cons x1 empty)); Intros; Inversion_clear H43; Inversion_clear H44; Inversion_clear H45; Clear H47.
Elim (H33 x5); Intros.
Apply sbupto with (S n2); Simpl; Apply op_nu with [v:name](p6 v x1) [v:name](x3 v x1); Intros; Auto.
Cut (nu [v:name](p6 v x1))=(x0 x1); [Intro; Rewrite H47; Auto | Rewrite <-H18; Trivial].
Elim (unsat (par (nu [u:name](nu [v:name](p6 v u))) (nu [u:name](nu [v:name](x3 v u)))) (cons x1 (cons x5 (cons z empty)))); Intros; Inversion_clear H49; Inversion_clear H50; Inversion_clear H51; Inversion_clear H53; Inversion_clear H54; Clear H55.
Change (NU_fun R n2 ([_:name](p6 _ x1) z) ([_:name](x3 _ x1) z)); Apply NU_FUN_RW with x6; Auto.
Inversion_clear H49; Apply notin_nu; Intros; Cut (notin x6 (nu [v:name](p6 v x1))); [Intro | Auto]; Inversion_clear H55; Auto.
Inversion_clear H52; Apply notin_nu; Intros; Cut (notin x6 (nu [v:name](x3 v x1))); [Intro | Auto]; Inversion_clear H55; Auto.
Apply NU_FUN_RW with x5; Auto.
Inversion_clear H43; Apply notin_nu; Intros; Cut (notin x5 (nu [v:name](p6 v z0))); [Intro | Auto]; Inversion_clear H55; Auto.
Inversion_clear H46; Apply notin_nu; Intros; Cut (notin x5 (nu [v:name](x3 v z0))); [Intro | Auto]; Inversion_clear H55; Auto.
Change (NU_fun R n2 ([_:name](p6 _ x5) x6) ([_:name](x3 _ x5) x6)); Apply NU_FUN_RW with x1; Auto.
Inversion_clear H24; Apply notin_nu; Intros; Cut (notin x1 (nu (p6 z0))); [Intro | Auto]; Inversion_clear H55; Auto.
Inversion_clear H34; Apply notin_nu; Intros; Cut (notin x1 (nu (x3 z0))); [Intro | Auto]; Inversion_clear H55; Auto.
Inversion_clear H49; Apply notin_nu; Intros; Cut (notin x6 (nu [v:name](p6 v x5))); [Intro | Auto]; Inversion_clear H55; Auto.
Inversion_clear H52; Apply notin_nu; Intros; Cut (notin x6 (nu [v:name](x3 v x5))); [Intro | Auto]; Inversion_clear H55; Auto.
Inversion_clear H24; Auto.
Inversion_clear H34; Auto.

Elim (H33 y); Intros.
Apply sbupto with (S n2); Simpl; Apply op_nu with [v:name](p6 v y) [v:name](x3 v y); Intros; Auto.
Cut (nu [v:name](p6 v y))=(x0 y); [Intro; Rewrite H44; Auto | Rewrite <-H18; Trivial].
Change (NU_fun R n2 ([_:name](p6 _ y) z) ([_:name](x3 _ y) z)); Apply NU_FUN_RW with x1; Auto.
Inversion_clear H24; Apply notin_nu; Intros; Cut (notin x1 (nu (p6 z0))); [Intro | Auto]; Inversion_clear H47; Auto.
Inversion_clear H34; Apply notin_nu; Intros; Cut (notin x1 (nu (x3 z0))); [Intro | Auto]; Inversion_clear H47; Auto.

Inversion_clear H4; Inversion_clear H9; Inversion_clear H10.
Elim (H9 x); Intros.
Elim (H12 q1 H8); Intros.
Inversion_clear H13; Inversion H14.
Elim (unsat (par (nu p') (par (nu q') (nu [z:name](nu (p5 z))))) (cons x l)); Intros.
Inversion_clear H19; Inversion_clear H20; Inversion_clear H21; Inversion_clear H22.
Cut (btrans (q' x1) (In x) (p5 x1)); [Intro; Cut (NU_fun R n1 (p' x1) (q' x1)); [Intro | Auto] | Apply H16; Auto; Try (Apply b_act_notin_In; Auto)].
Cut (Op_StBisim (SBUPTO R) (p' x1) (q' x1)); [Intro | Auto].
Inversion_clear H26; Inversion_clear H27; Inversion_clear H28.
Elim (H27 x); Intros.
Elim (H30 (p5 x1) H22); Intros; Inversion_clear H31.
Elim (ho_proc_exp x2 x1); Intros; Inversion_clear H31; Rewrite H35 in H32; Rewrite H35 in H33.
Inversion_clear H3; Inversion_clear H31; Inversion_clear H36.
Elim (H31 x); Intros.
Cut (btrans (nu p') (In x) [u:name](nu [v:name](x3 v u))); [Intro | Apply bRES with (cons x empty); Intros; Inversion_clear H42; Change (btrans (p' y) ([_:name](In x) y) (x3 y)); Apply BTR_L3 with x1; Auto; Try (Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto)].
Elim (H38 [u:name](nu [v:name](x3 v u)) H39); Intros; Inversion_clear H40.
Exists x4; Split; [Assumption | Idtac].
Intro; Elim (LEM_name x1 y); Intros.

Rewrite <-H40; Elim (unsat (par (nu [u:name](nu [v:name](p5 v u))) (nu [u:name](nu [v:name](x3 v u)))) (cons x1 empty)); Intros; Inversion_clear H43; Inversion_clear H44; Inversion_clear H45; Clear H47.
Elim (H33 x5); Intros.
Apply sbupto with (S n2); Simpl; Apply op_nu with [v:name](x3 v x1) [v:name](p5 v x1); Intros; Auto.
Cut (nu [v:name](p5 v x1))=(x0 x1); [Intro; Rewrite H47; Auto | Rewrite <-H18; Trivial].
Elim (unsat (par (nu [u:name](nu [v:name](p5 v u))) (nu [u:name](nu [v:name](x3 v u)))) (cons x1 (cons x5 (cons z empty)))); Intros; Inversion_clear H49; Inversion_clear H50; Inversion_clear H51; Inversion_clear H53; Inversion_clear H54; Clear H55.
Change (NU_fun R n2 ([_:name](x3 _ x1) z) ([_:name](p5 _ x1) z)); Apply NU_FUN_RW with x6; Auto.
Inversion_clear H52; Apply notin_nu; Intros; Cut (notin x6 (nu [v:name](x3 v x1))); [Intro | Auto]; Inversion_clear H55; Auto.
Inversion_clear H49; Apply notin_nu; Intros; Cut (notin x6 (nu [v:name](p5 v x1))); [Intro | Auto]; Inversion_clear H55; Auto.
Apply NU_FUN_RW with x5; Auto.
Inversion_clear H46; Apply notin_nu; Intros; Cut (notin x5 (nu [v:name](x3 v z0))); [Intro | Auto]; Inversion_clear H55; Auto.
Inversion_clear H43; Apply notin_nu; Intros; Cut (notin x5 (nu [v:name](p5 v z0))); [Intro | Auto]; Inversion_clear H55; Auto.
Change (NU_fun R n2 ([_:name](x3 _ x5) x6) ([_:name](p5 _ x5) x6)); Apply NU_FUN_RW with x1; Auto.
Inversion_clear H34; Apply notin_nu; Intros; Cut (notin x1 (nu (x3 z0))); [Intro | Auto]; Inversion_clear H55; Auto.
Inversion_clear H24; Apply notin_nu; Intros; Cut (notin x1 (nu (p5 z0))); [Intro | Auto]; Inversion_clear H55; Auto.
Inversion_clear H52; Apply notin_nu; Intros; Cut (notin x6 (nu [v:name](x3 v x5))); [Intro | Auto]; Inversion_clear H55; Auto.
Inversion_clear H49; Apply notin_nu; Intros; Cut (notin x6 (nu [v:name](p5 v x5))); [Intro | Auto]; Inversion_clear H55; Auto.
Inversion_clear H34; Auto.
Inversion_clear H24; Auto.

Elim (H33 y); Intros.
Apply sbupto with (S n2); Simpl; Apply op_nu with [v:name](x3 v y) [v:name](p5 v y); Intros; Auto.
Cut (nu [v:name](p5 v y))=(x0 y); [Intro; Rewrite H44; Auto | Rewrite <-H18; Trivial].
Change (NU_fun R n2 ([_:name](x3 _ y) z) ([_:name](p5 _ y) z)); Apply NU_FUN_RW with x1; Auto.
Inversion_clear H34; Apply notin_nu; Intros; Cut (notin x1 (nu (x3 z0))); [Intro | Auto]; Inversion_clear H47; Auto.
Inversion_clear H24; Apply notin_nu; Intros; Cut (notin x1 (nu (p5 z0))); [Intro | Auto]; Inversion_clear H47; Auto.

Split; Intros.

Inversion_clear H3; Inversion_clear H9; Inversion_clear H10.
Elim (H11 x); Intros.
Elim (H10 p4 H8); Intros.
Inversion_clear H13; Inversion H14.

Elim (unsat (par (nu p') (par (nu q') (par p0 (par p3 (nu [z:name](nu (p6 z))))))) (cons x l)); Intros.
Inversion_clear H19; Inversion_clear H20; Inversion_clear H21; Inversion_clear H22; Inversion_clear H24; Inversion_clear H25.
Cut (btrans (p' x1) (bOut x) (p6 x1)); [Intro; Cut (NU_fun R n1 (p' x1) (q' x1)); [Intro | Auto] | Apply H16; Auto; Try (Apply b_act_notin_bOut; Auto)].
Cut (Op_StBisim (SBUPTO R) (p' x1) (q' x1)); [Intro | Auto].
Inversion_clear H28; Inversion_clear H29; Inversion_clear H30.
Elim (H31 x); Intros.
Elim (H30 (p6 x1) H25); Intros; Inversion_clear H33.
Elim (ho_proc_exp x2 x1); Intros; Inversion_clear H33; Rewrite H37 in H34; Rewrite H37 in H35.
Inversion_clear H4; Inversion_clear H33; Inversion_clear H38.
Elim (H39 x); Intros.
Cut (btrans (nu q') (bOut x) [u:name](nu [v:name](x3 v u))); [Intro | Apply bRES with (cons x empty); Intros; Inversion_clear H44; Change (btrans (q' y) ([_:name](bOut x) y) (x3 y)); Apply BTR_L3 with x1; Auto; Try (Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto)].
Elim (H38 [u:name](nu [v:name](x3 v u)) H41); Intros; Inversion_clear H42.
Exists x4; Split; [Assumption | Idtac].
Intros; Elim (unsat (par (nu p4) (par (nu x4) (par (nu [u:name](nu [v:name](p6 v u))) (nu [u:name](nu [v:name](x3 v u)))))) (cons x1 empty)); Intros; Inversion_clear H46; Inversion_clear H47; Inversion_clear H48; Clear H50; Inversion_clear H49; Inversion_clear H50.
Elim (H35 x5); Intros.
Apply sbupto with (S n2); Apply NU_FUN_RW with x1; Auto.
Elim (BTR_L1 p0 p4 (bOut x) x1 H22 H8); Intros; Assumption.
Elim (BTR_L1 p3 x4 (bOut x) x1 H24 H43); Intros; Assumption.
Simpl; Apply op_nu with [v:name](p6 v x1) [v:name](x3 v x1); Intros; Auto.
Cut (nu [v:name](p6 v x1))=(x0 x1); [Intro; Rewrite H52; Apply H15; Auto | Rewrite <-H18; Trivial].
Elim (BTR_L1 p0 p4 (bOut x) x1 H22 H8); Intros; Assumption.
Elim (BTR_L1 (nu p') x0 (bOut x) x1 H19 H14); Intros; Assumption.
Apply H44; Auto.
Inversion_clear H36; Do 2 (Apply notin_nu; Intros); Cut (notin x1 (nu (x3 z0))); [Intro | Auto]; Inversion_clear H54; Auto.
Elim (BTR_L1 p3 x4 (bOut x) x1 H24 H43); Intros; Assumption.
Elim (unsat (par (nu [u:name](nu [v:name](p6 v u))) (nu [u:name](nu [v:name](x3 v u)))) (cons x1 (cons x5 (cons z empty)))); Intros; Inversion_clear H54; Inversion_clear H55; Inversion_clear H56; Inversion_clear H58; Inversion_clear H59; Clear H60.
Change (NU_fun R n2 ([_:name](p6 _ x1) z) ([_:name](x3 _ x1) z)); Apply NU_FUN_RW with x6; Auto.
Inversion_clear H54; Apply notin_nu; Intros; Cut (notin x6 (nu [v:name](p6 v x1))); [Intro | Auto]; Inversion_clear H60; Auto.
Inversion_clear H57; Apply notin_nu; Intros; Cut (notin x6 (nu [v:name](x3 v x1))); [Intro | Auto]; Inversion_clear H60; Auto.
Apply NU_FUN_RW with x5; Auto.
Inversion_clear H49; Apply notin_nu; Intros; Cut (notin x5 (nu [v:name](p6 v z0))); [Intro | Auto]; Inversion_clear H60; Auto.
Inversion_clear H51; Apply notin_nu; Intros; Cut (notin x5 (nu [v:name](x3 v z0))); [Intro | Auto]; Inversion_clear H60; Auto.
Change (NU_fun R n2 ([_:name](p6 _ x5) x6) ([_:name](x3 _ x5) x6)); Apply NU_FUN_RW with x1; Auto.
Inversion_clear H26; Apply notin_nu; Intros; Cut (notin x1 (nu (p6 z0))); [Intro | Auto]; Inversion_clear H60; Auto.
Inversion_clear H36; Apply notin_nu; Intros; Cut (notin x1 (nu (x3 z0))); [Intro | Auto]; Inversion_clear H60; Auto.
Inversion_clear H54; Apply notin_nu; Intros; Cut (notin x6 (nu [v:name](p6 v x5))); [Intro | Auto]; Inversion_clear H60; Auto.
Inversion_clear H57; Apply notin_nu; Intros; Cut (notin x6 (nu [v:name](x3 v x5))); [Intro | Auto]; Inversion_clear H60; Auto.
Inversion_clear H26; Auto.
Inversion_clear H36; Auto.
Inversion_clear H49; Apply notin_nu; Intros; Cut (notin x5 (nu [v:name](p6 v z))); [Intro | Auto]; Inversion_clear H52; Auto.
Inversion_clear H51; Apply notin_nu; Intros; Cut (notin x5 (nu [v:name](x3 v z))); [Intro | Auto]; Inversion_clear H52; Auto.

Elim (unsat (par (nu p') (par (nu q') (par p0 (par p3 (nu x0))))) (cons x l)); Intros.
Inversion_clear H19; Inversion_clear H20; Inversion_clear H21; Inversion_clear H22; Inversion_clear H24; Inversion_clear H25.
Cut (ftrans (p' x2) (Out x x2) (x0 x2)); [Intro; Cut (NU_fun R n1 (p' x2) (q' x2)); [Intro | Auto] | Apply H17; Auto; Try (Apply f_act_notin_Out; Auto)].
Cut (Op_StBisim (SBUPTO R) (p' x2) (q' x2)); [Intro | Auto].
Inversion_clear H28; Inversion_clear H29.
Elim (H28 (Out x x2)); Intros.
Elim (H29 (x0 x2) H25); Intros; Inversion_clear H32.
Elim (proc_exp x3 x2); Intros; Inversion_clear H32; Rewrite H36 in H33; Rewrite H36 in H34.
Inversion_clear H4; Inversion_clear H32; Inversion_clear H37.
Elim (H38 x); Intros.
Cut (btrans (nu q') (bOut x) x4); [Intro | Apply OPEN with (cons x empty); Intros; Inversion_clear H43;  Apply FTR_L3 with x2; Auto; Try (Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto)].
Elim (H37 x4 H40); Intros; Inversion_clear H41.
Exists x5; Split; [Assumption | Intros; Inversion_clear H34].
Generalize H45; Clear H45; Case n2; Intros; Simpl in H45; Inversion_clear H45.
Inversion_clear H34; Inversion_clear H45; Inversion_clear H46.

Apply sbupto with O; Apply NU_FUN_RW with x2; Auto.
Elim (BTR_L1 p0 p4 (bOut x) x2 H22 H8); Intros; Assumption.
Elim (BTR_L1 p3 x5 (bOut x) x2 H24 H42); Intros; Assumption.
Simpl; Unfold SB_R_SB; Exists x6; Exists x7; Split; [Apply TRANS with (x0 x2); Auto | Split; [Assumption | Apply TRANS with (x4 x2); Auto]].
Apply H15; Auto; Elim (BTR_L1 p0 p4 (bOut x) x2 H22 H8); Intros; Assumption.
Apply H43; Auto; Elim (BTR_L1 p3 x5 (bOut x) x2 H24 H42); Intros; Assumption.

Apply sbupto with (S n3); Apply NU_FUN_RW with x2; Auto.
Elim (BTR_L1 p0 p4 (bOut x) x2 H22 H8); Intros; Assumption.
Elim (BTR_L1 p3 x5 (bOut x) x2 H24 H42); Intros; Assumption.
Simpl; Apply op_nu with p'0 q'0; Intros; Auto.
Apply TRANS with (x0 x2); [Apply H15; Auto; Elim (BTR_L1 p0 p4 (bOut x) x2 H22 H8); Intros; Assumption | Assumption].
Apply TRANS with (x4 x2); [Assumption | Apply H43;  Auto; Elim (BTR_L1 p3 x5 (bOut x) x2 H24 H42); Intros; Assumption].

Inversion_clear H4; Inversion_clear H9; Inversion_clear H10.
Elim (H11 x); Intros.
Elim (H12 q1 H8); Intros.
Inversion_clear H13; Inversion H14.

Elim (unsat (par (nu p') (par (nu q') (par p0 (par p3 (nu [z:name](nu (p5 z))))))) (cons x l)); Intros.
Inversion_clear H19; Inversion_clear H20; Inversion_clear H21; Inversion_clear H22; Inversion_clear H24; Inversion_clear H25.
Cut (btrans (q' x1) (bOut x) (p5 x1)); [Intro; Cut (NU_fun R n1 (p' x1) (q' x1)); [Intro | Auto] | Apply H16; Auto; Try (Apply b_act_notin_bOut; Auto)].
Cut (Op_StBisim (SBUPTO R) (p' x1) (q' x1)); [Intro | Auto].
Inversion_clear H28; Inversion_clear H29; Inversion_clear H30.
Elim (H31 x); Intros.
Elim (H32 (p5 x1) H25); Intros; Inversion_clear H33.
Elim (ho_proc_exp x2 x1); Intros; Inversion_clear H33; Rewrite H37 in H34; Rewrite H37 in H35.
Inversion_clear H3; Inversion_clear H33; Inversion_clear H38.
Elim (H39 x); Intros.
Cut (btrans (nu p') (bOut x) [u:name](nu [v:name](x3 v u))); [Intro | Apply bRES with (cons x empty); Intros; Inversion_clear H44; Change (btrans (p' y) ([_:name](bOut x) y) (x3 y)); Apply BTR_L3 with x1; Auto; Try (Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto)].
Elim (H40 [u:name](nu [v:name](x3 v u)) H41); Intros; Inversion_clear H42.
Exists x4; Split; [Assumption | Idtac].
Intros; Elim (unsat (par (nu p4) (par (nu x4) (par (nu [u:name](nu [v:name](p5 v u))) (nu [u:name](nu [v:name](x3 v u)))))) (cons x1 empty)); Intros; Inversion_clear H46; Inversion_clear H47; Inversion_clear H48; Clear H50; Inversion_clear H49; Inversion_clear H50.
Elim (H35 x5); Intros.
Apply sbupto with (S n2); Apply NU_FUN_RW with x1; Auto.
Elim (BTR_L1 p0 x4 (bOut x) x1 H22 H43); Intros; Assumption.
Elim (BTR_L1 p3 q1 (bOut x) x1 H24 H8); Intros; Assumption.
Simpl; Apply op_nu with [v:name](x3 v x1) [v:name](p5 v x1); Intros; Auto.
Apply H44; Auto.
Elim (BTR_L1 p0 x4 (bOut x) x1 H22 H43); Intros; Assumption.
Inversion_clear H36; Do 2 (Apply notin_nu; Intros); Cut (notin x1 (nu (x3 z0))); [Intro | Auto]; Inversion_clear H54; Auto.
Cut (nu [v:name](p5 v x1))=(x0 x1); [Intro; Rewrite H52; Apply H15; Auto | Rewrite <-H18; Trivial].
Elim (BTR_L1 (nu q') x0 (bOut x) x1 H21 H14); Intros; Assumption.
Elim (BTR_L1 p3 q1 (bOut x) x1 H24 H8); Intros; Assumption.
Elim (unsat (par (nu [u:name](nu [v:name](p5 v u))) (nu [u:name](nu [v:name](x3 v u)))) (cons x1 (cons x5 (cons z empty)))); Intros; Inversion_clear H54; Inversion_clear H55; Inversion_clear H56; Inversion_clear H58; Inversion_clear H59; Clear H60.
Change (NU_fun R n2 ([_:name](x3 _ x1) z) ([_:name](p5 _ x1) z)); Apply NU_FUN_RW with x6; Auto.
Inversion_clear H57; Apply notin_nu; Intros; Cut (notin x6 (nu [v:name](x3 v x1))); [Intro | Auto]; Inversion_clear H60; Auto.
Inversion_clear H54; Apply notin_nu; Intros; Cut (notin x6 (nu [v:name](p5 v x1))); [Intro | Auto]; Inversion_clear H60; Auto.
Apply NU_FUN_RW with x5; Auto.
Inversion_clear H51; Apply notin_nu; Intros; Cut (notin x5 (nu [v:name](x3 v z0))); [Intro | Auto]; Inversion_clear H60; Auto.
Inversion_clear H49; Apply notin_nu; Intros; Cut (notin x5 (nu [v:name](p5 v z0))); [Intro | Auto]; Inversion_clear H60; Auto.
Change (NU_fun R n2 ([_:name](x3 _ x5) x6) ([_:name](p5 _ x5) x6)); Apply NU_FUN_RW with x1; Auto.
Inversion_clear H36; Apply notin_nu; Intros; Cut (notin x1 (nu (x3 z0))); [Intro | Auto]; Inversion_clear H60; Auto.
Inversion_clear H26; Apply notin_nu; Intros; Cut (notin x1 (nu (p5 z0))); [Intro | Auto]; Inversion_clear H60; Auto.
Inversion_clear H57; Apply notin_nu; Intros; Cut (notin x6 (nu [v:name](x3 v x5))); [Intro | Auto]; Inversion_clear H60; Auto.
Inversion_clear H54; Apply notin_nu; Intros; Cut (notin x6 (nu [v:name](p5 v x5))); [Intro | Auto]; Inversion_clear H60; Auto.
Inversion_clear H36; Auto.
Inversion_clear H26; Auto.
Inversion_clear H51; Apply notin_nu; Intros; Cut (notin x5 (nu [v:name](x3 v z))); [Intro | Auto]; Inversion_clear H52; Auto.
Inversion_clear H49; Apply notin_nu; Intros; Cut (notin x5 (nu [v:name](p5 v z))); [Intro | Auto]; Inversion_clear H52; Auto.

Elim (unsat (par (nu p') (par (nu q') (par p0 (par p3 (nu x0))))) (cons x l)); Intros.
Inversion_clear H19; Inversion_clear H20; Inversion_clear H21; Inversion_clear H22; Inversion_clear H24; Inversion_clear H25.
Cut (ftrans (q' x2) (Out x x2) (x0 x2)); [Intro; Cut (NU_fun R n1 (p' x2) (q' x2)); [Intro | Auto] | Apply H17; Auto; Try (Apply f_act_notin_Out; Auto)].
Cut (Op_StBisim (SBUPTO R) (p' x2) (q' x2)); [Intro | Auto].
Inversion_clear H28; Inversion_clear H29.
Elim (H28 (Out x x2)); Intros.
Elim (H31 (x0 x2) H25); Intros; Inversion_clear H32.
Elim (proc_exp x3 x2); Intros; Inversion_clear H32; Rewrite H36 in H33; Rewrite H36 in H34.
Inversion_clear H3; Inversion_clear H32; Inversion_clear H37.
Elim (H38 x); Intros.
Cut (btrans (nu p') (bOut x) x4); [Intro | Apply OPEN with (cons x empty); Intros; Inversion_clear H43;  Apply FTR_L3 with x2; Auto; Try (Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto)].
Elim (H39 x4 H40); Intros; Inversion_clear H41.
Exists x5; Split; [Assumption | Intros; Inversion_clear H34].
Generalize H45; Clear H45; Case n2; Intros; Simpl in H45; Inversion_clear H45.
Inversion_clear H34; Inversion_clear H45; Inversion_clear H46.

Apply sbupto with O; Apply NU_FUN_RW with x2; Auto.
Elim (BTR_L1 p0 x5 (bOut x) x2 H22 H42); Intros; Assumption.
Elim (BTR_L1 p3 q1 (bOut x) x2 H24 H8); Intros; Assumption.
Simpl; Unfold SB_R_SB; Exists x6; Exists x7; Split; [Apply TRANS with (x4 x2); Auto | Split; [Assumption | Apply TRANS with (x0 x2); Auto]].
Apply H43; Auto; Elim (BTR_L1 p0 x5 (bOut x) x2 H22 H42); Intros; Assumption.
Apply H15; Auto; Elim (BTR_L1 p3 q1 (bOut x) x2 H24 H8); Intros; Assumption.

Apply sbupto with (S n3); Apply NU_FUN_RW with x2; Auto.
Elim (BTR_L1 p0 x5 (bOut x) x2 H22 H42); Intros; Assumption.
Elim (BTR_L1 p3 q1 (bOut x) x2 H24 H8); Intros; Assumption.
Simpl; Apply op_nu with p'0 q'0; Intros; Auto.
Apply TRANS with (x4 x2); [Apply H43;  Auto; Elim (BTR_L1 p0 x5 (bOut x) x2 H22 H42); Intros; Assumption | Assumption].
Apply TRANS with (x0 x2); [Assumption | Apply H15; Auto; Elim (BTR_L1 p3 q1 (bOut x) x2 H24 H8); Intros; Assumption].

Qed.

(*******************************************************************)
(* Here is the definition of the strong bisimulation UpTo allowing *)
(* to prove the extrusion law of the restriction operator.         *)
(*******************************************************************)

Inductive BisNU_E: proc -> proc -> Prop :=
  bisnu_e: (p:name->proc)(q:proc)(BisNU_E (nu [y:name](par (p y) q)) (par (nu p) q))
| bisnu_e_ref: (p:proc)(BisNU_E p p).

Lemma NU_E_RW: (p,q:name->proc)(x:name)
               (notin x (nu p)) -> (notin x (nu q)) -> (BisNU_E (p x) (q x)) ->
               (y:name)(notin y (nu p)) -> (notin y (nu q)) -> (BisNU_E (p y) (q y)).

Proof.

Intros; Elim (LEM_name x y); Intro.

Rewrite <-H4; Assumption.

Inversion H1.
Elim (ho_proc_exp p0 x); Elim (proc_exp q0 x); Intros.
Inversion_clear H5; Inversion_clear H8.
Rewrite H10 in H6; Rewrite H10 in H7; Rewrite H11 in H6; Rewrite H11 in H7.
Cut p=([_:name](nu [y:name](par (x1 _ y) (x0 _)))); [Intro | Apply proc_ext with x; Auto].
Cut q=([_:name](par (nu (x1 _)) (x0 _))); [Intro | Apply proc_ext with x; Auto].
Rewrite H8; Rewrite H12; Apply bisnu_e.
Inversion_clear H9; Inversion_clear H5; Apply notin_nu; Intros; Apply notin_par; Auto.
Inversion_clear H9; Inversion_clear H5; Do 2 (Apply notin_nu; Intros); Apply notin_par; [Cut (notin x (nu (x1 z))); [Intro; Inversion_clear H13; Auto | Auto] | Auto].

Cut p=q; [Intro | Apply proc_ext with x; Auto].
Rewrite H6; Apply bisnu_e_ref.

Qed.

Lemma BisNU_E_FTR1: (p,q:proc)(BisNU_E p q) ->
                    (x,y:name)
                    ((p1:proc)(ftrans p (Out x y) p1)->
                     (Ex [q1:proc]((ftrans q (Out x y) q1) /\ (BisNU_E p1 q1)))
                    )
                    /\
                    ((q1:proc)(ftrans q (Out x y) q1)->
                     (Ex [p1:proc]((ftrans p (Out x y) p1) /\ (BisNU_E p1 q1)))
                    ).

Proof.

Intros; Inversion H; Split; Intros.

Inversion_clear H2.
Elim (unsat (par (nu [y:name](par (p0 y) q0)) (nu p3)) (cons x (cons y l))); Intros.
Inversion_clear H2.
Inversion_clear H4; Inversion_clear H5.
Inversion_clear H7.
Cut (ftrans (par (p0 x0) q0) (Out x y) (p3 x0)); [Intro | Apply H3; Auto].
Inversion H7.
Elim (proc_exp p5 x0); Intros.
Inversion_clear H14.
Rewrite H16 in H12.
Exists (par (nu x1) q0); Split; [Apply fPAR1; Apply fRES with l; Intros | Idtac].
Inversion_clear H19; Change (ftrans (p0 y0) ([_:name](Out x y) y0) (x1 y0)); Apply FTR_L3 with x0; Auto.
Inversion_clear H2; Apply notin_nu; Intros.
Cut (notin x0 (par (p0 z) q0)); [Intro; Inversion_clear H22; Auto | Auto].
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite H16 in H13; Cut p3=([_:name](par (x1 _) q0)); [Intro | Apply proc_ext with x0; Auto].
Rewrite H14; Apply bisnu_e.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H15; Auto | Inversion_clear H2; Cut (notin x0 (par (p0 z) q0)); [Intro; Inversion_clear H2; Assumption | Auto]].
Exists (par (nu p0) p5); Split; [Apply fPAR2; Assumption | Idtac].
Cut p3=([_:name](par (p0 _) p5)); [Intro | Apply proc_ext with x0; Auto].
Rewrite H14; Apply bisnu_e.
Inversion_clear H2; Apply notin_nu; Intros; Cut (notin x0 (par (p0 z) q0)); [Intro | Auto].
Inversion_clear H15; Apply notin_par; [Assumption | Elim (FTR_L1 q0 p5 (Out x y) x0 H17 H12); Intros; Assumption].
Apply f_act_notin_Out; Auto.

Inversion_clear H2.
Inversion H3.
Elim (unsat (par (nu p0) (par q0 (nu p2))) (cons x (cons y l))); Intros.
Inversion_clear H7.
Inversion_clear H8; Inversion_clear H9.
Inversion_clear H10; Inversion_clear H11.
Cut (ftrans (p0 x0) (Out x y) (p2 x0)); [Intro | Apply H4; Auto].
Exists (nu [y0:name](par (p2 y0) q0)); Split; [Apply fRES with l; Intros | Apply bisnu_e].
Change (ftrans ([_:name](par (p0 _) q0) y0) ([_:name](Out x y) y0) ([_:name](par (p2 _) q0) y0)); Apply FTR_L3 with x0; Auto.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H7; Auto | Assumption].
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H12; Auto | Assumption].
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Apply fPAR1; Assumption.
Inversion_clear H17; Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Apply f_act_notin_Out; Auto.
Exists (nu [y0:name](par (p0 y0) p3)); Split; [Apply fRES with empty; Intros; Apply fPAR2; Assumption | Apply bisnu_e].

Exists p1; Split; [Assumption | Apply bisnu_e_ref].

Exists q1; Split; [Assumption | Apply bisnu_e_ref].

Qed.

Lemma BisNU_E_IN  : (p,q:proc)(BisNU_E p q) ->
                    (x:name)
                    ((p1:name->proc)(btrans p (In x) p1)->
                     (Ex [q1:name->proc]((btrans q (In x) q1) /\ (y:name)(BisNU_E (p1 y) (q1 y))))
                    )
                    /\
                    ((q1:name->proc)(btrans q (In x) q1)->
                     (Ex [p1:name->proc]((btrans p (In x) p1) /\ (y:name)(BisNU_E (p1 y) (q1 y))))
                    ).

Proof.

Intros; Inversion H; Split; Intros.

Inversion H2.

Elim (unsat (par (nu [y:name](par (p0 y) q0)) (nu [z:name](nu (p3 z)))) (cons x l)); Intros.
Inversion_clear H7.
Inversion_clear H8; Inversion_clear H9.
Cut (btrans (par (p0 x0) q0) (In x) (p3 x0)); [Intro | Apply H4; Auto].
Inversion H9.
Elim (ho_proc_exp p6 x0); Intros.
Inversion_clear H17.
Rewrite H19 in H15; Exists [_:name](par (nu [u:name](x1 u _)) q0); Split; [Change (btrans (par (nu p0) q0) (In x) [_:name](par ([v:name](nu [u:name](x1 u v)) _) q0)); Apply bPAR1; Apply bRES with l; Intros | Idtac].
Inversion_clear H21; Change (btrans (p0 y) ([_:name](In x) y) (x1 y)); Apply BTR_L3 with x0; Auto.
Inversion_clear H7; Apply notin_nu; Intros; Cut (notin x0 (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H24; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Rewrite H19 in H16; Cut p3=([x0:name]([x:name](par (x1 x0 x) q0))); [Intro | Apply ho_proc_ext with x0; Auto].
Intro; Rewrite H17; Change (BisNU_E (nu [z:name](par ([_:name](x1 _ y) z) q0)) (par (nu [u:name](x1 u y)) q0)); Apply bisnu_e.
Do 2 (Apply notin_nu; Intros); Apply notin_par; [Inversion_clear H18; Cut (notin x0 (nu (x1 z))); [Intro | Auto]; Inversion_clear H18; Auto | Inversion_clear H7; Cut (notin x0 (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H7; Assumption].

Exists [_:name](par (nu p0) (p6 _)); Split; [Apply bPAR2; Assumption | Idtac].
Intro; Cut p3=([x0:name][x:name](par (p0 x0) (p6 x))); [Intro | Apply ho_proc_ext with x0; Auto].
Rewrite H17; Apply bisnu_e.
Inversion_clear H7; Do 2 (Apply notin_nu; Intros); Apply notin_par.
Cut (notin x0 (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H19; Assumption.
Cut (notin x0 (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H19; Elim (BTR_L1 q0 p6 (In x) x0 H21 H15); Intros.
Inversion_clear H22; Auto.
Apply b_act_notin_In; Auto.

Inversion H2.
Inversion H7.
Exists [_:name](nu [y:name](par (p5 y _) q0)); Split; [Change (btrans (nu [y:name](par (p0 y) q0)) (In x) [_:name](nu [y:name]([u:name][v:name](par (p5 u v) q0) y _))); Apply bRES with l; Intros; Apply bPAR1; Apply H9; Auto | Idtac].
Inversion_clear H12; Apply notin_nu; Intros; Cut (notin y (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H17; Auto.
Do 2 (Apply notin_nu; Intros); Inversion_clear H13; Cut (notin y (nu [v:name](par (p5 z v) q0))); [Intro | Auto]; Inversion_clear H13; Cut (notin y (par (p5 z z0) q0)); [Intro | Auto]; Inversion_clear H13; Assumption.
Intro; Change (BisNU_E (nu [y0:name](par ([_:name](p5 _ y) y0) q0)) (par (nu [z:name](p5 z y)) q0)); Apply bisnu_e.

Exists [_:name](nu [y:name](par (p0 y) (p3 _))); Split; [Change (btrans (nu [y:name](par (p0 y) q0)) (In x) [_:name](nu [y:name]([u:name][v:name](par (p0 u) (p3 v)) y _))); Apply bRES with empty; Intros; Apply bPAR2; Assumption | Intro; Apply bisnu_e].

Exists p1; Split; [Assumption | Intro; Apply bisnu_e_ref].

Exists q1; Split; [Assumption | Intro; Apply bisnu_e_ref].

Qed.

Lemma BisNU_E_bOUT: (p,q:proc)(BisNU_E p q) ->
                    (x:name)
                    ((p1:name->proc)(btrans p (bOut x) p1)->
                     (Ex [q1:name->proc]((btrans q (bOut x) q1) /\ (y:name)(notin y (nu p1))->(notin y (nu q1))->(BisNU_E (p1 y) (q1 y))))
                    )
                    /\
                    ((q1:name->proc)(btrans q (bOut x) q1)->
                     (Ex [p1:name->proc]((btrans p (bOut x) p1) /\ (y:name)(notin y (nu p1))->(notin y (nu q1))->(BisNU_E (p1 y) (q1 y))))
                    ).

Proof.

Intros; Inversion H; Split; Intros.

Inversion H2.

Elim (unsat (par (nu [y:name](par (p0 y) q0)) (nu [z:name](nu (p3 z)))) (cons x l)); Intros.
Inversion_clear H7.
Inversion_clear H8; Inversion_clear H9.
Cut (btrans (par (p0 x0) q0) (bOut x) (p3 x0)); [Intro | Apply H4; Auto].
Inversion H9.
Elim (ho_proc_exp p6 x0); Intros.
Inversion_clear H17.
Rewrite H19 in H15; Exists [_:name](par (nu [u:name](x1 u _)) q0); Split; [Change (btrans (par (nu p0) q0) (bOut x) [_:name](par ([v:name](nu [u:name](x1 u v)) _) q0)); Apply bPAR1; Apply bRES with l; Intros | Idtac].
Inversion_clear H21; Change (btrans (p0 y) ([_:name](bOut x) y) (x1 y)); Apply BTR_L3 with x0; Auto.
Inversion_clear H7; Apply notin_nu; Intros; Cut (notin x0 (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H24; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Rewrite H19 in H16; Cut p3=([x0:name]([x:name](par (x1 x0 x) q0))); [Intro | Apply ho_proc_ext with x0; Auto].
Intros; Rewrite H17; Change (BisNU_E (nu [z:name](par ([_:name](x1 _ y) z) q0)) (par (nu [u:name](x1 u y)) q0)); Apply bisnu_e.
Do 2 (Apply notin_nu; Intros); Apply notin_par; [Inversion_clear H18; Cut (notin x0 (nu (x1 z))); [Intro | Auto]; Inversion_clear H18; Auto | Inversion_clear H7; Cut (notin x0 (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H7; Assumption].

Exists [_:name](par (nu p0) (p6 _)); Split; [Apply bPAR2; Assumption | Idtac].
Intros; Cut p3=([x0:name][x:name](par (p0 x0) (p6 x))); [Intro | Apply ho_proc_ext with x0; Auto].
Rewrite H19; Apply bisnu_e.
Inversion_clear H7; Do 2 (Apply notin_nu; Intros); Apply notin_par.
Cut (notin x0 (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H21; Assumption.
Cut (notin x0 (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H21; Elim (BTR_L1 q0 p6 (bOut x) x0 H23 H15); Intros.
Inversion_clear H24; Auto.
Apply b_act_notin_bOut; Auto.

Elim (unsat (par (nu [y:name](par (p0 y) q0)) (nu p1)) (cons x l)); Intros.
Inversion_clear H7.
Inversion_clear H8; Inversion_clear H9.
Cut (ftrans (par (p0 x1) q0) (Out x x1) (p1 x1)); [Intro | Apply H5; Auto].
Inversion H9.
Elim (proc_exp p6 x1); Intros.
Inversion_clear H17; Exists [_:name](par (x2 _) q0); Split; [Apply bPAR1; Apply OPEN with l; Intros; Rewrite H19 in H15; Apply FTR_L3 with x1; Auto | Idtac].
Inversion_clear H7; Apply notin_nu; Intros; Cut (notin x1 (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H24; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite H19 in H16; Cut p1=([_:name](par (x2 _) q0)); [Intro | Apply proc_ext with x1; Auto].
Intros; Rewrite H17; Apply bisnu_e_ref.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H18; Auto | Inversion_clear H7; Cut (notin x1 (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H7; Assumption].

Inversion_clear H7; Cut (notin x1 (par (p0 x) q0)); [Intro | Auto]; Inversion_clear H7.
Elim (FTR_L1 q0 p6 (Out x x1) x1 H19 H15); Intros; Inversion_clear H7; Unfold not in H22; Elim (H22 (refl_equal name x1)).

Inversion H2.
Inversion H7.
Exists [_:name](nu [y:name](par (p5 y _) q0)); Split; [Change (btrans (nu [y:name](par (p0 y) q0)) (bOut x) [_:name](nu [y:name]([u:name][v:name](par (p5 u v) q0) y _))); Apply bRES with l; Intros; Apply bPAR1; Apply H9; Auto | Idtac].
Inversion_clear H12; Apply notin_nu; Intros; Cut (notin y (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H17; Auto.
Do 2 (Apply notin_nu; Intros); Inversion_clear H13; Cut (notin y (nu [v:name](par (p5 z v) q0))); [Intro | Auto]; Inversion_clear H13; Cut (notin y (par (p5 z z0) q0)); [Intro | Auto]; Inversion_clear H13; Assumption.
Intros; Change (BisNU_E (nu [y0:name](par ([_:name](p5 _ y) y0) q0)) (par (nu [z:name](p5 z y)) q0)); Apply bisnu_e.

Exists [_:name](par (p3 _) q0); Split; [Apply OPEN with l; Intros; Apply fPAR1; Apply H10; Auto | Intros; Apply bisnu_e_ref].
Inversion_clear H12; Apply notin_nu; Intros; Cut (notin y (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H17; Auto.
Inversion_clear H13; Apply notin_nu; Intros; Cut (notin y (par (p3 z) q0)); [Intro | Auto]; Inversion_clear H17; Auto.

Exists [_:name](nu [y:name](par (p0 y) (p3 _))); Split; [Change (btrans (nu [y:name](par (p0 y) q0)) (bOut x) [_:name](nu [y:name]([u:name][v:name](par (p0 u) (p3 v)) y _))); Apply bRES with empty; Intros; Apply bPAR2; Assumption | Intros; Apply bisnu_e].

Exists p1; Split; [Assumption | Intros; Apply bisnu_e_ref].

Exists q1; Split; [Assumption | Intros; Apply bisnu_e_ref].

Qed.

Lemma BisNU_E_FTR2: (p,q:proc)(BisNU_E p q) ->
                    ((p1:proc)(ftrans p tau p1)->
                     (Ex [q1:proc](ftrans q tau q1) /\
                         ((Ex [p1':proc](Ex [q1':proc](StBisim p1 p1')
                                                  /\ (BisNU_E p1' q1')
                                                  /\ (StBisim q1' q1)))
                          \/
                          (Ex [p1':name->proc](Ex [q1':name->proc](StBisim p1 (nu p1'))
                                                               /\ ((z:name)
                                                                   (notin z (nu p1'))->
                                                                   (notin z (nu q1'))->
                                                                   (BisNU_E (p1' z) (q1' z))
                                                                  )
                                                               /\ (StBisim (nu q1') q1)))
                         )
                     )
                    )
                    /\
                    ((q1:proc)(ftrans q tau q1)->
                     (Ex [p1:proc](ftrans p tau p1) /\
                         ((Ex [p1':proc](Ex [q1':proc](StBisim p1 p1')
                                                  /\ (BisNU_E p1' q1')
                                                  /\ (StBisim q1' q1)))
                          \/
                          (Ex [p1':name->proc](Ex [q1':name->proc](StBisim p1 (nu p1'))
                                                               /\ ((z:name)
                                                                   (notin z (nu p1'))->
                                                                   (notin z (nu q1'))->
                                                                   (BisNU_E (p1' z) (q1' z))
                                                                  )
                                                               /\ (StBisim (nu q1') q1)))
                         )
                     )
                    ).

Proof.

Intros; Inversion H; Split; Intros.

Inversion H2.
Elim (unsat (par (nu [y:name](par (p0 y) q0)) (nu p3)) l); Intros.
Inversion_clear H7.
Inversion_clear H8.
Cut (ftrans (par (p0 x) q0) tau (p3 x)); [Intro | Apply H4; Auto].
Inversion H8.
Elim (proc_exp p6 x); Intros.
Inversion_clear H16.
Rewrite H18 in H14; Exists (par (nu x0) q0); Split; [Apply fPAR1; Apply fRES with l; Intros | Idtac].
Clear H21; Change (ftrans (p0 y) ([_:name]tau y) (x0 y)); Apply FTR_L3 with x; Auto.
Inversion_clear H7; Apply notin_nu; Intros; Cut (notin x (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H22; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Left; Exists (nu [_:name](par (x0 _) q0)); Exists (par (nu x0) q0); Split; [Idtac | Split; [Apply bisnu_e | Apply REF]].
Rewrite H18 in H15; Cut p3=[_:name](par (x0 _) q0); [Intro | Apply proc_ext with x; Auto].
Rewrite H16; Apply REF.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H17; Auto | Inversion_clear H7; Cut (notin x (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H7; Assumption].

Exists (par (nu p0) p6); Split; [Apply fPAR2; Assumption | Left].
Exists (nu p3); Exists (par (nu p0) p6); Split; [Apply REF | Split; [Idtac | Apply REF]].
Cut p3=([_:name](par (p0 _) p6)); [Intro | Apply proc_ext with x; Auto].
Rewrite H16; Apply bisnu_e.
Inversion_clear H7; Apply notin_nu; Intros; Cut (notin x (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H17; Apply notin_par; [Assumption | Elim (FTR_L1 q0 p6 tau x H19 H14); Intros; Assumption].

Elim (ho_proc_exp q1 x); Intros.
Inversion_clear H17.
Rewrite H19 in H15; Exists (par (nu [_:name](x1 _ y)) q2); Split; [Change (ftrans (par (nu p0) q0) tau (par ([u:name](nu [_:name](x1 _ u)) y) q2)); Apply COM1 with x0; [Apply bRES with l; Intros | Assumption] | Left].
Change (btrans (p0 y0) ([_:name](In x0) y0) (x1 y0)); Apply BTR_L3 with x; Auto.
Inversion_clear H7; Apply notin_nu; Intros; Cut (notin x (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H24; Assumption.
Inversion_clear H7; Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Cut (notin x (par (p0 y1) q0)); [Intro | Auto]; Inversion_clear H24; Elim (FTR_L1 q0 q2 (Out x0 y) x H26 H16); Intros; Inversion_clear H24; Auto.
Inversion_clear H21; Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Exists (nu p3); Exists (par (nu [_:name](x1 _ y)) q2); Split; [Apply REF | Split; [Idtac | Apply REF]].
Rewrite H19 in H13; Cut p3=([_:name](par (x1 _ y) q2)); [Intro | Apply proc_ext with x; Auto].
Rewrite H17; Change (BisNU_E (nu [_:name](par ([u:name](x1 u y) _) q2)) (par (nu [_:name](x1 _ y)) q2)); Apply bisnu_e.
Apply notin_nu; Intros; Inversion_clear H7; Cut (notin x (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H7; Elim (FTR_L1 q0 q2 (Out x0 y) x H22 H16); Intros; Inversion_clear H7; Apply notin_par; [Inversion_clear H18; Cut (notin x (nu (x1 z))); [Intro | Auto]; Inversion_clear H18; Auto
 | Assumption].

Elim (proc_exp q1 x); Intros.
Inversion_clear H17; Elim (LEM_name x y); Intro.

Rewrite H19 in H15; Rewrite H17 in H15; Exists (nu [_:name](par (x1 _) (q2 _))); Split; [Apply CLOSE2 with x0; [Apply OPEN with l; Intros; Apply FTR_L3 with x; Auto | Assumption] | Left].
Inversion_clear H7; Apply notin_nu; Intros; Cut (notin x (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H25; Assumption.
Inversion_clear H7; Unfold f_act_notin_ho; Intros; Cut (notin x (par (p0 y1) q0)); [Intro | Auto]; Inversion_clear H25; Elim (BTR_L1 q0 q2 (In x0) x H27 H16); Intros; Inversion_clear H25; Apply f_act_notin_Out; Auto.
Rewrite H17; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Exists (nu p3); Exists (nu [_:name](par (x1 _) (q2 _))); Split; [Apply REF | Split; [Idtac | Apply REF]].
Rewrite <-H17 in H13; Rewrite H19 in H13; Cut p3=([_:name](par (x1 _) (q2 _))); [Intro | Apply proc_ext with x; Auto].
Rewrite H20; Apply bisnu_e_ref.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H18; Auto | Inversion_clear H7; Cut (notin x (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H7; Elim (BTR_L1 q0 q2 (In x0) x H23 H16); Intros; Inversion_clear H24; Auto].

Rewrite H19 in H15; Exists (par (nu [_:name](x1 _)) (q2 y)); Split; [Apply COM2 with x0; [Apply fRES with l; Intros | Assumption] | Left].
Change (ftrans (p0 y0) ([_:name](Out x0 y) y0) (x1 y0)); Apply FTR_L3 with x; Auto.
Inversion_clear H7; Apply notin_nu; Intros; Cut (notin x (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H25; Assumption.
Inversion_clear H7; Unfold f_act_notin_ho; Intros; Cut (notin x (par (p0 y1) q0)); [Intro | Auto]; Inversion_clear H25; Elim (BTR_L1 q0 q2 (In x0) x H27 H16); Intros; Inversion_clear H25; Apply f_act_notin_Out; Auto.
Inversion_clear H21; Apply notin_nu; Intros; Auto.
Unfold f_act_notin_ho; Intros; Assumption.
Exists (nu p3); Exists (par (nu [_:name](x1 _)) (q2 y)); Split; [Apply REF | Split; [Idtac | Apply REF]].
Rewrite H19 in H13; Cut p3=([_:name](par (x1 _) (q2 y))); [Intro | Apply proc_ext with x; Auto].
Rewrite H20; Change (BisNU_E (nu [_:name](par ([u:name](x1 u) _) (q2 y))) (par (nu [_:name](x1 _)) (q2 y))); Apply bisnu_e.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H18; Auto | Inversion_clear H7; Cut (notin x (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H7; Elim (BTR_L1 q0 q2 (In x0) x H23 H16); Intros; Inversion_clear H24; Auto].

Elim (ho_proc_exp q1 x); Intros.
Inversion_clear H17.
Rewrite H19 in H15; Exists (nu [_:name](par (nu [u:name](x1 u _)) (q2 _))); Split; [Change (ftrans (par (nu p0) q0) tau (nu [_:name](par ([v:name](nu [u:name](x1 u v)) _) (q2 _)))); Apply CLOSE1 with x0; [Apply bRES with l; Intros | Assumption] | Right].
Change (btrans (p0 y) ([_:name](In x0) y) (x1 y)); Apply BTR_L3 with x; Auto.
Inversion_clear H7; Apply notin_nu; Intros; Cut (notin x (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H24; Assumption.
Inversion_clear H7; Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Cut (notin x (par (p0 y0) q0)); [Intro | Auto]; Inversion_clear H24; Elim (BTR_L1 q0 q2 (bOut x0) x H26 H16); Intros; Inversion_clear H24; Auto.
Unfold b_act_notin_ho; Intros; Assumption.
Exists [z:name](nu [_:name](par (x1 _ z) (q2 z))); Exists [_:name](par (nu [u:name](x1 u _)) (q2 _)); Split; [Idtac | Split; [Intros | Apply REF]].
Rewrite H19 in H13; Cut p3=([_:name](nu [z:name](par (x1 _ z) (q2 z)))); [Intro | Apply proc_ext with x; Auto].
Rewrite H17; Change (StBisim (nu [_:name](nu [z:name]([u:name][v:name](par (x1 u v) (q2 v)) _ z))) (nu [z:name](nu [_:name]([u:name][v:name](par (x1 u v) (q2 v)) _ z)))); Apply NU_COMM.
Do 2 (Apply notin_nu; Intros); Apply notin_par; [Inversion_clear H18; Cut (notin x (nu (x1 z))); [Intro | Auto]; Inversion_clear H18; Auto | Inversion_clear H7; Cut (notin x (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H7; Elim (BTR_L1 q0 q2 (bOut x0) x H23 H16); Intros; Inversion_clear H24; Auto].
Change (BisNU_E (nu [_:name](par ([u:name](x1 u z) _) (q2 z))) (par (nu [u:name](x1 u z)) (q2 z))); Apply bisnu_e.

Elim (ho_proc_exp q1 x); Intros.
Inversion_clear H17.
Rewrite H19 in H15; Exists (nu [_:name](par (nu [u:name](x1 u _)) (q2 _))); Split; [Change (ftrans (par (nu p0) q0) tau (nu [_:name](par ([v:name](nu [u:name](x1 u v)) _) (q2 _)))); Apply CLOSE2 with x0; [Apply bRES with l; Intros | Assumption] | Right].
Change (btrans (p0 y) ([_:name](bOut x0) y) (x1 y)); Apply BTR_L3 with x; Auto.
Inversion_clear H7; Apply notin_nu; Intros; Cut (notin x (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H24; Assumption.
Inversion_clear H7; Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Cut (notin x (par (p0 y0) q0)); [Intro | Auto]; Inversion_clear H24; Elim (BTR_L1 q0 q2 (In x0) x H26 H16); Intros; Inversion_clear H24; Auto.
Unfold b_act_notin_ho; Intros; Assumption.
Exists [z:name](nu [_:name](par (x1 _ z) (q2 z))); Exists [_:name](par (nu [u:name](x1 u _)) (q2 _)); Split; [Idtac | Split; [Intros | Apply REF]].
Rewrite H19 in H13; Cut p3=([_:name](nu [z:name](par (x1 _ z) (q2 z)))); [Intro | Apply proc_ext with x; Auto].
Rewrite H17; Change (StBisim (nu [_:name](nu [z:name]([u:name][v:name](par (x1 u v) (q2 v)) _ z))) (nu [z:name](nu [_:name]([u:name][v:name](par (x1 u v) (q2 v)) _ z)))); Apply NU_COMM.
Do 2 (Apply notin_nu; Intros); Apply notin_par; [Inversion_clear H18; Cut (notin x (nu (x1 z))); [Intro | Auto]; Inversion_clear H18; Auto | Inversion_clear H7; Cut (notin x (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H7; Elim (BTR_L1 q0 q2 (In x0) x H23 H16); Intros; Inversion_clear H24; Auto].
Change (BisNU_E (nu [_:name](par ([u:name](x1 u z) _) (q2 z))) (par (nu [u:name](x1 u z)) (q2 z))); Apply bisnu_e.

Apply f_act_notin_tau.

Inversion H2.

Inversion H7.
Exists (nu [y:name](par (p5 y) q0)); Split; [Apply fRES with l; Intros; Apply fPAR1 | Left].
Clear H15; Apply H9; Auto.
Inversion_clear H12; Apply notin_nu; Intros; Cut (notin y (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H16; Assumption.
Inversion_clear H13; Apply notin_nu; Intros; Cut (notin y (par (p5 z) q0)); [Intro | Auto]; Inversion_clear H16; Assumption.
Apply f_act_notin_tau.
Exists (nu [y:name](par (p5 y) q0)); Exists (par (nu p5) q0); Split; [Apply REF | Split; [Apply bisnu_e | Apply REF]].

Exists (nu [y:name](par (p0 y) p3)); Split; [Apply fRES with empty; Intros; Apply fPAR2; Assumption | Left].
Exists (nu [y:name](par (p0 y) p3)); Exists (par (nu p0) p3); Split; [Apply REF | Split; [Apply bisnu_e | Apply REF]].

Inversion H5.
Exists (nu [y0:name](par (p4 y0 y) q2)); Split; [Apply fRES with l; Intros; Apply COM1 with x; [Apply H10; Auto | Assumption] | Left].
Inversion_clear H13; Apply notin_nu; Intros; Cut (notin y0 (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H18; Assumption.
Inversion_clear H14; Apply notin_nu; Intros; Apply proc_mono with y; Cut (notin y0 (par (p4 z y) q2)); [Intro | Auto]; Inversion_clear H18; Assumption.
Elim (unsat nil (cons y0 empty)); Intros; Inversion_clear H17; Inversion_clear H19; Clear H18; Clear H20; Inversion_clear H13; Cut (notin y0 (par (p0 x0) q0)); [Intro | Auto]; Inversion_clear H13; Elim (FTR_L1 q0 q2 (Out x y) y0 H20 H8); Intros; Inversion_clear H13; Apply b_act_notin_In; Auto.
Exists (nu [y0:name](par (p4 y0 y) q2)); Exists (par (nu [z:name](p4 z y)) q2); Split; [Apply REF | Split; [Change (BisNU_E (nu [y0:name](par ([_:name](p4 _ y) y0) q2)) (par (nu [z:name](p4 z y)) q2)); Apply bisnu_e | Apply REF]].

Inversion H5.
Exists (nu [y0:name](par (p4 y0) (q3 y))); Split; [Apply fRES with l; Intros; Apply COM2 with x; [Apply H10; Auto | Assumption] | Left].
Inversion_clear H13; Apply notin_nu; Intros; Cut (notin y0 (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H18; Assumption.
Inversion_clear H14; Apply notin_nu; Intros; Cut (notin y0 (par (p4 z) (q3 y))); [Intro | Auto]; Inversion_clear H18; Assumption.
Cut (notin y0 (nu p0)); [Intro; Elim (FTR_L1 (nu p0) q2 (Out x y) y0 H17 H5); Intros; Assumption | Inversion_clear H13; Apply notin_nu; Intros; Cut (notin y0 (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H18; Assumption].
Exists (nu [y0:name](par (p4 y0) (q3 y))); Exists (par (nu p4) (q3 y)); Split; [Apply REF | Split; [Apply bisnu_e | Apply REF]].

Inversion H5.
Exists (nu [y:name](nu [_:name](par (p4 y _) (q3 _)))); Split; [Apply fRES with l; Intros; Apply CLOSE1 with x; [Apply H10; Auto | Assumption] | Right].
Inversion_clear H13; Apply notin_nu; Intros; Cut (notin y (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H18; Assumption.
Inversion_clear H14; Do 2 (Apply notin_nu; Intros); Cut (notin y (nu [_:name](par (p4 z _) (q3 _)))); [Intro | Auto]; Inversion_clear H19; Cut (notin y (par (p4 z z0) (q3 z0))); [Intro | Auto]; Inversion_clear H19; Assumption.
Elim (unsat nil (cons y empty)); Intros; Inversion_clear H17; Inversion_clear H19; Clear H18; Clear H20; Inversion_clear H13; Cut (notin y (par (p0 x0) q0)); [Intro | Auto]; Inversion_clear H13; Elim (BTR_L1 q0 q3 (bOut x) y H20 H8); Intros; Inversion_clear H13; Apply b_act_notin_In; Auto.
Exists [z:name](nu [_:name](par (p4 _ z) (q3 z))); Exists [_:name](par (nu [u:name](p4 u _)) (q3 _)); Split; [Idtac | Split; [Intros | Apply REF]].
Change (StBisim (nu [_:name](nu [z:name]([u:name][v:name](par (p4 u v) (q3 v)) _ z))) (nu [z:name](nu [_:name]([u:name][v:name](par (p4 u v) (q3 v)) _ z)))); Apply NU_COMM.
Change (BisNU_E (nu [_:name](par ([u:name](p4 u z) _) (q3 z))) (par (nu [u:name](p4 u z)) (q3 z))); Apply bisnu_e.

Inversion H5.
Exists (nu [y:name](nu [_:name](par (p4 y _) (q3 _)))); Split; [Apply fRES with l; Intros; Apply CLOSE2 with x; [Apply H10; Auto | Assumption] | Right].
Inversion_clear H13; Apply notin_nu; Intros; Cut (notin y (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H18; Assumption.
Inversion_clear H14; Do 2 (Apply notin_nu; Intros); Cut (notin y (nu [_:name](par (p4 z _) (q3 _)))); [Intro | Auto]; Inversion_clear H19; Cut (notin y (par (p4 z z0) (q3 z0))); [Intro | Auto]; Inversion_clear H19; Assumption.
Elim (unsat nil (cons y empty)); Intros; Inversion_clear H17; Inversion_clear H19; Clear H18; Clear H20; Inversion_clear H13; Cut (notin y (par (p0 x0) q0)); [Intro | Auto]; Inversion_clear H13; Elim (BTR_L1 q0 q3 (In x) y H20 H8); Intros; Inversion_clear H13; Apply b_act_notin_bOut; Auto.
Exists [z:name](nu [_:name](par (p4 _ z) (q3 z))); Exists [_:name](par (nu [u:name](p4 u _)) (q3 _)); Split; [Idtac | Split; [Intros | Apply REF]].
Change (StBisim (nu [_:name](nu [z:name]([u:name][v:name](par (p4 u v) (q3 v)) _ z))) (nu [z:name](nu [_:name]([u:name][v:name](par (p4 u v) (q3 v)) _ z)))); Apply NU_COMM.
Change (BisNU_E (nu [_:name](par ([u:name](p4 u z) _) (q3 z))) (par (nu [u:name](p4 u z)) (q3 z))); Apply bisnu_e.

Exists (nu [y:name](par (q2 y) (q3 y))); Split; [Apply fRES with l; Intros; Apply COM2 with x; [Apply H11; Auto | Assumption] | Left].
Inversion_clear H13; Apply notin_nu; Intros; Cut (notin y (par (p0 z) q0)); [Intro | Auto]; Inversion_clear H18; Assumption.
Inversion_clear H14; Apply notin_nu; Intros; Cut (notin y (par (q2 z) (q3 z))); [Intro | Auto]; Inversion_clear H18; Assumption.
Elim (unsat nil (cons y empty)); Intros; Inversion_clear H17; Inversion_clear H19; Clear H18 H20; Inversion_clear H13; Cut (notin y (par (p0 x1) q0)); [Intro | Auto]; Inversion_clear H13; Elim (BTR_L1 q0 q3 (In x) y H20 H8); Intros; Inversion_clear H13; Auto.
Exists (nu [y:name](par (q2 y) (q3 y))); Exists (nu [z:name](par (q2 z) (q3 z))); Split; [Apply REF | Split; [Apply bisnu_e_ref | Apply REF]].

Exists p1; Split; [Assumption | Left; Exists p1; Exists p1; Split; [Apply REF | Split; [Apply bisnu_e_ref | Apply REF]]].

Exists q1; Split; [Assumption | Left; Exists q1; Exists q1; Split; [Apply REF | Split; [Apply bisnu_e_ref | Apply REF]]].

Qed.

(* BisNU_E is a strong bisimulation UpTo *)

Lemma BisNU_E_UPTO: (UpTo BisNU_E).

Proof.

Unfold UpTo; Intros.

Split; Intros.

Rewrite H2 in H; Rewrite H3 in H; Apply NU_E_RW with x; Auto.

Split; Intros.

Elim (BisNU_E_FTR1 p q H x y); Intros; Split; Intros.

Elim (H0 p1 H2); Intros; Inversion_clear H3.
Exists x0; Split; [Assumption | Unfold SB_R_SB; Exists p1; Exists x0; Split; [Apply REF | Split; [Assumption | Apply REF]]].

Elim (H1 q1 H2); Intros; Inversion_clear H3.
Exists x0; Split; [Assumption | Unfold SB_R_SB; Exists x0; Exists q1; Split; [Apply REF | Split; [Assumption | Apply REF]]].

Split; Intros.

Elim (BisNU_E_IN p q H x); Intros; Split; Intros.

Elim (H0 p1 H2); Intros; Inversion_clear H3.
Exists x0; Split; [Assumption | Intro; Unfold SB_R_SB; Exists (p1 y); Exists (x0 y); Split; [Apply REF | Split; [Auto | Apply REF]]].

Elim (H1 q1 H2); Intros; Inversion_clear H3.
Exists x0; Split; [Assumption | Intro; Unfold SB_R_SB; Exists (x0 y); Exists (q1 y); Split; [Apply REF | Split; [Auto | Apply REF]]].

Split; Intros.

Elim (BisNU_E_bOUT p q H x); Intros; Split; Intros.

Elim (H0 p1 H2); Intros; Inversion_clear H3.
Exists x0; Split; [Assumption | Intros; Unfold SB_R_SB; Exists (p1 y); Exists (x0 y); Split; [Apply REF | Split; [Auto | Apply REF]]].

Elim (H1 q1 H2); Intros; Inversion_clear H3.
Exists x0; Split; [Assumption | Intros; Unfold SB_R_SB; Exists (x0 y); Exists (q1 y); Split; [Apply REF | Split; [Auto | Apply REF]]].

Elim (BisNU_E_FTR2 p q H); Intros; Split; Intros.

Elim (H0 p1 H2); Intros; Inversion_clear H3; Inversion_clear H5.
Exists x; Split; [Assumption | Left; Inversion_clear H3; Inversion_clear H5; Inversion_clear H3; Inversion_clear H6; Unfold SB_R_SB; Exists x0; Exists x1; Split; [Assumption | Split; Assumption]].
Exists x; Split; [Assumption | Right; Inversion_clear H3; Inversion_clear H5; Inversion_clear H3; Inversion_clear H6; Exists x0; Exists x1; Split; [Assumption | Split; [Intros; Auto | Assumption]]].

Elim (H1 q1 H2); Intros; Inversion_clear H3; Inversion_clear H5.
Exists x; Split; [Assumption | Left; Inversion_clear H3; Inversion_clear H5; Inversion_clear H3; Inversion_clear H6; Unfold SB_R_SB; Exists x0; Exists x1; Split; [Assumption | Split; Assumption]].
Exists x; Split; [Assumption | Right; Inversion_clear H3; Inversion_clear H5; Inversion_clear H3; Inversion_clear H6; Exists x0; Exists x1; Split; [Assumption | Split; [Intros; Auto | Assumption]]].

Qed.

Lemma NU_EXTR: (p:name->proc)(q:proc)(StBisim (nu [y:name](par (p y) q)) (par (nu p) q)).

Proof.

Intros; Apply Completeness; Apply Co_Ind with (SBUPTO BisNU_E); [Apply UPTO_SB'; Exact BisNU_E_UPTO | Apply sbupto with O; Simpl; Unfold SB_R_SB; Exists (nu [y:name](par (p y) q)); Exists (par (nu p) q); Split; [Apply REF | Split; [Apply bisnu_e | Apply REF]]].

Qed.

(*******************************************************************)
(* Here is the definition of the strong bisimulation UpTo allowing *)
(* to prove the associativity of the parallel composition.         *)
(*******************************************************************)

Inductive BisPAR_A: proc -> proc -> Prop :=
  bispar_a: (p,q,r:proc)(BisPAR_A (par (par p q) r) (par p (par q r))).

Lemma PAR_A_RW: (p,q:name->proc)(x:name)
                (notin x (nu p)) -> (notin x (nu q)) -> (BisPAR_A (p x) (q x)) ->
                (y:name)(notin y (nu p)) -> (notin y (nu q)) -> (BisPAR_A (p y) (q y)).

Proof.

Intros; Elim (LEM_name x y); Intro.

Rewrite <-H4; Assumption.

Inversion H1.
Elim (proc_exp p0 x); Elim (proc_exp q0 x); Elim (proc_exp r x); Intros.
Inversion_clear H5; Inversion_clear H8; Inversion_clear H9.
Rewrite H11 in H6; Rewrite H11 in H7; Rewrite H12 in H6; Rewrite H12 in H7; Rewrite H13 in H6; Rewrite H13 in H7.
Cut p=([_:name](par (par (x2 _) (x1 _)) (x0 _))); [Intro | Apply proc_ext with x; Auto].
Cut q=([_:name](par (x2 _) (par (x1 _) (x0 _)))); [Intro | Apply proc_ext with x; Auto].
Rewrite H9; Rewrite H14; Apply bispar_a.
Inversion_clear H5; Inversion_clear H8; Inversion_clear H10; Apply notin_nu; Intros; Apply notin_par; [Auto | Apply notin_par; Auto].
Inversion_clear H5; Inversion_clear H8; Inversion_clear H10; Apply notin_nu; Intros; Apply notin_par; [Apply notin_par; Auto | Auto].

Qed.

Lemma BisPAR_A_FTR1: (p,q:proc)(BisPAR_A p q) ->
                     (x,y:name)
                     ((p1:proc)(ftrans p (Out x y) p1)->
                      (Ex [q1:proc]((ftrans q (Out x y) q1) /\ (BisPAR_A p1 q1)))
                     )
                     /\
                     ((q1:proc)(ftrans q (Out x y) q1)->
                      (Ex [p1:proc]((ftrans p (Out x y) p1) /\ (BisPAR_A p1 q1)))
                     ).

Proof.

Intros; Inversion H; Split; Intros.

Inversion H2.

Inversion H7; [Exists (par p7 (par q0 r)); Split; [Apply fPAR1; Assumption | Apply bispar_a] | Exists (par p0 (par p7 r)); Split; [Apply fPAR2; Apply fPAR1; Assumption | Apply bispar_a]].

Exists (par p0 (par q0 p4)); Split; [Apply fPAR2; Apply fPAR2; Assumption | Apply bispar_a].

Inversion H2.

Exists (par (par p3 q0) r); Split; [Apply fPAR1; Apply fPAR1; Assumption | Apply bispar_a].

Inversion H7; [Exists (par (par p0 p6) r); Split; [Apply fPAR1; Apply fPAR2; Assumption | Apply bispar_a] | Exists (par (par p0 q0) p6); Split; [Apply fPAR2; Assumption | Apply bispar_a]].

Qed.

Lemma BisPAR_A_IN  : (p,q:proc)(BisPAR_A p q) ->
                    (x:name)
                    ((p1:name->proc)(btrans p (In x) p1)->
                     (Ex [q1:name->proc]((btrans q (In x) q1) /\ (y:name)(BisPAR_A (p1 y) (q1 y))))
                    )
                    /\
                    ((q1:name->proc)(btrans q (In x) q1)->
                     (Ex [p1:name->proc]((btrans p (In x) p1) /\ (y:name)(BisPAR_A (p1 y) (q1 y))))
                    ).

Proof.

Intros; Inversion H; Split; Intros.

Inversion H2.

Inversion H7; [Exists [_:name](par (p7 _) (par q0 r)); Split; [Apply bPAR1; Assumption | Intro; Apply bispar_a] | Exists [_:name](par p0 (par (p7 _) r)); Split; [Change (btrans (par p0 (par q0 r)) (In x) [_:name](par p0 ([u:name](par (p7 u) r) _))); Apply bPAR2; Apply bPAR1; Assumption | Intro; Apply bispar_a]].

Exists [_:name](par p0 (par q0 (p4 _))); Split; [Change (btrans (par p0 (par q0 r)) (In x) [_:name](par p0 ([u:name](par q0 (p4 u)) _))); Apply bPAR2; Apply bPAR2; Assumption | Intro; Apply bispar_a].

Inversion H2.

Exists [_:name](par (par (p3 _) q0) r); Split; [Change (btrans (par (par p0 q0) r) (In x) [_:name](par ([u:name](par (p3 u) q0) _) r)); Apply bPAR1; Apply bPAR1; Assumption | Intro; Apply bispar_a].

Inversion H7; [Exists [_:name](par (par p0 (p6 _)) r); Split; [Change (btrans (par (par p0 q0) r) (In x) [_:name](par ([u:name](par p0 (p6 u)) _) r)); Apply bPAR1; Apply bPAR2; Assumption | Intro; Apply bispar_a] | Exists [_:name](par (par p0 q0) (p6 _)); Split; [Apply bPAR2; Assumption | Intro; Apply bispar_a]].

Qed.

Lemma BisPAR_A_bOUT: (p,q:proc)(BisPAR_A p q) ->
                     (x:name)
                     ((p1:name->proc)(btrans p (bOut x) p1)->
                      (Ex [q1:name->proc]((btrans q (bOut x) q1) /\ (y:name)(notin y (nu p1))->(notin y (nu q1))->(BisPAR_A (p1 y) (q1 y))))
                     )
                     /\
                     ((q1:name->proc)(btrans q (bOut x) q1)->
                      (Ex [p1:name->proc]((btrans p (bOut x) p1) /\ (y:name)(notin y (nu p1))->(notin y (nu q1))->(BisPAR_A (p1 y) (q1 y))))
                     ).

Proof.

Intros; Inversion H; Split; Intros.

Inversion H2.

Inversion H7; [Exists [_:name](par (p7 _) (par q0 r)); Split; [Apply bPAR1; Assumption | Intros; Apply bispar_a] | Exists [_:name](par p0 (par (p7 _) r)); Split; [Change (btrans (par p0 (par q0 r)) (bOut x) [_:name](par p0 ([u:name](par (p7 u) r) _))); Apply bPAR2; Apply bPAR1; Assumption | Intros; Apply bispar_a]].

Exists [_:name](par p0 (par q0 (p4 _))); Split; [Change (btrans (par p0 (par q0 r)) (bOut x) [_:name](par p0 ([u:name](par q0 (p4 u)) _))); Apply bPAR2; Apply bPAR2; Assumption | Intros; Apply bispar_a].

Inversion H2.

Exists [_:name](par (par (p3 _) q0) r); Split; [Change (btrans (par (par p0 q0) r) (bOut x) [_:name](par ([u:name](par (p3 u) q0) _) r)); Apply bPAR1; Apply bPAR1; Assumption | Intros; Apply bispar_a].

Inversion H7; [Exists [_:name](par (par p0 (p6 _)) r); Split; [Change (btrans (par (par p0 q0) r) (bOut x) [_:name](par ([u:name](par p0 (p6 u)) _) r)); Apply bPAR1; Apply bPAR2; Assumption | Intros; Apply bispar_a] | Exists [_:name](par (par p0 q0) (p6 _)); Split; [Apply bPAR2; Assumption | Intros; Apply bispar_a]].

Qed.

Lemma BisPAR_A_FTR2: (p,q:proc)(BisPAR_A p q) ->
                     ((p1:proc)(ftrans p tau p1)->
                      (Ex [q1:proc](ftrans q tau q1) /\
                          ((Ex [p1':proc](Ex [q1':proc](StBisim p1 p1')
                                                   /\ (BisPAR_A p1' q1')
                                                   /\ (StBisim q1' q1)))
                           \/
                           (Ex [p1':name->proc](Ex [q1':name->proc](StBisim p1 (nu p1'))
                                                                /\ ((z:name)
                                                                    (notin z (nu p1'))->
                                                                    (notin z (nu q1'))->
                                                                    (BisPAR_A (p1' z) (q1' z))
                                                                   )
                                                                /\ (StBisim (nu q1') q1)))
                          )
                      )
                     )
                     /\
                     ((q1:proc)(ftrans q tau q1)->
                      (Ex [p1:proc](ftrans p tau p1) /\
                          ((Ex [p1':proc](Ex [q1':proc](StBisim p1 p1')
                                                   /\ (BisPAR_A p1' q1')
                                                   /\ (StBisim q1' q1)))
                           \/
                           (Ex [p1':name->proc](Ex [q1':name->proc](StBisim p1 (nu p1'))
                                                                /\ ((z:name)
                                                                    (notin z (nu p1'))->
                                                                    (notin z (nu q1'))->
                                                                    (BisPAR_A (p1' z) (q1' z))
                                                                   )
                                                                /\ (StBisim (nu q1') q1)))
                          )
                      )
                     ).

Proof.

Intros; Inversion H; Split; Intros.

Inversion H2.

Inversion H7.
Exists (par p7 (par q0 r)); Split; [Apply fPAR1; Assumption | Left; Unfold SB_R_SB; Exists (par (par p7 q0) r); Exists (par p7 (par q0 r)); Split; [Apply REF | Split; [Apply bispar_a | Apply REF]]].
Exists (par p0 (par p7 r)); Split; [Apply fPAR2; Apply fPAR1; Assumption | Left; Unfold SB_R_SB; Exists (par (par p0 p7) r); Exists (par p0 (par p7 r)); Split; [Apply REF | Split; [Apply bispar_a | Apply REF]]].
Exists (par (q1 y) (par q2 r)); Split; [Apply COM1 with x; [Assumption | Apply fPAR1; Assumption] | Left; Unfold SB_R_SB; Exists (par (par (q1 y) q2) r); Exists (par (q1 y) (par q2 r)); Split; [Apply REF | Split; [Apply bispar_a | Apply REF]]].
Exists (par q1 (par (q2 y) r)); Split; [Change (ftrans (par p0 (par q0 r)) tau (par q1 ([_:name](par (q2 _) r) y))); Apply COM2 with x; [Assumption | Apply bPAR1; Assumption] | Left; Unfold SB_R_SB; Exists (par (par q1 (q2 y)) r); Exists (par q1 (par (q2 y) r)); Split; [Apply REF | Split; [Apply bispar_a | Apply REF]]].
Exists (nu [_:name](par (q1 _) (par (q2 _) r))); Split; [Change (ftrans (par p0 (par q0 r)) tau (nu [_:name](par (q1 _) ([u:name](par (q2 u) r) _)))); Apply CLOSE1 with x; [Assumption | Apply bPAR1; Assumption] | Right; Exists [z:name](par (par (q1 z) (q2 z)) r); Exists [_:name](par (q1 _) (par (q2 _) r)); Split; [Apply SYM; Change (StBisim (nu [z:name](par ([_:name](par (q1 _) (q2 _)) z) r)) (par (nu [z:name](par (q1 z) (q2 z))) r)); Apply NU_EXTR | Split; [Intros; Apply bispar_a | Apply REF]]].
Exists (nu [_:name](par (q1 _) (par (q2 _) r))); Split; [Change (ftrans (par p0 (par q0 r)) tau (nu [_:name](par (q1 _) ([u:name](par (q2 u) r) _)))); Apply CLOSE2 with x; [Assumption | Apply bPAR1; Assumption] | Right; Exists [z:name](par (par (q1 z) (q2 z)) r); Exists [_:name](par (q1 _) (par (q2 _) r)); Split; [Apply SYM; Change (StBisim (nu [z:name](par ([_:name](par (q1 _) (q2 _)) z) r)) (par (nu [z:name](par (q1 z) (q2 z))) r)); Apply NU_EXTR | Split; [Intros; Apply bispar_a | Apply REF]]].

Exists (par p0 (par q0 p4)); Split; [Apply fPAR2; Apply fPAR2; Assumption | Left; Exists (par (par p0 q0) p4); Exists (par p0 (par q0 p4)); Split; [Apply REF | Split; [Apply bispar_a | Apply REF]]].

Inversion H5.
Exists (par (p6 y) (par q0 q2)); Split; [Apply COM1 with x; [Assumption | Apply fPAR2; Assumption] | Left; Exists (par (par (p6 y) q0) q2); Exists (par (p6 y) (par q0 q2)); Split; [Apply REF | Split; [Apply bispar_a | Apply REF]]].
Exists (par p0 (par (p6 y) q2)); Split; [Apply fPAR2; Apply COM1 with x; Assumption | Left; Exists (par (par p0 (p6 y)) q2); Exists (par p0 (par (p6 y) q2)); Split; [Apply REF | Split; [Apply bispar_a | Apply REF]]].

Inversion H5.
Exists (par p6 (par q0 (q2 y))); Split; [Change (ftrans (par p0 (par q0 r)) tau (par p6 ([_:name](par q0 (q2 _)) y))); Apply COM2 with x; [Assumption | Apply bPAR2; Assumption] | Left; Exists (par (par p6 q0) (q2 y)); Exists (par p6 (par q0 (q2 y))); Split; [Apply REF | Split; [Apply bispar_a | Apply REF]]].
Exists (par p0 (par p6 (q2 y))); Split; [Apply fPAR2; Apply COM2 with x; Assumption | Left; Exists (par (par p0 p6) (q2 y)); Exists (par p0 (par p6 (q2 y))); Split; [Apply REF | Split; [Apply bispar_a | Apply REF]]].

Inversion H5.
Exists (nu [_:name](par (p6 _) (par q0 (q2 _)))); Split; [Change (ftrans (par p0 (par q0 r)) tau (nu [y:name](par (p6 y) ([_:name](par q0 (q2 _)) y)))); Apply CLOSE1 with x; [Assumption | Apply bPAR2; Assumption] | Right; Exists [z:name](par (par (p6 z) q0) (q2 z)); Exists [_:name](par (p6 _) (par q0 (q2 _))); Split; [Apply REF | Split; [Intros; Apply bispar_a | Apply REF]]].
Exists (par p0 (nu [_:name](par (p6 _) (q2 _)))); Split; [Apply fPAR2; Apply CLOSE1 with x; Assumption | Right; Exists [z:name](par (par p0 (p6 z)) (q2 z)); Exists [_:name](par p0 (par (p6 _) (q2 _))); Split; [Apply REF | Split; [Intros; Apply bispar_a | Apply TRANS with (par (nu [_:name](par (p6 _) (q2 _))) p0); [Apply TRANS with (nu [_:name](par (par (p6 _) (q2 _)) p0)); [Apply NU_S; Intros; Apply COMM_PAR | Change (StBisim (nu [_:name](par ([u:name](par (p6 u) (q2 u)) _) p0)) (par (nu [_:name](par (p6 _) (q2 _))) p0)); Apply NU_EXTR] | Apply COMM_PAR]]]].

Inversion H5.
Exists (nu [_:name](par (p6 _) (par q0 (q2 _)))); Split; [Change (ftrans (par p0 (par q0 r)) tau (nu [y:name](par (p6 y) ([_:name](par q0 (q2 _)) y)))); Apply CLOSE2 with x; [Assumption | Apply bPAR2; Assumption] | Right; Exists [z:name](par (par (p6 z) q0) (q2 z)); Exists [_:name](par (p6 _) (par q0 (q2 _))); Split; [Apply REF | Split; [Intros; Apply bispar_a | Apply REF]]].
Exists (par p0 (nu [_:name](par (p6 _) (q2 _)))); Split; [Apply fPAR2; Apply CLOSE2 with x; Assumption | Right; Exists [z:name](par (par p0 (p6 z)) (q2 z)); Exists [_:name](par p0 (par (p6 _) (q2 _))); Split; [Apply REF | Split; [Intros; Apply bispar_a | Apply TRANS with (par (nu [_:name](par (p6 _) (q2 _))) p0); [Apply TRANS with (nu [_:name](par (par (p6 _) (q2 _)) p0)); [Apply NU_S; Intros; Apply COMM_PAR | Change (StBisim (nu [_:name](par ([u:name](par (p6 u) (q2 u)) _) p0)) (par (nu [_:name](par (p6 _) (q2 _))) p0)); Apply NU_EXTR] | Apply COMM_PAR]]]].

Inversion H2.

Exists (par (par p3 q0) r); Split; [Apply fPAR1; Apply fPAR1; Assumption | Left; Exists (par (par p3 q0) r); Exists (par p3 (par q0 r)); Split; [Apply REF | Split; [Apply bispar_a | Apply REF]]].

Inversion H7.
Exists (par (par p0 p6) r); Split; [Apply fPAR1; Apply fPAR2; Assumption | Left; Exists (par (par p0 p6) r); Exists (par p0 (par p6 r)); Split; [Apply REF | Split; [Apply bispar_a | Apply REF]]].
Exists (par (par p0 q0) p6); Split; [Apply fPAR2; Assumption | Left; Exists (par (par p0 q0) p6); Exists (par p0 (par q0 p6)); Split; [Apply REF | Split; [Apply bispar_a | Apply REF]]].
Exists (par (par p0 (q3 y)) q2); Split; [Change (ftrans (par (par p0 q0) r) tau (par ([_:name](par p0 (q3 _)) y) q2)); Apply COM1 with x; [Apply bPAR2; Assumption | Assumption] | Left; Exists (par (par p0 (q3 y)) q2); Exists (par p0 (par (q3 y) q2)); Split; [Apply REF | Split; [Apply bispar_a | Apply REF]]].
Exists (par (par p0 q2) (q3 y)); Split; [Apply COM2 with x; [Apply fPAR2; Assumption | Assumption] | Left; Exists (par (par p0 q2) (q3 y)); Exists (par p0 (par q2 (q3 y))); Split; [Apply REF | Split; [Apply bispar_a | Apply REF]]].
Exists (nu [_:name](par (par p0 (q2 _)) (q3 _))); Split; [Change (ftrans (par (par p0 q0) r) tau (nu [_:name](par ([u:name](par p0 (q2 u)) _) (q3 _)))); Apply CLOSE1 with x; [Apply bPAR2; Assumption | Assumption] | Right; Exists [_:name](par (par p0 (q2 _)) (q3 _)); Exists [_:name](par p0 (par (q2 _) (q3 _))); Split; [Apply REF | Split; [Intros; Apply bispar_a | Apply TRANS with (nu [_:name](par (par (q2 _) (q3 _)) p0)); [Apply NU_S; Intros; Apply COMM_PAR | Apply TRANS with (par (nu [z:name](par (q2 z) (q3 z))) p0); [Change (StBisim (nu [_:name](par ([u:name](par (q2 u) (q3 u)) _) p0)) (par (nu [z:name](par (q2 z) (q3 z))) p0)); Apply NU_EXTR | Apply COMM_PAR]]]]].
Exists (nu [_:name](par (par p0 (q2 _)) (q3 _))); Split; [Change (ftrans (par (par p0 q0) r) tau (nu [_:name](par ([u:name](par p0 (q2 u)) _) (q3 _)))); Apply CLOSE2 with x; [Apply bPAR2; Assumption | Assumption] | Right; Exists [_:name](par (par p0 (q2 _)) (q3 _)); Exists [_:name](par p0 (par (q2 _) (q3 _))); Split; [Apply REF | Split; [Intros; Apply bispar_a | Apply TRANS with (nu [_:name](par (par (q2 _) (q3 _)) p0)); [Apply NU_S; Intros; Apply COMM_PAR | Apply TRANS with (par (nu [z:name](par (q2 z) (q3 z))) p0); [Change (StBisim (nu [_:name](par ([u:name](par (q2 u) (q3 u)) _) p0)) (par (nu [z:name](par (q2 z) (q3 z))) p0)); Apply NU_EXTR | Apply COMM_PAR]]]]].

Inversion H8.
Exists (par (par (q3 y) p5) r); Split; [Apply fPAR1; Apply COM1 with x; Assumption | Left; Exists (par (par (q3 y) p5) r); Exists (par (q3 y) (par p5 r)); Split; [Apply REF | Split; [Apply bispar_a | Apply REF]]].
Exists (par (par (q3 y) q0) p5); Split; [Change (ftrans (par (par p0 q0) r) tau (par ([_:name](par (q3 _) q0) y) p5)); Apply COM1 with x; [Apply bPAR1; Assumption | Assumption] | Left; Exists (par (par (q3 y) q0) p5); Exists (par (q3 y) (par q0 p5)); Split; [Apply REF | Split; [Apply bispar_a | Apply REF]]].

Inversion H8.
Exists (par (par q2 (p5 y)) r); Split; [Apply fPAR1; Apply COM2 with x; Assumption | Left; Exists (par (par q2 (p5 y)) r); Exists (par q2 (par (p5 y) r)); Split; [Apply REF | Split; [Apply bispar_a | Apply REF]]].
Exists (par (par q2 q0) (p5 y)); Split; [Apply COM2 with x; [Apply fPAR1; Assumption | Assumption] | Left; Exists (par (par q2 q0) (p5 y)); Exists (par q2 (par q0 (p5 y))); Split; [Apply REF | Split; [Apply bispar_a | Apply REF]]].

Inversion H8.
Exists (par (nu [_:name](par (q2 _) (p5 _))) r); Split; [Apply fPAR1; Apply CLOSE1 with x; Assumption | Right; Exists [_:name](par (par (q2 _) (p5 _)) r); Exists [z:name](par (q2 z) (par (p5 z) r)); Split; [Apply SYM; Change (StBisim 
     (nu [_:name](par ([u:name](par (q2 u) (p5 u)) _) r)) (par (nu [_:name](par (q2 _) (p5 _))) r)); Apply NU_EXTR | Split; [Intros; Apply bispar_a | Apply REF]]].
Exists (nu [_:name](par (par (q2 _) q0) (p5 _))); Split; [Change (ftrans (par (par p0 q0) r) tau (nu [_:name](par ([u:name](par (q2 u) q0) _) (p5 _)))); Apply CLOSE1 with x; [Apply bPAR1; Assumption | Assumption] | Right; Exists [_:name](par (par (q2 _) q0) (p5 _)); Exists [z:name](par (q2 z) (par q0 (p5 z))); Split; [Apply REF | Split; [Intros; Apply bispar_a | Apply REF]]].

Inversion H8.
Exists (par (nu [_:name](par (q2 _) (p5 _))) r); Split; [Apply fPAR1; Apply CLOSE2 with x; Assumption | Right; Exists [_:name](par (par (q2 _) (p5 _)) r); Exists [z:name](par (q2 z) (par (p5 z) r)); Split; [Apply SYM; Change (StBisim 
     (nu [_:name](par ([u:name](par (q2 u) (p5 u)) _) r)) (par (nu [_:name](par (q2 _) (p5 _))) r)); Apply NU_EXTR | Split; [Intros; Apply bispar_a | Apply REF]]].
Exists (nu [_:name](par (par (q2 _) q0) (p5 _))); Split; [Change (ftrans (par (par p0 q0) r) tau (nu [_:name](par ([u:name](par (q2 u) q0) _) (p5 _)))); Apply CLOSE2 with x; [Apply bPAR1; Assumption | Assumption] | Right; Exists [_:name](par (par (q2 _) q0) (p5 _)); Exists [z:name](par (q2 z) (par q0 (p5 z))); Split; [Apply REF | Split; [Intros; Apply bispar_a | Apply REF]]].

Qed.

(* BisPAR_A is a strong bisimulation UpTo *)

Lemma BisPAR_A_UPTO: (UpTo BisPAR_A).

Proof.

Unfold UpTo; Intros.

Split; Intros.

Rewrite H2 in H; Rewrite H3 in H; Apply PAR_A_RW with x; Auto.

Split; Intros.

Elim (BisPAR_A_FTR1 p q H x y); Intros; Split; Intros.

Elim (H0 p1 H2); Intros; Inversion_clear H3.
Exists x0; Split; [Assumption | Unfold SB_R_SB; Exists p1; Exists x0; Split; [Apply REF | Split; [Assumption | Apply REF]]].

Elim (H1 q1 H2); Intros; Inversion_clear H3.
Exists x0; Split; [Assumption | Unfold SB_R_SB; Exists x0; Exists q1; Split; [Apply REF | Split; [Assumption | Apply REF]]].

Split; Intros.

Elim (BisPAR_A_IN p q H x); Intros; Split; Intros.

Elim (H0 p1 H2); Intros; Inversion_clear H3.
Exists x0; Split; [Assumption | Intro; Unfold SB_R_SB; Exists (p1 y); Exists (x0 y); Split; [Apply REF | Split; [Auto | Apply REF]]].

Elim (H1 q1 H2); Intros; Inversion_clear H3.
Exists x0; Split; [Assumption | Intro; Unfold SB_R_SB; Exists (x0 y); Exists (q1 y); Split; [Apply REF | Split; [Auto | Apply REF]]].

Split; Intros.

Elim (BisPAR_A_bOUT p q H x); Intros; Split; Intros.

Elim (H0 p1 H2); Intros; Inversion_clear H3.
Exists x0; Split; [Assumption | Intros; Unfold SB_R_SB; Exists (p1 y); Exists (x0 y); Split; [Apply REF | Split; [Auto | Apply REF]]].

Elim (H1 q1 H2); Intros; Inversion_clear H3.
Exists x0; Split; [Assumption | Intros; Unfold SB_R_SB; Exists (x0 y); Exists (q1 y); Split; [Apply REF | Split; [Auto | Apply REF]]].

Elim (BisPAR_A_FTR2 p q H); Intros; Split; Intros.

Elim (H0 p1 H2); Intros; Inversion_clear H3; Inversion_clear H5.
Exists x; Split; [Assumption | Left; Inversion_clear H3; Inversion_clear H5; Inversion_clear H3; Inversion_clear H6; Unfold SB_R_SB; Exists x0; Exists x1; Split; [Assumption | Split; Assumption]].
Exists x; Split; [Assumption | Right; Inversion_clear H3; Inversion_clear H5; Inversion_clear H3; Inversion_clear H6; Exists x0; Exists x1; Split; [Assumption | Split; [Intros; Auto | Assumption]]].

Elim (H1 q1 H2); Intros; Inversion_clear H3; Inversion_clear H5.
Exists x; Split; [Assumption | Left; Inversion_clear H3; Inversion_clear H5; Inversion_clear H3; Inversion_clear H6; Unfold SB_R_SB; Exists x0; Exists x1; Split; [Assumption | Split; Assumption]].
Exists x; Split; [Assumption | Right; Inversion_clear H3; Inversion_clear H5; Inversion_clear H3; Inversion_clear H6; Exists x0; Exists x1; Split; [Assumption | Split; [Intros; Auto | Assumption]]].

Qed.

Lemma ASS_PAR: (p,q,r:proc)(StBisim (par (par p q) r) (par p (par q r))).

Proof.

Intros; Apply Completeness; Apply Co_Ind with (SBUPTO BisPAR_A); [Apply UPTO_SB'; Exact BisPAR_A_UPTO | Apply sbupto with O; Simpl; Unfold SB_R_SB; Exists (par (par p q) r); Exists (par p (par q r)); Split; [Apply REF | Split; [Apply bispar_a | Apply REF]]].

Qed.

End upto.

Section structural_congruence2.

Lemma PAR_S2: (p,q,r,s:proc)(StBisim p q)->(StBisim r s)->(StBisim (par p r) (par q s)).

Proof. (* This lemma depends on Lemmata TRANS, PAR_S and COMM_PAR. *)

Intros; Apply TRANS with (par s q);
 [Apply TRANS with (par r q);
   [Apply TRANS with (par q r);
     [Apply PAR_S; Assumption | Apply COMM_PAR]
   | Apply PAR_S; Assumption]
 | Apply COMM_PAR].

Qed.

End structural_congruence2.

Section restriction_laws2.

Lemma NU_P: (p:proc)(StBisim (nu [x:name]p) p).

Proof. (* This lemma depends on Lemmata REF, SYM, TRANS, NU_S,
        * ID_PAR, COMM_PAR, NU_EXTR, PAR_S2, NU_NIL.
        *)

Intro; Apply TRANS with (nu [_:name](par p nil)).
Apply NU_S; Intros; Apply SYM; Apply ID_PAR.
Apply TRANS with (nu [_:name](par nil p)).
Apply NU_S; Intros; Apply COMM_PAR.
Apply TRANS with (par nil p).
Apply TRANS with (par (nu [x:name]nil) p).
Change (StBisim (nu [_:name](par ([x:name]nil _) p)) (par (nu [x:name]nil) p)); Apply NU_EXTR.
Apply PAR_S2; [Apply NU_NIL | Apply REF].
Apply TRANS with (par p nil); [Apply COMM_PAR | Apply ID_PAR].

Qed.

Lemma NU_PAR1: (p:name->proc)(q:proc)(StBisim (nu [x:name](par (p x) q)) (par (nu p) (nu [x:name]q))).

Proof. (* This lemma depends on Lemmata REF, SYM, TRANS, NU_EXTR, NU_P, PAR_S2. *)

Intros; Apply TRANS with (par (nu p) q); [Apply NU_EXTR | Apply PAR_S2; [Apply REF | Apply SYM; Apply NU_P]].

Qed.

Lemma NU_PAR2: (p:proc)(q:name->proc)(StBisim (nu [x:name](par p (q x))) (par (nu [x:name]p) (nu q))).

Proof. (* This lemma depends on Lemmata REF, SYM, TRANS, NU_EXTR, NU_P, PAR_S2, COMM_PAR. *)

Intros; Apply TRANS with (nu [x:name](par (q x) p)); [Apply NU_S; Intros; Apply COMM_PAR | Apply TRANS with (par (nu q) p); [Apply NU_EXTR | Apply TRANS with (par (nu q) (nu [_:name]p)); [Apply PAR_S2; [Apply REF | Apply SYM; Apply NU_P] | Apply COMM_PAR]]].

Qed.

End restriction_laws2.

Section matching_laws.

Lemma MATCH1: (p:proc)(x,y:name)(x=y)->(StBisim (match x y p) p).

Proof. (* This lemma depends on Lemma REF. *)

Intros; Apply sb; Do 3 Try (Split; Intros).

Inversion_clear H0.
Exists p1; Split; [Assumption | Apply REF].
Exists q1; Split; [Rewrite H; Apply fMATCH; Assumption | Apply REF].

Inversion_clear H0.
Exists p1; Split; [Assumption | Intro; Apply REF].
Exists q1; Split; [Rewrite H; Apply bMATCH; Assumption | Intro; Apply REF].

Inversion_clear H0.
Exists p1; Split; [Assumption | Intros; Apply REF].
Exists q1; Split; [Rewrite H; Apply bMATCH; Assumption | Intros; Apply REF].

Qed.

Lemma MATCH2: (p:proc)(x,y:name)~(x=y)->(StBisim (match x y p) nil).

Proof. (* This lemma is proved from scratch. *)

Intros; Apply sb; Do 3 Try (Split; Intros).

Inversion H0; Absurd x=y; Assumption.
Inversion_clear H0.

Inversion H0; Absurd x=y; Assumption.
Inversion_clear H0.

Inversion H0; Absurd x=y; Assumption.
Inversion_clear H0.

Qed.

Lemma MISMATCH1: (p:proc)(x,y:name)~(x=y)->(StBisim (mismatch x y p) p).

Proof. (* This lemma depends on Lemma REF. *)

Intros; Apply sb; Do 3 Try (Split; Intros).

Inversion_clear H0.
Exists p1; Split; [Assumption | Apply REF].
Exists q1; Split; [Apply fMISMATCH; Assumption | Apply REF].

Inversion_clear H0.
Exists p1; Split; [Assumption | Intro; Apply REF].
Exists q1; Split; [Apply bMISMATCH; Assumption | Intro; Apply REF].

Inversion_clear H0.
Exists p1; Split; [Assumption | Intros; Apply REF].
Exists q1; Split; [Apply bMISMATCH; Assumption | Intros; Apply REF].

Qed.

Lemma MISMATCH2: (p:proc)(x,y:name)(x=y)->(StBisim (mismatch x y p) nil).

Proof. (* This lemma is proved from scratch. *)

Intros; Apply sb; Do 3 Try (Split; Intros).

Inversion H0; Absurd x=y; Assumption.
Inversion_clear H0.

Inversion H0; Absurd x=y; Assumption.
Inversion_clear H0.

Inversion H0; Absurd x=y; Assumption.
Inversion_clear H0.

Qed.

End matching_laws.
