(*******************************************************)
(*                                                     *)
(* Technical lemmata of the metatheory of pi-calculus  *)
(*    (Milner-Parrow-Walker's Lemmata 1, 3, 4, 6).     *)
(*                                                     *)
(*******************************************************)
(*                                                     *)
(* File   : basiclemmata.v                             *)
(* Author : Ivan Scagnetto (scagnett@dimi.uniud.it),   *)
(*          Dipartimento di Matematica e Informatica,  *)
(*          University of Udine                        *)
(* Date   : July 1998                                  *)
(*                                                     *)
(*******************************************************)
(*                                                     *)
(* This file is part of "Pi-calculus in (Co)Inductive  *)
(* Type Theory", a joint work with Furio Honsell and   *)
(* Marino Miculan (accepted for publication on TCS).   *)
(* If you are interested you can find more information *)
(* (and a gzipped PostScript version of the article)   *)
(* at the following URL:                               *)
(* http://www.dimi.uniud.it/~scagnett/pi-calculus.html *)
(*                                                     *)
(*******************************************************)

Require Export xadequacy.

Require Export hoaspack.

(*******************************************************)
(* Mutual induction principle needed in order to prove *)
(* results about LTS.                                  *)
(*******************************************************)

Scheme ftrans_btrans_ind:= Minimality for ftrans Sort Prop
with btrans_ftrans_ind:= Minimality for btrans Sort Prop.

Section MPW_Lemma1.

Lemma FTR_L1: (p,q:proc)(a:f_act)(x:name)
                (notin x p)
                  ->(ftrans p a q)
                    ->((f_act_notin x a)/\(notin x q)).

Proof.

Intros.
Generalize H0.
Generalize H.
Generalize x.
Pattern p a q;
Apply ftrans_btrans_ind with [p:proc][a:b_act][q:name->proc]((x:name)
                (notin x p)
                  ->(btrans p a q)
                    ->((b_act_notin x a)/\(notin x (nu q))));
Intros; Auto.

(* TAU Case *)

Split; [Apply f_act_notin_tau | Inversion H1; Assumption].

(* OUT Case *)

Inversion H1; Split; [Apply f_act_notin_Out; Assumption | Assumption].

(* fSUM1 Case *)

Inversion_clear H3.
Elim (H2 x0 H5 H1); Intros; Split; Assumption.

(* fSUM2 Case *)

Inversion_clear H3.
Elim (H2 x0 H6 H1); Intros; Split; Assumption.

(* fPAR1 Case *)

Inversion_clear H3.
Elim (H2 x0 H5 H1); Intros; Split; [Assumption | Apply notin_par; Assumption].

(* fPAR2 Case *)

Inversion_clear H3.
Elim (H2 x0 H6 H1); Intros; Split; [Assumption | Apply notin_par; Assumption].

(* fMATCH Case *)

Inversion_clear H3.
Elim (H2 x1 H7 H1); Intros; Split; Assumption.

(* fMISMATCH Case *)

Inversion_clear H4.
Elim (H3 x1 H8 H2); Intros; Split; Assumption.

(* fBANG Case *)

Inversion_clear H3.
Cut (notin x0 (par p0 (bang p0))); [Intro | Apply notin_par; [Assumption | Apply notin_bang; Assumption]].
Elim (H2 x0 H3 H1); Intros; Split; Assumption.

(* COM1 Case *)

Inversion_clear H5.
Elim (H2 x1 H7 H1); Intros; Elim (H4 x1 H8 H3); Intros; Split; [Apply f_act_notin_tau | Apply notin_par].
Inversion_clear H10; Inversion_clear H9; Auto.
Assumption.

(* COM2 Case *)

Inversion_clear H5.
Elim (H2 x1 H7 H1); Intros; Elim (H4 x1 H8 H3); Intros; Split; [Apply f_act_notin_tau | Apply notin_par].
Assumption.
Inversion_clear H5; Inversion_clear H11; Auto.

(* fRES Case *)

Generalize H1 H2; Elim a0; Intros.
Elim (unsat (par (nu p1) (nu p2)) (cons x0 l)); Intros.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H9.
Cut (ftrans (p1 x1) tau (p2 x1)); [Intro | Apply H5; Auto; Try Apply f_act_notin_tau].
Cut (f_act_notin x1 tau); [Intro | Apply f_act_notin_tau].
Cut (notin x0 (p1 x1)); [Intro | Inversion H3; Auto].
Elim (H6 x1 H7 H10 H11 H12 x0 H13 H9); Intros.
Split; [Assumption | Apply proc_mono with x1; Assumption].

Elim (unsat (par (nu p1) (nu p2)) (cons x0 (cons n (cons n0 l)))); Intros.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H9.
Inversion_clear H11.
Inversion_clear H12.
Cut (ftrans (p1 x1) (Out n n0) (p2 x1)); [Intro | Apply H5; Auto; Try Apply f_act_notin_Out; Auto].
Cut (f_act_notin x1 (Out n n0)); [Intro | Apply f_act_notin_Out; Auto].
Cut (notin x0 (p1 x1)); [Intro | Inversion H3; Auto].
Elim (H6 x1 H7 H10 H13 H14 x0 H15 H12); Intros.
Split; [Assumption | Apply proc_mono with x1; Assumption].

(* CLOSE1 Case *)

Split; [Apply f_act_notin_tau | Apply notin_nu; Intros].
Inversion_clear H5.
Elim (H2 x1 H8 H1); Intros; Elim (H4 x1 H9 H3); Intros.
Apply notin_par; [Inversion_clear H10; Auto | Inversion_clear H12; Auto].

(* CLOSE2 Case *)

Split; [Apply f_act_notin_tau | Apply notin_nu; Intros].
Inversion_clear H5.
Elim (H2 x1 H8 H1); Intros; Elim (H4 x1 H9 H3); Intros.
Apply notin_par; [Inversion_clear H10; Auto | Inversion_clear H12; Auto].

(* IN Case *)

Inversion_clear H1; Split; [Apply b_act_notin_In; Auto | Apply notin_nu; Intros; Auto].

(* bSUM1 Case *)

Inversion_clear H3.
Elim (H2 x0 H5 H1); Intros; Split; Assumption.

(* bSUM2 Case *)

Inversion_clear H3.
Elim (H2 x0 H6 H1); Intros; Split; Assumption.

(* bPAR1 Case *)

Inversion_clear H3.
Elim (H2 x0 H5 H1); Intros; Split; [Assumption | Apply notin_nu; Intros; Apply notin_par; Inversion H7; Auto].

(* bPAR2 Case *)

Inversion_clear H3.
Elim (H2 x0 H6 H1); Intros; Split; [Assumption | Apply notin_nu; Intros; Apply notin_par; Inversion H7; Auto].

(* bMATCH Case *)

Inversion_clear H3.
Elim (H2 x1 H7 H1); Intros; Split; Assumption.

(* bMISMATCH Case *)

Inversion_clear H4.
Elim (H3 x1 H8 H2); Intros; Split; Assumption.

(* bBANG Case *)

Inversion_clear H3.
Cut (notin x0 (par p0 (bang p0))); [Intro | Apply notin_par; [Assumption | Apply notin_bang; Assumption]].
Elim (H2 x0 H3 H1); Intros; Split; Assumption.

(* bRES Case *)

Generalize H1 H2; Elim a0; Intros.
Elim (unsat (par (nu p1) (nu [_:name](nu (p2 _)))) (cons x0 (cons n l))); Intros.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H9.
Inversion_clear H11.
Cut (btrans (p1 x1) (In n) (p2 x1)); [Intro | Apply H5; Auto; Try Apply b_act_notin_In; Auto].
Cut (b_act_notin x1 (In n)); [Intro | Apply b_act_notin_In; Auto].
Cut (notin x0 (p1 x1)); [Intro | Inversion H3; Auto].
Elim (H6 x1 H7 H10 H13 H12 x0 H14 H11); Intros.
Split; [Assumption | Apply notin_nu; Intros; Apply proc_mono with x1; Inversion H16; Auto].

Elim (unsat (par (nu p1) (nu [_:name](nu (p2 _)))) (cons x0 (cons n l))); Intros.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H9.
Inversion_clear H11.
Cut (btrans (p1 x1) (bOut n) (p2 x1)); [Intro | Apply H5; Auto; Try Apply b_act_notin_bOut; Auto].
Cut (b_act_notin x1 (bOut n)); [Intro | Apply b_act_notin_bOut; Auto].
Cut (notin x0 (p1 x1)); [Intro | Inversion H3; Auto].
Elim (H6 x1 H7 H10 H13 H12 x0 H14 H11); Intros.
Split; [Assumption | Apply notin_nu; Intros; Apply proc_mono with x1; Inversion H16; Auto].

(* OPEN Case *)

Elim (unsat (par (nu p1) (nu p2)) (cons x0 (cons x1 l))); Intros.
Inversion_clear H5.
Inversion_clear H6.
Inversion_clear H7.
Inversion_clear H9.
Cut (ftrans (p1 x2) (Out x0 x2) (p2 x2)); [Intro | Apply H1; Auto].
Cut (notin x1 (p1 x2)); [Intro | Inversion H3; Auto].
Cut ~x0=x2; [Intro | Auto].
Elim (H2 x2 H5 H8 H12 H10 x1 H11 H9); Intros.
Split; [Inversion_clear H13; Apply b_act_notin_bOut; Auto | Apply proc_mono with x2; Auto].

Qed.

Lemma BTR_L1: (p:proc)(q:name->proc)(a:b_act)(x:name)
                (notin x p)
                  ->(btrans p a q)
                    ->((b_act_notin x a)/\(notin x (nu q))).

Proof.

Intros.
Generalize H0.
Generalize H.
Generalize x.
Pattern p a q;
Apply btrans_ftrans_ind with [p:proc][a:f_act][q:proc]((x:name)
                (notin x p)
                  ->(ftrans p a q)
                    ->((f_act_notin x a)/\(notin x q)));
Intros; Auto.

(* TAU Case *)

Split; [Apply f_act_notin_tau | Inversion H1; Assumption].

(* OUT Case *)

Inversion H1; Split; [Apply f_act_notin_Out; Assumption | Assumption].

(* fSUM1 Case *)

Inversion_clear H3.
Elim (H2 x0 H5 H1); Intros; Split; Assumption.

(* fSUM2 Case *)

Inversion_clear H3.
Elim (H2 x0 H6 H1); Intros; Split; Assumption.

(* fPAR1 Case *)

Inversion_clear H3.
Elim (H2 x0 H5 H1); Intros; Split; [Assumption | Apply notin_par; Assumption].

(* fPAR2 Case *)

Inversion_clear H3.
Elim (H2 x0 H6 H1); Intros; Split; [Assumption | Apply notin_par; Assumption].

(* fMATCH Case *)

Inversion_clear H3.
Elim (H2 x1 H7 H1); Intros; Split; Assumption.

(* fMISMATCH Case *)

Inversion_clear H4.
Elim (H3 x1 H8 H2); Intros; Split; Assumption.

(* fBANG Case *)

Inversion_clear H3.
Cut (notin x0 (par p0 (bang p0))); [Intro | Apply notin_par; [Assumption | Apply notin_bang; Assumption]].
Elim (H2 x0 H3 H1); Intros; Split; Assumption.

(* COM1 Case *)

Inversion_clear H5.
Elim (H2 x1 H7 H1); Intros; Elim (H4 x1 H8 H3); Intros; Split; [Apply f_act_notin_tau | Apply notin_par].
Inversion_clear H10; Inversion_clear H9; Auto.
Assumption.

(* COM2 Case *)

Inversion_clear H5.
Elim (H2 x1 H7 H1); Intros; Elim (H4 x1 H8 H3); Intros; Split; [Apply f_act_notin_tau | Apply notin_par].
Assumption.
Inversion_clear H5; Inversion_clear H11; Auto.

(* fRES Case *)

Generalize H1 H2; Elim a0; Intros.
Elim (unsat (par (nu p1) (nu p2)) (cons x0 l)); Intros.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H9.
Cut (ftrans (p1 x1) tau (p2 x1)); [Intro | Apply H5; Auto; Try Apply f_act_notin_tau].
Cut (f_act_notin x1 tau); [Intro | Apply f_act_notin_tau].
Cut (notin x0 (p1 x1)); [Intro | Inversion H3; Auto].
Elim (H6 x1 H7 H10 H11 H12 x0 H13 H9); Intros.
Split; [Assumption | Apply proc_mono with x1; Assumption].

Elim (unsat (par (nu p1) (nu p2)) (cons x0 (cons n (cons n0 l)))); Intros.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H9.
Inversion_clear H11.
Inversion_clear H12.
Cut (ftrans (p1 x1) (Out n n0) (p2 x1)); [Intro | Apply H5; Auto; Try Apply f_act_notin_Out; Auto].
Cut (f_act_notin x1 (Out n n0)); [Intro | Apply f_act_notin_Out; Auto].
Cut (notin x0 (p1 x1)); [Intro | Inversion H3; Auto].
Elim (H6 x1 H7 H10 H13 H14 x0 H15 H12); Intros.
Split; [Assumption | Apply proc_mono with x1; Assumption].

(* CLOSE1 Case *)

Split; [Apply f_act_notin_tau | Apply notin_nu; Intros].
Inversion_clear H5.
Elim (H2 x1 H8 H1); Intros; Elim (H4 x1 H9 H3); Intros.
Apply notin_par; [Inversion_clear H10; Auto | Inversion_clear H12; Auto].

(* CLOSE2 Case *)

Split; [Apply f_act_notin_tau | Apply notin_nu; Intros].
Inversion_clear H5.
Elim (H2 x1 H8 H1); Intros; Elim (H4 x1 H9 H3); Intros.
Apply notin_par; [Inversion_clear H10; Auto | Inversion_clear H12; Auto].

(* IN Case *)

Inversion_clear H1; Split; [Apply b_act_notin_In; Auto | Apply notin_nu; Intros; Auto].

(* bSUM1 Case *)

Inversion_clear H3.
Elim (H2 x0 H5 H1); Intros; Split; Assumption.

(* bSUM2 Case *)

Inversion_clear H3.
Elim (H2 x0 H6 H1); Intros; Split; Assumption.

(* bPAR1 Case *)

Inversion_clear H3.
Elim (H2 x0 H5 H1); Intros; Split; [Assumption | Apply notin_nu; Intros; Apply notin_par; Inversion H7; Auto].

(* bPAR2 Case *)

Inversion_clear H3.
Elim (H2 x0 H6 H1); Intros; Split; [Assumption | Apply notin_nu; Intros; Apply notin_par; Inversion H7; Auto].

(* bMATCH Case *)

Inversion_clear H3.
Elim (H2 x1 H7 H1); Intros; Split; Assumption.

(* bMISMATCH Case *)

Inversion_clear H4.
Elim (H3 x1 H8 H2); Intros; Split; Assumption.

(* bBANG Case *)

Inversion_clear H3.
Cut (notin x0 (par p0 (bang p0))); [Intro | Apply notin_par; [Assumption | Apply notin_bang; Assumption]].
Elim (H2 x0 H3 H1); Intros; Split; Assumption.

(* bRES Case *)

Generalize H1 H2; Elim a0; Intros.
Elim (unsat (par (nu p1) (nu [_:name](nu (p2 _)))) (cons x0 (cons n l))); Intros.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H9.
Inversion_clear H11.
Cut (btrans (p1 x1) (In n) (p2 x1)); [Intro | Apply H5; Auto; Try Apply b_act_notin_In; Auto].
Cut (b_act_notin x1 (In n)); [Intro | Apply b_act_notin_In; Auto].
Cut (notin x0 (p1 x1)); [Intro | Inversion H3; Auto].
Elim (H6 x1 H7 H10 H13 H12 x0 H14 H11); Intros.
Split; [Assumption | Apply notin_nu; Intros; Apply proc_mono with x1; Inversion H16; Auto].

Elim (unsat (par (nu p1) (nu [_:name](nu (p2 _)))) (cons x0 (cons n l))); Intros.
Inversion_clear H7.
Inversion_clear H8.
Inversion_clear H9.
Inversion_clear H11.
Cut (btrans (p1 x1) (bOut n) (p2 x1)); [Intro | Apply H5; Auto; Try Apply b_act_notin_bOut; Auto].
Cut (b_act_notin x1 (bOut n)); [Intro | Apply b_act_notin_bOut; Auto].
Cut (notin x0 (p1 x1)); [Intro | Inversion H3; Auto].
Elim (H6 x1 H7 H10 H13 H12 x0 H14 H11); Intros.
Split; [Assumption | Apply notin_nu; Intros; Apply proc_mono with x1; Inversion H16; Auto].

(* OPEN Case *)

Elim (unsat (par (nu p1) (nu p2)) (cons x0 (cons x1 l))); Intros.
Inversion_clear H5.
Inversion_clear H6.
Inversion_clear H7.
Inversion_clear H9.
Cut (ftrans (p1 x2) (Out x0 x2) (p2 x2)); [Intro | Apply H1; Auto].
Cut (notin x1 (p1 x2)); [Intro | Inversion H3; Auto].
Cut ~x0=x2; [Intro | Auto].
Elim (H2 x2 H5 H8 H12 H10 x1 H11 H9); Intros.
Split; [Inversion_clear H13; Apply b_act_notin_bOut; Auto | Apply proc_mono with x2; Auto].

Qed.

Lemma Aux1: (p,q:proc)(a:f_act)(x:name)
             (f_act_isin x a)
               ->(ftrans p a q)
                 ->(isin x p).

Proof.

Intros.
Apply isin_to_notin.
(Unfold not; Intro).
Absurd x=x.
(Elim (FTR_L1 p q a x H1 H0); Intros).
(Apply Sep_f_act with a; Assumption).
Trivial.

Qed.

Lemma Aux2: (p:proc)(a:b_act)(q:name->proc)(x:name)
             (b_act_isin x a)
               ->(btrans p a q)
                 ->(isin x p).

Proof.

Intros.
Apply isin_to_notin.
(Unfold not; Intro).
Absurd x=x.
(Elim (BTR_L1 p q a x H1 H0); Intros).
(Apply Sep_b_act with a; Assumption).
Trivial.

Qed.

End MPW_Lemma1.

Section MPW_Lemma3.

(*************************************************)
(* Lemmata 3 and 4 are the same in our encoding. *)
(*************************************************)

Lemma L3: (p,q:proc)(a:f_act)
              (p',q':name->proc)(a':name->f_act)(x:name)
                (notin x (nu p'))->(notin x (nu q'))->(f_act_notin_ho x a') -> p=(p' x) -> q=(q' x) -> a=(a' x)
                  ->(ftrans p a q)
                    ->(y:name)(notin y (nu p'))->(notin y (nu q'))->(f_act_notin_ho y a')
                      ->(ftrans (p' y) (a' y) (q' y)).

Proof.

Do 14 Intro.
Generalize H5.
Generalize H4.
Generalize H3.
Generalize H2.
Generalize H1.
Generalize H0.
Generalize H.
Generalize x.
Generalize a'.
Generalize q'.
Generalize p'.
(Pattern p a q ; Apply ftrans_btrans_ind with
[p:proc][a:b_act][q:name->proc]
  ((p':name->proc)(q':name->name->proc)(a':name->b_act)(x:name)
                (notin x (nu p'))->(notin x (nu [z:name](nu (q' z))))->(b_act_notin_ho x a') -> p=(p' x) -> q=(q' x) -> a=(a' x)
                  ->(btrans p a q)
                    ->(y:name)(notin y (nu p'))->(notin y (nu [z:name](nu (q' z))))->(b_act_notin_ho y a')
                      ->(btrans (p' y) (a' y) (q' y)));
Intros; Auto).

(* TAU Case *)

Rewrite H10 in H9.
Cut (tau_pref (q'0 y))=(p'0 y); [Intro | Change ([_:name](tau_pref (q'0 _)) y)=(p'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H16.
Cut tau=(a'0 y); [Intro | Change ([_:name]tau y)=(a'0 y); Apply f_act_sub_congr with x0; Auto].
Rewrite <- H17.
Apply TAU.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply notin_nu; Intros; Apply notin_tau; Inversion H7; Auto.

(* OUT Case *)

Elim (LEM_name x0 x1); Elim (LEM_name x1 y); Intros.

(* Subcase x0=x1=y *)

Rewrite <- H16 in H9; Rewrite H17 in H9.
Rewrite <- H16 in H11; Rewrite H17 in H11.
Rewrite H10 in H9.
Cut (out_pref y0 y0 (q'0 y0))=(p'0 y0); [Intro | Change ([_:name](out_pref _ _ (q'0 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Rewrite <- H18.
Cut (Out y0 y0)=(a'0 y0); [Intro | Change ([_:name](Out _ _) y0)=(a'0 y0); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H19.
Apply OUT.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Apply notin_nu; Intros; Apply notin_out; Inversion H7; Auto.

(* Subcase x0=x1 /\ ~x1=y *)

Rewrite H17 in H9.
Rewrite H17 in H11.
Rewrite H10 in H9.
Cut (out_pref y0 y (q'0 y0))=(p'0 y0); [Intro | Change ([_:name](out_pref _ y (q'0 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Rewrite <- H18.
Cut (Out y0 y)=(a'0 y0); [Intro | Change ([_:name](Out _ y) y0)=(a'0 y0); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H19.
Apply OUT.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Apply notin_nu; Intros; Apply notin_out; Inversion H7; Auto.

(* Subcase ~x0=x1 /\ x1=y *)

Rewrite <- H16 in H9.
Rewrite <- H16 in H11.
Rewrite H10 in H9.
Cut (out_pref x0 y0 (q'0 y0))=(p'0 y0); [Intro | Change ([_:name](out_pref x0 _ (q'0 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Rewrite <- H18.
Cut (Out x0 y0)=(a'0 y0); [Intro | Change ([_:name](Out x0 _) y0)=(a'0 y0); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H19.
Apply OUT.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Apply notin_nu; Intros; Apply notin_out; Inversion H7; Auto.

(* Subcase ~x0=x1 /\ ~x1=y *)

Rewrite H10 in H9.
Cut (out_pref x0 y (q'0 y0))=(p'0 y0); [Intro | Change ([_:name](out_pref x0 y (q'0 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Rewrite <- H18.
Cut (Out x0 y)=(a'0 y0); [Intro | Change ([_:name](Out x0 y) y0)=(a'0 y0); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H19.
Apply OUT.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Apply notin_nu; Intros; Apply notin_out; Inversion H7; Auto.

(* fSUM1 Case *)

Elim (LEM_name x0 y); Intros.

(* Subcase x0=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=y *)

Elim (proc_exp p1 x0); Elim (proc_exp p2 x0); Intros.
Inversion_clear H19; Inversion_clear H20.
Rewrite H23 in H11; Rewrite H22 in H11.
Cut (sum (x2 y) (x1 y))=(p'0 y); [Intro | Change ([_:name](sum (x2 _) (x1 _)) y)=(p'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H20.
Apply fSUM1.
Apply H7 with x0; Auto.
Apply proc_mono with x0.
Inversion_clear H15.
Cut (notin y (p'0 x0)); [Intro | Auto].
Rewrite <- H11 in H15.
Inversion_clear H15; Assumption.
Apply notin_nu; Intros; Apply notin_sum; [Inversion_clear H19; Auto | Inversion_clear H21; Auto].

(* fSUM2 Case *)

Elim (LEM_name x0 y); Intros.

(* Subcase x0=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=y *)

Elim (proc_exp p1 x0); Elim (proc_exp p2 x0); Intros.
Inversion_clear H19; Inversion_clear H20.
Rewrite H23 in H11; Rewrite H22 in H11.
Cut (sum (x2 y) (x1 y))=(p'0 y); [Intro | Change ([_:name](sum (x2 _) (x1 _)) y)=(p'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H20.
Apply fSUM2.
Apply H7 with x0; Auto.
Apply proc_mono with x0.
Inversion_clear H15.
Cut (notin y (p'0 x0)); [Intro | Auto].
Rewrite <- H11 in H15.
Inversion_clear H15; Assumption.
Apply notin_nu; Intros; Apply notin_sum; [Inversion_clear H19; Auto | Inversion_clear H21; Auto].

(* fPAR1 Case *)

Elim (LEM_name x0 y); Intros.

(* Subcase x0=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=y *)

Elim (proc_exp p1 x0); Elim (proc_exp p2 x0); Elim (proc_exp p0 x0); Intros.
Inversion_clear H19; Inversion_clear H20; Inversion_clear H21.
Rewrite H23 in H12; Rewrite H24 in H11; Rewrite H24 in H12; Rewrite H25 in H11.
Cut (par (x1 y) (x2 y))=(q'0 y); [Intro | Change ([_:name](par (x1 _) (x2 _)) y)=(q'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H21.
Cut (par (x3 y) (x2 y))=(p'0 y); [Intro | Change ([_:name](par (x3 _) (x2 _)) y)=(p'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H26.
Apply fPAR1.
Apply H7 with x0; Auto.
Apply proc_mono with x0.
Inversion_clear H15.
Cut (notin y (p'0 x0)); [Intro | Auto].
Rewrite <- H11 in H15.
Inversion_clear H15; Assumption.
Apply proc_mono with x0.
Inversion_clear H16.
Cut (notin y (q'0 x0)); [Intro | Auto].
Rewrite <- H12 in H16.
Inversion_clear H16; Assumption.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H20; Auto | Inversion_clear H19; Auto].
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H22; Auto | Inversion_clear H19; Auto].

(* fPAR2 Case *)

Elim (LEM_name x0 y); Intros.

(* Subcase x0=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=y *)

Elim (proc_exp p1 x0); Elim (proc_exp p2 x0); Elim (proc_exp p0 x0); Intros.
Inversion_clear H19; Inversion_clear H20; Inversion_clear H21.
Rewrite H23 in H12; Rewrite H24 in H11; Rewrite H25 in H11; Rewrite H25 in H12.
Cut (par (x3 y) (x1 y))=(q'0 y); [Intro | Change ([_:name](par (x3 _) (x1 _)) y)=(q'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H21.
Cut (par (x3 y) (x2 y))=(p'0 y); [Intro | Change ([_:name](par (x3 _) (x2 _)) y)=(p'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H26.
Apply fPAR2.
Apply H7 with x0; Auto.
Apply proc_mono with x0.
Inversion_clear H15.
Cut (notin y (p'0 x0)); [Intro | Auto].
Rewrite <- H11 in H15.
Inversion_clear H15; Assumption.
Apply proc_mono with x0.
Inversion_clear H16.
Cut (notin y (q'0 x0)); [Intro | Auto].
Rewrite <- H12 in H16.
Inversion_clear H16; Assumption.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H20; Auto | Inversion_clear H19; Auto].
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H20; Auto | Inversion_clear H22; Auto].

(* fMATCH Case *)

Elim (LEM_name x0 x1); Elim (LEM_name x1 y); Intros.

(* Subcase x0=x1=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase x0=x1 /\ ~x1=y *)

Elim (proc_exp p0 x1); Intros.
Inversion_clear H20.
Rewrite H22 in H11.
Rewrite H19 in H11.
Cut (match y y (x2 y))=(p'0 y); [Intro | Change ([_:name](match _ _ (x2 _)) y)=(p'0 y); Apply proc_sub_congr with x1; Auto].
Rewrite <- H20.
Apply fMATCH.
Apply H7 with x1; Auto.
Apply proc_mono with x1.
Inversion_clear H15.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H11 in H15.
Inversion_clear H15; Assumption.
Apply notin_nu; Intros; Apply notin_match; Inversion H21; Auto.

(* Subcase ~x0=x1 /\ x1=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=x1 /\ ~x1=y *)

Elim (proc_exp p0 x1); Intros.
Inversion_clear H20.
Rewrite H22 in H11.
Cut (match x0 x0 (x2 y))=(p'0 y); [Intro | Change ([_:name](match x0 x0 (x2 _)) y)=(p'0 y); Apply proc_sub_congr with x1; Auto].
Rewrite <- H20.
Apply fMATCH.
Apply H7 with x1; Auto.
Apply proc_mono with x1.
Inversion_clear H15.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H11 in H15.
Inversion_clear H15; Assumption.
Apply notin_nu; Intros; Apply notin_match; Inversion H21; Auto.

(* fMISMATCH Case *)

Elim (LEM_name x1 y0); Intros.

(* Subcase x1=y0 *)

Rewrite <- H19; Rewrite <- H12; Rewrite <- H13; Rewrite <- H14; Assumption.

(* Subcase ~x1=y0 *)

Elim (LEM_name x0 x1); Elim (LEM_name y x1); Intros.

(* Subsubcase x0=x1=y *)

Absurd x0=y; [Assumption | Rewrite H20; Rewrite H21; Trivial].

(* Subsubcase x0=x1 /\ ~x1=y *)

Rewrite H21 in H12.
Elim (proc_exp p0 x1); Intros.
Inversion_clear H22.
Rewrite H24 in H12.
Cut (mismatch y0 y (x2 y0))=(p'0 y0); [Intro | Change ([_:name](mismatch _ y (x2 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Rewrite <- H22.
Apply fMISMATCH.
Cut ~y=y0; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H12; Apply isin_mismatch3.
Inversion_clear H16; Auto.
Apply H8 with x1; Auto.
Apply proc_mono with x1.
Inversion_clear H16.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H12 in H16.
Inversion_clear H16; Assumption.
Apply notin_nu; Intros; Apply notin_mismatch; Inversion H23; Auto.

(* Subsubcase ~x0=x1 /\ x1=y *)

Rewrite H20 in H12.
Elim (proc_exp p0 x1); Intros.
Inversion_clear H22.
Rewrite H24 in H12.
Cut (mismatch x0 y0 (x2 y0))=(p'0 y0); [Intro | Change ([_:name](mismatch x0 _ (x2 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Rewrite <- H22.
Apply fMISMATCH.
Apply Sep_proc with (p'0 x1).
Rewrite <- H12; Apply isin_mismatch2.
Inversion_clear H16; Auto.
Apply H8 with x1; Auto.
Apply proc_mono with x1.
Inversion_clear H16.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H12 in H16.
Inversion_clear H16; Assumption.
Apply notin_nu; Intros; Apply notin_mismatch; Inversion H23; Auto.

(* Subsubcase ~x0=x1 /\ ~x1=y *)

Elim (proc_exp p0 x1); Intros.
Inversion_clear H22.
Rewrite H24 in H12.
Cut (mismatch x0 y (x2 y0))=(p'0 y0); [Intro | Change ([_:name](mismatch x0 y (x2 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Rewrite <- H22.
Apply fMISMATCH.
Inversion H15; Assumption.
Apply H8 with x1; Auto.
Apply proc_mono with x1.
Inversion_clear H16.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H12 in H16.
Inversion_clear H16; Assumption.
Apply notin_nu; Intros; Apply notin_mismatch; Inversion H23; Auto.

(* fBANG Case *)

Elim (LEM_name x0 y); Intros.

(* Subcase x0=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=y *)

Elim (proc_exp p0 x0); Intros.
Inversion_clear H19.
Rewrite H21 in H11.
Cut (bang (x1 y))=(p'0 y); [Intro | Change ([_:name](bang (x1 _)) y)=(p'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H19; Apply fBANG.
Change (ftrans ([_:name](par (x1 _) (bang (x1 _))) y) (a'0 y) (q'0 y)); Apply H7 with x0; Auto.
Apply notin_nu; Intros; Inversion_clear H20; Apply notin_par; [Auto | Apply notin_bang; Auto].
Rewrite H21; Trivial.
Apply proc_mono with x0.
Inversion_clear H15.
Cut (notin y (p'0 x0)); [Intro | Auto].
Rewrite <- H11 in H15; Inversion H15; Apply notin_par; Assumption.
Apply notin_nu; Intros; Apply notin_bang; Inversion_clear H20; Auto.

(* COM1 Case *)

Elim (LEM_name x1 y0); Intros.

(* Subcase x1=y0 *)

Rewrite <- H20; Rewrite <- H13; Rewrite <- H14; Rewrite <- H15; Assumption.

(* Subcase ~x1=y0 *)

Elim (proc_exp p1 x1); Elim (proc_exp p2 x1); Elim (ho_proc_exp q1 x1); Elim (proc_exp q2 x1); Intros.
Inversion_clear H21; Inversion_clear H22; Inversion_clear H23; Inversion_clear H24.
Rewrite H26 in H14; Rewrite H27 in H14; Rewrite H28 in H13; Rewrite H29 in H13.
Elim (LEM_name x0 x1); Elim (LEM_name x1 y); Intros.

Rewrite <- H24 in H14.
Cut (par (x3 y0 y0) (x2 y0))=(q'0 y0); [Intro | Change ([_:name](par (x3 _ _) (x2 _)) y0)=(q'0 y0); Apply proc_sub_congr with x1; Auto].
Cut (par (x5 y0) (x4 y0))=(p'0 y0); [Intro | Change ([_:name](par (x5 _) (x4 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Cut tau=(a'0 y0); [Intro | Change ([_:name]tau y0)=(a'0 y0); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H31; Rewrite <- H32; Rewrite <- H33.
Apply COM1 with y0.
Change (btrans (x5 y0) ([_:name](In _) y0) (x3 y0)).
Apply H7 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Rewrite H30; Trivial.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Apply proc_mono with x1.
Inversion_clear H18; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Change (ftrans (x4 y0) ([_:name](Out _ _) y0) (x2 y0)).
Apply H9 with x1; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite <- H24; Rewrite H30; Trivial.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Inversion_clear H18; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H23; Auto | Inversion_clear H22; Auto].
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H21; Cut (notin x1 (nu (x3 z))); [Intro | Auto]; Inversion_clear H21; Auto | Inversion_clear H25; Auto].

Cut (par (x3 y0 y) (x2 y0))=(q'0 y0); [Intro | Change ([_:name](par (x3 _ y) (x2 _)) y0)=(q'0 y0); Apply proc_sub_congr with x1; Auto].
Cut (par (x5 y0) (x4 y0))=(p'0 y0); [Intro | Change ([_:name](par (x5 _) (x4 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Cut tau=(a'0 y0); [Intro | Change ([_:name]tau y0)=(a'0 y0); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H31; Rewrite <- H32; Rewrite <- H33.
Apply COM1 with y0.
Change (btrans (x5 y0) ([_:name](In _) y0) (x3 y0)).
Apply H7 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Rewrite H30; Trivial.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Apply proc_mono with y.
Inversion_clear H18; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Change (ftrans (x4 y0) ([_:name](Out _ y) y0) (x2 y0)).
Apply H9 with x1; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite H30; Trivial.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Inversion_clear H18; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Cut ~y=y0; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par2.
Rewrite <- H28.
Apply Aux1 with q2 (Out x0 y); [Apply f_act_isin_Out2 | Auto].
Inversion_clear H17; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H23; Auto | Inversion_clear H22; Auto].
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H21; Cut (notin x1 (nu (x3 z))); [Intro | Auto]; Inversion_clear H21; Auto | Inversion_clear H25; Auto].

Rewrite <- H24 in H14.
Cut (par (x3 y0 y0) (x2 y0))=(q'0 y0); [Intro | Change ([_:name](par (x3 _ _) (x2 _)) y0)=(q'0 y0); Apply proc_sub_congr with x1; Auto].
Cut (par (x5 y0) (x4 y0))=(p'0 y0); [Intro | Change ([_:name](par (x5 _) (x4 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Cut tau=(a'0 y0); [Intro | Change ([_:name]tau y0)=(a'0 y0); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H31; Rewrite <- H32; Rewrite <- H33.
Apply COM1 with x0.
Change (btrans (x5 y0) ([_:name](In x0) y0) (x3 y0)).
Apply H7 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Apply proc_mono with x1.
Inversion_clear H17; Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17; Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Apply proc_mono with x1.
Inversion_clear H18; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Cut ~x0=y0; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par2.
Rewrite <- H28.
Apply Aux1 with q2 (Out x0 y); [Apply f_act_isin_Out1 | Auto].
Inversion_clear H17; Auto.
Change (ftrans (x4 y0) ([_:name](Out x0 _) y0) (x2 y0)).
Apply H9 with x1; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite H24; Trivial.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Inversion_clear H18; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Cut ~x0=y0; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par2.
Rewrite <- H28.
Apply Aux1 with q2 (Out x0 y); [Apply f_act_isin_Out1 | Auto].
Inversion_clear H17; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H23; Auto | Inversion_clear H22; Auto].
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H21; Cut (notin x1 (nu (x3 z))); [Intro | Auto]; Inversion_clear H21; Auto | Inversion_clear H25; Auto].

Cut (par (x3 y0 y) (x2 y0))=(q'0 y0); [Intro | Change ([_:name](par (x3 _ y) (x2 _)) y0)=(q'0 y0); Apply proc_sub_congr with x1; Auto].
Cut (par (x5 y0) (x4 y0))=(p'0 y0); [Intro | Change ([_:name](par (x5 _) (x4 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Cut tau=(a'0 y0); [Intro | Change ([_:name]tau y0)=(a'0 y0); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H31; Rewrite <- H32; Rewrite <- H33.
Apply COM1 with x0.
Change (btrans (x5 y0) ([_:name](In x0) y0) (x3 y0)).
Apply H7 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Apply proc_mono with x1.
Inversion_clear H17; Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17; Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Apply proc_mono with y.
Inversion_clear H18; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Cut ~x0=y0; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par2.
Rewrite <- H28.
Apply Aux1 with q2 (Out x0 y); [Apply f_act_isin_Out1 | Auto].
Inversion_clear H17; Auto.
Change (ftrans (x4 y0) ([_:name](Out x0 y) y0) (x2 y0)).
Apply H9 with x1; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Inversion_clear H18; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Cut ~x0=y0; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par2.
Rewrite <- H28.
Apply Aux1 with q2 (Out x0 y); [Apply f_act_isin_Out1 | Auto].
Inversion_clear H17; Auto.
Cut ~y=y0; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par2.
Rewrite <- H28.
Apply Aux1 with q2 (Out x0 y); [Apply f_act_isin_Out2 | Auto].
Inversion_clear H17; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H23; Auto | Inversion_clear H22; Auto].
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H21; Cut (notin x1 (nu (x3 z))); [Intro | Auto]; Inversion_clear H21; Auto | Inversion_clear H25; Auto].

(* COM2 Case *)

Elim (LEM_name x1 y0); Intros.

(* Subcase x1=y0 *)

Rewrite <- H20; Rewrite <- H13; Rewrite <- H14; Rewrite <- H15; Assumption.

(* Subcase ~x1=y0 *)

Elim (proc_exp p1 x1); Elim (proc_exp p2 x1); Elim (proc_exp q1 x1); Elim (ho_proc_exp q2 x1); Intros.
Inversion_clear H21; Inversion_clear H22; Inversion_clear H23; Inversion_clear H24.
Rewrite H26 in H14; Rewrite H27 in H14; Rewrite H28 in H13; Rewrite H29 in H13.
Elim (LEM_name x0 x1); Elim (LEM_name x1 y); Intros.

Rewrite <- H24 in H14.
Cut (par (x3 y0) (x2 y0 y0))=(q'0 y0); [Intro | Change ([_:name](par (x3 _) (x2 _ _)) y0)=(q'0 y0); Apply proc_sub_congr with x1; Auto].
Cut (par (x5 y0) (x4 y0))=(p'0 y0); [Intro | Change ([_:name](par (x5 _) (x4 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Cut tau=(a'0 y0); [Intro | Change ([_:name]tau y0)=(a'0 y0); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H31; Rewrite <- H32; Rewrite <- H33.
Apply COM2 with y0.
Change (ftrans (x5 y0) ([_:name](Out _ _) y0) (x3 y0)).
Apply H7 with x1; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite <- H24; Rewrite H30; Trivial.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Inversion_clear H18; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Change (btrans (x4 y0) ([_:name](In _) y0) (x2 y0)).
Apply H9 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Rewrite H30; Trivial.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Apply proc_mono with x1.
Inversion_clear H18; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H23; Auto | Inversion_clear H22; Auto].
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H21; Auto | Inversion_clear H25; Cut (notin x1 (nu (x2 z))); [Intro | Auto]; Inversion_clear H25; Auto].

Cut (par (x3 y0) (x2 y0 y))=(q'0 y0); [Intro | Change ([_:name](par (x3 _) (x2 _ y)) y0)=(q'0 y0); Apply proc_sub_congr with x1; Auto].
Cut (par (x5 y0) (x4 y0))=(p'0 y0); [Intro | Change ([_:name](par (x5 _) (x4 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Cut tau=(a'0 y0); [Intro | Change ([_:name]tau y0)=(a'0 y0); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H31; Rewrite <- H32; Rewrite <- H33.
Apply COM2 with y0.

Change (ftrans (x5 y0) ([_:name](Out _ y) y0) (x3 y0)).
Apply H7 with x1; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite H30; Trivial.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Inversion_clear H18; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Cut ~y=y0; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par1.
Rewrite <- H29.
Apply Aux1 with q1 (Out x0 y); [Apply f_act_isin_Out2 | Auto].
Inversion_clear H17; Auto.
Change (btrans (x4 y0) ([_:name](In _) y0) (x2 y0)).
Apply H9 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Rewrite H30; Trivial.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Apply proc_mono with y.
Inversion_clear H18; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H23; Auto | Inversion_clear H22; Auto].
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H21; Auto | Inversion_clear H25; Cut (notin x1 (nu (x2 z))); [Intro | Auto]; Inversion_clear H25; Auto].

Rewrite <- H24 in H14.
Cut (par (x3 y0) (x2 y0 y0))=(q'0 y0); [Intro | Change ([_:name](par (x3 _) (x2 _ _)) y0)=(q'0 y0); Apply proc_sub_congr with x1; Auto].
Cut (par (x5 y0) (x4 y0))=(p'0 y0); [Intro | Change ([_:name](par (x5 _) (x4 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Cut tau=(a'0 y0); [Intro | Change ([_:name]tau y0)=(a'0 y0); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H31; Rewrite <- H32; Rewrite <- H33.
Apply COM2 with x0.
Change (ftrans (x5 y0) ([_:name](Out x0 _) y0) (x3 y0)).
Apply H7 with x1; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite H24; Trivial.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Inversion_clear H18; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Cut ~x0=y0; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par1.
Rewrite <- H29.
Apply Aux1 with q1 (Out x0 y); [Apply f_act_isin_Out1 | Auto].
Inversion_clear H17; Auto.
Change (btrans (x4 y0) ([_:name](In x0) y0) (x2 y0)).
Apply H9 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Apply proc_mono with x1.
Inversion_clear H17; Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17; Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Apply proc_mono with x1.
Inversion_clear H18; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Cut ~x0=y0; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par1.
Rewrite <- H29.
Apply Aux1 with q1 (Out x0 y); [Apply f_act_isin_Out1 | Auto].
Inversion_clear H17; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H23; Auto | Inversion_clear H22; Auto].
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H21; Auto | Inversion_clear H25; Cut (notin x1 (nu (x2 z))); [Intro | Auto]; Inversion_clear H25; Auto].

Cut (par (x3 y0) (x2 y0 y))=(q'0 y0); [Intro | Change ([_:name](par (x3 _) (x2 _ y)) y0)=(q'0 y0); Apply proc_sub_congr with x1; Auto].
Cut (par (x5 y0) (x4 y0))=(p'0 y0); [Intro | Change ([_:name](par (x5 _) (x4 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Cut tau=(a'0 y0); [Intro | Change ([_:name]tau y0)=(a'0 y0); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H31; Rewrite <- H32; Rewrite <- H33.
Apply COM2 with x0.
Change (ftrans (x5 y0) ([_:name](Out x0 y) y0) (x3 y0)).
Apply H7 with x1; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Inversion_clear H18; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Cut ~x0=y0; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par1.
Rewrite <- H29.
Apply Aux1 with q1 (Out x0 y); [Apply f_act_isin_Out1 | Auto].
Inversion_clear H17; Auto.
Cut ~y=y0; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par1.
Rewrite <- H29.
Apply Aux1 with q1 (Out x0 y); [Apply f_act_isin_Out2 | Auto].
Inversion_clear H17; Auto.
Change (btrans (x4 y0) ([_:name](In x0) y0) (x2 y0)).
Apply H9 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Apply proc_mono with x1.
Inversion_clear H17; Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17; Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Apply proc_mono with y.
Inversion_clear H18; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Cut ~x0=y0; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par1.
Rewrite <- H29.
Apply Aux1 with q1 (Out x0 y); [Apply f_act_isin_Out1 | Auto].
Inversion_clear H17; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H23; Auto | Inversion_clear H22; Auto].
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H21; Auto | Inversion_clear H25; Cut (notin x1 (nu (x2 z))); [Intro | Auto]; Inversion_clear H25; Auto].

(* fRES Case *)

Elim (LEM_name x0 y); Intros.

(* Subcase x0=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=y *)

Elim (ho_proc_exp p1 x0); Elim (ho_proc_exp p2 x0); Intros.
Inversion_clear H19; Inversion_clear H20.
Rewrite H22 in H12; Rewrite H23 in H11.
Cut (nu (x2 y))=(p'0 y); [Intro | Change ([_:name](nu (x2 _)) y)=(p'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H20.
Cut (nu (x1 y))=(q'0 y); [Intro | Change ([_:name](nu (x1 _)) y)=(q'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H24.
Apply fRES with (cons x0 (cons y l)); Intros.
Inversion_clear H27.
Inversion_clear H30.
Change (ftrans ([_:name](x2 _ y0) y) (a'0 y) ([_:name](x1 _ y0) y)).
Apply H7 with y0 x0; Auto.
Cut (notin y0 (nu [_:name](nu (x2 _)))); [Intro | Apply proc_mono with y; Assumption].
Rewrite H23; Inversion_clear H30; Auto.
Cut (notin y0 (nu [_:name](nu (x1 _)))); [Intro | Apply proc_mono with y; Assumption].
Rewrite H22; Inversion_clear H30; Auto.
Rewrite H13; Cut (f_act_notin_ho y0 a'0); [Intro | Apply f_act_mono with y; Assumption].
Unfold f_act_notin_ho in H30; Auto.
Apply notin_nu; Intros.
Inversion_clear H19.
Cut (notin x0 (nu (x2 z))); [Intro | Auto].
Inversion_clear H19; Auto.
Apply notin_nu; Intros.
Inversion_clear H21.
Cut (notin x0 (nu (x1 z))); [Intro | Auto].
Inversion_clear H21; Auto.
Rewrite H23; Trivial.
Rewrite H22; Trivial.
Apply H6; Auto.
Cut (notin y0 (nu [_:name](nu (x2 _)))); [Intro | Apply proc_mono with y; Auto].
Inversion_clear H30.
Cut (notin y0 (nu (x2 x0))); [Intro | Auto].
Rewrite H23; Assumption.
Cut (notin y0 (nu [_:name](nu (x1 _)))); [Intro | Apply proc_mono with y; Auto].
Inversion_clear H30.
Cut (notin y0 (nu (x1 x0))); [Intro | Auto].
Rewrite H22; Assumption.
Rewrite H13; Cut (f_act_notin_ho y0 a'0); [Intro | Apply f_act_mono with y; Assumption].
Unfold f_act_notin_ho in H30; Auto.
Cut (notin  y (p'0 x0)); [Intro | Inversion_clear H15; Auto].
Cut (notin y (nu [_:name](nu (x2 _)))); [Intro | Apply proc_mono with x0; Rewrite <- H11 in H30; Auto].
Apply notin_nu; Intros.
Inversion_clear H32.
Cut (notin y (nu (x2 z))); [Intro | Auto].
Inversion_clear H32; Auto.
Cut (notin  y (q'0 x0)); [Intro | Inversion_clear H16; Auto].
Cut (notin y (nu [_:name](nu (x1 _)))); [Intro | Apply proc_mono with x0; Rewrite <- H12 in H30; Auto].
Apply notin_nu; Intros.
Inversion_clear H32.
Cut (notin y (nu (x1 z))); [Intro | Auto].
Inversion_clear H32; Auto.

(* CLOSE1 Case *)

Elim (LEM_name x1 y); Intros.

(* Subcase x1=y *)

Rewrite <- H20; Rewrite <- H13; Rewrite <- H14; Rewrite <- H15; Assumption.

(* Subcase ~x1=y *)

Elim (proc_exp p1 x1); Elim (proc_exp p2 x1); Elim (ho_proc_exp q1 x1); Elim (ho_proc_exp q2 x1); Intros.
Inversion_clear H21; Inversion_clear H22; Inversion_clear H23; Inversion_clear H24.
Rewrite H26 in H14; Rewrite H27 in H14; Rewrite H28 in H13; Rewrite H29 in H13.
Elim (LEM_name x0 x1); Intros.
Cut (nu [z:name](par (x3 y z) (x2 y z)))=(q'0 y); [Intro | Change ([_:name](nu [z:name](par (x3 _ z) (x2 _ z))) y)=(q'0 y); Apply proc_sub_congr with x1; Auto].
Cut (par (x5 y) (x4 y))=(p'0 y); [Intro | Change ([_:name](par (x5 _) (x4 _)) y)=(p'0 y); Apply proc_sub_congr with x1; Auto].
Cut tau=(a'0 y); [Intro | Change ([_:name]tau y)=(a'0 y); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H30; Rewrite <- H31; Rewrite <- H32.
Apply CLOSE1 with y.
Apply H7 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Rewrite H24; Trivial.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y (par (x3 x1 x1) (x2 x1 x1))); [Intro | Auto].
Inversion_clear H18; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Apply H9 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Rewrite H24; Trivial.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Inversion_clear H18.
Cut (notin y (par (x3 x1 x1) (x2 x1 x1))); [Intro | Auto].
Inversion_clear H18; Apply proc_mono with x1; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H23; Auto | Inversion_clear H22; Auto].
Apply notin_nu; Intros; Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H21; Cut (notin x1 (nu (x3 z))); [Intro | Auto]; Inversion_clear H21; Auto | Inversion_clear H25; Cut (notin x1 (nu (x2 z))); [Intro | Auto]; Inversion_clear H25; Auto].

Cut (nu [z:name](par (x3 y z) (x2 y z)))=(q'0 y); [Intro | Change ([_:name](nu [z:name](par (x3 _ z) (x2 _ z))) y)=(q'0 y); Apply proc_sub_congr with x1; Auto].
Cut (par (x5 y) (x4 y))=(p'0 y); [Intro | Change ([_:name](par (x5 _) (x4 _)) y)=(p'0 y); Apply proc_sub_congr with x1; Auto].
Cut tau=(a'0 y); [Intro | Change ([_:name]tau y)=(a'0 y); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H30; Rewrite <- H31; Rewrite <- H32.
Apply CLOSE1 with x0.
Change (btrans (x5 y) ([_:name](In x0) y) (x3 y)).
Apply H7 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y (par (x3 x1 x1) (x2 x1 x1))); [Intro | Auto].
Inversion_clear H18; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Cut ~x0=y; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par2.
Rewrite <- H28.
Apply Aux2 with (bOut x0) q2; [Apply b_act_isin_bOut | Auto].
Inversion_clear H17; Auto.
Change (btrans (x4 y) ([_:name](bOut x0) y) (x2 y)).
Apply H9 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Inversion_clear H18.
Cut (notin y (par (x3 x1 x1) (x2 x1 x1))); [Intro | Auto].
Inversion_clear H18; Apply proc_mono with x1; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Cut ~x0=y; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par2.
Rewrite <- H28.
Apply Aux2 with (bOut x0) q2; [Apply b_act_isin_bOut | Auto].
Inversion_clear H17; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H23; Auto | Inversion_clear H22; Auto].
Apply notin_nu; Intros; Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H21; Cut (notin x1 (nu (x3 z))); [Intro | Auto]; Inversion_clear H21; Auto | Inversion_clear H25; Cut (notin x1 (nu (x2 z))); [Intro | Auto]; Inversion_clear H25; Auto].

(* CLOSE2 Case *)

Elim (LEM_name x1 y); Intros.

(* Subcase x1=y *)

Rewrite <- H20; Rewrite <- H13; Rewrite <- H14; Rewrite <- H15; Assumption.

(* Subcase ~x1=y *)

Elim (proc_exp p1 x1); Elim (proc_exp p2 x1); Elim (ho_proc_exp q1 x1); Elim (ho_proc_exp q2 x1); Intros.
Inversion_clear H21; Inversion_clear H22; Inversion_clear H23; Inversion_clear H24.
Rewrite H26 in H14; Rewrite H27 in H14; Rewrite H28 in H13; Rewrite H29 in H13.
Elim (LEM_name x0 x1); Intros.
Cut (nu [z:name](par (x3 y z) (x2 y z)))=(q'0 y); [Intro | Change ([_:name](nu [z:name](par (x3 _ z) (x2 _ z))) y)=(q'0 y); Apply proc_sub_congr with x1; Auto].
Cut (par (x5 y) (x4 y))=(p'0 y); [Intro | Change ([_:name](par (x5 _) (x4 _)) y)=(p'0 y); Apply proc_sub_congr with x1; Auto].
Cut tau=(a'0 y); [Intro | Change ([_:name]tau y)=(a'0 y); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H30; Rewrite <- H31; Rewrite <- H32.
Apply CLOSE2 with y.
Apply H7 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Rewrite H24; Trivial.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Inversion_clear H18.
Cut (notin y (par (x3 x1 x1) (x2 x1 x1))); [Intro | Auto].
Inversion_clear H18; Apply proc_mono with x1; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Apply H9 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Rewrite H24; Trivial.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y (par (x3 x1 x1) (x2 x1 x1))); [Intro | Auto].
Inversion_clear H18; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H23; Auto | Inversion_clear H22; Auto].
Apply notin_nu; Intros; Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H21; Cut (notin x1 (nu (x3 z))); [Intro | Auto]; Inversion_clear H21; Auto | Inversion_clear H25; Cut (notin x1 (nu (x2 z))); [Intro | Auto]; Inversion_clear H25; Auto].

Cut (nu [z:name](par (x3 y z) (x2 y z)))=(q'0 y); [Intro | Change ([_:name](nu [z:name](par (x3 _ z) (x2 _ z))) y)=(q'0 y); Apply proc_sub_congr with x1; Auto].
Cut (par (x5 y) (x4 y))=(p'0 y); [Intro | Change ([_:name](par (x5 _) (x4 _)) y)=(p'0 y); Apply proc_sub_congr with x1; Auto].
Cut tau=(a'0 y); [Intro | Change ([_:name]tau y)=(a'0 y); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H30; Rewrite <- H31; Rewrite <- H32.
Apply CLOSE2 with x0.
Change (btrans (x5 y) ([_:name](bOut x0) y) (x3 y)).
Apply H7 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y (par (x3 x1 x1) (x2 x1 x1))); [Intro | Auto].
Inversion_clear H18; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Cut ~x0=y; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par2.
Rewrite <- H28.
Apply Aux2 with (In x0) q2; [Apply b_act_isin_In | Auto].
Inversion_clear H17; Auto.
Change (btrans (x4 y) ([_:name](In x0) y) (x2 y)).
Apply H9 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Inversion_clear H18.
Cut (notin y (par (x3 x1 x1) (x2 x1 x1))); [Intro | Auto].
Inversion_clear H18; Apply proc_mono with x1; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Cut ~x0=y; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par2.
Rewrite <- H28.
Apply Aux2 with (In x0) q2; [Apply b_act_isin_In | Auto].
Inversion_clear H17; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H23; Auto | Inversion_clear H22; Auto].
Apply notin_nu; Intros; Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H21; Cut (notin x1 (nu (x3 z))); [Intro | Auto]; Inversion_clear H21; Auto | Inversion_clear H25; Cut (notin x1 (nu (x2 z))); [Intro | Auto]; Inversion_clear H25; Auto].

(* IN Case *)

Elim (LEM_name x1 y); Intros.

(* Subcase x1=y *)

Rewrite <- H16; Rewrite <- H9; Rewrite <- H10; Rewrite <- H11; Assumption.

(* Subcase ~x1=y *)

Elim (ho_proc_exp p0 x1); Intros.
Inversion_clear H17.
Rewrite H19 in H9; Rewrite H19 in H10.
Elim (LEM_name x0 x1); Intros.
Rewrite H17 in H9.
Rewrite H17 in H11.
Cut (in_pref y (x2 y))=(p'0 y); [Intro | Change ([_:name](in_pref _ (x2 _)) y)=(p'0 y); Apply proc_sub_congr with x1; Auto].
Cut (x2 y)=(q'0 y); [Intro | Apply ho_proc_sub_congr with x1; Auto].
Cut (In y)=(a'0 y); [Intro | Apply b_act_sub_congr with x1; Auto].
Rewrite <- H20; Rewrite <- H21; Rewrite <- H22; Apply IN.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Apply notin_nu; Intros; Apply notin_in; Intros; Auto.
Inversion_clear H18.
Cut (notin x1 (nu (x2 z))); [Intro | Auto].
Inversion_clear H18; Auto.

Cut (in_pref x0 (x2 y))=(p'0 y); [Intro | Change ([_:name](in_pref x0 (x2 _)) y)=(p'0 y); Apply proc_sub_congr with x1; Auto].
Cut (x2 y)=(q'0 y); [Intro | Apply ho_proc_sub_congr with x1; Auto].
Cut (In x0)=(a'0 y); [Intro | Change ([_:name](In x0) y)=(a'0 y); Apply b_act_sub_congr with x1; Auto].
Rewrite <- H20; Rewrite <- H21; Rewrite <- H22; Apply IN.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Apply notin_nu; Intros; Apply notin_in; Intros; Auto.
Inversion_clear H18.
Cut (notin x1 (nu (x2 z))); [Intro | Auto].
Inversion_clear H18; Auto.

(* bSUM1 Case *)

Elim (LEM_name x0 y); Intros.

(* Subcase x0=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=y *)

Elim (proc_exp p1 x0); Elim (proc_exp p2 x0); Intros.
Inversion_clear H19; Inversion_clear H20.
Rewrite H23 in H11; Rewrite H22 in H11.
Cut (sum (x2 y) (x1 y))=(p'0 y); [Intro | Change ([_:name](sum (x2 _) (x1 _)) y)=(p'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H20.
Apply bSUM1.
Apply H7 with x0; Auto.
Apply proc_mono with x0.
Inversion_clear H15.
Cut (notin y (p'0 x0)); [Intro | Auto].
Rewrite <- H11 in H15.
Inversion_clear H15; Assumption.
Apply notin_nu; Intros; Apply notin_sum; [Inversion_clear H19; Auto | Inversion_clear H21; Auto].

(* bSUM2 Case *)

Elim (LEM_name x0 y); Intros.

(* Subcase x0=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=y *)

Elim (proc_exp p1 x0); Elim (proc_exp p2 x0); Intros.
Inversion_clear H19; Inversion_clear H20.
Rewrite H23 in H11; Rewrite H22 in H11.
Cut (sum (x2 y) (x1 y))=(p'0 y); [Intro | Change ([_:name](sum (x2 _) (x1 _)) y)=(p'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H20.
Apply bSUM2.
Apply H7 with x0; Auto.
Apply proc_mono with x0.
Inversion_clear H15.
Cut (notin y (p'0 x0)); [Intro | Auto].
Rewrite <- H11 in H15.
Inversion_clear H15; Assumption.
Apply notin_nu; Intros; Apply notin_sum; [Inversion_clear H19; Auto | Inversion_clear H21; Auto].

(* bPAR1 Case *)

Elim (LEM_name x0 y); Intros.

(* Subcase x0=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=y *)

Elim (proc_exp p1 x0); Elim (proc_exp p2 x0); Elim (ho_proc_exp p0 x0); Intros.
Inversion_clear H19; Inversion_clear H20; Inversion_clear H21.
Rewrite H23 in H12; Rewrite H24 in H11; Rewrite H24 in H12; Rewrite H25 in H11.
Cut ([x:name](par (x1 y x) (x2 y)))=(q'0 y); [Intro | Change ([_:name]([x:name](par (x1 _ x) (x2 _))) y)=(q'0 y); Apply ho_proc_sub_congr with x0; Auto].
Rewrite <- H21.
Cut (par (x3 y) (x2 y))=(p'0 y); [Intro | Change ([_:name](par (x3 _) (x2 _)) y)=(p'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H26.
Apply bPAR1.
Apply H7 with x0; Auto.
Apply proc_mono with x0.
Inversion_clear H15.
Cut (notin y (p'0 x0)); [Intro | Auto].
Rewrite <- H11 in H15.
Inversion_clear H15; Assumption.
Apply proc_mono with x0.
Inversion_clear H16.
Cut (notin y (nu (q'0 x0))); [Intro | Auto].
Rewrite <- H12 in H16.
Apply proc_mono with x0; Inversion_clear H16; Cut (notin y (par (x1 x0 x0) (x2 x0))); [Intro | Auto]; Inversion_clear H16; Auto.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H20; Auto | Inversion_clear H19; Auto].
Apply notin_nu; Intros; Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H22; Cut (notin x0 (nu (x1 z))); [Intro | Auto]; Inversion_clear H22; Auto | Inversion_clear H19; Auto].

(* bPAR2 Case *)

Elim (LEM_name x0 y); Intros.

(* Subcase x0=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=y *)

Elim (proc_exp p1 x0); Elim (proc_exp p2 x0); Elim (ho_proc_exp p0 x0); Intros.
Inversion_clear H19; Inversion_clear H20; Inversion_clear H21.
Rewrite H23 in H12; Rewrite H24 in H11; Rewrite H25 in H11; Rewrite H25 in H12.
Cut ([x:name](par (x3 y) (x1 y x)))=(q'0 y); [Intro | Change ([_:name]([x:name](par (x3 _) (x1 _ x))) y)=(q'0 y); Apply ho_proc_sub_congr with x0; Auto].
Rewrite <- H21.
Cut (par (x3 y) (x2 y))=(p'0 y); [Intro | Change ([_:name](par (x3 _) (x2 _)) y)=(p'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H26.
Apply bPAR2.
Apply H7 with x0; Auto.
Apply proc_mono with x0.
Inversion_clear H15.
Cut (notin y (p'0 x0)); [Intro | Auto].
Rewrite <- H11 in H15.
Inversion_clear H15; Assumption.
Apply proc_mono with x0.
Inversion_clear H16.
Cut (notin y (nu (q'0 x0))); [Intro | Auto].
Rewrite <- H12 in H16.
Apply proc_mono with x0; Inversion_clear H16; Cut (notin y (par (x3 x0) (x1 x0 x0))); [Intro | Auto]; Inversion_clear H16; Auto.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H20; Auto | Inversion_clear H19; Auto].
Apply notin_nu; Intros; Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H20; Auto | Inversion_clear H22; Cut (notin x0 (nu (x1 z))); [Intro | Auto]; Inversion_clear H22; Auto].

(* bMATCH Case *)

Elim (LEM_name x0 x1); Elim (LEM_name x1 y); Intros.

(* Subcase x0=x1=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase x0=x1 /\ ~x1=y *)

Elim (proc_exp p0 x1); Intros.
Inversion_clear H20.
Rewrite H22 in H11.
Rewrite H19 in H11.
Cut (match y y (x2 y))=(p'0 y); [Intro | Change ([_:name](match _ _ (x2 _)) y)=(p'0 y); Apply proc_sub_congr with x1; Auto].
Rewrite <- H20.
Apply bMATCH.
Apply H7 with x1; Auto.
Apply proc_mono with x1.
Inversion_clear H15.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H11 in H15.
Inversion_clear H15; Assumption.
Apply notin_nu; Intros; Apply notin_match; Inversion H21; Auto.

(* Subcase ~x0=x1 /\ x1=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=x1 /\ ~x1=y *)

Elim (proc_exp p0 x1); Intros.
Inversion_clear H20.
Rewrite H22 in H11.
Cut (match x0 x0 (x2 y))=(p'0 y); [Intro | Change ([_:name](match x0 x0 (x2 _)) y)=(p'0 y); Apply proc_sub_congr with x1; Auto].
Rewrite <- H20.
Apply bMATCH.
Apply H7 with x1; Auto.
Apply proc_mono with x1.
Inversion_clear H15.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H11 in H15.
Inversion_clear H15; Assumption.
Apply notin_nu; Intros; Apply notin_match; Inversion H21; Auto.

(* bMISMATCH Case *)

Elim (LEM_name x1 y0); Intros.

(* Subcase x1=y0 *)

Rewrite <- H19; Rewrite <- H12; Rewrite <- H13; Rewrite <- H14; Assumption.

(* Subcase ~x1=y0 *)

Elim (LEM_name x0 x1); Elim (LEM_name y x1); Intros.

(* Subsubcase x0=x1=y *)

Absurd x0=y; [Assumption | Rewrite H20; Rewrite H21; Trivial].

(* Subsubcase x0=x1 /\ ~x1=y *)

Rewrite H21 in H12.
Elim (proc_exp p0 x1); Intros.
Inversion_clear H22.
Rewrite H24 in H12.
Cut (mismatch y0 y (x2 y0))=(p'0 y0); [Intro | Change ([_:name](mismatch _ y (x2 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Rewrite <- H22.
Apply bMISMATCH.
Cut ~y=y0; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H12; Apply isin_mismatch3.
Inversion_clear H16; Auto.
Apply H8 with x1; Auto.
Apply proc_mono with x1.
Inversion_clear H16.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H12 in H16.
Inversion_clear H16; Assumption.
Apply notin_nu; Intros; Apply notin_mismatch; Inversion H23; Auto.

(* Subsubcase ~x0=x1 /\ x1=y *)

Rewrite H20 in H12.
Elim (proc_exp p0 x1); Intros.
Inversion_clear H22.
Rewrite H24 in H12.
Cut (mismatch x0 y0 (x2 y0))=(p'0 y0); [Intro | Change ([_:name](mismatch x0 _ (x2 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Rewrite <- H22.
Apply bMISMATCH.
Apply Sep_proc with (p'0 x1).
Rewrite <- H12; Apply isin_mismatch2.
Inversion_clear H16; Auto.
Apply H8 with x1; Auto.
Apply proc_mono with x1.
Inversion_clear H16.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H12 in H16.
Inversion_clear H16; Assumption.
Apply notin_nu; Intros; Apply notin_mismatch; Inversion H23; Auto.

(* Subsubcase ~x0=x1 /\ ~x1=y *)

Elim (proc_exp p0 x1); Intros.
Inversion_clear H22.
Rewrite H24 in H12.
Cut (mismatch x0 y (x2 y0))=(p'0 y0); [Intro | Change ([_:name](mismatch x0 y (x2 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Rewrite <- H22.
Apply bMISMATCH.
Inversion H15; Assumption.
Apply H8 with x1; Auto.
Apply proc_mono with x1.
Inversion_clear H16.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H12 in H16.
Inversion_clear H16; Assumption.
Apply notin_nu; Intros; Apply notin_mismatch; Inversion H23; Auto.

(* bBANG Case *)

Elim (LEM_name x0 y); Intros.

(* Subcase x0=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=y *)

Elim (proc_exp p0 x0); Intros.
Inversion_clear H19.
Rewrite H21 in H11.
Cut (bang (x1 y))=(p'0 y); [Intro | Change ([_:name](bang (x1 _)) y)=(p'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H19; Apply bBANG.
Change (btrans ([_:name](par (x1 _) (bang (x1 _))) y) (a'0 y) (q'0 y)); Apply H7 with x0; Auto.
Apply notin_nu; Intros; Inversion_clear H20; Apply notin_par; [Auto | Apply notin_bang; Auto].
Rewrite H21; Trivial.
Apply proc_mono with x0.
Inversion_clear H15.
Cut (notin y (p'0 x0)); [Intro | Auto].
Rewrite <- H11 in H15; Inversion H15; Apply notin_par; Assumption.
Apply notin_nu; Intros; Apply notin_bang; Inversion_clear H20; Auto.

(* bRES Case *)

Elim (LEM_name x0 y); Intros.

(* Subcase x0=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=y *)

Elim (ho_proc_exp p1 x0); Elim (ho2_proc_exp p2 x0); Intros.
Inversion_clear H19; Inversion_clear H20.
Rewrite H22 in H12; Rewrite H23 in H11.
Cut (nu (x2 y))=(p'0 y); [Intro | Change ([_:name](nu (x2 _)) y)=(p'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H20.
Cut ([v:name](nu [z:name](x1 y z v)))=(q'0 y); [Intro | Change ([_:name]([v:name](nu [z:name](x1 _ z v))) y)=(q'0 y); Apply ho_proc_sub_congr with x0; Auto].
Rewrite <- H24.
Apply bRES with (cons x0 (cons y l)); Intros.
Inversion_clear H28.
Inversion_clear H30.
Change (btrans ([_:name](x2 _ y0) y) (a'0 y) ([_:name](x1 _ y0) y)).
Apply H7 with y0 x0; Auto.
Cut (notin y0 (nu [_:name](nu (x2 _)))); [Intro | Apply proc_mono with y; Assumption].
Rewrite H23; Inversion_clear H30; Auto.
Cut (notin y0 (nu [y:name](nu [z:name](nu (x1 y z))))); [Intro | Apply proc_mono with y; Assumption].
Rewrite H22; Inversion_clear H30; Auto.
Rewrite H13; Cut (b_act_notin_ho y0 a'0); [Intro | Apply b_act_mono with y; Assumption].
Unfold f_act_notin_ho in H30; Auto.
Apply notin_nu; Intros.
Inversion_clear H19.
Cut (notin x0 (nu (x2 z))); [Intro | Auto].
Inversion_clear H19; Auto.
Apply notin_nu; Intros.
Inversion_clear H21.
Cut (notin x0 (nu [z0:name](nu (x1 z z0)))); [Intro | Auto].
Inversion_clear H21.
Cut (notin x0 (nu (x1 z y0))); [Intro | Auto]; Assumption.
Rewrite H23; Trivial.
Rewrite H22; Trivial.
Apply H6; Auto.
Cut (notin y0 (nu [_:name](nu (x2 _)))); [Intro | Apply proc_mono with y; Auto].
Inversion_clear H30.
Cut (notin y0 (nu (x2 x0))); [Intro | Auto].
Rewrite H23; Assumption.
Cut (notin y0 (nu [y:name](nu [z:name](nu (x1 y z))))); [Intro | Apply proc_mono with y; Assumption].
Rewrite H22; Inversion_clear H30; Auto.
Rewrite H13; Cut (b_act_notin_ho y0 a'0); [Intro | Apply b_act_mono with y; Assumption].
Unfold b_act_notin_ho in H30; Auto.
Cut (notin  y (p'0 x0)); [Intro | Inversion_clear H15; Auto].
Cut (notin y (nu [_:name](nu (x2 _)))); [Intro | Apply proc_mono with x0; Rewrite <- H11 in H30; Auto].
Apply notin_nu; Intros.
Inversion_clear H32.
Cut (notin y (nu (x2 z))); [Intro | Auto].
Inversion_clear H32; Auto.
Cut (notin  y (nu (q'0 x0))); [Intro | Inversion_clear H16; Auto].
Rewrite <- H12 in H30; Inversion_clear H30.
Cut (notin y (nu [z0:name](x1 x0 z0 y0))); [Intro | Auto].
Inversion_clear H30.
Apply proc_mono with x0; Apply proc_mono with y0; Auto.
Do 3 (Apply notin_nu; Intros).
Inversion_clear H21.
Cut (notin x0 (nu [z0:name](nu (x1 z z0)))); [Intro | Auto].
Inversion_clear H21.
Cut (notin x0 (nu (x1 z z1))); [Intro | Auto].
Inversion_clear H21; Auto.

(* OPEN Case *)

Elim (LEM_name x1 y); Intros.

(* Subcase x0=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=y *)

Elim (ho_proc_exp p1 x1); Elim (ho_proc_exp p2 x1); Intros.
Inversion_clear H19; Inversion_clear H20.
Rewrite H23 in H11; Rewrite H22 in H12.
Cut (nu (x3 y))=(p'0 y); [Intro | Change ([_:name](nu (x3 _)) y)=(p'0 y); Apply proc_sub_congr with x1; Auto].
Rewrite <- H20.
Cut (x2 y)=(q'0 y); [Intro | Change ([_:name](x2 _) y)=(q'0 y); Apply ho_proc_sub_congr with x1; Auto].
Rewrite <- H24.
Elim (LEM_name x0 x1); Intros.
Rewrite H25 in H13.
Cut (bOut y)=(a'0 y); [Intro | Apply b_act_sub_congr with x1; Auto].
Rewrite <- H26.
Apply OPEN with (cons x1 (cons y l)); Intros.
Inversion_clear H30.
Inversion_clear H32.
Change (ftrans ([_:name](x3 _ y0) y) ([_:name](Out _ y0) y) ([_:name](x2 _ y0) y)); Apply H7 with y0 x1; Auto.
Cut (notin y0 (nu [_:name](nu (x3 _)))); [Intro | Apply proc_mono with y; Assumption].
Rewrite H23; Inversion_clear H32; Auto.
Cut (notin y0 (nu [_:name](nu (x2 _)))); [Intro | Apply proc_mono with y; Assumption].
Rewrite H22; Inversion_clear H32; Auto.
Rewrite H25; Auto.
Apply notin_nu; Intros; Inversion_clear H19.
Cut (notin x1 (nu (x3 z))); [Intro | Auto].
Inversion_clear H19; Auto.
Apply notin_nu; Intros; Inversion_clear H21.
Cut (notin x1 (nu (x2 z))); [Intro | Auto].
Inversion_clear H21; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite H23; Trivial.
Rewrite H22; Trivial.
Rewrite H25; Trivial.
Apply H6; Auto.
Cut (notin y0 (nu [_:name](nu (x3 _)))); [Intro | Apply proc_mono with y; Assumption].
Rewrite H23; Inversion_clear H32; Auto.
Cut (notin y0 (nu [_:name](nu (x2 _)))); [Intro | Apply proc_mono with y; Assumption].
Rewrite H22; Inversion_clear H32; Auto.
Rewrite H25; Auto.
Inversion_clear H15.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H11 in H15.
Apply proc_mono with x1; Inversion_clear H15; Auto.
Inversion_clear H16.
Cut (notin y (nu (q'0 x1))); [Intro | Auto].
Rewrite <- H12 in H16.
Apply proc_mono with x1; Inversion_clear H16; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.

Cut (bOut x0)=(a'0 y); [Intro | Change ([_:name](bOut x0) y)=(a'0 y); Apply b_act_sub_congr with x1; Auto].
Rewrite <- H26.
Apply OPEN with (cons x1 (cons y l)); Intros.
Inversion_clear H30.
Inversion_clear H32.
Change (ftrans ([_:name](x3 _ y0) y) ([_:name](Out x0 y0) y) ([_:name](x2 _ y0) y)); Apply H7 with y0 x1; Auto.
Cut (notin y0 (nu [_:name](nu (x3 _)))); [Intro | Apply proc_mono with y; Assumption].
Rewrite H23; Inversion_clear H32; Auto.
Cut (notin y0 (nu [_:name](nu (x2 _)))); [Intro | Apply proc_mono with y; Assumption].
Rewrite H22; Inversion_clear H32; Auto.
Apply notin_nu; Intros; Inversion_clear H19.
Cut (notin x1 (nu (x3 z))); [Intro | Auto].
Inversion_clear H19; Auto.
Apply notin_nu; Intros; Inversion_clear H21.
Cut (notin x1 (nu (x2 z))); [Intro | Auto].
Inversion_clear H21; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite H23; Trivial.
Rewrite H22; Trivial.
Apply H6; Auto.
Cut (notin y0 (nu [_:name](nu (x3 _)))); [Intro | Apply proc_mono with y; Assumption].
Rewrite H23; Inversion_clear H32; Auto.
Cut (notin y0 (nu [_:name](nu (x2 _)))); [Intro | Apply proc_mono with y; Assumption].
Rewrite H22; Inversion_clear H32; Auto.
Inversion_clear H15.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H11 in H15.
Apply proc_mono with x1; Inversion_clear H15; Auto.
Inversion_clear H16.
Cut (notin y (nu (q'0 x1))); [Intro | Auto].
Rewrite <- H12 in H16.
Apply proc_mono with x1; Inversion_clear H16; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Cut ~x0=y; [Intro; Auto | Apply Sep_proc with (nu p1)].
Apply Aux2 with (bOut x0) p2; [Apply b_act_isin_bOut | Auto].
Inversion_clear H15.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H11 in H15; Rewrite <- H23 in H15; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.

Qed.

Lemma L3bis: (p:proc)(a:b_act)(q:name->proc)
              (p':name->proc)(a':name->b_act)(q':name->name->proc)(x:name)
                (notin x (nu p'))->(notin x (nu [_:name](nu (q' _))))->(b_act_notin_ho x a') -> p=(p' x) -> q=(q' x) -> a=(a' x)
                  ->(btrans p a q)
                    ->(y:name)(notin y (nu p'))->(notin y (nu [_:name](nu (q' _))))->(b_act_notin_ho y a')
                      ->(btrans (p' y) (a' y) (q' y)).

Proof.

Do 14 Intro.
Generalize H5.
Generalize H4.
Generalize H3.
Generalize H2.
Generalize H1.
Generalize H0.
Generalize H.
Generalize x.
Generalize a'.
Generalize q'.
Generalize p'.
(Pattern p a q ; Apply btrans_ftrans_ind with
[p:proc][a:f_act][q:proc]
  ((p',q':name->proc)(a':name->f_act)(x:name)
                (notin x (nu p'))->(notin x (nu q'))->(f_act_notin_ho x a') -> p=(p' x) -> q=(q' x) -> a=(a' x)
                  ->(ftrans p a q)
                    ->(y:name)(notin y (nu p'))->(notin y (nu q'))->(f_act_notin_ho y a')
                      ->(ftrans (p' y) (a' y) (q' y)));
Intros; Auto).

(* TAU Case *)

Rewrite H10 in H9.
Cut (tau_pref (q'0 y))=(p'0 y); [Intro | Change ([_:name](tau_pref (q'0 _)) y)=(p'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H16.
Cut tau=(a'0 y); [Intro | Change ([_:name]tau y)=(a'0 y); Apply f_act_sub_congr with x0; Auto].
Rewrite <- H17.
Apply TAU.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply notin_nu; Intros; Apply notin_tau; Inversion H7; Auto.

(* OUT Case *)

Elim (LEM_name x0 x1); Elim (LEM_name x1 y); Intros.

(* Subcase x0=x1=y *)

Rewrite <- H16 in H9; Rewrite H17 in H9.
Rewrite <- H16 in H11; Rewrite H17 in H11.
Rewrite H10 in H9.
Cut (out_pref y0 y0 (q'0 y0))=(p'0 y0); [Intro | Change ([_:name](out_pref _ _ (q'0 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Rewrite <- H18.
Cut (Out y0 y0)=(a'0 y0); [Intro | Change ([_:name](Out _ _) y0)=(a'0 y0); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H19.
Apply OUT.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Apply notin_nu; Intros; Apply notin_out; Inversion H7; Auto.

(* Subcase x0=x1 /\ ~x1=y *)

Rewrite H17 in H9.
Rewrite H17 in H11.
Rewrite H10 in H9.
Cut (out_pref y0 y (q'0 y0))=(p'0 y0); [Intro | Change ([_:name](out_pref _ y (q'0 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Rewrite <- H18.
Cut (Out y0 y)=(a'0 y0); [Intro | Change ([_:name](Out _ y) y0)=(a'0 y0); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H19.
Apply OUT.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Apply notin_nu; Intros; Apply notin_out; Inversion H7; Auto.

(* Subcase ~x0=x1 /\ x1=y *)

Rewrite <- H16 in H9.
Rewrite <- H16 in H11.
Rewrite H10 in H9.
Cut (out_pref x0 y0 (q'0 y0))=(p'0 y0); [Intro | Change ([_:name](out_pref x0 _ (q'0 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Rewrite <- H18.
Cut (Out x0 y0)=(a'0 y0); [Intro | Change ([_:name](Out x0 _) y0)=(a'0 y0); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H19.
Apply OUT.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Apply notin_nu; Intros; Apply notin_out; Inversion H7; Auto.

(* Subcase ~x0=x1 /\ ~x1=y *)

Rewrite H10 in H9.
Cut (out_pref x0 y (q'0 y0))=(p'0 y0); [Intro | Change ([_:name](out_pref x0 y (q'0 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Rewrite <- H18.
Cut (Out x0 y)=(a'0 y0); [Intro | Change ([_:name](Out x0 y) y0)=(a'0 y0); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H19.
Apply OUT.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Apply notin_nu; Intros; Apply notin_out; Inversion H7; Auto.

(* fSUM1 Case *)

Elim (LEM_name x0 y); Intros.

(* Subcase x0=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=y *)

Elim (proc_exp p1 x0); Elim (proc_exp p2 x0); Intros.
Inversion_clear H19; Inversion_clear H20.
Rewrite H23 in H11; Rewrite H22 in H11.
Cut (sum (x2 y) (x1 y))=(p'0 y); [Intro | Change ([_:name](sum (x2 _) (x1 _)) y)=(p'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H20.
Apply fSUM1.
Apply H7 with x0; Auto.
Apply proc_mono with x0.
Inversion_clear H15.
Cut (notin y (p'0 x0)); [Intro | Auto].
Rewrite <- H11 in H15.
Inversion_clear H15; Assumption.
Apply notin_nu; Intros; Apply notin_sum; [Inversion_clear H19; Auto | Inversion_clear H21; Auto].

(* fSUM2 Case *)

Elim (LEM_name x0 y); Intros.

(* Subcase x0=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=y *)

Elim (proc_exp p1 x0); Elim (proc_exp p2 x0); Intros.
Inversion_clear H19; Inversion_clear H20.
Rewrite H23 in H11; Rewrite H22 in H11.
Cut (sum (x2 y) (x1 y))=(p'0 y); [Intro | Change ([_:name](sum (x2 _) (x1 _)) y)=(p'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H20.
Apply fSUM2.
Apply H7 with x0; Auto.
Apply proc_mono with x0.
Inversion_clear H15.
Cut (notin y (p'0 x0)); [Intro | Auto].
Rewrite <- H11 in H15.
Inversion_clear H15; Assumption.
Apply notin_nu; Intros; Apply notin_sum; [Inversion_clear H19; Auto | Inversion_clear H21; Auto].

(* fPAR1 Case *)

Elim (LEM_name x0 y); Intros.

(* Subcase x0=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=y *)

Elim (proc_exp p1 x0); Elim (proc_exp p2 x0); Elim (proc_exp p0 x0); Intros.
Inversion_clear H19; Inversion_clear H20; Inversion_clear H21.
Rewrite H23 in H12; Rewrite H24 in H11; Rewrite H24 in H12; Rewrite H25 in H11.
Cut (par (x1 y) (x2 y))=(q'0 y); [Intro | Change ([_:name](par (x1 _) (x2 _)) y)=(q'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H21.
Cut (par (x3 y) (x2 y))=(p'0 y); [Intro | Change ([_:name](par (x3 _) (x2 _)) y)=(p'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H26.
Apply fPAR1.
Apply H7 with x0; Auto.
Apply proc_mono with x0.
Inversion_clear H15.
Cut (notin y (p'0 x0)); [Intro | Auto].
Rewrite <- H11 in H15.
Inversion_clear H15; Assumption.
Apply proc_mono with x0.
Inversion_clear H16.
Cut (notin y (q'0 x0)); [Intro | Auto].
Rewrite <- H12 in H16.
Inversion_clear H16; Assumption.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H20; Auto | Inversion_clear H19; Auto].
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H22; Auto | Inversion_clear H19; Auto].

(* fPAR2 Case *)

Elim (LEM_name x0 y); Intros.

(* Subcase x0=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=y *)

Elim (proc_exp p1 x0); Elim (proc_exp p2 x0); Elim (proc_exp p0 x0); Intros.
Inversion_clear H19; Inversion_clear H20; Inversion_clear H21.
Rewrite H23 in H12; Rewrite H24 in H11; Rewrite H25 in H11; Rewrite H25 in H12.
Cut (par (x3 y) (x1 y))=(q'0 y); [Intro | Change ([_:name](par (x3 _) (x1 _)) y)=(q'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H21.
Cut (par (x3 y) (x2 y))=(p'0 y); [Intro | Change ([_:name](par (x3 _) (x2 _)) y)=(p'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H26.
Apply fPAR2.
Apply H7 with x0; Auto.
Apply proc_mono with x0.
Inversion_clear H15.
Cut (notin y (p'0 x0)); [Intro | Auto].
Rewrite <- H11 in H15.
Inversion_clear H15; Assumption.
Apply proc_mono with x0.
Inversion_clear H16.
Cut (notin y (q'0 x0)); [Intro | Auto].
Rewrite <- H12 in H16.
Inversion_clear H16; Assumption.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H20; Auto | Inversion_clear H19; Auto].
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H20; Auto | Inversion_clear H22; Auto].

(* fMATCH Case *)

Elim (LEM_name x0 x1); Elim (LEM_name x1 y); Intros.

(* Subcase x0=x1=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase x0=x1 /\ ~x1=y *)

Elim (proc_exp p0 x1); Intros.
Inversion_clear H20.
Rewrite H22 in H11.
Rewrite H19 in H11.
Cut (match y y (x2 y))=(p'0 y); [Intro | Change ([_:name](match _ _ (x2 _)) y)=(p'0 y); Apply proc_sub_congr with x1; Auto].
Rewrite <- H20.
Apply fMATCH.
Apply H7 with x1; Auto.
Apply proc_mono with x1.
Inversion_clear H15.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H11 in H15.
Inversion_clear H15; Assumption.
Apply notin_nu; Intros; Apply notin_match; Inversion H21; Auto.

(* Subcase ~x0=x1 /\ x1=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=x1 /\ ~x1=y *)

Elim (proc_exp p0 x1); Intros.
Inversion_clear H20.
Rewrite H22 in H11.
Cut (match x0 x0 (x2 y))=(p'0 y); [Intro | Change ([_:name](match x0 x0 (x2 _)) y)=(p'0 y); Apply proc_sub_congr with x1; Auto].
Rewrite <- H20.
Apply fMATCH.
Apply H7 with x1; Auto.
Apply proc_mono with x1.
Inversion_clear H15.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H11 in H15.
Inversion_clear H15; Assumption.
Apply notin_nu; Intros; Apply notin_match; Inversion H21; Auto.

(* fMISMATCH Case *)

Elim (LEM_name x1 y0); Intros.

(* Subcase x1=y0 *)

Rewrite <- H19; Rewrite <- H12; Rewrite <- H13; Rewrite <- H14; Assumption.

(* Subcase ~x1=y0 *)

Elim (LEM_name x0 x1); Elim (LEM_name y x1); Intros.

(* Subsubcase x0=x1=y *)

Absurd x0=y; [Assumption | Rewrite H20; Rewrite H21; Trivial].

(* Subsubcase x0=x1 /\ ~x1=y *)

Rewrite H21 in H12.
Elim (proc_exp p0 x1); Intros.
Inversion_clear H22.
Rewrite H24 in H12.
Cut (mismatch y0 y (x2 y0))=(p'0 y0); [Intro | Change ([_:name](mismatch _ y (x2 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Rewrite <- H22.
Apply fMISMATCH.
Cut ~y=y0; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H12; Apply isin_mismatch3.
Inversion_clear H16; Auto.
Apply H8 with x1; Auto.
Apply proc_mono with x1.
Inversion_clear H16.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H12 in H16.
Inversion_clear H16; Assumption.
Apply notin_nu; Intros; Apply notin_mismatch; Inversion H23; Auto.

(* Subsubcase ~x0=x1 /\ x1=y *)

Rewrite H20 in H12.
Elim (proc_exp p0 x1); Intros.
Inversion_clear H22.
Rewrite H24 in H12.
Cut (mismatch x0 y0 (x2 y0))=(p'0 y0); [Intro | Change ([_:name](mismatch x0 _ (x2 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Rewrite <- H22.
Apply fMISMATCH.
Apply Sep_proc with (p'0 x1).
Rewrite <- H12; Apply isin_mismatch2.
Inversion_clear H16; Auto.
Apply H8 with x1; Auto.
Apply proc_mono with x1.
Inversion_clear H16.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H12 in H16.
Inversion_clear H16; Assumption.
Apply notin_nu; Intros; Apply notin_mismatch; Inversion H23; Auto.

(* Subsubcase ~x0=x1 /\ ~x1=y *)

Elim (proc_exp p0 x1); Intros.
Inversion_clear H22.
Rewrite H24 in H12.
Cut (mismatch x0 y (x2 y0))=(p'0 y0); [Intro | Change ([_:name](mismatch x0 y (x2 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Rewrite <- H22.
Apply fMISMATCH.
Inversion H15; Assumption.
Apply H8 with x1; Auto.
Apply proc_mono with x1.
Inversion_clear H16.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H12 in H16.
Inversion_clear H16; Assumption.
Apply notin_nu; Intros; Apply notin_mismatch; Inversion H23; Auto.

(* fBANG Case *)

Elim (LEM_name x0 y); Intros.

(* Subcase x0=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=y *)

Elim (proc_exp p0 x0); Intros.
Inversion_clear H19.
Rewrite H21 in H11.
Cut (bang (x1 y))=(p'0 y); [Intro | Change ([_:name](bang (x1 _)) y)=(p'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H19; Apply fBANG.
Change (ftrans ([_:name](par (x1 _) (bang (x1 _))) y) (a'0 y) (q'0 y)); Apply H7 with x0; Auto.
Apply notin_nu; Intros; Inversion_clear H20; Apply notin_par; [Auto | Apply notin_bang; Auto].
Rewrite H21; Trivial.
Apply proc_mono with x0.
Inversion_clear H15.
Cut (notin y (p'0 x0)); [Intro | Auto].
Rewrite <- H11 in H15; Inversion H15; Apply notin_par; Assumption.
Apply notin_nu; Intros; Apply notin_bang; Inversion_clear H20; Auto.

(* COM1 Case *)

Elim (LEM_name x1 y0); Intros.

(* Subcase x1=y0 *)

Rewrite <- H20; Rewrite <- H13; Rewrite <- H14; Rewrite <- H15; Assumption.

(* Subcase ~x1=y0 *)

Elim (proc_exp p1 x1); Elim (proc_exp p2 x1); Elim (ho_proc_exp q1 x1); Elim (proc_exp q2 x1); Intros.
Inversion_clear H21; Inversion_clear H22; Inversion_clear H23; Inversion_clear H24.
Rewrite H26 in H14; Rewrite H27 in H14; Rewrite H28 in H13; Rewrite H29 in H13.
Elim (LEM_name x0 x1); Elim (LEM_name x1 y); Intros.

Rewrite <- H24 in H14.
Cut (par (x3 y0 y0) (x2 y0))=(q'0 y0); [Intro | Change ([_:name](par (x3 _ _) (x2 _)) y0)=(q'0 y0); Apply proc_sub_congr with x1; Auto].
Cut (par (x5 y0) (x4 y0))=(p'0 y0); [Intro | Change ([_:name](par (x5 _) (x4 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Cut tau=(a'0 y0); [Intro | Change ([_:name]tau y0)=(a'0 y0); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H31; Rewrite <- H32; Rewrite <- H33.
Apply COM1 with y0.
Change (btrans (x5 y0) ([_:name](In _) y0) (x3 y0)).
Apply H7 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Rewrite H30; Trivial.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Apply proc_mono with x1.
Inversion_clear H18; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Change (ftrans (x4 y0) ([_:name](Out _ _) y0) (x2 y0)).
Apply H9 with x1; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite <- H24; Rewrite H30; Trivial.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Inversion_clear H18; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H23; Auto | Inversion_clear H22; Auto].
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H21; Cut (notin x1 (nu (x3 z))); [Intro | Auto]; Inversion_clear H21; Auto | Inversion_clear H25; Auto].

Cut (par (x3 y0 y) (x2 y0))=(q'0 y0); [Intro | Change ([_:name](par (x3 _ y) (x2 _)) y0)=(q'0 y0); Apply proc_sub_congr with x1; Auto].
Cut (par (x5 y0) (x4 y0))=(p'0 y0); [Intro | Change ([_:name](par (x5 _) (x4 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Cut tau=(a'0 y0); [Intro | Change ([_:name]tau y0)=(a'0 y0); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H31; Rewrite <- H32; Rewrite <- H33.
Apply COM1 with y0.
Change (btrans (x5 y0) ([_:name](In _) y0) (x3 y0)).
Apply H7 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Rewrite H30; Trivial.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Apply proc_mono with y.
Inversion_clear H18; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Change (ftrans (x4 y0) ([_:name](Out _ y) y0) (x2 y0)).
Apply H9 with x1; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite H30; Trivial.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Inversion_clear H18; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Cut ~y=y0; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par2.
Rewrite <- H28.
Apply Aux1 with q2 (Out x0 y); [Apply f_act_isin_Out2 | Auto].
Inversion_clear H17; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H23; Auto | Inversion_clear H22; Auto].
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H21; Cut (notin x1 (nu (x3 z))); [Intro | Auto]; Inversion_clear H21; Auto | Inversion_clear H25; Auto].

Rewrite <- H24 in H14.
Cut (par (x3 y0 y0) (x2 y0))=(q'0 y0); [Intro | Change ([_:name](par (x3 _ _) (x2 _)) y0)=(q'0 y0); Apply proc_sub_congr with x1; Auto].
Cut (par (x5 y0) (x4 y0))=(p'0 y0); [Intro | Change ([_:name](par (x5 _) (x4 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Cut tau=(a'0 y0); [Intro | Change ([_:name]tau y0)=(a'0 y0); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H31; Rewrite <- H32; Rewrite <- H33.
Apply COM1 with x0.
Change (btrans (x5 y0) ([_:name](In x0) y0) (x3 y0)).
Apply H7 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Apply proc_mono with x1.
Inversion_clear H17; Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17; Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Apply proc_mono with x1.
Inversion_clear H18; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Cut ~x0=y0; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par2.
Rewrite <- H28.
Apply Aux1 with q2 (Out x0 y); [Apply f_act_isin_Out1 | Auto].
Inversion_clear H17; Auto.
Change (ftrans (x4 y0) ([_:name](Out x0 _) y0) (x2 y0)).
Apply H9 with x1; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite H24; Trivial.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Inversion_clear H18; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Cut ~x0=y0; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par2.
Rewrite <- H28.
Apply Aux1 with q2 (Out x0 y); [Apply f_act_isin_Out1 | Auto].
Inversion_clear H17; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H23; Auto | Inversion_clear H22; Auto].
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H21; Cut (notin x1 (nu (x3 z))); [Intro | Auto]; Inversion_clear H21; Auto | Inversion_clear H25; Auto].

Cut (par (x3 y0 y) (x2 y0))=(q'0 y0); [Intro | Change ([_:name](par (x3 _ y) (x2 _)) y0)=(q'0 y0); Apply proc_sub_congr with x1; Auto].
Cut (par (x5 y0) (x4 y0))=(p'0 y0); [Intro | Change ([_:name](par (x5 _) (x4 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Cut tau=(a'0 y0); [Intro | Change ([_:name]tau y0)=(a'0 y0); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H31; Rewrite <- H32; Rewrite <- H33.
Apply COM1 with x0.
Change (btrans (x5 y0) ([_:name](In x0) y0) (x3 y0)).
Apply H7 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Apply proc_mono with x1.
Inversion_clear H17; Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17; Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Apply proc_mono with y.
Inversion_clear H18; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Cut ~x0=y0; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par2.
Rewrite <- H28.
Apply Aux1 with q2 (Out x0 y); [Apply f_act_isin_Out1 | Auto].
Inversion_clear H17; Auto.
Change (ftrans (x4 y0) ([_:name](Out x0 y) y0) (x2 y0)).
Apply H9 with x1; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Inversion_clear H18; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Cut ~x0=y0; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par2.
Rewrite <- H28.
Apply Aux1 with q2 (Out x0 y); [Apply f_act_isin_Out1 | Auto].
Inversion_clear H17; Auto.
Cut ~y=y0; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par2.
Rewrite <- H28.
Apply Aux1 with q2 (Out x0 y); [Apply f_act_isin_Out2 | Auto].
Inversion_clear H17; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H23; Auto | Inversion_clear H22; Auto].
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H21; Cut (notin x1 (nu (x3 z))); [Intro | Auto]; Inversion_clear H21; Auto | Inversion_clear H25; Auto].

(* COM2 Case *)

Elim (LEM_name x1 y0); Intros.

(* Subcase x1=y0 *)

Rewrite <- H20; Rewrite <- H13; Rewrite <- H14; Rewrite <- H15; Assumption.

(* Subcase ~x1=y0 *)

Elim (proc_exp p1 x1); Elim (proc_exp p2 x1); Elim (proc_exp q1 x1); Elim (ho_proc_exp q2 x1); Intros.
Inversion_clear H21; Inversion_clear H22; Inversion_clear H23; Inversion_clear H24.
Rewrite H26 in H14; Rewrite H27 in H14; Rewrite H28 in H13; Rewrite H29 in H13.
Elim (LEM_name x0 x1); Elim (LEM_name x1 y); Intros.

Rewrite <- H24 in H14.
Cut (par (x3 y0) (x2 y0 y0))=(q'0 y0); [Intro | Change ([_:name](par (x3 _) (x2 _ _)) y0)=(q'0 y0); Apply proc_sub_congr with x1; Auto].
Cut (par (x5 y0) (x4 y0))=(p'0 y0); [Intro | Change ([_:name](par (x5 _) (x4 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Cut tau=(a'0 y0); [Intro | Change ([_:name]tau y0)=(a'0 y0); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H31; Rewrite <- H32; Rewrite <- H33.
Apply COM2 with y0.
Change (ftrans (x5 y0) ([_:name](Out _ _) y0) (x3 y0)).
Apply H7 with x1; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite <- H24; Rewrite H30; Trivial.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Inversion_clear H18; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Change (btrans (x4 y0) ([_:name](In _) y0) (x2 y0)).
Apply H9 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Rewrite H30; Trivial.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Apply proc_mono with x1.
Inversion_clear H18; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H23; Auto | Inversion_clear H22; Auto].
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H21; Auto | Inversion_clear H25; Cut (notin x1 (nu (x2 z))); [Intro | Auto]; Inversion_clear H25; Auto].

Cut (par (x3 y0) (x2 y0 y))=(q'0 y0); [Intro | Change ([_:name](par (x3 _) (x2 _ y)) y0)=(q'0 y0); Apply proc_sub_congr with x1; Auto].
Cut (par (x5 y0) (x4 y0))=(p'0 y0); [Intro | Change ([_:name](par (x5 _) (x4 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Cut tau=(a'0 y0); [Intro | Change ([_:name]tau y0)=(a'0 y0); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H31; Rewrite <- H32; Rewrite <- H33.
Apply COM2 with y0.

Change (ftrans (x5 y0) ([_:name](Out _ y) y0) (x3 y0)).
Apply H7 with x1; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite H30; Trivial.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Inversion_clear H18; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Cut ~y=y0; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par1.
Rewrite <- H29.
Apply Aux1 with q1 (Out x0 y); [Apply f_act_isin_Out2 | Auto].
Inversion_clear H17; Auto.
Change (btrans (x4 y0) ([_:name](In _) y0) (x2 y0)).
Apply H9 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Rewrite H30; Trivial.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Apply proc_mono with y.
Inversion_clear H18; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H23; Auto | Inversion_clear H22; Auto].
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H21; Auto | Inversion_clear H25; Cut (notin x1 (nu (x2 z))); [Intro | Auto]; Inversion_clear H25; Auto].

Rewrite <- H24 in H14.
Cut (par (x3 y0) (x2 y0 y0))=(q'0 y0); [Intro | Change ([_:name](par (x3 _) (x2 _ _)) y0)=(q'0 y0); Apply proc_sub_congr with x1; Auto].
Cut (par (x5 y0) (x4 y0))=(p'0 y0); [Intro | Change ([_:name](par (x5 _) (x4 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Cut tau=(a'0 y0); [Intro | Change ([_:name]tau y0)=(a'0 y0); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H31; Rewrite <- H32; Rewrite <- H33.
Apply COM2 with x0.
Change (ftrans (x5 y0) ([_:name](Out x0 _) y0) (x3 y0)).
Apply H7 with x1; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite H24; Trivial.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Inversion_clear H18; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Cut ~x0=y0; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par1.
Rewrite <- H29.
Apply Aux1 with q1 (Out x0 y); [Apply f_act_isin_Out1 | Auto].
Inversion_clear H17; Auto.
Change (btrans (x4 y0) ([_:name](In x0) y0) (x2 y0)).
Apply H9 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Apply proc_mono with x1.
Inversion_clear H17; Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17; Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Apply proc_mono with x1.
Inversion_clear H18; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Cut ~x0=y0; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par1.
Rewrite <- H29.
Apply Aux1 with q1 (Out x0 y); [Apply f_act_isin_Out1 | Auto].
Inversion_clear H17; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H23; Auto | Inversion_clear H22; Auto].
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H21; Auto | Inversion_clear H25; Cut (notin x1 (nu (x2 z))); [Intro | Auto]; Inversion_clear H25; Auto].

Cut (par (x3 y0) (x2 y0 y))=(q'0 y0); [Intro | Change ([_:name](par (x3 _) (x2 _ y)) y0)=(q'0 y0); Apply proc_sub_congr with x1; Auto].
Cut (par (x5 y0) (x4 y0))=(p'0 y0); [Intro | Change ([_:name](par (x5 _) (x4 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Cut tau=(a'0 y0); [Intro | Change ([_:name]tau y0)=(a'0 y0); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H31; Rewrite <- H32; Rewrite <- H33.
Apply COM2 with x0.
Change (ftrans (x5 y0) ([_:name](Out x0 y) y0) (x3 y0)).
Apply H7 with x1; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Inversion_clear H18; Assumption.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Cut ~x0=y0; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par1.
Rewrite <- H29.
Apply Aux1 with q1 (Out x0 y); [Apply f_act_isin_Out1 | Auto].
Inversion_clear H17; Auto.
Cut ~y=y0; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par1.
Rewrite <- H29.
Apply Aux1 with q1 (Out x0 y); [Apply f_act_isin_Out2 | Auto].
Inversion_clear H17; Auto.
Change (btrans (x4 y0) ([_:name](In x0) y0) (x2 y0)).
Apply H9 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Apply proc_mono with x1.
Inversion_clear H17; Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17; Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y0 (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Apply proc_mono with y.
Inversion_clear H18; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Cut ~x0=y0; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par1.
Rewrite <- H29.
Apply Aux1 with q1 (Out x0 y); [Apply f_act_isin_Out1 | Auto].
Inversion_clear H17; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H23; Auto | Inversion_clear H22; Auto].
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H21; Auto | Inversion_clear H25; Cut (notin x1 (nu (x2 z))); [Intro | Auto]; Inversion_clear H25; Auto].

(* fRES Case *)

Elim (LEM_name x0 y); Intros.

(* Subcase x0=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=y *)

Elim (ho_proc_exp p1 x0); Elim (ho_proc_exp p2 x0); Intros.
Inversion_clear H19; Inversion_clear H20.
Rewrite H22 in H12; Rewrite H23 in H11.
Cut (nu (x2 y))=(p'0 y); [Intro | Change ([_:name](nu (x2 _)) y)=(p'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H20.
Cut (nu (x1 y))=(q'0 y); [Intro | Change ([_:name](nu (x1 _)) y)=(q'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H24.
Apply fRES with (cons x0 (cons y l)); Intros.
Inversion_clear H27.
Inversion_clear H30.
Change (ftrans ([_:name](x2 _ y0) y) (a'0 y) ([_:name](x1 _ y0) y)).
Apply H7 with y0 x0; Auto.
Cut (notin y0 (nu [_:name](nu (x2 _)))); [Intro | Apply proc_mono with y; Assumption].
Rewrite H23; Inversion_clear H30; Auto.
Cut (notin y0 (nu [_:name](nu (x1 _)))); [Intro | Apply proc_mono with y; Assumption].
Rewrite H22; Inversion_clear H30; Auto.
Rewrite H13; Cut (f_act_notin_ho y0 a'0); [Intro | Apply f_act_mono with y; Assumption].
Unfold f_act_notin_ho in H30; Auto.
Apply notin_nu; Intros.
Inversion_clear H19.
Cut (notin x0 (nu (x2 z))); [Intro | Auto].
Inversion_clear H19; Auto.
Apply notin_nu; Intros.
Inversion_clear H21.
Cut (notin x0 (nu (x1 z))); [Intro | Auto].
Inversion_clear H21; Auto.
Rewrite H23; Trivial.
Rewrite H22; Trivial.
Apply H6; Auto.
Cut (notin y0 (nu [_:name](nu (x2 _)))); [Intro | Apply proc_mono with y; Auto].
Inversion_clear H30.
Cut (notin y0 (nu (x2 x0))); [Intro | Auto].
Rewrite H23; Assumption.
Cut (notin y0 (nu [_:name](nu (x1 _)))); [Intro | Apply proc_mono with y; Auto].
Inversion_clear H30.
Cut (notin y0 (nu (x1 x0))); [Intro | Auto].
Rewrite H22; Assumption.
Rewrite H13; Cut (f_act_notin_ho y0 a'0); [Intro | Apply f_act_mono with y; Assumption].
Unfold f_act_notin_ho in H30; Auto.
Cut (notin  y (p'0 x0)); [Intro | Inversion_clear H15; Auto].
Cut (notin y (nu [_:name](nu (x2 _)))); [Intro | Apply proc_mono with x0; Rewrite <- H11 in H30; Auto].
Apply notin_nu; Intros.
Inversion_clear H32.
Cut (notin y (nu (x2 z))); [Intro | Auto].
Inversion_clear H32; Auto.
Cut (notin  y (q'0 x0)); [Intro | Inversion_clear H16; Auto].
Cut (notin y (nu [_:name](nu (x1 _)))); [Intro | Apply proc_mono with x0; Rewrite <- H12 in H30; Auto].
Apply notin_nu; Intros.
Inversion_clear H32.
Cut (notin y (nu (x1 z))); [Intro | Auto].
Inversion_clear H32; Auto.

(* CLOSE1 Case *)

Elim (LEM_name x1 y); Intros.

(* Subcase x1=y *)

Rewrite <- H20; Rewrite <- H13; Rewrite <- H14; Rewrite <- H15; Assumption.

(* Subcase ~x1=y *)

Elim (proc_exp p1 x1); Elim (proc_exp p2 x1); Elim (ho_proc_exp q1 x1); Elim (ho_proc_exp q2 x1); Intros.
Inversion_clear H21; Inversion_clear H22; Inversion_clear H23; Inversion_clear H24.
Rewrite H26 in H14; Rewrite H27 in H14; Rewrite H28 in H13; Rewrite H29 in H13.
Elim (LEM_name x0 x1); Intros.
Cut (nu [z:name](par (x3 y z) (x2 y z)))=(q'0 y); [Intro | Change ([_:name](nu [z:name](par (x3 _ z) (x2 _ z))) y)=(q'0 y); Apply proc_sub_congr with x1; Auto].
Cut (par (x5 y) (x4 y))=(p'0 y); [Intro | Change ([_:name](par (x5 _) (x4 _)) y)=(p'0 y); Apply proc_sub_congr with x1; Auto].
Cut tau=(a'0 y); [Intro | Change ([_:name]tau y)=(a'0 y); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H30; Rewrite <- H31; Rewrite <- H32.
Apply CLOSE1 with y.
Apply H7 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Rewrite H24; Trivial.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y (par (x3 x1 x1) (x2 x1 x1))); [Intro | Auto].
Inversion_clear H18; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Apply H9 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Rewrite H24; Trivial.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Inversion_clear H18.
Cut (notin y (par (x3 x1 x1) (x2 x1 x1))); [Intro | Auto].
Inversion_clear H18; Apply proc_mono with x1; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H23; Auto | Inversion_clear H22; Auto].
Apply notin_nu; Intros; Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H21; Cut (notin x1 (nu (x3 z))); [Intro | Auto]; Inversion_clear H21; Auto | Inversion_clear H25; Cut (notin x1 (nu (x2 z))); [Intro | Auto]; Inversion_clear H25; Auto].

Cut (nu [z:name](par (x3 y z) (x2 y z)))=(q'0 y); [Intro | Change ([_:name](nu [z:name](par (x3 _ z) (x2 _ z))) y)=(q'0 y); Apply proc_sub_congr with x1; Auto].
Cut (par (x5 y) (x4 y))=(p'0 y); [Intro | Change ([_:name](par (x5 _) (x4 _)) y)=(p'0 y); Apply proc_sub_congr with x1; Auto].
Cut tau=(a'0 y); [Intro | Change ([_:name]tau y)=(a'0 y); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H30; Rewrite <- H31; Rewrite <- H32.
Apply CLOSE1 with x0.
Change (btrans (x5 y) ([_:name](In x0) y) (x3 y)).
Apply H7 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y (par (x3 x1 x1) (x2 x1 x1))); [Intro | Auto].
Inversion_clear H18; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Cut ~x0=y; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par2.
Rewrite <- H28.
Apply Aux2 with (bOut x0) q2; [Apply b_act_isin_bOut | Auto].
Inversion_clear H17; Auto.
Change (btrans (x4 y) ([_:name](bOut x0) y) (x2 y)).
Apply H9 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Inversion_clear H18.
Cut (notin y (par (x3 x1 x1) (x2 x1 x1))); [Intro | Auto].
Inversion_clear H18; Apply proc_mono with x1; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Cut ~x0=y; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par2.
Rewrite <- H28.
Apply Aux2 with (bOut x0) q2; [Apply b_act_isin_bOut | Auto].
Inversion_clear H17; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H23; Auto | Inversion_clear H22; Auto].
Apply notin_nu; Intros; Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H21; Cut (notin x1 (nu (x3 z))); [Intro | Auto]; Inversion_clear H21; Auto | Inversion_clear H25; Cut (notin x1 (nu (x2 z))); [Intro | Auto]; Inversion_clear H25; Auto].

(* CLOSE2 Case *)

Elim (LEM_name x1 y); Intros.

(* Subcase x1=y *)

Rewrite <- H20; Rewrite <- H13; Rewrite <- H14; Rewrite <- H15; Assumption.

(* Subcase ~x1=y *)

Elim (proc_exp p1 x1); Elim (proc_exp p2 x1); Elim (ho_proc_exp q1 x1); Elim (ho_proc_exp q2 x1); Intros.
Inversion_clear H21; Inversion_clear H22; Inversion_clear H23; Inversion_clear H24.
Rewrite H26 in H14; Rewrite H27 in H14; Rewrite H28 in H13; Rewrite H29 in H13.
Elim (LEM_name x0 x1); Intros.
Cut (nu [z:name](par (x3 y z) (x2 y z)))=(q'0 y); [Intro | Change ([_:name](nu [z:name](par (x3 _ z) (x2 _ z))) y)=(q'0 y); Apply proc_sub_congr with x1; Auto].
Cut (par (x5 y) (x4 y))=(p'0 y); [Intro | Change ([_:name](par (x5 _) (x4 _)) y)=(p'0 y); Apply proc_sub_congr with x1; Auto].
Cut tau=(a'0 y); [Intro | Change ([_:name]tau y)=(a'0 y); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H30; Rewrite <- H31; Rewrite <- H32.
Apply CLOSE2 with y.
Apply H7 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Rewrite H24; Trivial.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Inversion_clear H18.
Cut (notin y (par (x3 x1 x1) (x2 x1 x1))); [Intro | Auto].
Inversion_clear H18; Apply proc_mono with x1; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Apply H9 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Rewrite H24; Trivial.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y (par (x3 x1 x1) (x2 x1 x1))); [Intro | Auto].
Inversion_clear H18; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H23; Auto | Inversion_clear H22; Auto].
Apply notin_nu; Intros; Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H21; Cut (notin x1 (nu (x3 z))); [Intro | Auto]; Inversion_clear H21; Auto | Inversion_clear H25; Cut (notin x1 (nu (x2 z))); [Intro | Auto]; Inversion_clear H25; Auto].

Cut (nu [z:name](par (x3 y z) (x2 y z)))=(q'0 y); [Intro | Change ([_:name](nu [z:name](par (x3 _ z) (x2 _ z))) y)=(q'0 y); Apply proc_sub_congr with x1; Auto].
Cut (par (x5 y) (x4 y))=(p'0 y); [Intro | Change ([_:name](par (x5 _) (x4 _)) y)=(p'0 y); Apply proc_sub_congr with x1; Auto].
Cut tau=(a'0 y); [Intro | Change ([_:name]tau y)=(a'0 y); Apply f_act_sub_congr with x1; Auto].
Rewrite <- H30; Rewrite <- H31; Rewrite <- H32.
Apply CLOSE2 with x0.
Change (btrans (x5 y) ([_:name](bOut x0) y) (x3 y)).
Apply H7 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y (par (x3 x1 x1) (x2 x1 x1))); [Intro | Auto].
Inversion_clear H18; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.
Cut ~x0=y; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par2.
Rewrite <- H28.
Apply Aux2 with (In x0) q2; [Apply b_act_isin_In | Auto].
Inversion_clear H17; Auto.
Change (btrans (x4 y) ([_:name](In x0) y) (x2 y)).
Apply H9 with x1; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Apply proc_mono with x1.
Inversion_clear H17.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H13 in H17.
Inversion_clear H17; Assumption.
Apply proc_mono with x1.
Inversion_clear H18.
Cut (notin y (q'0 x1)); [Intro | Auto].
Rewrite <- H14 in H18.
Inversion_clear H18.
Cut (notin y (par (x3 x1 x1) (x2 x1 x1))); [Intro | Auto].
Inversion_clear H18; Apply proc_mono with x1; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Cut ~x0=y; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H13.
Apply isin_par2.
Rewrite <- H28.
Apply Aux2 with (In x0) q2; [Apply b_act_isin_In | Auto].
Inversion_clear H17; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_tau.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H23; Auto | Inversion_clear H22; Auto].
Apply notin_nu; Intros; Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H21; Cut (notin x1 (nu (x3 z))); [Intro | Auto]; Inversion_clear H21; Auto | Inversion_clear H25; Cut (notin x1 (nu (x2 z))); [Intro | Auto]; Inversion_clear H25; Auto].

(* IN Case *)

Elim (LEM_name x1 y); Intros.

(* Subcase x1=y *)

Rewrite <- H16; Rewrite <- H9; Rewrite <- H10; Rewrite <- H11; Assumption.

(* Subcase ~x1=y *)

Elim (ho_proc_exp p0 x1); Intros.
Inversion_clear H17.
Rewrite H19 in H9; Rewrite H19 in H10.
Elim (LEM_name x0 x1); Intros.
Rewrite H17 in H9.
Rewrite H17 in H11.
Cut (in_pref y (x2 y))=(p'0 y); [Intro | Change ([_:name](in_pref _ (x2 _)) y)=(p'0 y); Apply proc_sub_congr with x1; Auto].
Cut (x2 y)=(q'0 y); [Intro | Apply ho_proc_sub_congr with x1; Auto].
Cut (In y)=(a'0 y); [Intro | Apply b_act_sub_congr with x1; Auto].
Rewrite <- H20; Rewrite <- H21; Rewrite <- H22; Apply IN.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Apply notin_nu; Intros; Apply notin_in; Intros; Auto.
Inversion_clear H18.
Cut (notin x1 (nu (x2 z))); [Intro | Auto].
Inversion_clear H18; Auto.

Cut (in_pref x0 (x2 y))=(p'0 y); [Intro | Change ([_:name](in_pref x0 (x2 _)) y)=(p'0 y); Apply proc_sub_congr with x1; Auto].
Cut (x2 y)=(q'0 y); [Intro | Apply ho_proc_sub_congr with x1; Auto].
Cut (In x0)=(a'0 y); [Intro | Change ([_:name](In x0) y)=(a'0 y); Apply b_act_sub_congr with x1; Auto].
Rewrite <- H20; Rewrite <- H21; Rewrite <- H22; Apply IN.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Auto.
Apply notin_nu; Intros; Apply notin_in; Intros; Auto.
Inversion_clear H18.
Cut (notin x1 (nu (x2 z))); [Intro | Auto].
Inversion_clear H18; Auto.

(* bSUM1 Case *)

Elim (LEM_name x0 y); Intros.

(* Subcase x0=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=y *)

Elim (proc_exp p1 x0); Elim (proc_exp p2 x0); Intros.
Inversion_clear H19; Inversion_clear H20.
Rewrite H23 in H11; Rewrite H22 in H11.
Cut (sum (x2 y) (x1 y))=(p'0 y); [Intro | Change ([_:name](sum (x2 _) (x1 _)) y)=(p'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H20.
Apply bSUM1.
Apply H7 with x0; Auto.
Apply proc_mono with x0.
Inversion_clear H15.
Cut (notin y (p'0 x0)); [Intro | Auto].
Rewrite <- H11 in H15.
Inversion_clear H15; Assumption.
Apply notin_nu; Intros; Apply notin_sum; [Inversion_clear H19; Auto | Inversion_clear H21; Auto].

(* bSUM2 Case *)

Elim (LEM_name x0 y); Intros.

(* Subcase x0=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=y *)

Elim (proc_exp p1 x0); Elim (proc_exp p2 x0); Intros.
Inversion_clear H19; Inversion_clear H20.
Rewrite H23 in H11; Rewrite H22 in H11.
Cut (sum (x2 y) (x1 y))=(p'0 y); [Intro | Change ([_:name](sum (x2 _) (x1 _)) y)=(p'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H20.
Apply bSUM2.
Apply H7 with x0; Auto.
Apply proc_mono with x0.
Inversion_clear H15.
Cut (notin y (p'0 x0)); [Intro | Auto].
Rewrite <- H11 in H15.
Inversion_clear H15; Assumption.
Apply notin_nu; Intros; Apply notin_sum; [Inversion_clear H19; Auto | Inversion_clear H21; Auto].

(* bPAR1 Case *)

Elim (LEM_name x0 y); Intros.

(* Subcase x0=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=y *)

Elim (proc_exp p1 x0); Elim (proc_exp p2 x0); Elim (ho_proc_exp p0 x0); Intros.
Inversion_clear H19; Inversion_clear H20; Inversion_clear H21.
Rewrite H23 in H12; Rewrite H24 in H11; Rewrite H24 in H12; Rewrite H25 in H11.
Cut ([x:name](par (x1 y x) (x2 y)))=(q'0 y); [Intro | Change ([_:name]([x:name](par (x1 _ x) (x2 _))) y)=(q'0 y); Apply ho_proc_sub_congr with x0; Auto].
Rewrite <- H21.
Cut (par (x3 y) (x2 y))=(p'0 y); [Intro | Change ([_:name](par (x3 _) (x2 _)) y)=(p'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H26.
Apply bPAR1.
Apply H7 with x0; Auto.
Apply proc_mono with x0.
Inversion_clear H15.
Cut (notin y (p'0 x0)); [Intro | Auto].
Rewrite <- H11 in H15.
Inversion_clear H15; Assumption.
Apply proc_mono with x0.
Inversion_clear H16.
Cut (notin y (nu (q'0 x0))); [Intro | Auto].
Rewrite <- H12 in H16.
Apply proc_mono with x0; Inversion_clear H16; Cut (notin y (par (x1 x0 x0) (x2 x0))); [Intro | Auto]; Inversion_clear H16; Auto.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H20; Auto | Inversion_clear H19; Auto].
Apply notin_nu; Intros; Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H22; Cut (notin x0 (nu (x1 z))); [Intro | Auto]; Inversion_clear H22; Auto | Inversion_clear H19; Auto].

(* bPAR2 Case *)

Elim (LEM_name x0 y); Intros.

(* Subcase x0=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=y *)

Elim (proc_exp p1 x0); Elim (proc_exp p2 x0); Elim (ho_proc_exp p0 x0); Intros.
Inversion_clear H19; Inversion_clear H20; Inversion_clear H21.
Rewrite H23 in H12; Rewrite H24 in H11; Rewrite H25 in H11; Rewrite H25 in H12.
Cut ([x:name](par (x3 y) (x1 y x)))=(q'0 y); [Intro | Change ([_:name]([x:name](par (x3 _) (x1 _ x))) y)=(q'0 y); Apply ho_proc_sub_congr with x0; Auto].
Rewrite <- H21.
Cut (par (x3 y) (x2 y))=(p'0 y); [Intro | Change ([_:name](par (x3 _) (x2 _)) y)=(p'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H26.
Apply bPAR2.
Apply H7 with x0; Auto.
Apply proc_mono with x0.
Inversion_clear H15.
Cut (notin y (p'0 x0)); [Intro | Auto].
Rewrite <- H11 in H15.
Inversion_clear H15; Assumption.
Apply proc_mono with x0.
Inversion_clear H16.
Cut (notin y (nu (q'0 x0))); [Intro | Auto].
Rewrite <- H12 in H16.
Apply proc_mono with x0; Inversion_clear H16; Cut (notin y (par (x3 x0) (x1 x0 x0))); [Intro | Auto]; Inversion_clear H16; Auto.
Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H20; Auto | Inversion_clear H19; Auto].
Apply notin_nu; Intros; Apply notin_nu; Intros; Apply notin_par; [Inversion_clear H20; Auto | Inversion_clear H22; Cut (notin x0 (nu (x1 z))); [Intro | Auto]; Inversion_clear H22; Auto].

(* bMATCH Case *)

Elim (LEM_name x0 x1); Elim (LEM_name x1 y); Intros.

(* Subcase x0=x1=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase x0=x1 /\ ~x1=y *)

Elim (proc_exp p0 x1); Intros.
Inversion_clear H20.
Rewrite H22 in H11.
Rewrite H19 in H11.
Cut (match y y (x2 y))=(p'0 y); [Intro | Change ([_:name](match _ _ (x2 _)) y)=(p'0 y); Apply proc_sub_congr with x1; Auto].
Rewrite <- H20.
Apply bMATCH.
Apply H7 with x1; Auto.
Apply proc_mono with x1.
Inversion_clear H15.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H11 in H15.
Inversion_clear H15; Assumption.
Apply notin_nu; Intros; Apply notin_match; Inversion H21; Auto.

(* Subcase ~x0=x1 /\ x1=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=x1 /\ ~x1=y *)

Elim (proc_exp p0 x1); Intros.
Inversion_clear H20.
Rewrite H22 in H11.
Cut (match x0 x0 (x2 y))=(p'0 y); [Intro | Change ([_:name](match x0 x0 (x2 _)) y)=(p'0 y); Apply proc_sub_congr with x1; Auto].
Rewrite <- H20.
Apply bMATCH.
Apply H7 with x1; Auto.
Apply proc_mono with x1.
Inversion_clear H15.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H11 in H15.
Inversion_clear H15; Assumption.
Apply notin_nu; Intros; Apply notin_match; Inversion H21; Auto.

(* bMISMATCH Case *)

Elim (LEM_name x1 y0); Intros.

(* Subcase x1=y0 *)

Rewrite <- H19; Rewrite <- H12; Rewrite <- H13; Rewrite <- H14; Assumption.

(* Subcase ~x1=y0 *)

Elim (LEM_name x0 x1); Elim (LEM_name y x1); Intros.

(* Subsubcase x0=x1=y *)

Absurd x0=y; [Assumption | Rewrite H20; Rewrite H21; Trivial].

(* Subsubcase x0=x1 /\ ~x1=y *)

Rewrite H21 in H12.
Elim (proc_exp p0 x1); Intros.
Inversion_clear H22.
Rewrite H24 in H12.
Cut (mismatch y0 y (x2 y0))=(p'0 y0); [Intro | Change ([_:name](mismatch _ y (x2 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Rewrite <- H22.
Apply bMISMATCH.
Cut ~y=y0; [Intro; Auto | Apply Sep_proc with (p'0 x1)].
Rewrite <- H12; Apply isin_mismatch3.
Inversion_clear H16; Auto.
Apply H8 with x1; Auto.
Apply proc_mono with x1.
Inversion_clear H16.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H12 in H16.
Inversion_clear H16; Assumption.
Apply notin_nu; Intros; Apply notin_mismatch; Inversion H23; Auto.

(* Subsubcase ~x0=x1 /\ x1=y *)

Rewrite H20 in H12.
Elim (proc_exp p0 x1); Intros.
Inversion_clear H22.
Rewrite H24 in H12.
Cut (mismatch x0 y0 (x2 y0))=(p'0 y0); [Intro | Change ([_:name](mismatch x0 _ (x2 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Rewrite <- H22.
Apply bMISMATCH.
Apply Sep_proc with (p'0 x1).
Rewrite <- H12; Apply isin_mismatch2.
Inversion_clear H16; Auto.
Apply H8 with x1; Auto.
Apply proc_mono with x1.
Inversion_clear H16.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H12 in H16.
Inversion_clear H16; Assumption.
Apply notin_nu; Intros; Apply notin_mismatch; Inversion H23; Auto.

(* Subsubcase ~x0=x1 /\ ~x1=y *)

Elim (proc_exp p0 x1); Intros.
Inversion_clear H22.
Rewrite H24 in H12.
Cut (mismatch x0 y (x2 y0))=(p'0 y0); [Intro | Change ([_:name](mismatch x0 y (x2 _)) y0)=(p'0 y0); Apply proc_sub_congr with x1; Auto].
Rewrite <- H22.
Apply bMISMATCH.
Inversion H15; Assumption.
Apply H8 with x1; Auto.
Apply proc_mono with x1.
Inversion_clear H16.
Cut (notin y0 (p'0 x1)); [Intro | Auto].
Rewrite <- H12 in H16.
Inversion_clear H16; Assumption.
Apply notin_nu; Intros; Apply notin_mismatch; Inversion H23; Auto.

(* bBANG Case *)

Elim (LEM_name x0 y); Intros.

(* Subcase x0=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=y *)

Elim (proc_exp p0 x0); Intros.
Inversion_clear H19.
Rewrite H21 in H11.
Cut (bang (x1 y))=(p'0 y); [Intro | Change ([_:name](bang (x1 _)) y)=(p'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H19; Apply bBANG.
Change (btrans ([_:name](par (x1 _) (bang (x1 _))) y) (a'0 y) (q'0 y)); Apply H7 with x0; Auto.
Apply notin_nu; Intros; Inversion_clear H20; Apply notin_par; [Auto | Apply notin_bang; Auto].
Rewrite H21; Trivial.
Apply proc_mono with x0.
Inversion_clear H15.
Cut (notin y (p'0 x0)); [Intro | Auto].
Rewrite <- H11 in H15; Inversion H15; Apply notin_par; Assumption.
Apply notin_nu; Intros; Apply notin_bang; Inversion_clear H20; Auto.

(* bRES Case *)

Elim (LEM_name x0 y); Intros.

(* Subcase x0=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=y *)

Elim (ho_proc_exp p1 x0); Elim (ho2_proc_exp p2 x0); Intros.
Inversion_clear H19; Inversion_clear H20.
Rewrite H22 in H12; Rewrite H23 in H11.
Cut (nu (x2 y))=(p'0 y); [Intro | Change ([_:name](nu (x2 _)) y)=(p'0 y); Apply proc_sub_congr with x0; Auto].
Rewrite <- H20.
Cut ([v:name](nu [z:name](x1 y z v)))=(q'0 y); [Intro | Change ([_:name]([v:name](nu [z:name](x1 _ z v))) y)=(q'0 y); Apply ho_proc_sub_congr with x0; Auto].
Rewrite <- H24.
Apply bRES with (cons x0 (cons y l)); Intros.
Inversion_clear H28.
Inversion_clear H30.
Change (btrans ([_:name](x2 _ y0) y) (a'0 y) ([_:name](x1 _ y0) y)).
Apply H7 with y0 x0; Auto.
Cut (notin y0 (nu [_:name](nu (x2 _)))); [Intro | Apply proc_mono with y; Assumption].
Rewrite H23; Inversion_clear H30; Auto.
Cut (notin y0 (nu [y:name](nu [z:name](nu (x1 y z))))); [Intro | Apply proc_mono with y; Assumption].
Rewrite H22; Inversion_clear H30; Auto.
Rewrite H13; Cut (b_act_notin_ho y0 a'0); [Intro | Apply b_act_mono with y; Assumption].
Unfold f_act_notin_ho in H30; Auto.
Apply notin_nu; Intros.
Inversion_clear H19.
Cut (notin x0 (nu (x2 z))); [Intro | Auto].
Inversion_clear H19; Auto.
Apply notin_nu; Intros.
Inversion_clear H21.
Cut (notin x0 (nu [z0:name](nu (x1 z z0)))); [Intro | Auto].
Inversion_clear H21.
Cut (notin x0 (nu (x1 z y0))); [Intro | Auto]; Assumption.
Rewrite H23; Trivial.
Rewrite H22; Trivial.
Apply H6; Auto.
Cut (notin y0 (nu [_:name](nu (x2 _)))); [Intro | Apply proc_mono with y; Auto].
Inversion_clear H30.
Cut (notin y0 (nu (x2 x0))); [Intro | Auto].
Rewrite H23; Assumption.
Cut (notin y0 (nu [y:name](nu [z:name](nu (x1 y z))))); [Intro | Apply proc_mono with y; Assumption].
Rewrite H22; Inversion_clear H30; Auto.
Rewrite H13; Cut (b_act_notin_ho y0 a'0); [Intro | Apply b_act_mono with y; Assumption].
Unfold b_act_notin_ho in H30; Auto.
Cut (notin  y (p'0 x0)); [Intro | Inversion_clear H15; Auto].
Cut (notin y (nu [_:name](nu (x2 _)))); [Intro | Apply proc_mono with x0; Rewrite <- H11 in H30; Auto].
Apply notin_nu; Intros.
Inversion_clear H32.
Cut (notin y (nu (x2 z))); [Intro | Auto].
Inversion_clear H32; Auto.
Cut (notin  y (nu (q'0 x0))); [Intro | Inversion_clear H16; Auto].
Rewrite <- H12 in H30; Inversion_clear H30.
Cut (notin y (nu [z0:name](x1 x0 z0 y0))); [Intro | Auto].
Inversion_clear H30.
Apply proc_mono with x0; Apply proc_mono with y0; Auto.
Do 3 (Apply notin_nu; Intros).
Inversion_clear H21.
Cut (notin x0 (nu [z0:name](nu (x1 z z0)))); [Intro | Auto].
Inversion_clear H21.
Cut (notin x0 (nu (x1 z z1))); [Intro | Auto].
Inversion_clear H21; Auto.

(* OPEN Case *)

Elim (LEM_name x1 y); Intros.

(* Subcase x0=y *)

Rewrite <- H18; Rewrite <- H11; Rewrite <- H12; Rewrite <- H13; Assumption.

(* Subcase ~x0=y *)

Elim (ho_proc_exp p1 x1); Elim (ho_proc_exp p2 x1); Intros.
Inversion_clear H19; Inversion_clear H20.
Rewrite H23 in H11; Rewrite H22 in H12.
Cut (nu (x3 y))=(p'0 y); [Intro | Change ([_:name](nu (x3 _)) y)=(p'0 y); Apply proc_sub_congr with x1; Auto].
Rewrite <- H20.
Cut (x2 y)=(q'0 y); [Intro | Change ([_:name](x2 _) y)=(q'0 y); Apply ho_proc_sub_congr with x1; Auto].
Rewrite <- H24.
Elim (LEM_name x0 x1); Intros.
Rewrite H25 in H13.
Cut (bOut y)=(a'0 y); [Intro | Apply b_act_sub_congr with x1; Auto].
Rewrite <- H26.
Apply OPEN with (cons x1 (cons y l)); Intros.
Inversion_clear H30.
Inversion_clear H32.
Change (ftrans ([_:name](x3 _ y0) y) ([_:name](Out _ y0) y) ([_:name](x2 _ y0) y)); Apply H7 with y0 x1; Auto.
Cut (notin y0 (nu [_:name](nu (x3 _)))); [Intro | Apply proc_mono with y; Assumption].
Rewrite H23; Inversion_clear H32; Auto.
Cut (notin y0 (nu [_:name](nu (x2 _)))); [Intro | Apply proc_mono with y; Assumption].
Rewrite H22; Inversion_clear H32; Auto.
Rewrite H25; Auto.
Apply notin_nu; Intros; Inversion_clear H19.
Cut (notin x1 (nu (x3 z))); [Intro | Auto].
Inversion_clear H19; Auto.
Apply notin_nu; Intros; Inversion_clear H21.
Cut (notin x1 (nu (x2 z))); [Intro | Auto].
Inversion_clear H21; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite H23; Trivial.
Rewrite H22; Trivial.
Rewrite H25; Trivial.
Apply H6; Auto.
Cut (notin y0 (nu [_:name](nu (x3 _)))); [Intro | Apply proc_mono with y; Assumption].
Rewrite H23; Inversion_clear H32; Auto.
Cut (notin y0 (nu [_:name](nu (x2 _)))); [Intro | Apply proc_mono with y; Assumption].
Rewrite H22; Inversion_clear H32; Auto.
Rewrite H25; Auto.
Inversion_clear H15.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H11 in H15.
Apply proc_mono with x1; Inversion_clear H15; Auto.
Inversion_clear H16.
Cut (notin y (nu (q'0 x1))); [Intro | Auto].
Rewrite <- H12 in H16.
Apply proc_mono with x1; Inversion_clear H16; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.

Cut (bOut x0)=(a'0 y); [Intro | Change ([_:name](bOut x0) y)=(a'0 y); Apply b_act_sub_congr with x1; Auto].
Rewrite <- H26.
Apply OPEN with (cons x1 (cons y l)); Intros.
Inversion_clear H30.
Inversion_clear H32.
Change (ftrans ([_:name](x3 _ y0) y) ([_:name](Out x0 y0) y) ([_:name](x2 _ y0) y)); Apply H7 with y0 x1; Auto.
Cut (notin y0 (nu [_:name](nu (x3 _)))); [Intro | Apply proc_mono with y; Assumption].
Rewrite H23; Inversion_clear H32; Auto.
Cut (notin y0 (nu [_:name](nu (x2 _)))); [Intro | Apply proc_mono with y; Assumption].
Rewrite H22; Inversion_clear H32; Auto.
Apply notin_nu; Intros; Inversion_clear H19.
Cut (notin x1 (nu (x3 z))); [Intro | Auto].
Inversion_clear H19; Auto.
Apply notin_nu; Intros; Inversion_clear H21.
Cut (notin x1 (nu (x2 z))); [Intro | Auto].
Inversion_clear H21; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Rewrite H23; Trivial.
Rewrite H22; Trivial.
Apply H6; Auto.
Cut (notin y0 (nu [_:name](nu (x3 _)))); [Intro | Apply proc_mono with y; Assumption].
Rewrite H23; Inversion_clear H32; Auto.
Cut (notin y0 (nu [_:name](nu (x2 _)))); [Intro | Apply proc_mono with y; Assumption].
Rewrite H22; Inversion_clear H32; Auto.
Inversion_clear H15.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H11 in H15.
Apply proc_mono with x1; Inversion_clear H15; Auto.
Inversion_clear H16.
Cut (notin y (nu (q'0 x1))); [Intro | Auto].
Rewrite <- H12 in H16.
Apply proc_mono with x1; Inversion_clear H16; Auto.
Unfold f_act_notin_ho; Intros; Apply f_act_notin_Out; Auto.
Cut ~x0=y; [Intro; Auto | Apply Sep_proc with (nu p1)].
Apply Aux2 with (bOut x0) p2; [Apply b_act_isin_bOut | Auto].
Inversion_clear H15.
Cut (notin y (p'0 x1)); [Intro | Auto].
Rewrite <- H11 in H15; Rewrite <- H23 in H15; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Auto.

Qed.

Lemma FTR_L3: (p,q:name->proc)
        (a:name->f_act)
         (x:name)
          (notin x (nu p))
           ->(notin x (nu q))
              ->(f_act_notin_ho x a)
                 ->(ftrans (p x) (a x) (q x))
                    ->(y:name)
                       (notin y (nu p))
                        ->(notin y (nu q))
                           ->(f_act_notin_ho y a)
                              ->(ftrans (p y) (a y) (q y)).

Proof.

Intros; Apply L3 with (p x) (q x) (a x) x; Auto.

Qed.

Lemma BTR_L3: (p:name->proc)(q:name->name->proc)(a:name->b_act)(x:name)
                (notin x (nu p))->(notin x (nu [z:name](nu (q z))))->(b_act_notin_ho x a)
                  ->(btrans (p x) (a x) (q x))
                    ->(y:name)(notin y (nu p))->(notin y (nu [z:name](nu (q z))))->(b_act_notin_ho y a)
                      ->(btrans (p y) (a y) (q y)).

Proof.

Intros; Apply L3bis with (p x) (a x) (q x) x; Auto.

Qed.

End MPW_Lemma3.

Section MPW_Lemma6.

(***************************************************************)
(* Since a coinductive proof of Lemma 6 is not straightforward *)
(* (indeed all the attempts made using the Cofix tactic lead   *)
(* to guardedness problems), we will prove it by means of      *)
(* the GFP encoding approach and the Cross Adequacy result.    *)
(***************************************************************)

Inductive Op_S [R: proc -> proc -> Prop] : proc -> proc -> Prop :=
  op_s: (p,q:name->proc)(z:name)
         (notin z (nu p)) -> (notin z (nu q)) -> (R (p z) (q z)) ->
 (w:name)~(w=z)->(notin w (nu p)) -> (notin w (nu q)) ->
         ((Op_S R) (p w) (q w)).

Fixpoint Bfun [n:nat]: (proc->proc->Prop) := Case n of
                                               StBisim
                                               [m:nat](Op_S (Bfun m))
                                             end.

Inductive BL6 [p,q:proc] : Prop :=
   bl6 : (n:nat)((Bfun n) p q)->(BL6 p q).

Lemma BisimL6: (Inclus BL6 (Op_StBisim BL6)).

Proof.

(Unfold Inclus; Intros).
Inversion_clear H.
(Cut (Bfun n p1 p2);
 [ Generalize p1 p2; Generalize n ; Induction n0; Intros | Assumption ]).

(* Base Case *)

(Cut (StBisim p0 p3); [ Intro | Auto ]).
(Apply op_sb; Split; Intros; Split; Intros).

Inversion_clear H1.
Inversion_clear H3.
(Elim (H1 a); Intros).
(Elim (H3 p4 H2); Intros).
(Exists  x; Split; Try Tauto).
Inversion_clear H6.
(Cut (Bfun O p4 x); [ Intro | Auto ]).
(Apply bl6 with O; Auto).

Inversion_clear H1.
Inversion_clear H3.
(Elim (H1 a); Intros).
(Elim (H5 q1 H2); Intros).
(Exists  x; Split; Try Tauto).
Inversion_clear H6.
(Cut (Bfun O x q1); [ Intro | Auto ]).
(Apply bl6 with O; Auto).

Split; Intros.

Inversion_clear H1.
Inversion_clear H3.
Inversion_clear H4.
(Elim (H3 x); Intros).
(Elim (H4 p4 H2); Intros).
(Exists  x0; Split; Try Tauto).
Inversion_clear H7.
Intro.
(Cut (StBisim (p4 y) (x0 y)); [ Intro | Auto ]).
(Cut (Bfun O (p4 y) (x0 y)); [ Intro | Auto ]).
(Apply bl6 with O; Auto).

Inversion_clear H1.
Inversion_clear H3.
Inversion_clear H4.
(Elim (H3 x); Intros).
(Elim (H6 q1 H2); Intros).
(Exists  x0; Split; Try Tauto).
Inversion_clear H7.
Intro.
(Cut (StBisim (x0 y) (q1 y)); [ Intro | Auto ]).
(Cut (Bfun O (x0 y) (q1 y)); [ Intro | Auto ]).
(Apply bl6 with O; Auto).

Split; Intros.

Inversion_clear H1.
Inversion_clear H3.
Inversion_clear H4.
(Elim (H5 x); Intros).
(Elim (H4 p4 H2); Intros).
(Exists  x0; Split; Try Tauto).
Inversion_clear H7.
Intros.
(Cut (StBisim (p4 y) (x0 y)); [ Intro | Auto ]).
(Cut (Bfun O (p4 y) (x0 y)); [ Intro | Auto ]).
(Apply bl6 with O; Auto).

Inversion_clear H1.
Inversion_clear H3.
Inversion_clear H4.
(Elim (H5 x); Intros).
(Elim (H6 q1 H2); Intros).
(Exists  x0; Split; Try Tauto).
Inversion_clear H7.
Intros.
(Cut (StBisim (x0 y) (q1 y)); [ Intro | Auto ]).
(Cut (Bfun O (x0 y) (q1 y)); [ Intro | Auto ]).
(Apply bl6 with O; Auto).

(* Inductive Step *)

Simpl in H1.
Inversion H1.
Cut (Op_StBisim BL6 (p z) (q z)); [Intro | Auto].
Inversion_clear H10.
Apply op_sb; Split; Intros; Split; Intros.

Elim (proc_exp p4 w); Intros.
Inversion_clear H12.
Rewrite H14 in H10.
Rewrite H14.
Elim (f_act_exp a w); Intros.
Inversion_clear H12.
Rewrite H16 in H10.
Rewrite H16.
Cut (ftrans (p z) (x0 z) (x z)); [Intro | Idtac].
Inversion_clear H11.
Elim (H17 (x0 z)); Intros.
Elim (H11 (x z) H12); Intros.
Elim (proc_exp x1 z); Intros.
Inversion_clear H21.
Rewrite H23 in H20.
Inversion_clear H20.
Cut (ftrans (q w) (x0 w) (x2 w)); [Intro | Idtac].
Exists (x2 w); Split; Auto.
Inversion_clear H24.
Apply bl6 with (S n2).
Simpl.
Apply op_s with z; Auto.
Cut (notin z (p w)); [Intro | Inversion H2; Auto].
Elim (FTR_L1 (p w) (x w) (x0 w) z H24 H10); Intros; Auto.
Apply proc_mono with w; Auto.
Cut (notin w (q z)); [Intro | Inversion H7; Auto].
Elim (FTR_L1 (q z) (x2 z) (x0 z) w H24 H21); Intros; Auto.
Apply proc_mono with z; Auto.
Apply FTR_L3 with z; Auto.
Cut (notin z (p w)); [Intro | Inversion H2; Auto].
Elim (FTR_L1 (p w) (x w) (x0 w) z H20 H10); Intros; Auto.
Apply f_act_mono with w; Auto.
Cut (notin w (q z)); [Intro | Inversion H7; Auto].
Elim (FTR_L1 (q z) (x2 z) (x0 z) w H20 H21); Intros; Auto.
Apply proc_mono with z; Auto.
Apply FTR_L3 with w; Auto.
Cut (notin z (p w)); [Intro | Inversion H2; Auto].
Elim (FTR_L1 (p w) (x w) (x0 w) z H12 H10); Intros; Auto.
Apply proc_mono with w; Auto.
Cut (notin z (p w)); [Intro | Inversion H2; Auto].
Elim (FTR_L1 (p w) (x w) (x0 w) z H12 H10); Intros; Auto.
Apply f_act_mono with w; Auto.

Elim (proc_exp q1 w); Intros.
Inversion_clear H12.
Rewrite H14 in H10.
Rewrite H14.
Elim (f_act_exp a w); Intros.
Inversion_clear H12.
Rewrite H16 in H10.
Rewrite H16.
Cut (ftrans (q z) (x0 z) (x z)); [Intro | Idtac].
Inversion_clear H11.
Elim (H17 (x0 z)); Intros.
Elim (H19 (x z) H12); Intros.
Elim (proc_exp x1 z); Intros.
Inversion_clear H21.
Rewrite H23 in H20.
Inversion_clear H20.
Cut (ftrans (p w) (x0 w) (x2 w)); [Intro | Idtac].
Exists (x2 w); Split; Auto.
Inversion_clear H24.
Apply bl6 with (S n2).
Simpl.
Apply op_s with z; Auto.
Cut (notin z (q w)); [Intro | Inversion H3; Auto].
Elim (FTR_L1 (q w) (x w) (x0 w) z H24 H10); Intros; Auto.
Apply proc_mono with w; Auto.
Cut (notin w (p z)); [Intro | Inversion H6; Auto].
Elim (FTR_L1 (p z) (x2 z) (x0 z) w H24 H21); Intros; Auto.
Apply proc_mono with z; Auto.
Apply FTR_L3 with z; Auto.
Cut (notin z (q w)); [Intro | Inversion H3; Auto].
Elim (FTR_L1 (q w) (x w) (x0 w) z H20 H10); Intros; Auto.
Apply f_act_mono with w; Auto.
Cut (notin w (p z)); [Intro | Inversion H6; Auto].
Elim (FTR_L1 (p z) (x2 z) (x0 z) w H20 H21); Intros; Auto.
Apply proc_mono with z; Auto.
Apply FTR_L3 with w; Auto.
Cut (notin z (q w)); [Intro | Inversion H3; Auto].
Elim (FTR_L1 (q w) (x w) (x0 w) z H12 H10); Intros; Auto.
Apply proc_mono with w; Auto.
Cut (notin z (q w)); [Intro | Inversion H3; Auto].
Elim (FTR_L1 (q w) (x w) (x0 w) z H12 H10); Intros; Auto.
Apply f_act_mono with w; Auto.

(* Input-actions *)

Split; Intros.
(Elim (LEM_name w x); Intros).

(* Case: w=x *)

Rewrite <- H12 in H10.
Rewrite <- H12.
Elim (ho_proc_exp p4 w); Intros.
Inversion_clear H13.
Rewrite H15 in H10.
(Cut (btrans (p z) (In z) (x0 z)); [ Intro | Idtac ]).
Inversion_clear H11.
Inversion_clear H17.
(Elim (H11 z); Intros).
(Elim (H17 (x0 z) H13); Intros).
Inversion_clear H20.
(Elim (ho_proc_exp x1 z); Intros).
Inversion_clear H20.
Rewrite H24 in H21.
Rewrite H24 in H22.
(Cut (btrans (q w) (In w) (x2 w)); [ Intro | Idtac ]).
(Exists  (x2 w); Split).
Assumption.
Rewrite H15.
Intro.
Elim (LEM_name y w); Elim (LEM_name y z); Intros.

(* Subcase (y=w/\y=z) *)

Cut w=z; [Intro | Rewrite <- H25; Rewrite <- H26; Trivial].
Rewrite H27; Auto.

(* Subcase (y=w/\~y=z) *)

Rewrite H26.
Cut (BL6 (x0 z z) (x2 z z)); [Intro | Auto].
Inversion_clear H27.
Apply bl6 with (S n2).
Simpl.
Change (Op_S (Bfun n2) ([_:name](x0 _ _) w) ([_:name](x2 _ _) w)).
Apply op_s with z; Auto.
Cut (notin z (p w)); [Intro | Inversion H2; Auto].
Elim (BTR_L1 (p w) (x0 w) (In w) z H27 H10); Intros.
Apply proc_mono with w.
Inversion_clear H30; Auto.
Apply notin_nu; Intros.
Inversion_clear H23.
Cut (notin z (nu (x2 z0))); [Intro | Auto].
Inversion_clear H23.
Auto.
Apply notin_nu; Intros.
Inversion_clear H14.
Cut (notin w (nu (x0 z0))); [Intro | Auto].
Inversion_clear H14.
Auto.
Cut (notin w (q z)); [Intro | Inversion H7; Auto].
Elim (BTR_L1 (q z) (x2 z) (In z) w H27 H21); Intros.
Apply proc_mono with z.
Inversion_clear H30; Auto.

(* Subcase (~y=w/\y=z) *)

Rewrite H25.
Elim (unsat (par (nu [y:name](nu (x0 y))) (nu [y:name](nu (x2 y)))) (cons w (cons z empty))); Intros.
Inversion_clear H27.
Inversion_clear H28.
Inversion_clear H29.
Inversion_clear H31.
Clear H32.
Cut (BL6 (x0 z x3) (x2 z x3)); [Intro | Auto].
Inversion_clear H31.
Apply bl6 with (S (S n2)).
Simpl.
Apply op_s with x3; Auto.
Apply notin_nu; Intros.
Inversion_clear H27.
Cut (notin x3 (nu (x0 w))); [Intro | Auto].
Inversion_clear H27.
Auto.
Apply notin_nu; Intros.
Inversion_clear H30.
Cut (notin x3 (nu (x2 w))); [Intro | Auto].
Inversion_clear H30.
Auto.
Change (Op_S (Bfun n2) ([_:name](x0 _ x3) w) ([_:name](x2 _ x3) w)).
Apply op_s with z; Auto.
Cut (notin z (p w)); [Intro | Inversion H2; Auto].
Elim (BTR_L1 (p w) (x0 w) (In w) z H31 H10); Intros.
Apply proc_mono with w.
Inversion_clear H34; Auto.
Apply notin_nu; Intros.
Inversion_clear H23.
Cut (notin z (nu (x2 z0))); [Intro | Auto].
Inversion_clear H23.
Auto.
Apply notin_nu; Intros.
Inversion_clear H14.
Cut (notin w (nu (x0 z0))); [Intro | Auto].
Inversion_clear H14.
Auto.
Cut (notin w (q z)); [Intro | Inversion H7; Auto].
Elim (BTR_L1 (q z) (x2 z) (In z) w H31 H21); Intros.
Apply proc_mono with z.
Inversion_clear H34; Auto.
Cut (notin z (p w)); [Intro | Inversion H2; Auto].
Elim (BTR_L1 (p w) (x0 w) (In w) z H31 H10); Intros.
Apply proc_mono with w.
Inversion_clear H34; Auto.
Apply notin_nu; Intros.
Inversion_clear H23.
Cut (notin z (nu (x2 w))); [Intro | Auto].
Inversion_clear H23.
Auto.

(* Subcase (~y=w/\~y=z) *)

Cut (BL6 (x0 z y) (x2 z y)); [Intro | Auto].
Inversion_clear H27.
Apply bl6 with (S n2).
Simpl.
Change (Op_S (Bfun n2) ([_:name](x0 _ y) w) ([_:name](x2 _ y) w)).
Apply op_s with z; Auto.
Cut (notin z (p w)); [Intro | Inversion H2; Auto].
Elim (BTR_L1 (p w) (x0 w) (In w) z H27 H10); Intros.
Apply proc_mono with w.
Inversion_clear H30; Auto.
Apply notin_nu; Intros.
Inversion_clear H23.
Cut (notin z (nu (x2 z0))); [Intro | Auto].
Inversion_clear H23.
Auto.
Apply notin_nu; Intros.
Inversion_clear H14.
Cut (notin w (nu (x0 z0))); [Intro | Auto].
Inversion_clear H14.
Auto.
Cut (notin w (q z)); [Intro | Inversion H7; Auto].
Elim (BTR_L1 (q z) (x2 z) (In z) w H27 H21); Intros.
Apply proc_mono with z.
Inversion_clear H30; Auto.

(Apply BTR_L3 with z; Auto).
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Assumption.
Cut (notin w (q z)); [Intro | Inversion H7; Auto].
Elim (BTR_L1 (q z) (x2 z) (In z) w H20 H21); Intros.
Apply proc_mono with z.
Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Assumption.

(Apply BTR_L3 with w; Auto).
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Assumption.
Cut (notin z (p w)); [Intro | Inversion H2; Auto].
Elim (BTR_L1 (p w) (x0 w) (In w) z H13 H10); Intros.
Apply proc_mono with w.
Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Assumption.

(* Case ~w=x *)

Elim (ho_proc_exp p4 w); Intros.
Inversion_clear H13.
Rewrite H15 in H10.
(Cut (btrans (p z) (In x) (x0 z)); [ Intro | Idtac ]).
Inversion_clear H11.
Inversion_clear H17.
(Elim (H11 x); Intros).
(Elim (H17 (x0 z) H13); Intros).
Inversion_clear H20.
(Elim (ho_proc_exp x1 z); Intros).
Inversion_clear H20.
Rewrite H24 in H21.
Rewrite H24 in H22.
(Cut (btrans (q w) (In x) (x2 w)); [ Intro | Idtac ]).
(Exists  (x2 w); Split).
Assumption.
Rewrite H15.
Intro.
Elim (LEM_name y w); Elim (LEM_name y z); Intros.

(* Subcase (y=w/\y=z) *)

Cut w=z; [Intro | Rewrite <- H25; Rewrite <- H26; Trivial].
Rewrite H27; Auto.

(* Subcase (y=w/\~y=z) *)

Rewrite H26.
Cut (BL6 (x0 z z) (x2 z z)); [Intro | Auto].
Inversion_clear H27.
Apply bl6 with (S n2).
Simpl.
Change (Op_S (Bfun n2) ([_:name](x0 _ _) w) ([_:name](x2 _ _) w)).
Apply op_s with z; Auto.
Cut (notin z (p w)); [Intro | Inversion H2; Auto].
Elim (BTR_L1 (p w) (x0 w) (In x) z H27 H10); Intros.
Apply proc_mono with w.
Inversion_clear H30; Auto.
Apply notin_nu; Intros.
Inversion_clear H23.
Cut (notin z (nu (x2 z0))); [Intro | Auto].
Inversion_clear H23.
Auto.
Apply notin_nu; Intros.
Inversion_clear H14.
Cut (notin w (nu (x0 z0))); [Intro | Auto].
Inversion_clear H14.
Auto.
Cut (notin w (q z)); [Intro | Inversion H7; Auto].
Elim (BTR_L1 (q z) (x2 z) (In x) w H27 H21); Intros.
Apply proc_mono with z.
Inversion_clear H30; Auto.

(* Subcase (~y=w/\y=z) *)

Rewrite H25.
Elim (unsat (par (nu [y:name](nu (x0 y))) (nu [y:name](nu (x2 y)))) (cons w (cons z empty))); Intros.
Inversion_clear H27.
Inversion_clear H28.
Inversion_clear H29.
Inversion_clear H31.
Clear H32.
Cut (BL6 (x0 z x3) (x2 z x3)); [Intro | Auto].
Inversion_clear H31.
Apply bl6 with (S (S n2)).
Simpl.
Apply op_s with x3; Auto.
Apply notin_nu; Intros.
Inversion_clear H27.
Cut (notin x3 (nu (x0 w))); [Intro | Auto].
Inversion_clear H27.
Auto.
Apply notin_nu; Intros.
Inversion_clear H30.
Cut (notin x3 (nu (x2 w))); [Intro | Auto].
Inversion_clear H30.
Auto.
Change (Op_S (Bfun n2) ([_:name](x0 _ x3) w) ([_:name](x2 _ x3) w)).
Apply op_s with z; Auto.
Cut (notin z (p w)); [Intro | Inversion H2; Auto].
Elim (BTR_L1 (p w) (x0 w) (In x) z H31 H10); Intros.
Apply proc_mono with w.
Inversion_clear H34; Auto.
Apply notin_nu; Intros.
Inversion_clear H23.
Cut (notin z (nu (x2 z0))); [Intro | Auto].
Inversion_clear H23.
Auto.
Apply notin_nu; Intros.
Inversion_clear H14.
Cut (notin w (nu (x0 z0))); [Intro | Auto].
Inversion_clear H14.
Auto.
Cut (notin w (q z)); [Intro | Inversion H7; Auto].
Elim (BTR_L1 (q z) (x2 z) (In x) w H31 H21); Intros.
Apply proc_mono with z.
Inversion_clear H34; Auto.
Cut (notin z (p w)); [Intro | Inversion H2; Auto].
Elim (BTR_L1 (p w) (x0 w) (In x) z H31 H10); Intros.
Apply proc_mono with w.
Inversion_clear H34; Auto.
Apply notin_nu; Intros.
Inversion_clear H23.
Cut (notin z (nu (x2 w))); [Intro | Auto].
Inversion_clear H23.
Auto.

(* Subcase (~y=w/\~y=z) *)

Cut (BL6 (x0 z y) (x2 z y)); [Intro | Auto].
Inversion_clear H27.
Apply bl6 with (S n2).
Simpl.
Change (Op_S (Bfun n2) ([_:name](x0 _ y) w) ([_:name](x2 _ y) w)).
Apply op_s with z; Auto.
Cut (notin z (p w)); [Intro | Inversion H2; Auto].
Elim (BTR_L1 (p w) (x0 w) (In x) z H27 H10); Intros.
Apply proc_mono with w.
Inversion_clear H30; Auto.
Apply notin_nu; Intros.
Inversion_clear H23.
Cut (notin z (nu (x2 z0))); [Intro | Auto].
Inversion_clear H23.
Auto.
Apply notin_nu; Intros.
Inversion_clear H14.
Cut (notin w (nu (x0 z0))); [Intro | Auto].
Inversion_clear H14.
Auto.
Cut (notin w (q z)); [Intro | Inversion H7; Auto].
Elim (BTR_L1 (q z) (x2 z) (In x) w H27 H21); Intros.
Apply proc_mono with z.
Inversion_clear H30; Auto.

Cut (notin z (p w)); [Intro | Inversion H2; Auto].
Elim (BTR_L1 (p w) (x0 w) (In x) z H20 H10); Intros.
Change (btrans (q w) ([_:name](In x) w) (x2 w)).
(Apply BTR_L3 with z; Auto).
Unfold b_act_notin_ho; Intros; Assumption.
Cut (notin w (q z)); [Intro | Inversion H7; Auto].
Elim (BTR_L1 (q z) (x2 z) (In x) w H27 H21); Intros.
Apply proc_mono with z; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Assumption.

Cut (notin z (p w)); [Intro | Inversion H2; Auto].
Elim (BTR_L1 (p w) (x0 w) (In x) z H13 H10); Intros.
Change (btrans (p z) ([_:name](In x) z) (x0 z)).
(Apply BTR_L3 with w; Auto).
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Assumption.
Apply proc_mono with w; Assumption.
Unfold b_act_notin_ho; Intros; Assumption.

(* The way back (symmetric case) *)

(Elim (LEM_name w x); Intros).

(* Case: w=x *)

Rewrite <- H12 in H10.
Rewrite <- H12.
Elim (ho_proc_exp q1 w); Intros.
Inversion_clear H13.
Rewrite H15 in H10.
(Cut (btrans (q z) (In z) (x0 z)); [ Intro | Idtac ]).
Inversion_clear H11.
Inversion_clear H17.
(Elim (H11 z); Intros).
(Elim (H19 (x0 z) H13); Intros).
Inversion_clear H20.
(Elim (ho_proc_exp x1 z); Intros).
Inversion_clear H20.
Rewrite H24 in H21.
Rewrite H24 in H22.
(Cut (btrans (p w) (In w) (x2 w)); [ Intro | Idtac ]).
(Exists  (x2 w); Split).
Assumption.
Rewrite H15.
Intro.
Elim (LEM_name y w); Elim (LEM_name y z); Intros.

(* Subcase (y=w/\y=z) *)

Cut w=z; [Intro | Rewrite <- H25; Rewrite <- H26; Trivial].
Rewrite H27; Auto.

(* Subcase (y=w/\~y=z) *)

Rewrite H26.
Cut (BL6 (x2 z z) (x0 z z)); [Intro | Auto].
Inversion_clear H27.
Apply bl6 with (S n2).
Simpl.
Change (Op_S (Bfun n2) ([_:name](x2 _ _) w) ([_:name](x0 _ _) w)).
Apply op_s with z; Auto.
Apply notin_nu; Intros.
Inversion_clear H23.
Cut (notin z (nu (x2 z0))); [Intro | Auto].
Inversion_clear H23.
Auto.
Cut (notin z (q w)); [Intro | Inversion H3; Auto].
Elim (BTR_L1 (q w) (x0 w) (In w) z H27 H10); Intros.
Apply proc_mono with w.
Inversion_clear H30; Auto.
Cut (notin w (p z)); [Intro | Inversion H6; Auto].
Elim (BTR_L1 (p z) (x2 z) (In z) w H27 H21); Intros.
Apply proc_mono with z.
Inversion_clear H30; Auto.
Apply notin_nu; Intros.
Inversion_clear H14.
Cut (notin w (nu (x0 z0))); [Intro | Auto].
Inversion_clear H14.
Auto.

(* Subcase (~y=w/\y=z) *)

Rewrite H25.
Elim (unsat (par (nu [y:name](nu (x0 y))) (nu [y:name](nu (x2 y)))) (cons w (cons z empty))); Intros.
Inversion_clear H27.
Inversion_clear H28.
Inversion_clear H29.
Inversion_clear H31.
Clear H32.
Cut (BL6 (x2 z x3) (x0 z x3)); [Intro | Auto].
Inversion_clear H31.
Apply bl6 with (S (S n2)).
Simpl.
Apply op_s with x3; Auto.
Apply notin_nu; Intros.
Inversion_clear H30.
Cut (notin x3 (nu (x2 w))); [Intro | Auto].
Inversion_clear H30.
Auto.
Apply notin_nu; Intros.
Inversion_clear H27.
Cut (notin x3 (nu (x0 w))); [Intro | Auto].
Inversion_clear H27.
Auto.
Change (Op_S (Bfun n2) ([_:name](x2 _ x3) w) ([_:name](x0 _ x3) w)).
Apply op_s with z; Auto.
Apply notin_nu; Intros.
Inversion_clear H23.
Cut (notin z (nu (x2 z0))); [Intro | Auto].
Inversion_clear H23.
Auto.
Cut (notin z (q w)); [Intro | Inversion H3; Auto].
Elim (BTR_L1 (q w) (x0 w) (In w) z H31 H10); Intros.
Apply proc_mono with w.
Inversion_clear H34; Auto.
Cut (notin w (p z)); [Intro | Inversion H6; Auto].
Elim (BTR_L1 (p z) (x2 z) (In z) w H31 H21); Intros.
Apply proc_mono with z.
Inversion_clear H34; Auto.
Apply notin_nu; Intros.
Inversion_clear H14.
Cut (notin w (nu (x0 z0))); [Intro | Auto].
Inversion_clear H14.
Auto.
Apply notin_nu; Intros.
Inversion_clear H23.
Cut (notin z (nu (x2 w))); [Intro | Auto].
Inversion_clear H23.
Auto.
Cut (notin z (q w)); [Intro | Inversion H3; Auto].
Elim (BTR_L1 (q w) (x0 w) (In w) z H31 H10); Intros.
Apply proc_mono with w.
Inversion_clear H34; Auto.

(* Subcase (~y=w/\~y=z) *)

Cut (BL6 (x2 z y) (x0 z y)); [Intro | Auto].
Inversion_clear H27.
Apply bl6 with (S n2).
Simpl.
Change (Op_S (Bfun n2) ([_:name](x2 _ y) w) ([_:name](x0 _ y) w)).
Apply op_s with z; Auto.
Apply notin_nu; Intros.
Inversion_clear H23.
Cut (notin z (nu (x2 z0))); [Intro | Auto].
Inversion_clear H23.
Auto.
Cut (notin z (q w)); [Intro | Inversion H3; Auto].
Elim (BTR_L1 (q w) (x0 w) (In w) z H27 H10); Intros.
Apply proc_mono with w.
Inversion_clear H30; Auto.
Cut (notin w (p z)); [Intro | Inversion H6; Auto].
Elim (BTR_L1 (p z) (x2 z) (In z) w H27 H21); Intros.
Apply proc_mono with z.
Inversion_clear H30; Auto.
Apply notin_nu; Intros.
Inversion_clear H14.
Cut (notin w (nu (x0 z0))); [Intro | Auto].
Inversion_clear H14.
Auto.

(Apply BTR_L3 with z; Auto).
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Assumption.
Cut (notin w (p z)); [Intro | Inversion H6; Auto].
Elim (BTR_L1 (p z) (x2 z) (In z) w H20 H21); Intros.
Apply proc_mono with z.
Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Assumption.

(Apply BTR_L3 with w; Auto).
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Assumption.
Cut (notin z (q w)); [Intro | Inversion H3; Auto].
Elim (BTR_L1 (q w) (x0 w) (In w) z H13 H10); Intros.
Apply proc_mono with w.
Auto.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Assumption.

(* Case ~w=x *)

Elim (ho_proc_exp q1 w); Intros.
Inversion_clear H13.
Rewrite H15 in H10.
(Cut (btrans (q z) (In x) (x0 z)); [ Intro | Idtac ]).
Inversion_clear H11.
Inversion_clear H17.
(Elim (H11 x); Intros).
(Elim (H19 (x0 z) H13); Intros).
Inversion_clear H20.
(Elim (ho_proc_exp x1 z); Intros).
Inversion_clear H20.
Rewrite H24 in H21.
Rewrite H24 in H22.
(Cut (btrans (p w) (In x) (x2 w)); [ Intro | Idtac ]).
(Exists  (x2 w); Split).
Assumption.
Rewrite H15.
Intro.
Elim (LEM_name y w); Elim (LEM_name y z); Intros.

(* Subcase (y=w/\y=z) *)

Cut w=z; [Intro | Rewrite <- H25; Rewrite <- H26; Trivial].
Rewrite H27; Auto.

(* Subcase (y=w/\~y=z) *)

Rewrite H26.
Cut (BL6 (x2 z z) (x0 z z)); [Intro | Auto].
Inversion_clear H27.
Apply bl6 with (S n2).
Simpl.
Change (Op_S (Bfun n2) ([_:name](x2 _ _) w) ([_:name](x0 _ _) w)).
Apply op_s with z; Auto.
Apply notin_nu; Intros.
Inversion_clear H23.
Cut (notin z (nu (x2 z0))); [Intro | Auto].
Inversion_clear H23.
Auto.
Cut (notin z (q w)); [Intro | Inversion H3; Auto].
Elim (BTR_L1 (q w) (x0 w) (In x) z H27 H10); Intros.
Apply proc_mono with w.
Inversion_clear H30; Auto.
Cut (notin w (p z)); [Intro | Inversion H6; Auto].
Elim (BTR_L1 (p z) (x2 z) (In x) w H27 H21); Intros.
Apply proc_mono with z.
Inversion_clear H30; Auto.
Apply notin_nu; Intros.
Inversion_clear H14.
Cut (notin w (nu (x0 z0))); [Intro | Auto].
Inversion_clear H14.
Auto.

(* Subcase (~y=w/\y=z) *)

Rewrite H25.
Elim (unsat (par (nu [y:name](nu (x0 y))) (nu [y:name](nu (x2 y)))) (cons w (cons z empty))); Intros.
Inversion_clear H27.
Inversion_clear H28.
Inversion_clear H29.
Inversion_clear H31.
Clear H32.
Cut (BL6 (x2 z x3) (x0 z x3)); [Intro | Auto].
Inversion_clear H31.
Apply bl6 with (S (S n2)).
Simpl.
Apply op_s with x3; Auto.
Apply notin_nu; Intros.
Inversion_clear H30.
Cut (notin x3 (nu (x2 w))); [Intro | Auto].
Inversion_clear H30.
Auto.
Apply notin_nu; Intros.
Inversion_clear H27.
Cut (notin x3 (nu (x0 w))); [Intro | Auto].
Inversion_clear H27.
Auto.
Change (Op_S (Bfun n2) ([_:name](x2 _ x3) w) ([_:name](x0 _ x3) w)).
Apply op_s with z; Auto.
Apply notin_nu; Intros.
Inversion_clear H23.
Cut (notin z (nu (x2 z0))); [Intro | Auto].
Inversion_clear H23.
Auto.
Cut (notin z (q w)); [Intro | Inversion H3; Auto].
Elim (BTR_L1 (q w) (x0 w) (In x) z H31 H10); Intros.
Apply proc_mono with w.
Inversion_clear H34; Auto.
Cut (notin w (p z)); [Intro | Inversion H6; Auto].
Elim (BTR_L1 (p z) (x2 z) (In x) w H31 H21); Intros.
Apply proc_mono with z.
Inversion_clear H34; Auto.
Apply notin_nu; Intros.
Inversion_clear H14.
Cut (notin w (nu (x0 z0))); [Intro | Auto].
Inversion_clear H14.
Auto.
Apply notin_nu; Intros.
Inversion_clear H23.
Cut (notin z (nu (x2 w))); [Intro | Auto].
Inversion_clear H23.
Auto.
Cut (notin z (q w)); [Intro | Inversion H3; Auto].
Elim (BTR_L1 (q w) (x0 w) (In x) z H31 H10); Intros.
Apply proc_mono with w.
Inversion_clear H34; Auto.

(* Subcase (~y=w/\~y=z) *)

Cut (BL6 (x2 z y) (x0 z y)); [Intro | Auto].
Inversion_clear H27.
Apply bl6 with (S n2).
Simpl.
Change (Op_S (Bfun n2) ([_:name](x2 _ y) w) ([_:name](x0 _ y) w)).
Apply op_s with z; Auto.
Apply notin_nu; Intros.
Inversion_clear H23.
Cut (notin z (nu (x2 z0))); [Intro | Auto].
Inversion_clear H23.
Auto.
Cut (notin z (q w)); [Intro | Inversion H3; Auto].
Elim (BTR_L1 (q w) (x0 w) (In x) z H27 H10); Intros.
Apply proc_mono with w.
Inversion_clear H30; Auto.
Cut (notin w (p z)); [Intro | Inversion H6; Auto].
Elim (BTR_L1 (p z) (x2 z) (In x) w H27 H21); Intros.
Apply proc_mono with z.
Inversion_clear H30; Auto.
Apply notin_nu; Intros.
Inversion_clear H14.
Cut (notin w (nu (x0 z0))); [Intro | Auto].
Inversion_clear H14.
Auto.

Cut (notin z (q w)); [Intro | Inversion H3; Auto].
Elim (BTR_L1 (q w) (x0 w) (In x) z H20 H10); Intros.
Change (btrans (p w) ([_:name](In x) w) (x2 w)).
(Apply BTR_L3 with z; Auto).
Unfold b_act_notin_ho; Intros; Assumption.
Cut (notin w (p z)); [Intro | Inversion H6; Auto].
Elim (BTR_L1 (p z) (x2 z) (In x) w H27 H21); Intros.
Apply proc_mono with z; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Assumption.

Cut (notin z (q w)); [Intro | Inversion H3; Auto].
Elim (BTR_L1 (q w) (x0 w) (In x) z H13 H10); Intros.
Change (btrans (q z) ([_:name](In x) z) (x0 z)).
(Apply BTR_L3 with w; Auto).
Unfold b_act_notin_ho; Intros; Apply b_act_notin_In; Assumption.
Apply proc_mono with w; Assumption.
Unfold b_act_notin_ho; Intros; Assumption.

(* Bound-output actions *)

Split; Intros.
(Elim (LEM_name w x); Intros).

(* Case w=x *)

Rewrite <- H12 in H10.
Rewrite <- H12.
Elim (ho_proc_exp p4 w); Intros.
Inversion_clear H13.
Rewrite H15 in H10.
(Cut (btrans (p z) (bOut z) (x0 z)); [ Intro | Idtac ]).
Inversion_clear H11.
Inversion_clear H17.
(Elim (H18 z); Intros).
(Elim (H17 (x0 z) H13); Intros).
Inversion_clear H20.
(Elim (ho_proc_exp x1 z); Intros).
Inversion_clear H20.
Rewrite H24 in H21.
Rewrite H24 in H22.
(Cut (btrans (q w) (bOut w) (x2 w)); [ Intro | Idtac ]).
(Exists  (x2 w); Split).
Assumption.
Rewrite H15.
Intros.
Elim (unsat (par (nu [_:name](nu (x0 _))) (nu [_:name](nu (x2 _)))) (cons y (cons z (cons w empty)))); Intros.
Inversion_clear H27.
Inversion_clear H28.
Inversion_clear H29.
Inversion_clear H31.
Inversion_clear H32.
Clear H33.
Cut (BL6 (x0 z x3) (x2 z x3)); [Intro | Apply H22; Auto].
Inversion_clear H32.
Apply bl6 with (S (S n2)).
Simpl.
Apply op_s with x3; Intros.
Apply notin_nu; Intros.
Inversion_clear H27.
Cut (notin x3 (nu (x0 w))); [Intro | Auto].
Inversion_clear H27.
Auto.
Apply notin_nu; Intros.
Inversion_clear H30.
Cut (notin x3 (nu (x2 w))); [Intro | Auto].
Inversion_clear H30.
Auto.
Change (Op_S (Bfun n2) ([_:name](x0 _ x3) w) ([_:name](x2 _ x3) w)).
Apply op_s with z; Intros.
Cut (notin z (p w)); [Intro | Inversion H2; Auto].
Elim (BTR_L1 (p w) (x0 w) (bOut w) z H32 H10); Intros.
Apply proc_mono with w.
Inversion_clear H35; Auto.
Apply notin_nu; Intros.
Inversion_clear H23.
Cut (notin z (nu (x2 z0))); [Intro | Auto].
Inversion_clear H23.
Auto.
Assumption.
Auto.
Apply notin_nu; Intros.
Inversion_clear H14.
Cut (notin w (nu (x0 z0))); [Intro | Auto].
Inversion_clear H14.
Auto.
Cut (notin w (q z)); [Intro | Inversion H7; Auto].
Elim (BTR_L1 (q z) (x2 z) (bOut z) w H32 H21); Intros.
Apply proc_mono with z.
Inversion_clear H35; Auto.
Auto.
Auto.
Auto.
Apply notin_nu; Intros.
Inversion_clear H27.
Cut (notin x3 (nu (x0 z))); [Intro | Auto].
Inversion_clear H27.
Auto.
Apply notin_nu; Intros.
Inversion_clear H30.
Cut (notin x3 (nu (x2 z))); [Intro | Auto].
Inversion_clear H30.
Auto.

Cut (notin z (p w)); [Intro | Inversion H2; Auto].
Elim (BTR_L1 (p w) (x0 w) (bOut w) z H20 H10); Intros.
(Apply BTR_L3 with z; Auto).
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Assumption.
Cut (notin w (q z)); [Intro | Inversion H7; Auto].
Elim (BTR_L1 (q z) (x2 z) (bOut z) w H27 H21); Intros.
Apply proc_mono with z; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Assumption.

Cut (notin z (p w)); [Intro | Inversion H2; Auto].
Elim (BTR_L1 (p w) (x0 w) (bOut w) z H13 H10); Intros.
(Apply BTR_L3 with w; Auto).
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Assumption.
Apply proc_mono with w; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Assumption.

(* Case ~w=x *)

Elim (ho_proc_exp p4 w); Intros.
Inversion_clear H13.
Rewrite H15 in H10.
(Cut (btrans (p z) (bOut x) (x0 z)); [ Intro | Idtac ]).
Inversion_clear H11.
Inversion_clear H17.
(Elim (H18 x); Intros).
(Elim (H17 (x0 z) H13); Intros).
Inversion_clear H20.
(Elim (ho_proc_exp x1 z); Intros).
Inversion_clear H20.
Rewrite H24 in H21.
Rewrite H24 in H22.
(Cut (btrans (q w) (bOut x) (x2 w)); [ Intro | Idtac ]).
(Exists  (x2 w); Split).
Assumption.
Rewrite H15.
Intros.
Elim (unsat (par (nu [_:name](nu (x0 _))) (nu [_:name](nu (x2 _)))) (cons y (cons z (cons w empty)))); Intros.
Inversion_clear H27.
Inversion_clear H28.
Inversion_clear H29.
Inversion_clear H31.
Inversion_clear H32.
Clear H33.
Cut (BL6 (x0 z x3) (x2 z x3)); [Intro | Apply H22; Auto].
Inversion_clear H32.
Apply bl6 with (S (S n2)).
Simpl.
Apply op_s with x3; Intros.
Apply notin_nu; Intros.
Inversion_clear H27.
Cut (notin x3 (nu (x0 w))); [Intro | Auto].
Inversion_clear H27.
Auto.
Apply notin_nu; Intros.
Inversion_clear H30.
Cut (notin x3 (nu (x2 w))); [Intro | Auto].
Inversion_clear H30.
Auto.
Change (Op_S (Bfun n2) ([_:name](x0 _ x3) w) ([_:name](x2 _ x3) w)).
Apply op_s with z; Intros.
Cut (notin z (p w)); [Intro | Inversion H2; Auto].
Elim (BTR_L1 (p w) (x0 w) (bOut x) z H32 H10); Intros.
Apply proc_mono with w.
Inversion_clear H35; Auto.
Apply notin_nu; Intros.
Inversion_clear H23.
Cut (notin z (nu (x2 z0))); [Intro | Auto].
Inversion_clear H23.
Auto.
Assumption.
Auto.
Apply notin_nu; Intros.
Inversion_clear H14.
Cut (notin w (nu (x0 z0))); [Intro | Auto].
Inversion_clear H14.
Auto.
Cut (notin w (q z)); [Intro | Inversion H7; Auto].
Elim (BTR_L1 (q z) (x2 z) (bOut x) w H32 H21); Intros.
Apply proc_mono with z.
Inversion_clear H35; Auto.
Auto.
Auto.
Auto.
Apply notin_nu; Intros.
Inversion_clear H27.
Cut (notin x3 (nu (x0 z))); [Intro | Auto].
Inversion_clear H27.
Auto.
Apply notin_nu; Intros.
Inversion_clear H30.
Cut (notin x3 (nu (x2 z))); [Intro | Auto].
Inversion_clear H30.
Auto.

Cut (notin z (p w)); [Intro | Inversion H2; Auto].
Elim (BTR_L1 (p w) (x0 w) (bOut x) z H20 H10); Intros.
Change (btrans (q w) ([_:name](bOut x) w) (x2 w)).
(Apply BTR_L3 with z; Auto).
Unfold b_act_notin_ho; Intros; Assumption.
Cut (notin w (q z)); [Intro | Inversion H7; Auto].
Elim (BTR_L1 (q z) (x2 z) (bOut x) w H27 H21); Intros.
Apply proc_mono with z; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Assumption.

Cut (notin z (p w)); [Intro | Inversion H2; Auto].
Elim (BTR_L1 (p w) (x0 w) (bOut x) z H13 H10); Intros.
Change (btrans (p z) ([_:name](bOut x) z) (x0 z)).
(Apply BTR_L3 with w; Auto).
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Assumption.
Apply proc_mono with w; Assumption.
Unfold b_act_notin_ho; Intros; Assumption.

(* The symmetric case. *)

(Elim (LEM_name w x); Intros).

(* Case w=x *)

Rewrite <- H12 in H10.
Rewrite <- H12.
Elim (ho_proc_exp q1 w); Intros.
Inversion_clear H13.
Rewrite H15 in H10.
(Cut (btrans (q z) (bOut z) (x0 z)); [ Intro | Idtac ]).
Inversion_clear H11.
Inversion_clear H17.
(Elim (H18 z); Intros).
(Elim (H19 (x0 z) H13); Intros).
Inversion_clear H20.
(Elim (ho_proc_exp x1 z); Intros).
Inversion_clear H20.
Rewrite H24 in H21.
Rewrite H24 in H22.
(Cut (btrans (p w) (bOut w) (x2 w)); [ Intro | Idtac ]).
(Exists  (x2 w); Split).
Assumption.
Rewrite H15.
Intros.
Elim (unsat (par (nu [_:name](nu (x0 _))) (nu [_:name](nu (x2 _)))) (cons y (cons z (cons w empty)))); Intros.
Inversion_clear H27.
Inversion_clear H28.
Inversion_clear H29.
Inversion_clear H31.
Inversion_clear H32.
Clear H33.
Cut (BL6 (x2 z x3) (x0 z x3)); [Intro | Apply H22; Auto].
Inversion_clear H32.
Apply bl6 with (S (S n2)).
Simpl.
Apply op_s with x3; Intros.
Apply notin_nu; Intros.
Inversion_clear H30.
Cut (notin x3 (nu (x2 w))); [Intro | Auto].
Inversion_clear H30.
Auto.
Apply notin_nu; Intros.
Inversion_clear H27.
Cut (notin x3 (nu (x0 w))); [Intro | Auto].
Inversion_clear H27.
Auto.
Change (Op_S (Bfun n2) ([_:name](x2 _ x3) w) ([_:name](x0 _ x3) w)).
Apply op_s with z; Intros.
Apply notin_nu; Intros.
Inversion_clear H23.
Cut (notin z (nu (x2 z0))); [Intro | Auto].
Inversion_clear H23.
Auto.
Cut (notin z (q w)); [Intro | Inversion H3; Auto].
Elim (BTR_L1 (q w) (x0 w) (bOut w) z H32 H10); Intros.
Apply proc_mono with w.
Inversion_clear H35; Auto.
Assumption.
Auto.
Cut (notin w (p z)); [Intro | Inversion H6; Auto].
Elim (BTR_L1 (p z) (x2 z) (bOut z) w H32 H21); Intros.
Apply proc_mono with z.
Inversion_clear H35; Auto.
Apply notin_nu; Intros.
Inversion_clear H14.
Cut (notin w (nu (x0 z0))); [Intro | Auto].
Inversion_clear H14.
Auto.
Auto.
Auto.
Auto.
Apply notin_nu; Intros.
Inversion_clear H30.
Cut (notin x3 (nu (x2 z))); [Intro | Auto].
Inversion_clear H30.
Auto.
Apply notin_nu; Intros.
Inversion_clear H27.
Cut (notin x3 (nu (x0 z))); [Intro | Auto].
Inversion_clear H27.
Auto.

Cut (notin z (q w)); [Intro | Inversion H3; Auto].
Elim (BTR_L1 (q w) (x0 w) (bOut w) z H20 H10); Intros.
(Apply BTR_L3 with z; Auto).
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Assumption.
Cut (notin w (p z)); [Intro | Inversion H6; Auto].
Elim (BTR_L1 (p z) (x2 z) (bOut z) w H27 H21); Intros.
Apply proc_mono with z; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Assumption.

Cut (notin z (q w)); [Intro | Inversion H3; Auto].
Elim (BTR_L1 (q w) (x0 w) (bOut w) z H13 H10); Intros.
(Apply BTR_L3 with w; Auto).
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Assumption.
Apply proc_mono with w; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Assumption.

(* Case ~w=x *)

Elim (ho_proc_exp q1 w); Intros.
Inversion_clear H13.
Rewrite H15 in H10.
(Cut (btrans (q z) (bOut x) (x0 z)); [ Intro | Idtac ]).
Inversion_clear H11.
Inversion_clear H17.
(Elim (H18 x); Intros).
(Elim (H19 (x0 z) H13); Intros).
Inversion_clear H20.
(Elim (ho_proc_exp x1 z); Intros).
Inversion_clear H20.
Rewrite H24 in H21.
Rewrite H24 in H22.
(Cut (btrans (p w) (bOut x) (x2 w)); [ Intro | Idtac ]).
(Exists  (x2 w); Split).
Assumption.
Rewrite H15.
Intros.
Elim (unsat (par (nu [_:name](nu (x0 _))) (nu [_:name](nu (x2 _)))) (cons y (cons z (cons w empty)))); Intros.
Inversion_clear H27.
Inversion_clear H28.
Inversion_clear H29.
Inversion_clear H31.
Inversion_clear H32.
Clear H33.
Cut (BL6 (x2 z x3) (x0 z x3)); [Intro | Apply H22; Auto].
Inversion_clear H32.
Apply bl6 with (S (S n2)).
Simpl.
Apply op_s with x3; Intros.
Apply notin_nu; Intros.
Inversion_clear H30.
Cut (notin x3 (nu (x2 w))); [Intro | Auto].
Inversion_clear H30.
Auto.
Apply notin_nu; Intros.
Inversion_clear H27.
Cut (notin x3 (nu (x0 w))); [Intro | Auto].
Inversion_clear H27.
Auto.
Change (Op_S (Bfun n2) ([_:name](x2 _ x3) w) ([_:name](x0 _ x3) w)).
Apply op_s with z; Intros.
Apply notin_nu; Intros.
Inversion_clear H23.
Cut (notin z (nu (x2 z0))); [Intro | Auto].
Inversion_clear H23.
Auto.
Cut (notin z (q w)); [Intro | Inversion H3; Auto].
Elim (BTR_L1 (q w) (x0 w) (bOut x) z H32 H10); Intros.
Apply proc_mono with w.
Inversion_clear H35; Auto.
Assumption.
Auto.
Cut (notin w (p z)); [Intro | Inversion H6; Auto].
Elim (BTR_L1 (p z) (x2 z) (bOut x) w H32 H21); Intros.
Apply proc_mono with z.
Inversion_clear H35; Auto.
Apply notin_nu; Intros.
Inversion_clear H14.
Cut (notin w (nu (x0 z0))); [Intro | Auto].
Inversion_clear H14.
Auto.
Auto.
Auto.
Auto.
Apply notin_nu; Intros.
Inversion_clear H30.
Cut (notin x3 (nu (x2 z))); [Intro | Auto].
Inversion_clear H30.
Auto.
Apply notin_nu; Intros.
Inversion_clear H27.
Cut (notin x3 (nu (x0 z))); [Intro | Auto].
Inversion_clear H27.
Auto.

Cut (notin z (q w)); [Intro | Inversion H3; Auto].
Elim (BTR_L1 (q w) (x0 w) (bOut x) z H20 H10); Intros.
Change (btrans (p w) ([_:name](bOut x) w) (x2 w)).
(Apply BTR_L3 with z; Auto).
Unfold b_act_notin_ho; Intros; Assumption.
Cut (notin w (p z)); [Intro | Inversion H6; Auto].
Elim (BTR_L1 (p z) (x2 z) (bOut x) w H27 H21); Intros.
Apply proc_mono with z; Assumption.
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Assumption.

Cut (notin z (q w)); [Intro | Inversion H3; Auto].
Elim (BTR_L1 (q w) (x0 w) (bOut x) z H13 H10); Intros.
Change (btrans (q z) ([_:name](bOut x) z) (x0 z)).
(Apply BTR_L3 with w; Auto).
Unfold b_act_notin_ho; Intros; Apply b_act_notin_bOut; Assumption.
Apply proc_mono with w; Assumption.
Unfold b_act_notin_ho; Intros; Assumption.

Qed.

(****************************)
(* Now, the "real" Lemma 6. *)
(****************************)

Lemma L6: (p,q:name->proc)(z:name)
         (notin z (nu p)) -> (notin z (nu q)) ->
         (StBisim (p z) (q z)) ->
 (w:name)~(w=z)->(notin w (nu p)) -> (notin w (nu q)) ->
         (StBisim (p w) (q w)).

Proof.

Intros.
Apply Completeness.
Apply Co_Ind with BL6.
Exact BisimL6.
Cut (BL6 (p z) (q z)); [Intro | Apply bl6 with O; Auto].
Inversion_clear H5.
Apply bl6 with (S n).
Simpl.
Apply op_s with z; Auto.

Qed.

(*****************************************************)
(* A lighter version of the same lemma, often useful *)
(* in proving algebraic laws.                        *)
(*****************************************************)

Lemma L6_Light: (p,q:name->proc)(z:name)
         (notin z (nu p)) -> (notin z (nu q)) ->
         (StBisim (p z) (q z)) ->
 (w:name)(notin w (nu p)) -> (notin w (nu q)) ->
         (StBisim (p w) (q w)).

Proof.

Intros.
Elim (LEM_name w z); Intros.
Rewrite -> H4; Auto.
Apply L6 with z; Auto.

Qed.

End MPW_Lemma6.
